/*
                        Confirmation Dialog

	(I mispelled confirmation, could someone please fix it?)


	Functions:

	int ComfDialogYesPBCB(void *ptr)
        int ComfDialogNoPBCB(void *ptr)
        int ComfDialogAllPBCB(void *ptr)


	int ComfDialogInit()
	int ComfDialogDraw()
	int ComfDialogManage(event_t *event)
	int ComfDialogHaltTillButton()


	File or pointer operation confirmation:

	void ComfDialogMapQueryMovement(
		void *tar_ptr,
		int tar_ptr_type,
		void *src_ptr,
		int src_ptr_type,
		int tar_drag_type,
		int src_drag_type,
	        int tar_item_num,
	        int op_code
	)
	int ComfDialogMapQueryOverWrite(
	        char *tar_path,
	        char *src_path,
	        int op_code
	)


	Recycle bin confirmations:

        int ComfDialogMapQueryPurge(
		recbin_win_struct *rb,
                char *object_name
        )
        int ComfDialogMapQueryPurgeAll(
                recbin_win_struct *rb
        )


	Mount or unmount device confirmation:

	int ComfDialogMapQueryMount(
		char *src_path
	)
        int ComfDialogMapQueryUnmount(
                char *src_path
        )


	void ComfDialogResetQuery()
	void ComfDialogUnmap()
	void ComfDialogDestroy()


 */

#include "endeavour.h"

#define FIXED_WIDTH		500
#define FIXED_HEIGHT		200

#define BUTTON_WIDTH		70
#define BUTTON_HEIGHT		28


/*
 *	Halt confirmation records.
 */
static bool_t halt_comf_dialog;
static int last_pbtn_pressed;


/* *********************************************************************
 *
 *
 */
int ComfDialogYesPBCB(void *ptr)
{
	char stringa[256];


	/* ********************************************************** */
	/* Confirm pointer motion. */
	if(comf_dialog.query_type == CDIALOG_QUERY_CONFIRM_MOTION)
        {
	    /*   Call appropriate movement processor function by
             *   the target pointer type.
             */
	    switch(comf_dialog.tar_ptr_type)
	    {
	      case POINTER_TYPE_EDW:
	        EDWProcessMovement(
		    comf_dialog.tar_ptr,
                    comf_dialog.tar_ptr_type,
		    comf_dialog.src_ptr,
		    comf_dialog.src_ptr_type,
		    comf_dialog.tar_drag_type,
		    comf_dialog.src_drag_type,
		    comf_dialog.tar_item_num,
		    comf_dialog.op_code,
		    True	/* Confirmed! */
	        );
		break;

#ifdef USE_IMLIB
              case POINTER_TYPE_IB:
                IMBRProcessMovement(
                    comf_dialog.tar_ptr,
                    comf_dialog.tar_ptr_type,
                    comf_dialog.src_ptr,
                    comf_dialog.src_ptr_type,
                    comf_dialog.tar_drag_type,
                    comf_dialog.src_drag_type,
                    comf_dialog.tar_item_num,
                    comf_dialog.op_code,
                    True        /* Confirmed! */
                );
                break;
#endif /* USE_IMLIB */

	      default:
		sprintf(stringa,
"Internal non-critical error:\n\n\
    Confirmation target pointer type %i unsupported.\n\n",
		    comf_dialog.tar_ptr_type
		);
		printdw(&err_dw, stringa);
		break;
	    }

            ComfDialogUnmap();
        }
	/* ********************************************************* */
	/* Overwrite. */
        else if(comf_dialog.query_type == CDIALOG_QUERY_OVERWRITE)
        {
	    /* Adjust global halt_comf_dialog to False to stop halting. */
	    halt_comf_dialog = False;

	    /* Record last pbtn. */
	    last_pbtn_pressed = YES_PBTN_NUM;

            ComfDialogUnmap();
        }
        /* ********************************************************* */
        /* Purge. */
        else if(comf_dialog.query_type == CDIALOG_QUERY_RECBIN_PURGE)
        {
            /* Adjust global halt_comf_dialog to False to stop halting. */
            halt_comf_dialog = False;
         
            /* Record last pbtn. */
            last_pbtn_pressed = YES_PBTN_NUM;

            ComfDialogUnmap();
	}
        /* ********************************************************* */
        /* Purge all. */
        else if(comf_dialog.query_type == CDIALOG_QUERY_RECBIN_PURGE_ALL)
        {
            /* Adjust global halt_comf_dialog to False to stop halting. */
            halt_comf_dialog = False;

            /* Record last pbtn. */
            last_pbtn_pressed = YES_PBTN_NUM;

            ComfDialogUnmap(); 
        }
        /* ********************************************************* */
        /* Device mounting. */
        else if((comf_dialog.query_type == CDIALOG_QUERY_DEVICE_MOUNT) ||
                (comf_dialog.query_type == CDIALOG_QUERY_DEVICE_UNMOUNT)
        )
        {
            /* Adjust global halt_comf_dialog to False to stop halting. */
            halt_comf_dialog = False;

            /* Record last pbtn. */
            last_pbtn_pressed = YES_PBTN_NUM;

            ComfDialogUnmap();
	}

	return(0);
}



int ComfDialogNoPBCB(void *ptr)
{
        /* ********************************************************* */
        /* Pointer motion. */
	if(comf_dialog.query_type == CDIALOG_QUERY_CONFIRM_MOTION)
	{
	    ComfDialogUnmap();
	}
        /* ********************************************************* */
        /* Overwrite. */
	else if(comf_dialog.query_type == CDIALOG_QUERY_OVERWRITE)
	{
            /* Adjust global halt_comf_dialog to False to stop halting. */
            halt_comf_dialog = False;

            /* Record last pbtn. */
            last_pbtn_pressed = NO_PBTN_NUM;

	    ComfDialogUnmap();
	}
        /* ********************************************************* */
        /* Purge. */
        else if(comf_dialog.query_type == CDIALOG_QUERY_RECBIN_PURGE)
        {
            /* Adjust global halt_comf_dialog to False to stop halting. */
            halt_comf_dialog = False;

            /* Record last pbtn. */
            last_pbtn_pressed = NO_PBTN_NUM;

            ComfDialogUnmap();
        }
        /* ********************************************************* */
        /* Purge all. */
        else if(comf_dialog.query_type == CDIALOG_QUERY_RECBIN_PURGE_ALL)
        {
            /* Adjust global halt_comf_dialog to False to stop halting. */
            halt_comf_dialog = False;
        
            /* Record last pbtn. */
            last_pbtn_pressed = NO_PBTN_NUM;

            ComfDialogUnmap();
        }
        /* ********************************************************* */
        /* Device mounting. */
        else if((comf_dialog.query_type == CDIALOG_QUERY_DEVICE_MOUNT) ||
                (comf_dialog.query_type == CDIALOG_QUERY_DEVICE_UNMOUNT)
        )
        {
            /* Adjust global halt_comf_dialog to False to stop halting. */
            halt_comf_dialog = False;

            /* Record last pbtn. */
            last_pbtn_pressed = NO_PBTN_NUM;

            ComfDialogUnmap();
        }

	return(0);
}



int ComfDialogAllPBCB(void *ptr)
{
        /* ********************************************************* */
        /* Pointer motion. */
        if(comf_dialog.query_type == CDIALOG_QUERY_CONFIRM_MOTION)
        {
            /* Not applicateable. */
        }
        /* ********************************************************* */
        /* Overwrite. */
        else if(comf_dialog.query_type == CDIALOG_QUERY_OVERWRITE)
        {
            /* Adjust global halt_comf_dialog to False to stop halting. */
            halt_comf_dialog = False;

            /* Record last pbtn. */
            last_pbtn_pressed = ALL_PBTN_NUM;

            ComfDialogUnmap();
        }
        /* ********************************************************* */
        /* Purge. */
        else if(comf_dialog.query_type == CDIALOG_QUERY_RECBIN_PURGE)
        {
            /* Adjust global halt_comf_dialog to False to stop halting. */
            halt_comf_dialog = False;

            /* Record last pbtn. */
            last_pbtn_pressed = ALL_PBTN_NUM;

            ComfDialogUnmap();
        }
        /* ********************************************************* */
        /* Purge all. */
        else if(comf_dialog.query_type == CDIALOG_QUERY_RECBIN_PURGE_ALL)
        {
            /* Not applicateable. */
        }
        /* ********************************************************* */
        /* Device mounting. */
        else if((comf_dialog.query_type == CDIALOG_QUERY_DEVICE_MOUNT) ||
                (comf_dialog.query_type == CDIALOG_QUERY_DEVICE_UNMOUNT)
        )
        {
            /* Adjust global halt_comf_dialog to False to stop halting. */
            halt_comf_dialog = False;

            /* Record last pbtn. */
            last_pbtn_pressed = ALL_PBTN_NUM;
            
            ComfDialogUnmap();
        }

 
	return(0);
}
   


/* ***********************************************************************
 *
 *
 *
 */
int ComfDialogInit()
{
        win_attr_t wattr;
        pixmap_t pixmap;


	/* Error checks. */
	if(!IDC())
	    return(-1);


	/* *********************************************************** */
        comf_dialog.map_state = 0;
        comf_dialog.visibility_state = VisibilityFullyObscured;
        comf_dialog.x = 0;
        comf_dialog.y = 0;
        comf_dialog.width = FIXED_WIDTH;
        comf_dialog.height = FIXED_HEIGHT;
        comf_dialog.is_in_focus = 0;
        comf_dialog.disabled = False;
 

	/* Toplevel. */
	if(
	    OSWCreateWindow(
		&comf_dialog.toplevel,
		osw_gui[0].root_win,
		comf_dialog.x,
		comf_dialog.y,
		comf_dialog.width,
		comf_dialog.height
	    )
	)
	    return(-1);

        OSWSetWindowInput(comf_dialog.toplevel, KeyPressMask |
            KeyReleaseMask | ExposureMask | StructureNotifyMask |
            FocusChangeMask | VisibilityChangeMask
        );
        OSWSetWindowBkg(comf_dialog.toplevel, 0, widget_global.std_bkg_pm);

        WidgetCenterWindow(comf_dialog.toplevel, WidgetCenterWindowToRoot);
        OSWGetWindowAttributes(comf_dialog.toplevel, &wattr);
        comf_dialog.x = wattr.x;
        comf_dialog.y = wattr.y;
        comf_dialog.width = wattr.width;
        comf_dialog.height = wattr.height;


        pixmap = OSWCreatePixmapFromImage(endeavour_image.endeavour_icon);

        /* WM properties. */
	OSWSetWindowWMProperties(
	    comf_dialog.toplevel,
	    "Endeavour",
            "Endeavour",
            ((pixmap == 0) ? widget_global.std_icon_pm : pixmap),
            False,
	    comf_dialog.x, comf_dialog.y,
            comf_dialog.width, comf_dialog.height,
            comf_dialog.width, comf_dialog.height,
            WindowFrameStyleFixedFrameOnly,
            NULL, 0
	);
        pixmap = 0;     /* Do not destroy pixmap. */


	/* Buffer for toplevel. */
	OSWGetWindowAttributes(comf_dialog.toplevel, &wattr);
	if(
	    OSWCreatePixmap(
		&comf_dialog.toplevel_buf,
		wattr.width, wattr.height
	    )
	)
	    return(-1);


	/* ************************************************************** */

	/* Yes button. */
	if(
	    PBtnInit(
		&comf_dialog.yes_btn,
		comf_dialog.toplevel,
		(int)comf_dialog.width - (3 * 10) - (3 * BUTTON_WIDTH),
		(int)comf_dialog.height - 10 - BUTTON_HEIGHT,
		BUTTON_WIDTH, BUTTON_HEIGHT,
		"Yes",
		PBTN_TALIGN_CENTER,
		NULL,
		(void *)&comf_dialog.yes_btn,
		ComfDialogYesPBCB
	    )
	)
	    return(-1);
	PBtnSetHotKeys(&comf_dialog.yes_btn, "y");

        /* No button. */
        if(
            PBtnInit(
                &comf_dialog.no_btn,
                comf_dialog.toplevel,
                (int)comf_dialog.width - (2 * 10) - (2 * BUTTON_WIDTH),
                (int)comf_dialog.height - 10 - BUTTON_HEIGHT,
                BUTTON_WIDTH, BUTTON_HEIGHT,
                "No",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&comf_dialog.no_btn,
                ComfDialogNoPBCB
            )
        )
            return(-1);
        PBtnSetHotKeys(&comf_dialog.no_btn, "n");

        /* Yes to all button. */
        if(
            PBtnInit(
                &comf_dialog.all_btn,
                comf_dialog.toplevel,
                (int)comf_dialog.width - (1 * 10) - (1 * BUTTON_WIDTH),
                (int)comf_dialog.height - 10 - BUTTON_HEIGHT,
                BUTTON_WIDTH, BUTTON_HEIGHT,
                "All", 
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&comf_dialog.all_btn,
                ComfDialogAllPBCB
            )
        )
            return(-1);
        PBtnSetHotKeys(&comf_dialog.all_btn, "a");


	/* Reset query values. */
	comf_dialog.query_type = CDIALOG_QUERY_NONE;
	comf_dialog.op_code = OP_CODE_NONE;

	comf_dialog.tar_path = NULL;
	comf_dialog.src_path = NULL;

	comf_dialog.tar_ptr = NULL;
	comf_dialog.tar_ptr_type = POINTER_TYPE_UNDEFINED;
	comf_dialog.src_ptr = NULL;
        comf_dialog.src_ptr_type = POINTER_TYPE_UNDEFINED;
        comf_dialog.tar_drag_type = DRAG_TYPE_NONE;
        comf_dialog.src_drag_type = DRAG_TYPE_NONE;
        comf_dialog.tar_item_num = 0;


	return(0);
}



int ComfDialogDraw()
{
	int x;
	int x_pos, y_pos;
	win_attr_t wattr;
	struct stat stat_buf;
	char stringa[PATH_MAX + NAME_MAX + 128];
	char tmp_path[PATH_MAX + NAME_MAX];
	image_t *imgptr, *image;

	endeavour_win_struct *src_edw, *tar_edw;
	recbin_win_struct *src_rb, *tar_rb;
#ifdef USE_IMLIB
	img_browser_struct *src_ib, *tar_ib;
#endif


	/* Get toplevel attributes. */
	OSWGetWindowAttributes(comf_dialog.toplevel, &wattr);

	/* Map as needed. */
	if(!comf_dialog.map_state)
	{
	    /* Need to reset button states. */
	    comf_dialog.yes_btn.state = PBTN_UNARMED;
	    comf_dialog.no_btn.state = PBTN_UNARMED;
	    comf_dialog.all_btn.state = PBTN_UNARMED;


	    OSWMapRaised(comf_dialog.toplevel);

	    /* Confirmations for yes or no. */
	    if((comf_dialog.query_type == CDIALOG_QUERY_CONFIRM_MOTION) ||
               (comf_dialog.query_type == CDIALOG_QUERY_RECBIN_PURGE_ALL)
	    )
	    {
		OSWMoveWindow(
		    comf_dialog.yes_btn.toplevel,
		    (int)wattr.width - (2 * BUTTON_WIDTH) - (2 * 10),
		    (int)wattr.height - BUTTON_HEIGHT - 10
		);
                OSWMoveWindow(
                    comf_dialog.no_btn.toplevel,
                    (int)wattr.width - (1 * BUTTON_WIDTH) - (1 * 10),
                    (int)wattr.height - BUTTON_HEIGHT - 10
                );

		PBtnUnmap(&comf_dialog.all_btn);
                PBtnMap(&comf_dialog.yes_btn);
                PBtnMap(&comf_dialog.no_btn);
	    }
	    /* Confirmations for yes, no, and all. */
	    else
	    {
                OSWMoveWindow(
                    comf_dialog.yes_btn.toplevel,
                    (int)wattr.width - (3 * BUTTON_WIDTH) - (3 * 10),
                    (int)wattr.height - BUTTON_HEIGHT - 10
                );
                OSWMoveWindow(
                    comf_dialog.no_btn.toplevel,
                    (int)wattr.width - (2 * BUTTON_WIDTH) - (2 * 10),
                    (int)wattr.height - BUTTON_HEIGHT - 10
                );
                OSWMoveWindow(
                    comf_dialog.all_btn.toplevel,
                    (int)wattr.width - (1 * BUTTON_WIDTH) - (1 * 10),
                    (int)wattr.height - BUTTON_HEIGHT - 10
                );

		PBtnMap(&comf_dialog.all_btn);
                PBtnMap(&comf_dialog.yes_btn);
                PBtnMap(&comf_dialog.no_btn);
	    }

	    comf_dialog.map_state = 1;
	}

	/* *********************************************************** */
        /* Get source and target pointers. */
        tar_edw = NULL;
        src_edw = NULL;
	tar_rb = NULL;
	src_rb = NULL;
#ifdef USE_IMLIB
	tar_ib = NULL;
	src_ib = NULL;
#endif

	/* Get target pointer. */
        switch(comf_dialog.tar_ptr_type)
        {
          case POINTER_TYPE_EDW:
            tar_edw = (endeavour_win_struct *)comf_dialog.tar_ptr;
            break;
          case POINTER_TYPE_RB:
            tar_rb = (recbin_win_struct *)comf_dialog.tar_ptr;
            break;
#ifdef USE_IMLIB
          case POINTER_TYPE_IB:
            tar_ib = (img_browser_struct *)comf_dialog.tar_ptr;
            break;
#endif
          default:
/* This is okay, over-write for instance has no pointer types.
            fprintf(stderr,
 "ComfDialogDraw(): %i: Unsupported target pointer type.\n",
                comf_dialog.tar_ptr_type);
*/
            break;
        }
        /* Get source pointer. */
        switch(comf_dialog.src_ptr_type)
        {
          case POINTER_TYPE_EDW:
            src_edw = (endeavour_win_struct *)comf_dialog.src_ptr;
            break;
          case POINTER_TYPE_RB:
            src_rb = (recbin_win_struct *)comf_dialog.src_ptr;
            break;
#ifdef USE_IMLIB  
          case POINTER_TYPE_IB:
            src_ib = (img_browser_struct *)comf_dialog.src_ptr;
            break;
#endif
          default:
/* This is okay, over-write for instance has no pointer types.
            fprintf(stderr,
 "ComfDialogDraw(): %i: Unsupported source pointer type.\n",
                comf_dialog.src_ptr_type);
*/
            break;
        }


	/* *********************************************************** */

	/* Tile toplevel background. */
	WidgetPutImageTile(
	    comf_dialog.toplevel_buf, widget_global.std_bkg_img,
	    wattr.width, wattr.height
	);


	/* *********************************************************** */
	/* Draw message. */

	OSWSetFgPix(widget_global.normal_text_pix);

	switch(comf_dialog.query_type)
	{
          /* ********************************************************* */
	  case CDIALOG_QUERY_CONFIRM_MOTION:

            imgptr = endeavour_image.warning_icon;
            if(imgptr != NULL)
                x_pos = (int)imgptr->width + 25;
            else
                x_pos = 15;
            y_pos = 20;

	    /* Line 1: Confirmation subject */
	    OSWDrawString(comf_dialog.toplevel_buf,
		x_pos + 3, y_pos + 5,
		"Confirm pointer motion:"
	    );


	    /* Line 2: Source */
	    stringa[0] = '\0';
	    y_pos += 25;

	    /* Source: Endeavour Browser */
	    if(src_edw != NULL)
	    {
		/* Copy. */
		if(comf_dialog.op_code == OP_CODE_COPY)
		{
		    if(comf_dialog.src_drag_type == DRAG_TYPE_DIR)
		        sprintf(stringa,
		            "    Copy: %i director%s",
			    src_edw->total_sel_dirs,
			    (src_edw->total_sel_dirs > 1) ?
				"ies" : "y"
			);
		    else if(comf_dialog.src_drag_type == DRAG_TYPE_FILE)
                        sprintf(stringa,
                            "    Copy: %i file%s",
                            src_edw->total_sel_files,
                            (src_edw->total_sel_files > 1) ?
                                "s" : ""
                        );
                    else
                        sprintf(stringa,
                            "    Copy: ???"
                        );
		}
		/* Move. */
		else if(comf_dialog.op_code == OP_CODE_MOVE)
                {   
                    if(comf_dialog.src_drag_type == DRAG_TYPE_DIR)
                        sprintf(stringa,
                            "    Move: %i director%s",
                            src_edw->total_sel_dirs,
                            (src_edw->total_sel_dirs > 1) ?
                                "ies" : "y"
                        );
                    else if(comf_dialog.src_drag_type == DRAG_TYPE_FILE)
                        sprintf(stringa,
                            "    Move: %i file%s",
                            src_edw->total_sel_files,
                            (src_edw->total_sel_files > 1) ?
                                "s" : ""
                        );
                    else
                        sprintf(stringa,
                            "    Move: ???"
                        );
                }
	    }
	    /* Source: Recycle Bin */
            else if(src_rb != NULL)
            {
                if(comf_dialog.src_drag_type == DRAG_TYPE_RECYCLE_BIN)
                    sprintf(stringa,
                        "    Recover recycled objects"
                    );
                else
                    sprintf(stringa,
                        "    Recover: ???"
                    );
            }
#ifdef USE_IMLIB
            /* Source: Image Browser */
            else if(src_ib != NULL)
            {
                /* Copy. */
                if(comf_dialog.op_code == OP_CODE_COPY)
                {
                    sprintf(stringa,
                        "    Copy: %i object%s",
                        src_ib->total_sel_thumbs,
                        (src_ib->total_sel_thumbs > 1) ? "s" : ""
                    );
		} 
                /* Move. */
                else if(comf_dialog.op_code == OP_CODE_MOVE)
                {
                    sprintf(stringa,
                        "    Move: %i object%s",
                        src_ib->total_sel_thumbs,
                        (src_ib->total_sel_thumbs > 1) ? "s" : "" 
                    );
                }
            }
#endif
            OSWDrawString(comf_dialog.toplevel_buf,
                x_pos, y_pos + 5,
                stringa
            );


	    /* Line 3. */
	    stringa[0] = '\0';
	    y_pos += 25;
	    /* Target: Endeavour Browser */
            if(tar_edw != NULL)
            {
		/* Directory. */
		if(comf_dialog.tar_drag_type == DRAG_TYPE_DIR)
		{
		    x = comf_dialog.tar_item_num;
		    if((x >= 0) && (x < (int)tar_edw->total_dirs))
		    {
			if(tar_edw->dir[x] != NULL)
			{
			    strncpy(
			        tmp_path,
                                ((tar_edw->dir[x]->fullname == NULL) ?
				    "(null)" : tar_edw->dir[x]->fullname),
			        PATH_MAX + NAME_MAX
			    );
			}
		    }
		}
		/* File. */
		else if(comf_dialog.tar_drag_type == DRAG_TYPE_FILE)
		{
		    strncpy(
			tmp_path,
			((tar_edw->loc_prompt.buf == NULL) ?
                            "(null)" : tar_edw->loc_prompt.buf),
                        PATH_MAX + NAME_MAX
		    );
		}
		tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
		SimplifyPath(tmp_path);

		sprintf(stringa, "    To: %s", tmp_path);
	    }
            /* Target: Recycle Bin */
            else if(tar_rb != NULL)
	    {
		/* Recycle confirmation is made in overwrite comfermation. */
	    }
#ifdef USE_IMLIB
            /* Target: Image Browser */
            else if(tar_ib != NULL)
            {
		tmp_path[0] = '\0';

		/* current_dir must be valid on ib. */
		if(tar_ib->current_dir != NULL)
		{
		    x = comf_dialog.tar_item_num;
                    if((x >= 0) && (x < (int)tar_ib->total_thumbs))
		    {
		        if(tar_ib->thumb[x] != NULL)
                        {
			    const char *cstrptr;


                            cstrptr = (const char *)PrefixPaths(
				tar_ib->current_dir,
				tar_ib->thumb[x]->filename
			    );
			    strncpy(
				tmp_path,
				(cstrptr == NULL) ? "/" : cstrptr,
				PATH_MAX + NAME_MAX
			    );
			    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

			    if(!ISPATHDIR(tmp_path))
			    {
				/* Target on thumb not dir, use current. */
				strncpy(tmp_path, tar_ib->current_dir,
				    PATH_MAX + NAME_MAX);
				tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
			    }
			}
		    }
		    else
		    {
			/* No selection or invalid selection, use current. */
                        strncpy(tmp_path, tar_ib->current_dir,
                            PATH_MAX + NAME_MAX);
                        tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
		    }
		}
		SimplifyPath(tmp_path);
		sprintf(stringa, "    To: %s", tmp_path);
            }
#endif
            OSWDrawString(comf_dialog.toplevel_buf,
                x_pos, y_pos + 5,
                stringa
            );

	    break;

	  /* ************************************************************** */
	  case CDIALOG_QUERY_OVERWRITE:

            imgptr = endeavour_image.warning_icon;
            if(imgptr != NULL)
                x_pos = (int)imgptr->width + 25;
            else
                x_pos = 15;
            y_pos = 20;

	    /* Move, copy, or rename operations. */
	    if((comf_dialog.op_code == OP_CODE_COPY) ||
               (comf_dialog.op_code == OP_CODE_MOVE) ||
               (comf_dialog.op_code == OP_CODE_RENAME)
	    )
            {
		OSWDrawString(comf_dialog.toplevel_buf,
                    x_pos, y_pos + 5,
                    "Confirm overwrite:"
                );

                /* What is about to be overwritten. */
                if(comf_dialog.tar_path != NULL)
                {
		    y_pos += 30;

		    lstat(comf_dialog.tar_path, &stat_buf);

		    /* Label. */
                    sprintf(stringa,
                        "    Replace:     %s",
                        comf_dialog.tar_path
                    );
                    OSWDrawString(comf_dialog.toplevel_buf,
                        x_pos, y_pos + 5,
                        stringa
                    );

                    /* Icon. */
                    if(S_ISLNK(stat_buf.st_mode))
                        image = widget_global.linkicon_selected_img;
                    else if(S_ISDIR(stat_buf.st_mode))
                        image = widget_global.diricon_selected_img;
                    else if((S_IXUSR & stat_buf.st_mode) ||
                            (S_IXGRP & stat_buf.st_mode) ||
                            (S_IXOTH & stat_buf.st_mode)
                    )
                        image = widget_global.execicon_selected_img;
                    else if(S_ISREG(stat_buf.st_mode))
                        image = widget_global.fileicon_selected_img;
                    else if(S_ISCHR(stat_buf.st_mode) ||
                            S_ISBLK(stat_buf.st_mode) ||
                            S_ISFIFO(stat_buf.st_mode) ||
                            S_ISSOCK(stat_buf.st_mode)
                    )
                        image = widget_global.pipeicon_selected_img;
                    else
                        image = widget_global.fileicon_selected_img;

                    WidgetPutImageRaised(
                        comf_dialog.toplevel_buf,
                        image,
                        x_pos + 94, y_pos - 12,
                        10
                    );

		    /* Date and time label. */
                    y_pos += 22;
                    sprintf(stringa, "        %ld bytes   %s",
                        stat_buf.st_size,
                        EDW_GETTIMESTR(stat_buf.st_mtime)
		    );
                    OSWDrawString(comf_dialog.toplevel_buf,
                        x_pos, y_pos + 5,
                        stringa
                    );
	        }

	        /* What will we replace it with. */
	        if(comf_dialog.src_path != NULL)
                {
                    y_pos += 30;

		    lstat(comf_dialog.src_path, &stat_buf);

		    /* Label. */
                    sprintf(stringa,
                        "    With:     %s",
                        comf_dialog.src_path
                    );
                    OSWDrawString(comf_dialog.toplevel_buf,
                        x_pos, y_pos + 5,
                        stringa
                    );

                    /* Icon. */
                    if(S_ISLNK(stat_buf.st_mode))
                        image = widget_global.linkicon_selected_img;
                    else if(S_ISDIR(stat_buf.st_mode))
                        image = widget_global.diricon_selected_img;
                    else if((S_IXUSR & stat_buf.st_mode) ||
                            (S_IXGRP & stat_buf.st_mode) ||
                            (S_IXOTH & stat_buf.st_mode)
                    )
                        image = widget_global.execicon_selected_img;
                    else if(S_ISREG(stat_buf.st_mode))
                        image = widget_global.fileicon_selected_img;
                    else if(S_ISCHR(stat_buf.st_mode) ||
                            S_ISBLK(stat_buf.st_mode) ||
                            S_ISFIFO(stat_buf.st_mode) ||
                            S_ISSOCK(stat_buf.st_mode)
                    )
                        image = widget_global.pipeicon_selected_img;
                    else
                        image = widget_global.fileicon_selected_img;
             
                    WidgetPutImageRaised(
                        comf_dialog.toplevel_buf,
                        image,
                        x_pos + 72, y_pos - 12,
                        10
                    );

		    /* Date and time label. */
                    y_pos += 22;
                    sprintf(stringa, "        %ld bytes   %s",
                        stat_buf.st_size,
                        EDW_GETTIMESTR(stat_buf.st_mtime)
		    );
                    OSWDrawString(comf_dialog.toplevel_buf,
                        x_pos, y_pos + 5,
                        stringa
                    );
		}
            }
	    /* Recycle. */
	    else if(comf_dialog.op_code == OP_CODE_RECYCLE)
	    {
                OSWDrawString(comf_dialog.toplevel_buf,
                    x_pos, y_pos + 5,
                    "Confirm recycle:"
                );

                /* What is about to be overwritten. */
                if(comf_dialog.src_path != NULL)
                {
		    y_pos += 30;


		    lstat(comf_dialog.src_path, &stat_buf);

		    /* Name. */
                    sprintf(stringa, "      %s",
                        comf_dialog.src_path
                    );
                    OSWDrawString(comf_dialog.toplevel_buf,
                        x_pos, y_pos + 5,
                        stringa
                    );

		    /* If it is a dir, warn about its contents. */
		    if(S_ISDIR(stat_buf.st_mode))
		    {
                        OSWDrawString(comf_dialog.toplevel_buf,
			    x_pos, y_pos + 5 + 30,
                            "And all of its contents."
                        );
		    }

		    /* Icon. */
                    if(S_ISLNK(stat_buf.st_mode))
                        image = widget_global.linkicon_selected_img;
                    else if(S_ISDIR(stat_buf.st_mode))
                        image = widget_global.diricon_selected_img;
                    else if((S_IXUSR & stat_buf.st_mode) ||
                            (S_IXGRP & stat_buf.st_mode) ||
                            (S_IXOTH & stat_buf.st_mode)
                    )
                        image = widget_global.execicon_selected_img;
                    else if(S_ISREG(stat_buf.st_mode))
                        image = widget_global.fileicon_selected_img;
                    else if(S_ISCHR(stat_buf.st_mode) ||
                            S_ISBLK(stat_buf.st_mode) ||
                            S_ISFIFO(stat_buf.st_mode) ||
                            S_ISSOCK(stat_buf.st_mode)
                    )
                        image = widget_global.pipeicon_selected_img;
                    else
                        image = widget_global.fileicon_selected_img;

		    WidgetPutImageRaised(
			comf_dialog.toplevel_buf,
			image,
			x_pos + 11, y_pos - 12,
			10
		    );

                } 
	    }
            break;

          /* ************************************************************** */
          case CDIALOG_QUERY_RECBIN_PURGE:

            imgptr = endeavour_image.purge_icon;
            if(imgptr != NULL)
                x_pos = (int)imgptr->width + 25;
            else
                x_pos = 15;
            y_pos = 20;

            /* Purge or purge all operation. */
            if(comf_dialog.op_code == OP_CODE_RECBIN_PURGE)
            {
                OSWDrawString(comf_dialog.toplevel_buf,
                    x_pos, y_pos + 5,
                    "Confirm purge of object in recycle bin:"
                );

		x = comf_dialog.tar_item_num;
	        if((int)total_recbin_wins > 0)
		{
                    y_pos += 25;

		    sprintf(stringa, "    %s",
                        comf_dialog.src_path
                    );
                    OSWDrawString(comf_dialog.toplevel_buf,
                        x_pos, y_pos + 5,
                        stringa
                    );
		}
	    }
            break;

          /* ************************************************************** */
          case CDIALOG_QUERY_RECBIN_PURGE_ALL:

            imgptr = endeavour_image.purge_icon;
            if(imgptr != NULL)
                x_pos = (int)imgptr->width + 25;
            else
                x_pos = 15;
            y_pos = 20;

            if(comf_dialog.op_code == OP_CODE_RECBIN_PURGE_ALL)
	    {
                OSWDrawString(comf_dialog.toplevel_buf,
                    x_pos, y_pos + 5,
                    "Purge all objects in this recycle bin?"
                );

		y_pos += 50;
                OSWDrawString(comf_dialog.toplevel_buf,
                    x_pos, y_pos + 5,
                    "CAUTION: CURRENT RECYCLE BIN CONTENTS WILL BE"
		);
                y_pos += 25;
                OSWDrawString(comf_dialog.toplevel_buf,
                    x_pos, y_pos + 5,
                    "PERMANENTLY DESTROYED BEYOND RECOVERY!"
                );
	    }
	    break;

          /* ************************************************************** */
          case CDIALOG_QUERY_DEVICE_MOUNT:

            imgptr = endeavour_image.warning_icon;
            if(imgptr != NULL)
                x_pos = (int)imgptr->width + 25;
            else
                x_pos = 15;
            y_pos = 20;

	    OSWDrawString(comf_dialog.toplevel_buf,
                x_pos, y_pos + 5,
                "Mount device:"
            );
	    y_pos += 25;
	    sprintf(stringa, "    %s", comf_dialog.src_path);
            OSWDrawString(comf_dialog.toplevel_buf,
                x_pos, y_pos + 5,
                stringa  
            );
            break;

          /* ************************************************************** */
          case CDIALOG_QUERY_DEVICE_UNMOUNT:

            imgptr = endeavour_image.warning_icon;
            if(imgptr != NULL)
                x_pos = (int)imgptr->width + 25;
            else
                x_pos = 15;
            y_pos = 20;

            OSWDrawString(comf_dialog.toplevel_buf,
                x_pos, y_pos + 5,
                "Unmount device:"
            );
            y_pos += 25;
            sprintf(stringa, "    %s", comf_dialog.src_path);
            OSWDrawString(comf_dialog.toplevel_buf,
                x_pos, y_pos + 5,
                stringa
            );
            break;

          /* ************************************************************** */
	  default:

            imgptr = endeavour_image.warning_icon;
            if(imgptr != NULL)
                x_pos = (int)imgptr->width + 25;
            else
                x_pos = 15;
            y_pos = 20;

            OSWDrawString(comf_dialog.toplevel_buf,
                x_pos, y_pos + 5,
                "Confirm:"
            );
            break;
	}

	/* Draw image. */
	if(imgptr != NULL)
	{
	    WidgetPutImageRaised(
		(drawable_t)comf_dialog.toplevel_buf,
		imgptr,
		10,
		(((int)wattr.height - BUTTON_HEIGHT - 40) / 2) -
		    ((int)imgptr->height / 2),
		16	/* Altitude. */
	    );
	}

	/* Draw horizontal rule. */
	OSWSetFgPix(widget_global.surface_highlight_pix);
	OSWDrawLine(comf_dialog.toplevel_buf,
	    0, (int)wattr.height - BUTTON_HEIGHT - 24,
	    (int)wattr.width, (int)wattr.height - BUTTON_HEIGHT - 24
	);
	OSWSetFgPix(widget_global.surface_shadow_pix);
        OSWDrawLine(comf_dialog.toplevel_buf,
            0, (int)wattr.height - BUTTON_HEIGHT - 25,
            (int)wattr.width, (int)wattr.height - BUTTON_HEIGHT - 25
        );

	/* Put buffer. */
	OSWPutBufferToWindow(comf_dialog.toplevel, comf_dialog.toplevel_buf);



	return(0);
}



int ComfDialogManage(event_t *event)
{
	int events_handled = 0;


	if((event == NULL) ||
           !comf_dialog.map_state
	)
	    return(0);


	switch(event->type)
	{
	  /* ******************************************************** */
	  case KeyPress:
            /* Skip KeyPress if edw not in focus. */
            if(!comf_dialog.is_in_focus)
                return(events_handled);

	    /* Turn auto repeat off. */
	    OSWKBAutoRepeatOff();

            break;


          /* ********************************************************* */
          case KeyRelease:
            if(!comf_dialog.is_in_focus)
                return(events_handled);

            /* Turn auto repeat on. */
            OSWKBAutoRepeatOn();

	    /* Yes. */
	    if((event->xkey.keycode == osw_keycode.space) ||
               (event->xkey.keycode == osw_keycode.enter) ||
               (event->xkey.keycode == osw_keycode.np_enter)
	    )
	    {
		ComfDialogYesPBCB(NULL);
		events_handled++;
		return(events_handled);
	    }
	    /* No. */
            else if(event->xkey.keycode == osw_keycode.esc)
            {
                ComfDialogNoPBCB(NULL);
                events_handled++;
		return(events_handled);
            }

            break;

	  /* ******************************************************** */
	  case Expose:
	    if(event->xany.window == comf_dialog.toplevel)
	    {
		events_handled++;
	    }
	    break;

          /* ******************************************************** */
          case VisibilityNotify:
            if(event->xany.window == comf_dialog.toplevel)
            {
                comf_dialog.visibility_state = event->xvisibility.state;

                events_handled++;
                return(events_handled);
            }
            break;
        
          /* ******************************************************** */
          case FocusIn:
            if(event->xany.window == comf_dialog.toplevel)
            {
                comf_dialog.is_in_focus = 1;
                events_handled++;
            }
            else
            {
                comf_dialog.is_in_focus = 0;
                /* Do not report any event for focus out. */
            }
            return(events_handled);
            break;

          /* ******************************************************** */
          case FocusOut:
            if(event->xany.window == comf_dialog.toplevel)
            {
                comf_dialog.is_in_focus = 0;
                events_handled++;
            }
            return(events_handled); 
            break;

	  /* ********************************************************** */
          case ClientMessage:
            if(OSWIsEventDestroyWindow(comf_dialog.toplevel, event))
            {
		/* Interprite wm delete as a no. */
                ComfDialogNoPBCB(NULL);

		ComfDialogUnmap();
  
                events_handled++;
		return(events_handled);
            }
	}

	if(events_handled > 0)
	{
	    ComfDialogDraw();
	    return(events_handled);
	}

	/* *********************************************************** */
	/* Manage widgets. */

	if(events_handled == 0)
	{
	    events_handled += PBtnManage(&comf_dialog.yes_btn, event);
	}
        if(events_handled == 0)
        {
            events_handled += PBtnManage(&comf_dialog.no_btn, event);
        }    
        if(events_handled == 0)
        {
            events_handled += PBtnManage(&comf_dialog.all_btn, event);
        }



	return(events_handled);
}



/* *************************************************************************
 *
 *      Manage confirmation dialog untill one of its  push buttons is press,
 *      then return with code of which was pressed (yes, no or all button).
 *      
 *      This function is intended for the query type CDIALOG_QUERY_OVERWRITE.
 */
int ComfDialogHaltTillButton()
{
	int events_handled;
	event_t event;
	int map_raised;
	long t;


        /* Change global halt_comf_dialog to True. */
        halt_comf_dialog = True;
        
   
        /* ******************************************************* */
        /* Manage events only for comf_dialog. */
        
        map_raised = 0;

	/* Clear all events. */
        OSWPurgeAllEvents();
        
        while(1)
        {
            /* Update timming. */
            t = MilliTime();

            if(t < cur_millitime) 
            {
                /* Need to reset timmings. */
                ResetGlobalTimming();
            } 
            cur_millitime = t;


            /* Must keep dialog on top for 25 retries. */
            if(map_raised < 25)
            {
                OSWMapRaised(comf_dialog.toplevel);
                map_raised++;
            }


	    /* Manage events, LOCALLY. */
	    while(OSWEventsPending() > 0)
	    {
                events_handled = 0;
                OSWWaitNextEvent(&event);


                /* ClientMessage must be handled here. */
                if((event.type == ClientMessage) &&
                   OSWIsEventDestroyWindow(comf_dialog.toplevel, &event)
                )
                {
		    /* Interprite the destroy window event as a no. */
                    ComfDialogNoPBCB(NULL);

                    events_handled++;

                    if(!halt_comf_dialog)
                        break;
                    else
                        continue;
                }
                if(!halt_comf_dialog)
                    break;

                /* Give event for confirmation dialog for processing. */
                ComfDialogManage(&event);

                if(!halt_comf_dialog)
                    break;
	    }
            if(!halt_comf_dialog)
                break;


            /* Manage timmers. */
            TimmerManageAll();

  
#ifdef USE_IMLIB
            /* Image viewers repeat callback. */
            ImgViewerRepeatCB();
#endif

#ifdef USE_IMLIB
            /* Image browsers repeat callback. */
            IMBRRepeatCB();
#endif
            
            /* Call widget manager. */
            WidgetManage(NULL);

	    /* DO NOT CALL ENDEAVROUR's EVENT MANAGER. */


            /* Redraws. */
            ManagePadlockDraw();

            /* Sanity check. */
            SanityCheck();
            
            
            /* Sleep. */
	    usleep(16000);
        }


        /* Return the value of the last button pressed. */
        return(last_pbtn_pressed);
}



/* ************************************************************************
 *
 *          Confirmation: Pointer Movement on Endeavour browser
 *
 *	Map the confirmation dialog to query `confirm movement'.
 */
void ComfDialogMapQueryMovement(
	void *tar_ptr,   
        int tar_ptr_type,
        void *src_ptr,   
        int src_ptr_type,
        int tar_drag_type,
        int src_drag_type,
        int tar_item_num,
        int op_code
)
{
	/* Error checks. */
	if(!IDC())
	    return;

	UnfocusAllTopLvlWins();


	/* Reset buttons (this is apparently needed). */
	comf_dialog.yes_btn.state = PBTN_UNARMED;
        comf_dialog.no_btn.state = PBTN_UNARMED;
        comf_dialog.all_btn.state = PBTN_UNARMED;


	/* Set query type and operation code. */
	comf_dialog.query_type = CDIALOG_QUERY_CONFIRM_MOTION;
        comf_dialog.op_code = op_code;

	/* Set up need to know values. */
	comf_dialog.tar_ptr = tar_ptr;
	comf_dialog.tar_ptr_type = tar_ptr_type;
	comf_dialog.src_ptr = src_ptr;
        comf_dialog.src_ptr_type = src_ptr_type;
        comf_dialog.tar_drag_type = tar_drag_type;
        comf_dialog.src_drag_type = src_drag_type;
	comf_dialog.tar_item_num = tar_item_num;


	/* Map and draw. */
        PBtnUnmap(&comf_dialog.all_btn);

        comf_dialog.map_state = 0;
        ComfDialogDraw();
        comf_dialog.is_in_focus = 1;
        comf_dialog.visibility_state = VisibilityUnobscured;

	/* Need to purge all events after mapping. */
/*
        OSWPurgeAllEvents();
 */

	return;
}


/*
 *	Map the confirmation dialog and query for 'overwrite'.
 *
 *	The function ComfDialogHaltTillButton() is called in this function
 *	to isolate the management to only the confirmation dialog.
 */
int ComfDialogMapQueryOverWrite(
	char *tar_path,
	char *src_path,
	int op_code
)
{
	char tmp_path1[PATH_MAX + NAME_MAX];
	char tmp_path2[PATH_MAX + NAME_MAX];


        /* Error checks. */
        if(!IDC())
            return(NO_PBTN_NUM);

        UnfocusAllTopLvlWins();


        /* Reset buttons (this is apparently needed). */
        comf_dialog.yes_btn.state = PBTN_UNARMED;
        comf_dialog.no_btn.state = PBTN_UNARMED;
        comf_dialog.all_btn.state = PBTN_UNARMED;


        /* Set query type and operation code. */
        comf_dialog.query_type = CDIALOG_QUERY_OVERWRITE;
        comf_dialog.op_code = op_code;

	/* Set up need to know values. */
	free(comf_dialog.tar_path);
	comf_dialog.tar_path = NULL;
	if(tar_path != NULL)
	    comf_dialog.tar_path = StringCopyAlloc(tar_path);

        free(comf_dialog.src_path);
        comf_dialog.src_path = NULL;
        if(src_path != NULL)
            comf_dialog.src_path = StringCopyAlloc(src_path);


	/* If the target path is a directory, then must post pend
	 * file name of source to it.
	 */
	if((comf_dialog.src_path != NULL) &&
           (comf_dialog.tar_path != NULL)
	)
	{
	    if(ISPATHDIR(comf_dialog.tar_path))
	    {
		strncpy(tmp_path1,
		    comf_dialog.src_path,
		    PATH_MAX + NAME_MAX
		);
		StripAbsolutePath(tmp_path1);
		strncpy(tmp_path2,
                    comf_dialog.tar_path,
                    PATH_MAX + NAME_MAX
                );

		free(comf_dialog.tar_path);
		comf_dialog.tar_path = NULL;
		comf_dialog.tar_path = StringCopyAlloc(
		    PrefixPaths(tmp_path2, tmp_path1)
		);
	    }
	}


	/* *********************************************************** */

        /* Map and draw. */
	PBtnMap(&comf_dialog.all_btn);

	comf_dialog.map_state = 0;
	ComfDialogDraw();
	comf_dialog.is_in_focus = 1;
	comf_dialog.visibility_state = VisibilityUnobscured;


	/* Enter halted event management for overwrite query. */
	ComfDialogHaltTillButton();


	/* Return a code for yes, no or all. */
	return(last_pbtn_pressed);
}



/* **********************************************************************
 *
 *                 Confirmation: Recycle Bin Operations
 *
 *      Map the confirmation dialog and query for purging of an item
 *	in the recycle bin (not purge all).
 *
 *      The function ComfDialogHaltTillButton() is called in this function
 *      to isolate the management to only the confirmation dialog.
 */
int ComfDialogMapQueryPurge(  
	recbin_win_struct *rb,
	char *src_path
)
{
        /* Reset buttons (this is apparently needed). */
        comf_dialog.yes_btn.state = PBTN_UNARMED;
        comf_dialog.no_btn.state = PBTN_UNARMED;
        comf_dialog.all_btn.state = PBTN_UNARMED;



	/* Error checks. */
        if(!IDC())
            return(NO_PBTN_NUM);
	if(rb == NULL)
	    return(NO_PBTN_NUM);
	if(src_path == NULL)
	    return(NO_PBTN_NUM);


        UnfocusAllTopLvlWins();


        /* Set query type and operation code. */
        comf_dialog.query_type = CDIALOG_QUERY_RECBIN_PURGE;
	comf_dialog.op_code = OP_CODE_RECBIN_PURGE;


        free(comf_dialog.src_path);
        comf_dialog.src_path = NULL;
        if(src_path != NULL)
            comf_dialog.src_path = StringCopyAlloc(src_path);


        /* *********************************************************** */

        /* Map and draw. */
        PBtnMap(&comf_dialog.all_btn);
                  
        comf_dialog.map_state = 0;
        ComfDialogDraw();
        comf_dialog.is_in_focus = 1;
        comf_dialog.visibility_state = VisibilityUnobscured;


        /* Enter halted event management for overwrite query. */
        ComfDialogHaltTillButton();

        /* Return a code for yes, no or all. */
        return(last_pbtn_pressed);
}



int ComfDialogMapQueryPurgeAll(
                recbin_win_struct *rb 
)
{
        /* Reset buttons (this is apparently needed). */
        comf_dialog.yes_btn.state = PBTN_UNARMED;
        comf_dialog.no_btn.state = PBTN_UNARMED;
        comf_dialog.all_btn.state = PBTN_UNARMED;
        
        
        /* Error checks. */
        if(!IDC())
            return(NO_PBTN_NUM);
        if(rb == NULL)
            return(NO_PBTN_NUM);

        UnfocusAllTopLvlWins();


        /* Set query type and operation code. */
        comf_dialog.query_type = CDIALOG_QUERY_RECBIN_PURGE_ALL;
        comf_dialog.op_code = OP_CODE_RECBIN_PURGE_ALL;


        /* *********************************************************** */

        /* Map and draw. */
        PBtnUnmap(&comf_dialog.all_btn);
 
        comf_dialog.map_state = 0;
        ComfDialogDraw();
        comf_dialog.is_in_focus = 1;
        comf_dialog.visibility_state = VisibilityUnobscured;
        
 
        /* Enter halted event management for overwrite query. */
        ComfDialogHaltTillButton();


        /* Return a code for yes, no or all. */
        return(last_pbtn_pressed);
}



/* ************************************************************************
 *
 *                      Confirmation: Device Mounting
 */
int ComfDialogMapQueryMount(
	char *src_path
)
{
        /* Reset buttons (this is apparently needed). */   
        comf_dialog.yes_btn.state = PBTN_UNARMED;
        comf_dialog.no_btn.state = PBTN_UNARMED;
        comf_dialog.all_btn.state = PBTN_UNARMED;
        

        /* Error checks. */
        if(!IDC())
            return(NO_PBTN_NUM);
        if(src_path == NULL)
            return(NO_PBTN_NUM);

        UnfocusAllTopLvlWins();


        /* Set query type and operation code. */
        comf_dialog.query_type = CDIALOG_QUERY_DEVICE_MOUNT;
        comf_dialog.op_code = OP_CODE_NONE;


	free(comf_dialog.tar_path);
	comf_dialog.tar_path = NULL;
        free(comf_dialog.src_path);
        comf_dialog.src_path = NULL;

	/* Set source path. */
	comf_dialog.src_path = StringCopyAlloc(src_path);


        /* *********************************************************** */

        /* Map and draw. */
        PBtnMap(&comf_dialog.all_btn);
        
        comf_dialog.map_state = 0;
        ComfDialogDraw();
        comf_dialog.is_in_focus = 1;
        comf_dialog.visibility_state = VisibilityUnobscured;

        /* Enter halted event management for overwrite query. */
        ComfDialogHaltTillButton();


        /* Return a code for yes, no or all. */
        return(last_pbtn_pressed);
}



int ComfDialogMapQueryUnmount(
	char *src_path
)
{
        /* Reset buttons (this is apparently needed). */
        comf_dialog.yes_btn.state = PBTN_UNARMED;
        comf_dialog.no_btn.state = PBTN_UNARMED;
        comf_dialog.all_btn.state = PBTN_UNARMED;


        /* Error checks. */
        if(!IDC())
            return(NO_PBTN_NUM);
        if(src_path == NULL)
            return(NO_PBTN_NUM);

        UnfocusAllTopLvlWins();


        /* Set query type and operation code. */
        comf_dialog.query_type = CDIALOG_QUERY_DEVICE_UNMOUNT;
        comf_dialog.op_code = OP_CODE_NONE;


        free(comf_dialog.tar_path);
        comf_dialog.tar_path = NULL;
        free(comf_dialog.src_path);
        comf_dialog.src_path = NULL;
 
        /* Set source path. */
        comf_dialog.src_path = StringCopyAlloc(src_path);


        /* *********************************************************** */

        /* Map and draw. */
        PBtnMap(&comf_dialog.all_btn);
   
        comf_dialog.map_state = 0;
        ComfDialogDraw();
        comf_dialog.is_in_focus = 1;
        comf_dialog.visibility_state = VisibilityUnobscured;


        /* Enter halted event management for overwrite query. */
        ComfDialogHaltTillButton();


        /* Return a code for yes, no or all. */
        return(last_pbtn_pressed);

	return(0);
}



/* ************************************************************************
 *
 *
 */
void ComfDialogResetQuery()
{
        comf_dialog.query_type = CDIALOG_QUERY_NONE;

        comf_dialog.tar_ptr = NULL;
        comf_dialog.tar_ptr_type = POINTER_TYPE_UNDEFINED;
        comf_dialog.src_ptr = NULL;
        comf_dialog.src_ptr_type = POINTER_TYPE_UNDEFINED;
        comf_dialog.tar_drag_type = DRAG_TYPE_NONE;
        comf_dialog.src_drag_type = DRAG_TYPE_NONE;
        comf_dialog.tar_item_num = 0;
        comf_dialog.op_code = OP_CODE_NONE;

        free(comf_dialog.tar_path);   
        comf_dialog.tar_path = NULL;
 
        free(comf_dialog.src_path);
        comf_dialog.src_path = NULL;


	return;
}


void ComfDialogUnmap()
{
	/* Reset confirmation dialog values. */
	ComfDialogResetQuery();


	/* Unmap confirmation dialog. */
	OSWUnmapWindow(comf_dialog.toplevel);
	comf_dialog.map_state = 0;
	comf_dialog.visibility_state = VisibilityFullyObscured;
	comf_dialog.is_in_focus = 0;

	/* This is apparently needed. */
/*
	OSWPurgeAllEvents();
 */

	return;
}



void ComfDialogDestroy()
{

	comf_dialog.map_state = 0;
	comf_dialog.visibility_state = VisibilityFullyObscured;
	comf_dialog.x = 0;
        comf_dialog.y = 0;
        comf_dialog.width = 0;
        comf_dialog.height = 0;
        comf_dialog.is_in_focus = 0;
        comf_dialog.disabled = False;

	comf_dialog.query_type = CDIALOG_QUERY_NONE;
	comf_dialog.op_code = OP_CODE_NONE;

	free(comf_dialog.tar_path);
	comf_dialog.tar_path = NULL;

        free(comf_dialog.src_path);
        comf_dialog.src_path = NULL;


        if(IDC())
	{
            PBtnDestroy(&comf_dialog.yes_btn);
            PBtnDestroy(&comf_dialog.no_btn);
            PBtnDestroy(&comf_dialog.all_btn);

	    OSWDestroyPixmap(&comf_dialog.toplevel_buf);
	    OSWDestroyWindow(&comf_dialog.toplevel);
	}


	return;
}
