/*

                Configurable Default Compile-time Values


	You may edit the values #defined here as needed, do not change
	any value you are unsure of.

 */

#ifndef CONFIG_H
#define CONFIG_H



/*
 *	Default directories and files:
 */
#define LOCAL_ENDEAVOUR_DIR	".endeavour"	/* HOME env will be prefixed. */
#define DEF_RCFILE		"endeavourrc"


/*
 *	Program output spool file:
 *
 *	Programs runned by this program will have their stdout and stderr
 *	redirected to these files.
 *
 *	The pid of this process (not the forked child processed used to
 *	issue the command if applicateable) is postpended to the name.
 *
 *	These files will be placed in the LOCAL_ENDEAVOUR_DIR directory.
 */
#define DEF_PROG_OUTPUT		"prog_output"
#define DEF_LIST_OUTPUT		"list_output"


/*
 *	Program communications FIFO pipe:
 *
 *	For multiple endeavour processes (runned by the same user only)
 *	to communicate with one another.
 *
 *	This FIFO pipe will be placed in the LOCAL_ENDEAVOUR_DIR
 *	directory.
 */
#define DEF_ENDEAVOUR_COMPIPE	"compipe"


/*
 *	Default write protect startup:
 *
 *	0 means no.
 *	1 means yes.
 *
 *	Note: Configuration file values at runtime overrides this.
 */
#define DEF_WRITE_PROTECT	1


/*
 *	Default conferm pointer motion:
 *
 *      0 means no.
 *      1 means yes.
 *
 *      Note: Configuration file values at runtime overrides this.
 */
#define DEF_COMF_POINTER_MOTION_COPY	1
#define DEF_COMF_POINTER_MOTION_MOVE	1


/*
 *      Default conferm overwrite:
 * 
 *      0 means no.
 *      1 means yes.
 * 
 *      Note: Configuration file values at runtime overrides this.
 */
#define DEF_COMF_OVERWRITE	1

/*
 *      Default conferm recycle object:
 *
 *      0 means no.
 *      1 means yes.
 *
 *      Note: Configuration file values at runtime overrides this.
 */
#define DEF_COMF_RECYCLE	1


/*
 *      Default conferm recycle bin purge:
 *
 *      0 means no.
 *      1 means yes.
 *      
 *      Note: Configuration file values at runtime overrides this.
 */
#define DEF_COMF_PURGE		1
#define DEF_COMF_PURGE_ALL	1


/*
 *      Default conferm unmount removeable devices on exit:
 *
 *      0 means no.
 *      1 means yes.
 *
 *	Remember that this applies only to devices that are set
 *	Unmountable = Yes.
 *      Note: Configuration file values at runtime overrides this.
 */
#define DEF_COMF_UNMOUNT_ONEXIT		1


/*
 *	Default device find output command:
 *
 *	This is the command to output a POSIX standard list of devices
 *	and their statistics to stdout WITHOUT the redirect.
 *
 *	On most platforms, it is "df -P"
 *
 *      Note: Configuration file values at runtime overrides this.
 */
#define SYSTEM_CMD_DEVICE_FIND	"df -P"


/*
 *	Default mount device timeout:
 *
 *	In seconds.
 */
#define DEF_MOUNT_TIMEOUT	3

/*
 *	Default maximum number of external program executions per
 *	one 'movement' by either keystroke or pointer.
 *
 *	This is so that if you selected several files and it will
 *	execute too many programs, this will prevent your system from
 *	being hogged down.
 */
#define DEF_EXECUTION_PURST_MAX		5


/*
 *	GUI Interface Preferances:
 */
#define DRAG_TOLORANCE          5	/* In delta pixels FROM epsilon. */


/*
 *      Select auto-scroll shift delay, in milliseconds.
 *      When an item is selected near the edge of the list, the list
 *      is scrolled after a short delay of this many milliseconds.
 */
#define SELECT_SHIFT_DELAY      250


/* 
 *      If a name pointer is NULL or "", then use this string instead
 *      as its name.
 */
#define NO_NAME_STR             "(null)"


/*
 *	Default command lines:
 *
 *	These programs should be standard on your platform for performing
 *	most file operations.
 *
 *	"%t" will be replaced with the target path.
 *	"%s" will be replaced with the source path.
 *	"%a" will be replaced with the argument (for chmod and chown).
 *
 *	Note: Configuration file values at runtime override these.
 */
#define SYSTEM_CMD_MOVE_FILE	"mv \"%s\" \"%t\""
#define SYSTEM_CMD_MOVE_DIR	"mv \"%s\" \"%t\""

#define SYSTEM_CMD_COPY_FILE	"cp \"%s\" \"%t\""
#define SYSTEM_CMD_COPY_DIR	"cp -aR \"%s\" \"%t\""

#define SYSTEM_CMD_RENAME_FILE	"mv \"%s\" \"%t\""
#define SYSTEM_CMD_RENAME_DIR	"mv \"%s\" \"%t\""

#define SYSTEM_CMD_CHMOD_FILE	"chmod %a \"%t\""	/* Obsolete. */
#define SYSTEM_CMD_CHMOD_DIR	"chmod %a \"%t\""

#define SYSTEM_CMD_CHOWN_FILE	"chown %a \"%t\""	/* Obsolete. */
#define SYSTEM_CMD_CHOWN_DIR	"chown %a \"%t\""

#define SYSTEM_CMD_TOUCH_FILE	"touch -c \"%t\""
#define SYSTEM_CMD_TOUCH_DIR	"touch -c \"%t\""

#define SYSTEM_CMD_TEXT_EDITOR	"nedit \"%t\""


/*
 *	Event skip intervals, in milliseconds.
 *
 *	If an event in question occures and another occures between
 *	this interval, it is then discarded.
 *
 *	This allows the program to have a breather between events
 *	without sleeping.
 */
#define DEF_MOTIONNOTIFY_INT		75
#define DEF_CONFIGURENOTIFY_INT		75




#endif /* CONFIG_H */
