/*
                          Pointer Drag Management

	Functions:

	void DragDoAllWindowsActive()
	void DragCleanup()

	int DragManageButtonPress(event_t *event)
	int DragManageButtonRelease(event_t *event)
	int DragManageMotion(event_t *event)

	int DragManage(event_t *event)

	---

 */

#include "endeavour.h"


/*
 *      Makes all interested windows grab for pointer for a drag, does
 *      NOT set up drag records.
 */
void DragDoAllWindowsActive()
{
        int x;
                
        if(!IDC())
            return;
                
                
#ifdef X_H
        /* Make active pointer grabs on needed windows. */
         
        /* Endeavour windows. */
        for(x = 0; x < total_endeavour_wins; x++)
        {
            if(endeavour_win[x] == NULL) continue;
            if(endeavour_win[x]->toplevel == 0) continue;

            OSWGrabPointer(
                endeavour_win[x]->dir_win,
                False,
                ButtonPressMask | ButtonReleaseMask |
                    PointerMotionMask,
                GrabModeAsync, GrabModeAsync,
                None,
                widget_global.drag_file_wcr->cursor
            );

            OSWGrabPointer(
                endeavour_win[x]->file_win,
                False,
                ButtonPressMask | ButtonReleaseMask |
                    PointerMotionMask,
                GrabModeAsync, GrabModeAsync,
                None,
                widget_global.drag_file_wcr->cursor 
            );
        }       

        /* Recycle bin windows. */
        for(x = 0; x < total_recbin_wins; x++)
        {
            if(recbin_win[x] == NULL) continue;
            if(recbin_win[x]->toplevel == 0) continue;

            OSWGrabPointer(
                recbin_win[x]->list,
                False,
                ButtonPressMask | ButtonReleaseMask |
                    PointerMotionMask,
                GrabModeAsync, GrabModeAsync,
                None,
                widget_global.drag_file_wcr->cursor 
            );
        }
              
#ifdef USE_IMLIB
        /* Image browser windows. */
        for(x = 0; x < total_img_browsers; x++)
        {
            if(!IMBRIsInit(x)) continue;

            OSWGrabPointer(
                img_browser[x]->list,
                False,
                ButtonPressMask | ButtonReleaseMask |
                    PointerMotionMask,
                GrabModeAsync, GrabModeAsync,
                None,
                widget_global.drag_file_wcr->cursor
            );
        }
#endif /* USE_IMLIB */
            

#ifdef USE_IMLIB
        /* Image viewer windows. */ 
        for(x = 0; x < total_img_viewers; x++)
        {
            if(img_viewer[x] == NULL) continue;
            if(img_viewer[x]->toplevel == 0) continue;
            if(img_viewer[x]->view_screen == 0) continue;

            OSWGrabPointer(
                img_viewer[x]->view_screen,
                False,
                ButtonPressMask | ButtonReleaseMask |
                    PointerMotionMask,
                GrabModeAsync, GrabModeAsync,
                None,
                widget_global.drag_file_wcr->cursor
            );
        }
#endif /* USE_IMLIB */
                
#endif /* X_H */
                
        return;
                
}



/*          
 *      Resets drag records and ungrabs pointer.
 */
void DragCleanup()
{
        drag_record[0].button1_state = False;
        drag_record[0].button2_state = False;  
        drag_record[0].button3_state = False;
            
/* Do not change key states after drag, user may still be holding one down.
        drag_record[0].shift_key_state = False;
        drag_record[0].alt_key_state = False;
        drag_record[0].ctrl_key_state = False;
*/
        drag_record[0].start_x = 0;
        drag_record[0].start_y = 0;
                    
        drag_record[0].state = False;
        drag_record[0].type = DRAG_TYPE_NONE;
        drag_record[0].origin_ptr_type = POINTER_TYPE_UNDEFINED;
        drag_record[0].origin_ptr = NULL;
        drag_record[0].heading_col_flag = 0;

	OSWUngrabPointer();              
         
        return;
}           



/* ******************************************************************* */

/*
 *	Drag ButtonPress management.
 */
int DragManageButtonPress(event_t *event)
{
        int events_handled = 0;


        /* Error checks. */
        if(event == NULL)
	    return(events_handled);


	/* Record ButtonPress states and coordinates. */
        if(event->xbutton.button == Button1)
            drag_record[0].button1_state = True;
        else if(event->xbutton.button == Button2)
            drag_record[0].button2_state = True;
        else if(event->xbutton.button == Button3)
            drag_record[0].button3_state = True;

        drag_record[0].start_x = event->xbutton.x;
        drag_record[0].start_y = event->xbutton.y;


	/* Do not count event for this. */

	return(events_handled);
}



/*
 *      Drag ButtonRelease management.
 *
 *	Operations for drags are performed when a button is
 *	released.
 */
int DragManageButtonRelease(event_t *event)
{
        int i, x;
        win_t w;
	win_attr_t wattr;
        int events_handled = 0;


        /* Error checks. */
        if(event == NULL) return(events_handled);


        /* Record ButtonRelease states. */
        if(event->xbutton.button == Button1)
            drag_record[0].button1_state = False;
        else if(event->xbutton.button == Button2)
            drag_record[0].button2_state = False;
        else if(event->xbutton.button == Button3)
            drag_record[0].button3_state = False;


        /* Get window the event occured on. */
        w = event->xany.window;


	/* ************************************************************ */
	/* Check if drag is active. */
	if(drag_record[0].state)
	{
	    /* ******************************************************** */
	    /* Endeavour browser. */
            for(i = 0; i < (int)total_endeavour_wins; i++)
            {
                if(!EDWIsInit(i)) continue;

		/* Directory window. */
                if(endeavour_win[i]->dir_win == w)
                {
                    OSWGetWindowAttributes(endeavour_win[i]->dir_win, &wattr);

                    /* Get dir item button was released on. */
                    x = EDW_GETDIRNUMBYPOS(
                        endeavour_win[i],
                        event->xbutton.x,
                        event->xbutton.y
                    );

                    /* Did we get a valid position? */
                    if((x > -1) && (x < (int)endeavour_win[i]->total_dirs))
                    {   
                        EDWSETALLCURSORS(endeavour_win[i],
			    EDW_CURSORS_CODE_WORKING);

                        EDWProcessMovement(
                            endeavour_win[i],		/* Target. */
                            POINTER_TYPE_EDW,
                            drag_record[0].origin_ptr,  /* Source. */
                            drag_record[0].origin_ptr_type,

                            DRAG_TYPE_DIR,              /* Target type. */
                            drag_record[0].type,        /* Source type. */
                            x,                          /* Target number. */
                            (drag_record[0].ctrl_key_state) ? 
                                OP_CODE_COPY : OP_CODE_MOVE,
                            False                       /* Comfermed? */
                        );
                        
                        EDWSETALLCURSORS(endeavour_win[i],
                            EDW_CURSORS_CODE_NORMAL);
                    }       

                    events_handled++;
		    break;
		}
		/* File window. */
		else if(endeavour_win[i]->file_win == w)
                {
                    OSWGetWindowAttributes(endeavour_win[i]->file_win, &wattr);

                    /*   Get file item button was released on.
                     *   (This information will not be used.)
                     */
                    x = EDW_GETFILENUMBYPOS(
                        endeavour_win[i],
                        event->xbutton.x,
                        event->xbutton.y   
                    );

                    EDWSETALLCURSORS(endeavour_win[i],
                        EDW_CURSORS_CODE_WORKING);

                    /* We don't care about the file item number x. */
                    EDWProcessMovement(
                            endeavour_win[i],           /* Target. */
                            POINTER_TYPE_EDW,
                            drag_record[0].origin_ptr,  /* Source. */
                            drag_record[0].origin_ptr_type,

                            DRAG_TYPE_FILE,             /* Target type. */
                            drag_record[0].type,        /* Source type. */
                            x,                          /* Target number. */
                            (drag_record[0].ctrl_key_state) ?
                                OP_CODE_COPY : OP_CODE_MOVE,
                            False                       /* Comfermed? */
                    );

                    EDWSETALLCURSORS(endeavour_win[i],
                        EDW_CURSORS_CODE_NORMAL);

		    events_handled++;
                    break;
		}
	    }

            /* ******************************************************** */
            /* Recycle bin. */
            for(i = 0; i < (int)total_recbin_wins; i++)
            {
                if(recbin_win[i] == NULL) continue;

		/* List. */
                if(recbin_win[i]->list == w)
                {
                    RECBINSETALLCURSORS(recbin_win[i],
                        EDW_CURSORS_CODE_WORKING);

                    RecBinProcessMovement(
                        recbin_win[i], POINTER_TYPE_RB, /* Target. */
                        drag_record[0].origin_ptr,      /* Source. */
                        drag_record[0].origin_ptr_type,
                        DRAG_TYPE_RECYCLE_BIN,
                        drag_record[0].type,
                        -1,                     /* None. */
                        OP_CODE_RECYCLE,
                        False
                    );

                    RECBINSETALLCURSORS(recbin_win[i],
                        EDW_CURSORS_CODE_NORMAL);

		    events_handled++;
		    break;
		}
	    }

#ifdef USE_IMLIB
            /* ******************************************************** */
            /* Image browser. */
            for(i = 0; i < (int)total_img_browsers; i++)
            {
                if(!IMBRIsInit(i)) continue;

                /* Thumbs list. */
                if(img_browser[i]->list == w)
                {
		    x = IMBR_GETTHUMBFROMPOS(img_browser[i],
			event->xbutton.x,
			event->xbutton.y
		    );

		    IMBRProcessMovement(
                        img_browser[i], POINTER_TYPE_IB, /* Target. */
                        drag_record[0].origin_ptr,       /* Source. */
                        drag_record[0].origin_ptr_type,
                        DRAG_TYPE_THUMBS_LIST,		/* Target. */
                        drag_record[0].type,		/* Source. */
                        x,
                        (drag_record[0].ctrl_key_state) ?
                            OP_CODE_COPY : OP_CODE_MOVE,
                        False
		    );

                    events_handled++;
                    break;
                }
	    }
#endif /* USE_IMLIB */




	    /* Always clean up drag after ButtonRelease. */
	    DragCleanup();
	}

        return(events_handled);
}



/*
 *      Drag Motion management.
 *
 *	This is where drag is started when the button is pressed
 *	and the pointer is moved from its starting position with
 *	tolorance.
 */
int DragManageMotion(event_t *event)
{
	int i;
	win_t w;
        int events_handled = 0;


	/* Error checks. */
	if(event == NULL) return(events_handled);

	/* Get window the event occured on. */
	w = event->xany.window;

	/* Check if drag state was active. */
        if(!drag_record[0].state)
        {
            /*   Check drag start when motion brings us out of
	     *   DRAG_TOLORANCE of drag start coordinates.
             */
            if(drag_record[0].button1_state &&
               (((drag_record[0].start_x - DRAG_TOLORANCE) >=
                 event->xmotion.x) ||
                ((drag_record[0].start_x + DRAG_TOLORANCE) <=
                 event->xmotion.x) ||
                ((drag_record[0].start_y - DRAG_TOLORANCE) >=
                 event->xmotion.y) ||
                ((drag_record[0].start_y + DRAG_TOLORANCE) <=
                 event->xmotion.y)
               )
	    )
	    {
		/* Drag starts, check which window it occured on
                 * and set up drag values accordingly.
		 */

		/* Endeavour browser. */
		for(i = 0; i < (int)total_endeavour_wins; i++)
		{
		    if(!EDWIsInit(i)) continue;

		    if(endeavour_win[i]->dir_win == w)
		    {
			drag_record[0].state = True;
                        drag_record[0].type = DRAG_TYPE_DIR;
                        drag_record[0].origin_ptr_type = POINTER_TYPE_EDW;
                        drag_record[0].origin_ptr = endeavour_win[i];

			break;
		    }
                    else if(endeavour_win[i]->file_win == w)   
                    {
                        drag_record[0].state = True;
                        drag_record[0].type = DRAG_TYPE_FILE;
                        drag_record[0].origin_ptr_type = POINTER_TYPE_EDW;
                        drag_record[0].origin_ptr = endeavour_win[i];

                        break;
                    }
		}

                /* Recycle bins. */
                for(i = 0; i < (int)total_recbin_wins; i++)
                {
                    if(recbin_win[i] == NULL) continue;

                    if(recbin_win[i]->list == w)
                    {
                        drag_record[0].state = True;
                        drag_record[0].type = DRAG_TYPE_RECYCLE_BIN;
                        drag_record[0].origin_ptr_type = POINTER_TYPE_RB;
                        drag_record[0].origin_ptr = recbin_win[i];

                        break;
                    }
		}

#ifdef USE_IMLIB
                /* Image browsers. */
                for(i = 0; i < (int)total_img_browsers; i++)
                {
                    if(!IMBRIsInit(i)) continue;

                    if(img_browser[i]->list == w)
                    {
                        drag_record[0].state = True;
                        drag_record[0].type = DRAG_TYPE_THUMBS_LIST;
                        drag_record[0].origin_ptr_type = POINTER_TYPE_IB;
                        drag_record[0].origin_ptr = img_browser[i];

                        break;
                    }
                }
#endif



		/* Set all drag windows active. */
		if(drag_record[0].state)
                    DragDoAllWindowsActive();

		events_handled++;
	    }
	}


        return(events_handled);
}



/* ******************************************************************* */

/*
 *	Manages drag.
 */
int DragManage(event_t *event)
{

	int events_handled = 0;


	/* Error checks. */
	if(!IDC() || (event == NULL))
	    return(events_handled);


        /* *********************************************************** */
        switch(event->type)
        {
          /* ********************************************************* */
          case KeyPress:
            /* Shift key. */
            if((event->xkey.keycode == osw_keycode.shift_left) ||
               (event->xkey.keycode == osw_keycode.shift_right)
            )                 
                drag_record[0].shift_key_state = True;
            /* Alt key. */
            else if((event->xkey.keycode == osw_keycode.alt_left) ||
                    (event->xkey.keycode == osw_keycode.alt_right)
            )
	    drag_record[0].alt_key_state = True;
            /* Ctrl key. */
            else if((event->xkey.keycode == osw_keycode.ctrl_left) ||
                    (event->xkey.keycode == osw_keycode.ctrl_right)
            )
                drag_record[0].ctrl_key_state = True;

	    break;

          /* ********************************************************* */
          case KeyRelease:
            /* Shift key. */
            if((event->xkey.keycode == osw_keycode.shift_left) ||
               (event->xkey.keycode == osw_keycode.shift_right)
            )
                drag_record[0].shift_key_state = False;
            /* Alt key. */
            else if((event->xkey.keycode == osw_keycode.alt_left) ||
                    (event->xkey.keycode == osw_keycode.alt_right)
            )
                drag_record[0].alt_key_state = False;
            /* Ctrl key. */
            else if((event->xkey.keycode == osw_keycode.ctrl_left) ||
                    (event->xkey.keycode == osw_keycode.ctrl_right)
            )
                drag_record[0].ctrl_key_state = False;

	    break;

          /* ********************************************************* */
          case ButtonPress:
	    /* Manage drag for ButtonPress. */
	    DragManageButtonPress(event);
	    break;

          /* ********************************************************* */
          case ButtonRelease:
            /* Manage drag for ButtonRelease. */
            DragManageButtonRelease(event);
            break;

          /* ********************************************************* */
          case MotionNotify:
            /* Manage drag for Motion. */
            DragManageMotion(event);
            break;
	}



	return(events_handled);
}
