/*
              Endeavour Browser Window: Memory allocating

	Functions:

	int EDWIsAllocated(int edw_num)
	int EDWIsInit(int edw_num)
	int EDWAllocate()
	void EDWReclaimMemory()
	void EDWDelete(int edw_num)
	void EDWDeleteAll()

 */

#include "endeavour.h"
#include "edwdefaults.h"

#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))


int EDWIsAllocated(int edw_num)
{
	if( (endeavour_win == NULL) ||
            (edw_num < 0) ||
            (edw_num >= (int)total_endeavour_wins)
	)
	{
	    return(0);
	}
	else if(endeavour_win[edw_num] == NULL)
	{
	    return(0);
	}
	else
	{
	    return(1);
	}
}



int EDWIsInit(int edw_num)
{
	if(!EDWIsAllocated(edw_num))
	{
	    return(0);
	}
	else if(endeavour_win[edw_num]->toplevel == 0)
	{
	    return(0);
	}
	else
	{
	    return(1);
	}
}



int EDWAllocate()
{
	int i, n;


	/* Sanitize global total. */
	if((int)total_endeavour_wins < 0)
	    total_endeavour_wins = 0;


	/* Check for available entry. */
	for(i = 0; i < (int)total_endeavour_wins; i++)
	{
	    if(endeavour_win[i] == NULL)
		break;
	}
	if(i < (int)total_endeavour_wins)
	{
	    /* Found an unallocated pointer. */
	    n = i;
	}
	else
	{
	    /* Need to allocate more pointers. */
            n = total_endeavour_wins;
            total_endeavour_wins = n + 1;

            endeavour_win = (endeavour_win_struct **)realloc(
                endeavour_win,
		total_endeavour_wins * sizeof(endeavour_win_struct *)
            );
            if(endeavour_win == NULL)
            {
                total_endeavour_wins = 0;
                return(-1);
            }
	}


	/* Allocate a new edw structure for that pointer. */
        endeavour_win[n] = (endeavour_win_struct *)calloc(
	    1, sizeof(endeavour_win_struct)
	);
        if(endeavour_win[n] == NULL)
        {
            total_endeavour_wins--;
            return(-1);
        }

	/* Return new endeavour window index number. */
	return(n);
}



void EDWReclaimMemory()
{
        int i, highest_edw_num;


	/* Skip if nothing is allocated. */
	if( (endeavour_win == NULL) ||
	    ((int)total_endeavour_wins <= 0)
	)
	    return;


	/* ******************************************************** */

        /* Reclaim memory. */
        highest_edw_num = -1;
        for(i = 0; i < (int)total_endeavour_wins; i++)
        {
            if(endeavour_win[i] == NULL)
                continue;
            highest_edw_num = i;
        }
  
        if(highest_edw_num < ((int)total_endeavour_wins - 1))
        {
	    /* Adjust global variable total_endeavour_wins. */
	    total_endeavour_wins = highest_edw_num + 1;

	    /* Reallocate or free pointers. */
	    if(total_endeavour_wins <= 0)
	    {
		/* total_endeavour_wins has dropped to 0, free all pointers. */
		free(endeavour_win);
		endeavour_win = NULL;
		total_endeavour_wins = 0;
	    }
	    else
	    {
		endeavour_win = (endeavour_win_struct **)realloc(
		    endeavour_win,
		    total_endeavour_wins * sizeof(endeavour_win_struct *)
		);
		if(endeavour_win == NULL)
		{
		    total_endeavour_wins = 0;
		}
	    }
        }


	return;
}


/*
 *	Deallocates all substructures, windows, widgets, etc...
 *	plus the structure itself!
 */
void EDWDelete(int edw_num)
{
	/* Error checks. */
	if(!EDWIsAllocated(edw_num))
	    return;


	/* Deallocate all substructures, windows, widgets, etc... */
	EDWDestroy(endeavour_win[edw_num]);


	/* Deallocate the structure itself. */
	free(endeavour_win[edw_num]);
	endeavour_win[edw_num] = NULL;


	/*   Must clean up drag records after each edw delete,
	 *   so that no drag record pointers referance deallocated memory.
	 */
	DragCleanup();


	return;
}



/*
 *	Deletes all allocated edw's.
 */
void EDWDeleteAll()
{
        int i;
        
        for(i = 0; i < (int)total_endeavour_wins; i++)
	{
	    /* This will deallocate all substructures, widgets, windows, and
	     * then the structure itself.
	     */
            EDWDelete(i);
	}

        free(endeavour_win);
        endeavour_win = NULL;
        total_endeavour_wins = 0;


        return;
}
