/*
               Endeavour Browser Window: Event Management

	Functions:

	int EDWInit(endeavour_win_struct *edw)
	int EDWResize(endeavour_win_struct *edw)
	int EDWDraw(endeavour_win_struct *edw, int amount)
	int EDWManage(endeavour_win_struct *edw, event_t *event)
	int EDWManageAll(event_t *event)
	void EDWMap(endeavour_win_struct *edw)
	void EDWUnmap(endeavour_win_struct *edw)
	void EDWDestroy(endeavour_win_struct *edw)



 */

#include "endeavour.h"
#include "edwdefaults.h"


#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))



/*
 *	Initializes an already allocated Endeavour window structure.
 */
int EDWInit(
	endeavour_win_struct *edw,
	int x, int y,
	unsigned int width, unsigned int height,
	char *start_path
)
{
	int i, btn_x, btn_y, status;
	char *strptr;
        win_attr_t wattr;
	pixmap_t pixmap;
	char title[(NAME_MAX * 3) + 256];


        if(edw == NULL)
            return(-1);

        /* Clear structure. */
        memset(edw, 0x00, sizeof(endeavour_win_struct));

	/* Set default values. */
	edw->map_state = 0;
	edw->visibility_state = VisibilityFullyObscured;
	if(osw_gui[0].def_geometry_set)
	{
            edw->x = osw_gui[0].def_toplevel_x;
            edw->y = osw_gui[0].def_toplevel_y;
            edw->width = osw_gui[0].def_toplevel_width;
	    edw->height = osw_gui[0].def_toplevel_height;
	}
	else
	{
            edw->x = osw_gui[0].def_toplevel_x;
            edw->y = osw_gui[0].def_toplevel_y;
            edw->width = DEF_WIDTH;
            edw->height = DEF_HEIGHT;
	}
	if((int)edw->width < MIN_WIDTH)
	    edw->width = MIN_WIDTH;
	if((int)edw->height < MIN_HEIGHT)
            edw->height = MIN_HEIGHT;

        edw->is_in_focus = 0;
        edw->disabled = False;
        edw->dir_win_width = (double)DEF_DIR_WIN_WIDTH / 100;

	edw->file_headers_flag = EDW_COL_NAME_MASK | EDW_COL_SIZE_MASK |
	    EDW_COL_PERMISSION_MASK | EDW_COL_OWNER_MASK |
            EDW_COL_GROUP_MASK | EDW_COL_DATE_MASK;

	edw->show_options = EDW_SHOW_EXPANDABLE_MASK |
	    EDW_SHOW_BRANCH_MASK;

	edw->col_name_pos = DEF_COL_NAME_POS;
        edw->col_size_pos = DEF_COL_SIZE_POS;
        edw->col_permission_pos = DEF_COL_PERMISSION_POS;
        edw->col_owner_pos = DEF_COL_OWNER_POS;
        edw->col_group_pos = DEF_COL_GROUP_POS;
        edw->col_date_pos = DEF_COL_DATE_POS;

	edw->cv_src_path = NULL;


	if((int)width > MIN_WIDTH)
	    edw->width = width;
	if((int)height > MIN_HEIGHT)
	    edw->height = height;



	/* ************************************************************** */

        /* Create toplevel. */
        status = OSWCreateWindow(
            &edw->toplevel,
            osw_gui[0].root_win,
            edw->x, edw->y,
            edw->width, edw->height 
        );
        if(status)
            return(-1);
        OSWSetWindowInput(edw->toplevel, KeyPressMask |
            KeyReleaseMask | ButtonPressMask | ButtonReleaseMask |
            PointerMotionMask | ExposureMask | StructureNotifyMask |
            FocusChangeMask | VisibilityChangeMask | LeaveWindowMask
        );
        OSWSetWindowBkg(edw->toplevel, 0, widget_global.std_bkg_pm);

	WidgetCenterWindow(edw->toplevel, WidgetCenterWindowToRoot);
        OSWGetWindowAttributes(edw->toplevel, &wattr);
	edw->x = wattr.x;
	edw->y = wattr.y;
	edw->width = wattr.width;
	edw->height = wattr.height;

	/* Get icon. */
	pixmap = OSWCreatePixmapFromImage(endeavour_image.endeavour_icon);

	/* Set title. */
	if(userinfo.uid != userinfo.euid)
	{
	    char username[NAME_MAX], eusername[NAME_MAX], hostname[NAME_MAX];


	    strptr = IDGetNameByUID(userinfo.uid);
	    strncpy(
		username, ((strptr == NULL) ? "" : strptr), NAME_MAX
	    );
	    username[NAME_MAX - 1] = '\0';

	    strptr = IDGetNameByUID(userinfo.euid);
            strncpy(
                eusername, ((strptr == NULL) ? "" : strptr), NAME_MAX
            );
            eusername[NAME_MAX - 1] = '\0';

	    strptr = getenv("HOSTNAME");
	    if(strptr == NULL)
		strptr = getenv("HOST");
	    strncpy(hostname, ((strptr == NULL) ? "" : strptr), NAME_MAX);
	    hostname[NAME_MAX - 1] = '\0';

	    sprintf(title, "Endeavour: %s(%s)@%s",
		username, eusername, hostname
	    );
	}
	else
        {
            char username[NAME_MAX], hostname[NAME_MAX];
        
            strptr = IDGetNameByUID(userinfo.uid);
            strncpy(
                username, ((strptr == NULL) ? "" : strptr), NAME_MAX
            );
            username[NAME_MAX - 1] = '\0';
         
            strptr = getenv("HOSTNAME");  
            if(strptr == NULL)
                strptr = getenv("HOST");
            strncpy(hostname, ((strptr == NULL) ? "" : strptr), NAME_MAX);
            hostname[NAME_MAX - 1] = '\0';

            sprintf(title, "Endeavour: %s@%s",
                username, hostname
            );
        }


	/* WM Properties. */
	OSWSetWindowWMProperties(
	    edw->toplevel,
            title,
            "Endeavour",
            ((pixmap == 0) ? widget_global.std_icon_pm : pixmap),
            !osw_gui[0].def_geometry_set,
            edw->x, edw->y,
	    MIN_WIDTH, MIN_HEIGHT,
            osw_gui[0].display_width, osw_gui[0].display_height,
            WindowFrameStyleStandard,
            NULL, 0
	);
	pixmap = 0;	/* Do not destroy pixmap. */


        /* Tool buttons ribbon. */
	if(
	    OSWCreateWindow(
		&edw->tool_ribbon_win,
		edw->toplevel,
		0, 0,
		edw->width,
		TOOL_RIBBON_HEIGHT
	    )
	)
	    return(-1);
	OSWSetWindowInput(edw->tool_ribbon_win, ButtonPressMask |
            ButtonReleaseMask | PointerMotionMask | ExposureMask);
        OSWSetWindowBkg(edw->tool_ribbon_win, 0,
	    widget_global.std_bkg_pm);
        OSWGetWindowAttributes(edw->tool_ribbon_win, &wattr);
	if(
	    OSWCreatePixmap(
		&edw->tool_ribbon_win_buf,
		wattr.width,
		wattr.height
	    )
	)
	    return(-1);

        /* Command ribbon. */
        if(
            OSWCreateWindow(
                &edw->cmd_tool_ribbon,
                edw->toplevel,
                0, TOOL_RIBBON_HEIGHT,
                edw->width, PROMPT_HEIGHT + 4
            )
        )
            return(-1);
        OSWSetWindowInput(edw->cmd_tool_ribbon,
	    ButtonPressMask | ButtonReleaseMask | ExposureMask);
        OSWSetWindowBkg(edw->cmd_tool_ribbon, 0,
            widget_global.std_bkg_pm);
        OSWGetWindowAttributes(edw->cmd_tool_ribbon, &wattr);
        if(
            OSWCreatePixmap(
                &edw->cmd_tool_ribbon_buf,
                wattr.width,
                wattr.height
            )
        )
            return(-1);
	/* Command prompt. */
        status = PromptInit(
            &edw->cmd_prompt,
            edw->cmd_tool_ribbon,
            2, 2, 420, PROMPT_HEIGHT,
            PROMPT_STYLE_FLUSHED,
            "Command:",
            PATH_MAX + NAME_MAX,
            25,
            NULL
        );
        if(status)
            return(-1);
	/* Output button. */
        btn_x = 440;
        btn_y = 5;
	if(
	    PBtnInit( 
                &edw->output_btn,
                edw->cmd_tool_ribbon,
                btn_x, btn_y,
                56, 24,
                "Output",
                PBTN_TALIGN_CENTER,
                NULL,
                &edw->output_btn,
                NULL
	    )
        )
	    return(-1);
        PBtnSetHintMessage(
            &edw->output_btn,
            HINT_MESG_OUTPUT_BTN
        );
        /* Kill button. */
        btn_x += (56 + 10);
        if(
	    PBtnInit( 
                &edw->kill_btn,
                edw->cmd_tool_ribbon,
                btn_x, btn_y,
                56, 24,
                "Kill",
                PBTN_TALIGN_CENTER,
                NULL,
                &edw->kill_btn,
                NULL
	    )
        )
	    return(-1);
        PBtnSetHintMessage(   
            &edw->kill_btn,   
            HINT_MESG_KILL_BTN
        );
	/* View button. */
	btn_x += (56 + 15);
        if(
	    PBtnInit( 
                &edw->view_btn,
                edw->cmd_tool_ribbon,
                btn_x, btn_y,
                56, 24,
                "View",
                PBTN_TALIGN_CENTER,
                NULL,
	        &edw->view_btn,
                NULL
            )
	)
	    return(-1);
        PBtnSetHintMessage(   
            &edw->view_btn,
            HINT_MESG_VIEW_BTN
        );
        /* Edit button. */
        btn_x += (56 + 10);
        if(
	    PBtnInit(
                &edw->edit_btn,
                edw->cmd_tool_ribbon,
                btn_x, btn_y,
                56, 24,
                "Edit",
                PBTN_TALIGN_CENTER,
                NULL,
                &edw->edit_btn,
                NULL
	    )
        )
	    return(-1);
        PBtnSetHintMessage(
            &edw->edit_btn,
            HINT_MESG_EDIT_BTN
        );


        /* Location ribbon. */
        if(
            OSWCreateWindow(
                &edw->loc_tool_ribbon,
                edw->toplevel,
                0, TOOL_RIBBON_HEIGHT + PROMPT_HEIGHT + 4,
                edw->width, PROMPT_HEIGHT + 4
            )
        )
            return(-1);
        OSWSetWindowInput(edw->loc_tool_ribbon,
	    ButtonPressMask | ButtonReleaseMask | ExposureMask);
        OSWSetWindowBkg(edw->loc_tool_ribbon, 0,
            widget_global.std_bkg_pm);
        OSWGetWindowAttributes(edw->loc_tool_ribbon, &wattr);
        if(
            OSWCreatePixmap(
                &edw->loc_tool_ribbon_buf,
                wattr.width,
                wattr.height
            )
        )
            return(-1);
	/* Location prompt. */
        status = PromptInit(
            &edw->loc_prompt,
            edw->loc_tool_ribbon,
            2, 2, 420, PROMPT_HEIGHT,
            PROMPT_STYLE_FLUSHED,  
            "Location:",
            PATH_MAX + NAME_MAX, 
            25,
            NULL
        );
        if(status)
            return(-1);
        
        if(start_path != NULL)
        {   
            strncpy(edw->loc_prompt.buf, start_path,
                PATH_MAX + NAME_MAX);
            edw->loc_prompt.buf[PATH_MAX + NAME_MAX - 1] = '\0';
            edw->loc_prompt.buf_pos = strlen(edw->loc_prompt.buf);
        }


        /* Create directory window. */
        status = OSWCreateWindow(
            &edw->dir_win,
            edw->toplevel,
            0, TOOL_RIBBON_HEIGHT + (2 * (PROMPT_HEIGHT + 4)),
	    (unsigned int)MAX(
                (double)edw->width * ((double)DEF_DIR_WIN_WIDTH / 100),
		4
	    ),
	    (unsigned int)MAX(
	        (int)edw->height - (2 * (PROMPT_HEIGHT + 4)) - STATUS_HEIGHT -
                TOOL_RIBBON_HEIGHT,
		4
	    )
        );
        if(status)     
            return(-1);
        OSWSetWindowInput(edw->dir_win, ButtonPressMask |
            ButtonReleaseMask | PointerMotionMask | ExposureMask
        );
        OSWSetWindowBkg(edw->dir_win, widget_global.surface_editable_pix, 0);
        OSWGetWindowAttributes(edw->dir_win, &wattr);

	status = OSWCreatePixmap(
	    &(edw->dir_win_buf),
	    wattr.width, wattr.height
	);
        if(status)
            return(-1);

	status = SBarInit(&edw->dir_win_sb,
	    edw->dir_win, wattr.width, wattr.height);
	if(status)
            return(-1);


        /* Create file window. */
        status = OSWCreateWindow(
            &edw->file_win,
            edw->toplevel,
            (int)((double)edw->width * ((double)DEF_DIR_WIN_WIDTH / 100)) +
                VERT_SEP_WIDTH,
            FILE_HEADING_HEIGHT + TOOL_RIBBON_HEIGHT +
	    (2 * (PROMPT_HEIGHT + 4)),
	    (unsigned int)MAX(
                ((double)edw->width * ((double)(100 - DEF_DIR_WIN_WIDTH) / 100)) -
                VERT_SEP_WIDTH, 4
	    ),
            (unsigned int)MAX(
		(int)edw->height - (2 * (PROMPT_HEIGHT + 4)) - STATUS_HEIGHT -
		FILE_HEADING_HEIGHT - TOOL_RIBBON_HEIGHT, 4
	    )
        );
        if(status)
            return(-1);
        OSWSetWindowInput(edw->file_win, ButtonPressMask |
            ButtonReleaseMask | PointerMotionMask | ExposureMask
        );
        OSWSetWindowBkg(edw->file_win, widget_global.surface_editable_pix, 0);
	OSWGetWindowAttributes(edw->file_win, &wattr);

        status = OSWCreatePixmap(
            &(edw->file_win_buf),
            wattr.width, wattr.height
        );
        if(status)
            return(-1);

        status = SBarInit(&edw->file_win_sb,
            edw->file_win, wattr.width, wattr.height);
        if(status)
            return(-1);

	/* File window heading. */
	if(
	    OSWCreateWindow(
		&edw->file_heading_win,
		edw->toplevel,
                (int)((double)edw->width * ((double)DEF_DIR_WIN_WIDTH / 100)) +
                    VERT_SEP_WIDTH,
		TOOL_RIBBON_HEIGHT + (2 * (PROMPT_HEIGHT + 2)),
                (unsigned int)MAX(
                    ((double)edw->width * ((double)(100 - DEF_DIR_WIN_WIDTH) / 100)) -
                    VERT_SEP_WIDTH, 1
                ),
		(unsigned int)FILE_HEADING_HEIGHT
	    )
	)
	    return(-1);
        OSWSetWindowInput(edw->file_heading_win, ButtonPressMask |
            ButtonReleaseMask | PointerMotionMask | ExposureMask
        );
        OSWSetWindowBkg(edw->file_heading_win, 0,
	    widget_global.std_bkg_pm);
        OSWGetWindowAttributes(edw->file_heading_win, &wattr);

        status = OSWCreatePixmap(
            &(edw->file_heading_win_buf),
            wattr.width, wattr.height
        );
        if(status)
            return(-1);

        /* Directory and file list split cursor window. */
        if(
            OSWCreateWindow(
                &edw->split_cur,
                edw->toplevel,
                0, 0, VERT_SEP_WIDTH, 10
            )
        )
            return(-1);
        OSWSetWindowBkg(edw->split_cur, osw_gui[0].white_pix, 0);



	/* Status window. */
	if(
	    OSWCreateWindow(
		&edw->status_win,
		edw->toplevel,
		0,
		(int)edw->height - STATUS_HEIGHT,
		edw->width,
		STATUS_HEIGHT
	    )
	)
            return(-1);
        OSWSetWindowInput(edw->status_win, ButtonPressMask |
            ButtonReleaseMask | PointerMotionMask | ExposureMask |
	    EnterWindowMask | LeaveWindowMask
        );
        OSWSetWindowBkg(edw->status_win, 0, widget_global.std_bkg_pm);
        OSWGetWindowAttributes(edw->status_win, &wattr);  
        if(
	    OSWCreatePixmap(
                &edw->status_win_buf,
                wattr.width, wattr.height
            )
	)
	    return(-1);

	OSWGetWindowAttributes(edw->toplevel, &wattr);
        status = HintWinAddMessage(
            edw->status_win,
            edw->toplevel,
            (int)wattr.width - 275,
	    (int)wattr.height - 50,
            HINT_MESG_PADLOCK
        );
        if(status < 0)
            return(-1);
 


	/* ********************************************************* */

	/* Change values prompt. */
	edw->cv_prompt_mode = CV_PROMPT_MODE_NONE;
	status = PromptInit(
            &edw->cv_prompt,
            edw->toplevel,
            0, (int)edw->height / 2,
            edw->width,
	    ROW_SPACING,
            PROMPT_STYLE_NOBORDER,	/* Height must be size of row. */
            NULL,			/* No name. */
            PATH_MAX + NAME_MAX,
            0,
            NULL
        );
        if(status)
            return(-1);


	/* Tool ribbon push buttons. */
	btn_x = 10;
	btn_y = 5;
        status = PBtnInit(
            &edw->folder_back_btn,
            edw->tool_ribbon_win,
            btn_x, btn_y,
            EDW_BTN_WIDTH, EDW_BTN_HEIGHT,
            (widget_global.force_mono) ? "<-" : NULL,
            PBTN_TALIGN_CENTER,     
            endeavour_image.folder_back,
            (void *)&edw->folder_back_btn,
            NULL 
        );
	PBtnSetHintMessage(
	    &edw->folder_back_btn,
	    HINT_MESG_BACK_BTN
	);

        btn_x += 37;
        status = PBtnInit(
            &edw->folder_refresh_btn,
            edw->tool_ribbon_win,
            btn_x, btn_y,
            EDW_BTN_WIDTH, EDW_BTN_HEIGHT,
            (widget_global.force_mono) ? "RF" : NULL,
            PBTN_TALIGN_CENTER,
            endeavour_image.refresh,
            (void *)&edw->folder_refresh_btn,
            NULL
        );
        PBtnSetHintMessage(
            &edw->folder_refresh_btn,
            HINT_MESG_REFRESH_BTN
        );

        btn_x += 50;	/* More margin. */
	status = PBtnInit(
	    &edw->new_folder_btn,
	    edw->tool_ribbon_win,
	    btn_x, btn_y,
            EDW_BTN_WIDTH, EDW_BTN_HEIGHT,
	    (widget_global.force_mono) ? "NF" : NULL,
	    PBTN_TALIGN_CENTER,
	    endeavour_image.folder_new,
            (void *)&edw->new_folder_btn,
	    NULL
	);
	if(status)
	    return(-1);
        PBtnSetHintMessage(
            &edw->new_folder_btn,
            HINT_MESG_NEWFOLDER_BTN
        );

	btn_x += 37;
        status = PBtnInit(
            &edw->properties_btn,
            edw->tool_ribbon_win,   
            btn_x, btn_y,
            EDW_BTN_WIDTH, EDW_BTN_HEIGHT,
            (widget_global.force_mono) ? "Prop" : NULL,
            PBTN_TALIGN_CENTER,
            endeavour_image.properties,
            (void *)&edw->properties_btn,
            NULL
        );
        if(status)
            return(-1);
        PBtnSetHintMessage(
            &edw->properties_btn,
            HINT_MESG_PROPERTIES_BTN
        );

        btn_x += 37;
        status = PBtnInit(
            &edw->move_btn, 
            edw->tool_ribbon_win,
            btn_x, btn_y,
            EDW_BTN_WIDTH, EDW_BTN_HEIGHT,
            (widget_global.force_mono) ? "MV" : NULL,
            PBTN_TALIGN_CENTER,
            endeavour_image.move,
            (void *)&edw->move_btn,
            NULL
        );
        if(status)
            return(-1);
        PBtnSetHintMessage(
            &edw->move_btn,
            HINT_MESG_MOVE_BTN
        );

        btn_x += 37;
        status = PBtnInit(
            &edw->copy_btn,
            edw->tool_ribbon_win,
            btn_x, btn_y,
            EDW_BTN_WIDTH, EDW_BTN_HEIGHT,
            (widget_global.force_mono) ? "CP" : NULL,
            PBTN_TALIGN_CENTER,
            endeavour_image.copy,
            (void *)&edw->copy_btn,
            NULL
        );
        if(status)
            return(-1);
        PBtnSetHintMessage(
            &edw->copy_btn,
            HINT_MESG_COPY_BTN
        );

        btn_x += 37;
        status = PBtnInit(
            &edw->recycle_btn,      
            edw->tool_ribbon_win,
            btn_x, btn_y,
            EDW_BTN_WIDTH, EDW_BTN_HEIGHT,
            (widget_global.force_mono) ? "REC" : NULL,
            PBTN_TALIGN_CENTER,
            endeavour_image.recycle,
            (void *)&edw->recycle_btn,
            NULL
        );
        if(status)
            return(-1);
        PBtnSetHintMessage(
            &edw->recycle_btn,
            HINT_MESG_RECYCLE_BTN
        );

        btn_x += 50;	/* Margin. */
        status = PBtnInit(
            &edw->new_browser_btn,
            edw->tool_ribbon_win,
            btn_x, btn_y,
            EDW_BTN_WIDTH, EDW_BTN_HEIGHT,
            (widget_global.force_mono) ? "NE" : NULL,
            PBTN_TALIGN_CENTER,  
            endeavour_image.new_browser,
            (void *)&edw->new_browser_btn,
            NULL
        );
        if(status)
            return(-1);
        PBtnSetHintMessage(
            &edw->new_browser_btn,
            HINT_MESG_NEWBROWSER_BTN
        );

        btn_x += 37;
        status = PBtnInit(
            &edw->tile_btn,
            edw->tool_ribbon_win,
            btn_x, btn_y,
            EDW_BTN_WIDTH, EDW_BTN_HEIGHT,
            (widget_global.force_mono) ? "Tile" : NULL,
            PBTN_TALIGN_CENTER,
            endeavour_image.tile,
            (void *)&edw->tile_btn,
            NULL
        );  
        if(status)
            return(-1);
        PBtnSetHintMessage(
            &edw->tile_btn,
            HINT_MESG_TILE_BTN
        );  

        btn_x += 37;
        status = PBtnInit(
            &edw->recbin_map_btn,
            edw->tool_ribbon_win,
            btn_x, btn_y,
            EDW_BTN_WIDTH, EDW_BTN_HEIGHT,
            (widget_global.force_mono) ? "RB" : NULL,
            PBTN_TALIGN_CENTER,
            endeavour_image.recycle_bin_small0,
            (void *)&edw->recbin_map_btn,
            NULL
        );
        if(status)
            return(-1);
        PBtnSetHintMessage(
            &edw->recbin_map_btn,
            HINT_MESG_MAPRECBIN_BTN
        );


        /* ********************************************************** */
	/* Directory window quick menu. */

        if(
            MenuInit(  
                &edw->dir_qmenu,
                osw_gui[0].root_win,
                EDWMenuCB,
                (void *)edw
            )
        )
            return(-1);

        MenuAddItem(
            &edw->dir_qmenu,
            "Expand/Colapse",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            EDW_MC_DIR_EXPAND_TOGGLE,
            -1
        );
        MenuAddItem(
            &edw->dir_qmenu,
            "",
            MENU_ITEM_TYPE_HR,
            NULL,
            EDW_MC_NONE,
            -1
        );
        MenuAddItem(
            &edw->dir_qmenu,
            "Copy", 
            MENU_ITEM_TYPE_ENTRY,   
            NULL,
            EDW_MC_COPY,
            -1
        );
        MenuAddItem(
            &edw->dir_qmenu,
            "Move",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            EDW_MC_MOVE,
            -1
        );
        MenuAddItem(
            &edw->dir_qmenu,
            "Rename",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            EDW_MC_RENAME,
            -1
        );
        MenuAddItem(
            &edw->dir_qmenu,
            "Recycle",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            EDW_MC_RECYCLE,
            -1
        );
        MenuAddItem(
            &edw->dir_qmenu,
            "",
            MENU_ITEM_TYPE_HR,
            NULL,
            EDW_MC_NONE,
            -1   
        );
        MenuAddItem(
            &edw->dir_qmenu,
            "Properties",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            EDW_MC_PROPERTIES,
            -1
        );


        /* ********************************************************** */
        /* File window quick menu. */

        if(
            MenuInit(   
                &edw->file_qmenu,
                osw_gui[0].root_win,
                EDWMenuCB,
                (void *)edw 
            )
        )
            return(-1);

        MenuAddItem(
            &edw->file_qmenu,
            "Execute/view",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            EDW_MC_FILE_EXECUTE,
            -1   
        );
        MenuAddItem(
            &edw->file_qmenu,
            "",
            MENU_ITEM_TYPE_HR,   
            NULL,
            EDW_MC_NONE,
            -1
        );
        MenuAddItem(
            &edw->file_qmenu,
            "Copy",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            EDW_MC_COPY,
            -1
        );
        MenuAddItem(
            &edw->file_qmenu,
            "Move",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            EDW_MC_MOVE,
            -1
        );
        MenuAddItem(
            &edw->file_qmenu,
            "Rename",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            EDW_MC_RENAME,
            -1
        );
        MenuAddItem(
            &edw->file_qmenu,  
            "Recycle",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            EDW_MC_RECYCLE,
            -1
        );
        MenuAddItem(
            &edw->file_qmenu, 
            "",
            MENU_ITEM_TYPE_HR,
            NULL,
            EDW_MC_NONE,
            -1
        );
        MenuAddItem(
            &edw->file_qmenu,
            "Properties",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            EDW_MC_PROPERTIES,
            -1
        );
        MenuAddItem(
            &edw->file_qmenu,
            "",
            MENU_ITEM_TYPE_HR,
            NULL,
            EDW_MC_NONE,
            -1
        );
        MenuAddItem(
            &edw->file_qmenu,
            "Text Editor",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            EDW_MC_EDITOR,
            -1
        );


	/* ********************************************************** */

	/* Add toplevel entry to edw's dir list. */
	EDWInitToplevelDir(edw);

	i = EDWMatchDirFromPath(edw, start_path);
	EDWDoSelectDir(edw, i);
	EDWScrollToDirItem(edw, MAX(i - 1, 0));

	return(0);
}



/*
 *	Resizes and updates all sub windows and resources on edw.
 */
int EDWResize(endeavour_win_struct *edw)
{
	int btn_x, btn_y;
	win_attr_t wattr;
	static double last_dir_win_width;	/* In coefficent. */


	if(edw == NULL)
	    return(-1);


	/* Get toplevel attributes. */
	OSWGetWindowAttributes(edw->toplevel, &wattr);

	/* Skip if no change. */
	if(((int)edw->width == wattr.width) &&
	   ((int)edw->height == wattr.height) &&
           (edw->dir_win_width == last_dir_win_width)
	)
	    return(0);

	/* Set new size values of toplevel. */
	edw->x = wattr.x;
        edw->y = wattr.y;
        edw->width = wattr.width;
        edw->height = wattr.height;


	/* Tool ribbon. */
	OSWMoveResizeWindow(
            edw->tool_ribbon_win,
            0, 0,
            edw->width,
	    TOOL_RIBBON_HEIGHT
        );
        OSWGetWindowAttributes(edw->tool_ribbon_win, &wattr);
        OSWDestroyPixmap(&edw->tool_ribbon_win_buf);
        if(
            OSWCreatePixmap(
                &edw->tool_ribbon_win_buf,
                wattr.width, wattr.height
            )
        )
            return(-1);

	/* Tool ribbon push buttons. */
	btn_x = 10; 
        btn_y = 5;
	OSWMoveWindow(
	    edw->folder_back_btn.toplevel,
	    btn_x, btn_y
	);
        btn_x += 37;
        OSWMoveWindow(
            edw->folder_refresh_btn.toplevel,
            btn_x, btn_y
        );
        btn_x += 50;
	 OSWMoveWindow(
            edw->new_folder_btn.toplevel,
            btn_x, btn_y
        );
	btn_x += 37;
        OSWMoveWindow(
            edw->properties_btn.toplevel,
            btn_x, btn_y
        );
        btn_x += 37;
        OSWMoveWindow(
            edw->move_btn.toplevel,
            btn_x, btn_y
        );
        btn_x += 37;
        OSWMoveWindow(
            edw->copy_btn.toplevel,
            btn_x, btn_y
        );
        btn_x += 37;
        OSWMoveWindow(  
            edw->recycle_btn.toplevel,
            btn_x, btn_y
        );
        btn_x += 50;
        OSWMoveWindow(  
            edw->new_browser_btn.toplevel,
            btn_x, btn_y
        );
        btn_x += 37;
        OSWMoveWindow(
            edw->tile_btn.toplevel,
            btn_x, btn_y
        );
        btn_x += 37;
        OSWMoveWindow(
            edw->recbin_map_btn.toplevel,
            btn_x, btn_y
        );


        /* Command ribbon. */
        OSWMoveResizeWindow(
            edw->cmd_tool_ribbon,
            0, TOOL_RIBBON_HEIGHT,
            edw->width,
            PROMPT_HEIGHT + 4
        );
        OSWGetWindowAttributes(edw->cmd_tool_ribbon, &wattr);
        OSWDestroyPixmap(&edw->cmd_tool_ribbon_buf);
        if(
            OSWCreatePixmap(
                &edw->cmd_tool_ribbon_buf,
                wattr.width, wattr.height
            )
        )
            return(-1);

        /* Buttons on command ribbon. */
        btn_x = 440;
        btn_y = 5;
        OSWMoveWindow(
            edw->output_btn.toplevel,
            btn_x, btn_y
        );
	btn_x += (56 + 10);
        OSWMoveWindow(
            edw->kill_btn.toplevel,
            btn_x, btn_y
        );
        btn_x += (56 + 15);
        OSWMoveWindow(
            edw->view_btn.toplevel,
            btn_x, btn_y
        );
        btn_x += (56 + 10);
        OSWMoveWindow(
            edw->edit_btn.toplevel,
            btn_x, btn_y
        );
	/* Command prompt. */
	OSWMoveWindow(
            edw->cmd_prompt.toplevel,
	    2, 2
	);


        /* Location ribbon. */
        OSWMoveResizeWindow(
            edw->loc_tool_ribbon,
            0, TOOL_RIBBON_HEIGHT + PROMPT_HEIGHT + 4,
            edw->width,
            PROMPT_HEIGHT + 4
        );
        OSWGetWindowAttributes(edw->loc_tool_ribbon, &wattr);
        OSWDestroyPixmap(&edw->loc_tool_ribbon_buf);
        if( 
            OSWCreatePixmap(
                &edw->loc_tool_ribbon_buf,
                wattr.width, wattr.height
            )
        )
            return(-1);
	/* Location prompt. */
        OSWMoveWindow(
            edw->loc_prompt.toplevel,
            2, 2
        );


	/* Directory window. */
	OSWMoveResizeWindow(
            edw->dir_win,
            0, TOOL_RIBBON_HEIGHT + (2 * (PROMPT_HEIGHT + 4)),
            (unsigned int)((double)edw->width * edw->dir_win_width),
	    (unsigned int)MAX(
                (int)edw->height - (2 * (PROMPT_HEIGHT + 4)) - STATUS_HEIGHT -
		TOOL_RIBBON_HEIGHT,
		4
	    )
	);
	OSWGetWindowAttributes(edw->dir_win, &wattr);
	OSWDestroyPixmap(&edw->dir_win_buf);
	if(
	    OSWCreatePixmap(
		&edw->dir_win_buf,
		wattr.width, wattr.height
	    )
	)
	    return(-1);
	SBarResize(&edw->dir_win_sb, wattr.width, wattr.height);

	/* Update the maximum dir window width. */
	EDWUpdateDirWinWidth(edw);


        /* File window. */
        OSWMoveResizeWindow(
            edw->file_win,
            (int)((double)edw->width * edw->dir_win_width) + VERT_SEP_WIDTH,
            (int)(FILE_HEADING_HEIGHT + TOOL_RIBBON_HEIGHT +
		(2 * (PROMPT_HEIGHT + 4))),
	    (unsigned int)MAX(
                ((double)edw->width * (1 - edw->dir_win_width)) -
                VERT_SEP_WIDTH,
		4
	    ),
	    (unsigned int)MAX(
                (int)edw->height - (2 * (PROMPT_HEIGHT + 4)) - STATUS_HEIGHT -
		FILE_HEADING_HEIGHT - TOOL_RIBBON_HEIGHT,
		4
	    )
        );
        OSWGetWindowAttributes(edw->file_win, &wattr);
        OSWDestroyPixmap(&edw->file_win_buf);
        if(
            OSWCreatePixmap(
                &edw->file_win_buf,
                wattr.width, wattr.height
            )
        )
            return(-1);
        SBarResize(&edw->file_win_sb, wattr.width, wattr.height);

	/* File heading window. */
        OSWMoveResizeWindow(
            edw->file_heading_win,
            (int)((double)edw->width * edw->dir_win_width) + VERT_SEP_WIDTH,
            TOOL_RIBBON_HEIGHT + (2 * (PROMPT_HEIGHT + 2)),
            (unsigned int)MAX(
                ((double)edw->width * (1 - edw->dir_win_width)) -
                VERT_SEP_WIDTH,
		1
            ),
            FILE_HEADING_HEIGHT
        );

        OSWGetWindowAttributes(edw->file_heading_win, &wattr);
        OSWDestroyPixmap(&edw->file_heading_win_buf);
        if(
            OSWCreatePixmap(
                &edw->file_heading_win_buf,
                wattr.width, wattr.height
            )
        ) 
            return(-1);
	




	/* Status window. */
        OSWMoveResizeWindow(
            edw->status_win,
	    0, (int)edw->height - STATUS_HEIGHT,
	    (int)edw->width,
	    STATUS_HEIGHT
        );
        OSWGetWindowAttributes(edw->status_win, &wattr);
        OSWDestroyPixmap(&edw->status_win_buf);
        if(
            OSWCreatePixmap(
                &edw->status_win_buf,
                wattr.width, wattr.height
            )
        )
            return(-1);

	OSWGetWindowAttributes(edw->toplevel, &wattr);


	/* ************************************************************** */

	/* Prompts. */
/*
	OSWMoveResizeWindow(
	    edw->loc_prompt.toplevel,
            0, TOOL_RIBBON_HEIGHT + PROMPT_HEIGHT,
            edw->width, PROMPT_HEIGHT
	);
	PromptMap(&edw->loc_prompt);
*/

/*
        OSWMoveResizeWindow(
            edw->cmd_prompt.toplevel,
            0, TOOL_RIBBON_HEIGHT,
            edw->width, PROMPT_HEIGHT
        );
        PromptMap(&edw->cmd_prompt);
*/

        OSWMoveResizeWindow(
            edw->cv_prompt.toplevel,
            0, (int)edw->height / 2,
            edw->width, PROMPT_HEIGHT
        );
	if(edw->cv_prompt_mode == CV_PROMPT_MODE_NONE)
	    PromptUnmap(&edw->cv_prompt);
	else
            PromptMap(&edw->cv_prompt);



	/* Record last_dir_win_width. */
	last_dir_win_width = edw->dir_win_width;



	return(0);
}



/*
 *	Redraws the edw.
 */
int EDWDraw(endeavour_win_struct *edw, int amount)
{
	int outline_num;
	int pointer_x, pointer_y;
	int x, y;
	int i, n, len;
	win_attr_t wattr_toplevel, wattr;

        int dlist_y_pos, flist_y_pos;

	int dir_num, file_num;
	int items_visiable, items_drawn;

	char *strptr;

        image_t *ximage;
	char tmp_path[PATH_MAX + NAME_MAX];
        char stringa[NAME_MAX + 256];



	if(edw == NULL)
	    return(-1);


	/* Get toplevel attributes. */
	OSWGetWindowAttributes(edw->toplevel, &wattr_toplevel);

	/* Map as needed. */
	if(!edw->map_state)
	{
	    /* Refresh list. */
	    EDWRefreshAllLists(edw);

	    /* Map windows. */
	    OSWMapRaised(edw->toplevel);

	    OSWMapWindow(edw->tool_ribbon_win);
            OSWMapWindow(edw->cmd_tool_ribbon);
            OSWMapWindow(edw->loc_tool_ribbon);
	    OSWMapWindow(edw->dir_win);
	    OSWMapWindow(edw->file_win);
	    OSWMapWindow(edw->file_heading_win);
	    OSWMapWindow(edw->status_win);

	    PromptMap(&edw->loc_prompt);
	    PromptMap(&edw->cmd_prompt);

	    if(edw->cv_prompt_mode == CV_PROMPT_MODE_NONE)
		PromptUnmap(&edw->cv_prompt);
	    else
		PromptMap(&edw->cv_prompt);

            PBtnMap(&edw->folder_back_btn);
            PBtnMap(&edw->folder_refresh_btn);

            PBtnMap(&edw->new_folder_btn);
            PBtnMap(&edw->properties_btn);
            PBtnMap(&edw->move_btn);
            PBtnMap(&edw->copy_btn);
            PBtnMap(&edw->recycle_btn);

	    PBtnMap(&edw->new_browser_btn);
            PBtnMap(&edw->tile_btn);
	    PBtnMap(&edw->recbin_map_btn);

	    PBtnMap(&edw->output_btn);
	    PBtnMap(&edw->kill_btn);
	    PBtnMap(&edw->view_btn);
	    PBtnMap(&edw->edit_btn);


	    /* Set map state. */
	    edw->map_state = 1;
	    edw->visibility_state = VisibilityUnobscured;

	    /* Change draw amount to complete. */
	    amount = DRAW_AMOUNT_COMPLETE;
	}


	/* Tool ribbon. */
	if(amount == DRAW_AMOUNT_COMPLETE)
	{
	    OSWGetWindowAttributes(edw->tool_ribbon_win, &wattr);
	    WidgetPutImageTile(
		edw->tool_ribbon_win_buf,
		widget_global.std_bkg_img,
		wattr.width, wattr.height
	    );

	    /* Need to redraw recycle bin map button. */
	    PBtnDraw(&edw->recbin_map_btn);

	    /* Device records. */
	    x = DEVREC_X_POS_START;
	    for(i = 0; i < (int)total_devrecs; i++)
	    {
		if(devrec[i] == NULL) continue;

		switch(devrec[i]->image_code)
		{
		  case RECDEV_IMG_NFS:
		    ximage = (devrec[i]->is_mounted) ?
                        endeavour_image.dev_hd1 :
                        endeavour_image.dev_hd0;
		    break;

                  case RECDEV_IMG_MEMORY:
                    ximage = (devrec[i]->is_mounted) ?
                        endeavour_image.dev_hd1 :
                        endeavour_image.dev_hd0;
                    break;

                  case RECDEV_IMG_ZIP:
                    ximage = (devrec[i]->is_mounted) ?
                        endeavour_image.dev_zip1 :
                        endeavour_image.dev_zip0;
                    break;

                  case RECDEV_IMG_TAPE:
                    ximage = (devrec[i]->is_mounted) ?
                        endeavour_image.dev_fd1 :
                        endeavour_image.dev_fd0;
                    break;

                  case RECDEV_IMG_CDROM:
                    ximage = (devrec[i]->is_mounted) ?
                        endeavour_image.dev_cdrom1 :
                        endeavour_image.dev_cdrom0;
                    break;

                  case RECDEV_IMG_FD:
                    ximage = (devrec[i]->is_mounted) ?
                        endeavour_image.dev_fd1 :
                        endeavour_image.dev_fd0;
                    break;

                  default:	/* Default to hard drive. */
                    ximage = (devrec[i]->is_mounted) ?
                        endeavour_image.dev_hd1 :
                        endeavour_image.dev_hd0;
                    break;
		}
		if(ximage != NULL)
		{
		    WidgetPutImageNormal(
		        edw->tool_ribbon_win_buf,
		        ximage,
			x + (DEVREC_ENTRY_SPACING / 2) -
			    ((int)ximage->width / 2),
			4,
			True
		    );
		}

		if(devrec[i]->label != NULL)
		{
		    len = strlen(devrec[i]->label);
		    OSWSetFgPix(widget_global.normal_text_pix);
		    OSWDrawString(
		        (drawable_t)edw->tool_ribbon_win_buf,
		        x + ((int)DEVREC_ENTRY_SPACING / 2) -
			    (((int)len * CHAR_WIDTH) / 2),
		        ((int)wattr.height / 2) + 14,
		        devrec[i]->label
		    );
		}

		x += DEVREC_ENTRY_SPACING;
	    }

            /* Draw frame around tool ribbon window. */
            WidgetFrameButtonPixmap(edw->tool_ribbon_win_buf, False,
                wattr.width, wattr.height,
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix
            );

	    OSWPutBufferToWindow(
		edw->tool_ribbon_win,
		edw->tool_ribbon_win_buf
	    );

	    /* ********************************************************* */
	    /* Command ribbon. */
            OSWGetWindowAttributes(edw->cmd_tool_ribbon, &wattr);
            WidgetPutImageTile(
                edw->cmd_tool_ribbon_buf,
                widget_global.std_bkg_img, 
                wattr.width, wattr.height
            );

            WidgetFrameButtonPixmap(edw->cmd_tool_ribbon_buf, False,
                wattr.width, wattr.height,
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix
            );

            OSWPutBufferToWindow(
		edw->cmd_tool_ribbon,
                edw->cmd_tool_ribbon_buf
	    );

            /* ********************************************************* */
            /* Location ribbon. */
            OSWGetWindowAttributes(edw->loc_tool_ribbon, &wattr);
            WidgetPutImageTile(
                edw->loc_tool_ribbon_buf, 
                widget_global.std_bkg_img,
                wattr.width, wattr.height
            );
         
            WidgetFrameButtonPixmap(edw->loc_tool_ribbon_buf, False,
                wattr.width, wattr.height,
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix
            );
         
            OSWPutBufferToWindow(
		edw->loc_tool_ribbon,
                edw->loc_tool_ribbon_buf
	    );
	}


        /* ********************************************************** */

	/* Redraw scroll bars. */
	if(amount == DRAW_AMOUNT_COMPLETE)
	{
            OSWGetWindowAttributes(edw->dir_win, &wattr);
            SBarDraw(
                &edw->dir_win_sb,
                wattr.width,
                wattr.height,
                edw->max_dir_width,
                (edw->total_dirs + 2) * ROW_SPACING
            );
            OSWGetWindowAttributes(edw->file_win, &wattr);
            SBarDraw(
                &edw->file_win_sb,
                wattr.width,
                wattr.height,
                wattr.width,
		(edw->total_files + 2) * ROW_SPACING
            );

	    /* Redraw prompts. */
	    PromptDraw(&edw->loc_prompt, PROMPT_DRAW_AMOUNT_COMPLETE);
            PromptDraw(&edw->cmd_prompt, PROMPT_DRAW_AMOUNT_COMPLETE);

	    if(edw->cv_prompt_mode != CV_PROMPT_MODE_NONE)
                PromptDraw(&edw->cv_prompt, PROMPT_DRAW_AMOUNT_COMPLETE);
	}



        /* ********************************************************** */
        /* Draw directory list. */

	if((amount == DRAW_AMOUNT_COMPLETE) ||
           (amount == DRAW_AMOUNT_LISTS) ||
           (amount == DRAW_AMOUNT_DIR_LIST) ||
           (amount == DRAW_AMOUNT_DIR_LIST_OUTLINE)
	)
	{
	    /* Get window attributes. */
            OSWGetWindowAttributes(edw->dir_win, &wattr);

	    /* Clear buffer. */
            OSWClearPixmap(edw->dir_win_buf,
                wattr.width, wattr.height,
                widget_global.surface_editable_pix  
            );

            /* Get selection outline number. */
	    outline_num = -1;
            if(amount == DRAW_AMOUNT_DIR_LIST_OUTLINE)
            {
                OSWGetPointerCoords(edw->dir_win,
                    &x, &y, &pointer_x, &pointer_y
                );
                outline_num = EDW_GETDIRNUMBYPOS(edw, pointer_x, pointer_y);
            }


	    /* Check if there are any directories. */
	    if((edw->dir != NULL) &&
               ((int)edw->total_dirs > 0)
	    )
	    {
                /* Calculate starting position for gradiant scrolling. */
                dlist_y_pos = 5 - ((int)edw->dir_win_sb.y_win_pos %
                    ROW_SPACING);

		/* Calculate starting dir_num. */
                dir_num = edw->dir_win_sb.y_win_pos / ROW_SPACING;
		if(dir_num < 0)
		    dir_num = 0;

		/* Calculate items_visiable. */
                items_visiable = ((int)wattr.height / ROW_SPACING) + 1;

		items_drawn = 0;

		while(dir_num < (int)edw->total_dirs)
		{
		    /* Don't draw too many then visable. */
		    if(items_drawn > items_visiable)
			break;

		    /* Skip unallocated entries. */
		    if(edw->dir[dir_num] == NULL)
                    {
                        /* Skip unallocated dir entries. */
                        dir_num++; continue;
                    }

		    /* Allocate name for items with no name. */
                    if(edw->dir[dir_num]->name == NULL)
                    {
                        edw->dir[dir_num]->name = StringCopyAlloc(NO_NAME_STR);
                        if(edw->dir[dir_num]->name == NULL)
                        {
                            dir_num++; continue;
                        }
                    }


		    /* Check if item is selected. */
                    if(EDW_ISDIRSELECTED(edw, dir_num))
                    {
                        /* Selected directory entry. */
                        OSWSetFgPix(widget_global.surface_selected_pix);
                        OSWDrawSolidRectangle(edw->dir_win_buf,
			    LIST_MARGIN + edw->dir[dir_num]->x +
                                DIR_ICON_WIDTH + 4 -
				edw->dir_win_sb.x_win_pos,
			    dlist_y_pos,
			    (   (strlen(edw->dir[dir_num]->name) + 1) +
                                ((edw->dir[dir_num]->link_dest == NULL) ?
                                 0 : strlen(edw->dir[dir_num]->link_dest) + 4
				)
			    ) * CHAR_WIDTH,
			    ROW_SPACING
                        );

                        /* See which ximage to use. */
			ximage = edw->dir[dir_num]->sel_image;
                        OSWSetFgPix(widget_global.selected_text_pix);
		    }
		    /* Expanded? */
		    else if(edw->dir[dir_num]->expanded)
		    {
			 /* See which ximage to use. */
                        ximage = edw->dir[dir_num]->sel_image;
                        OSWSetFgPix(widget_global.editable_text_pix);
		    }
		    /* Not selected or expanded. */
		    else
		    {
                        /* See which ximage to use. */
                        ximage = edw->dir[dir_num]->unsel_image;
                        OSWSetFgPix(widget_global.editable_text_pix);
		    }

		    /* Draw image. */
                    OSWPutImageToDrawablePos(
                        ximage,
                        (drawable_t)edw->dir_win_buf,
                        LIST_MARGIN + edw->dir[dir_num]->x -
                            edw->dir_win_sb.x_win_pos,
			dlist_y_pos + 2
                    );

		    /* Name (print link name if needed). */
		    if(edw->dir[dir_num]->link_dest != NULL)
		    {
			sprintf(tmp_path, "%s -> %s",
			    edw->dir[dir_num]->name,
			    edw->dir[dir_num]->link_dest
			);
		    }
                    else
                    {
			strncpy(tmp_path, edw->dir[dir_num]->name, NAME_MAX);
                    }
		    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

                    OSWDrawString(edw->dir_win_buf,
                        LIST_MARGIN + edw->dir[dir_num]->x +
                            DIR_ICON_WIDTH + 8 -
                            edw->dir_win_sb.x_win_pos,
                        dlist_y_pos + 18,
                        tmp_path
                    );

                    /* Draw branch route lines. */
		    OSWSetFgPix(widget_global.surface_selected_pix);
                    for(i = 1; i < edw->dir[dir_num]->level; i++)
                    {
			n = LIST_MARGIN + (i * DIR_LEVEL_INC)
                            + (DIR_LEVEL_INC / 2) - DIR_ROUTE_WIDTH -
			    edw->dir_win_sb.x_win_pos;

                        OSWDrawLine(edw->dir_win_buf,
			    n, dlist_y_pos,
			    n, dlist_y_pos + ROW_SPACING
			);
                    }
		    if(edw->dir[dir_num]->level > 0)
		    {
		        if(edw->dir[dir_num]->expandable &&
                           (edw->show_options & EDW_SHOW_EXPANDABLE_MASK)
			)
		        {
                            n = LIST_MARGIN + (edw->dir[dir_num]->level *
                                DIR_LEVEL_INC) + (DIR_LEVEL_INC / 2) -
			        DIR_ROUTE_WIDTH -
                                edw->dir_win_sb.x_win_pos;

			    /* Top line. */
                            OSWDrawLine(edw->dir_win_buf,
                                n, dlist_y_pos,
                                n, (int)(dlist_y_pos + (ROW_SPACING * 0.30) + 1)
                            );
			    /* Bottom line. */
                            if((dir_num + 1) < edw->total_dirs)
                            {
                                if(edw->dir[dir_num + 1]->level >=
                                   edw->dir[dir_num]->level
                                )
                                    OSWDrawLine(edw->dir_win_buf,
                                        n, (int)(dlist_y_pos + (ROW_SPACING * 0.70)),
                                        n, dlist_y_pos + ROW_SPACING
                                    );
                            }
			    /* Right line. */
                            OSWDrawLine(edw->dir_win_buf,
                                (int)(n + (DIR_LEVEL_INC * 0.30) - 2),
                                dlist_y_pos + (ROW_SPACING / 2),
                                n + (DIR_LEVEL_INC / 2),
                                dlist_y_pos + (ROW_SPACING / 2)
                            );
			    /* Rectangle. */
			    x = (int)(n - (DIR_LEVEL_INC * 0.20) + 1);
			    y = (int)(dlist_y_pos + (ROW_SPACING * 0.30) + 2);
                            OSWDrawRectangle(edw->dir_win_buf,
                                x, y,
				(unsigned int)((DIR_LEVEL_INC * 0.40) - 1),
                                (unsigned int)((ROW_SPACING * 0.40) - 3)
                            );

			    if(edw->dir[dir_num]->expanded)
			    {
				/* Draw "-". */
				OSWDrawLine(
				    edw->dir_win_buf,
				    x + 2, y + 4,
				    x + 6, y + 4
				);
			    }
			    else
			    {
				/* Draw "+". */
				OSWDrawLine(
                                    edw->dir_win_buf,
                                    x + 2, y + 4,
                                    x + 6, y + 4
                                );
                                OSWDrawLine(
                                    edw->dir_win_buf,
                                    x + 4, y + 2,
                                    x + 4, y + 6
   				);
			    }
		        }
			/* Dir tree route when not expandable. */
		        else
		        {
                            n = LIST_MARGIN + (edw->dir[dir_num]->level *
                                DIR_LEVEL_INC) + (DIR_LEVEL_INC / 2) -
                                DIR_ROUTE_WIDTH - edw->dir_win_sb.x_win_pos;

			    if((dir_num + 1) < edw->total_dirs)
			    {
				if(edw->dir[dir_num + 1]->level <
                                   edw->dir[dir_num]->level
				)
                                    OSWDrawLine(edw->dir_win_buf,
                                        n, dlist_y_pos,
                                        n, dlist_y_pos + (ROW_SPACING / 2)
                                    );
				else
				    OSWDrawLine(edw->dir_win_buf,
                                        n, dlist_y_pos,
                                        n, dlist_y_pos + ROW_SPACING
                                    );
			    }
			    else
			    {
                                OSWDrawLine(edw->dir_win_buf,
                                    n, dlist_y_pos,
                                    n, dlist_y_pos + (ROW_SPACING / 2)
                                );
			    }

                            OSWDrawLine(edw->dir_win_buf,
                                n,
		                dlist_y_pos + (ROW_SPACING / 2),
                                n + (DIR_LEVEL_INC / 2),
                                dlist_y_pos + (ROW_SPACING / 2)
                            );
		        }
                    }

		    /* Draw outline the pointer cursor is over. */
		    if(dir_num == outline_num)
		    {
			OSWSetFgPix(widget_global.surface_selected_pix);
                        OSWDrawRectangle(edw->dir_win_buf,
                            LIST_MARGIN + edw->dir[dir_num]->x -
                                2 - edw->dir_win_sb.x_win_pos,
                            dlist_y_pos,
                            ((  (strlen(edw->dir[dir_num]->name) + 1) +
                                ((edw->dir[dir_num]->link_dest == NULL) ?
                                 0 : strlen(edw->dir[dir_num]->link_dest) + 4
                                )
                            ) * CHAR_WIDTH) + 29,
                            ROW_SPACING
                        );
		    }


		    /* Increment positions. */
		    dlist_y_pos += ROW_SPACING;
		    items_drawn++;
		    dir_num++;
		}
	    }

	    /* Draw frame around dir window. */
            WidgetFrameButtonPixmap(edw->dir_win_buf, True,
                wattr.width, wattr.height,
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix
            );

	    /* Put buffer. */
            OSWPutBufferToWindow(edw->dir_win, edw->dir_win_buf);
	}



        /* ********************************************************** */
        /* Draw file list. */
 
        if((amount == DRAW_AMOUNT_COMPLETE) ||
           (amount == DRAW_AMOUNT_LISTS) ||
           (amount == DRAW_AMOUNT_FILE_LIST) ||
	   (amount == DRAW_AMOUNT_FILE_LIST_OUTLINE)
        )
        {
	    /* Get window attributes (file heading window). */
	    OSWGetWindowAttributes(edw->file_heading_win, &wattr);
	    WidgetPutImageTile(
		edw->file_heading_win_buf,
		widget_global.std_bkg_img,
		wattr.width, wattr.height
	    );

	    /* File heading. */
	    OSWSetFgPix(widget_global.surface_shadow_pix);
	    OSWDrawLine(edw->file_heading_win_buf,
		edw->col_name_pos,
		0,
		edw->col_name_pos,
		wattr.height
	    );
            OSWDrawLine(edw->file_heading_win_buf,
                edw->col_size_pos,
                0,
                edw->col_size_pos,
                wattr.height
            );
            OSWDrawLine(edw->file_heading_win_buf,
                edw->col_permission_pos,
                0,
                edw->col_permission_pos,
                wattr.height
            );
            OSWDrawLine(edw->file_heading_win_buf,
                edw->col_owner_pos,
                0,
                edw->col_owner_pos,
                wattr.height
            );
            OSWDrawLine(edw->file_heading_win_buf,
                edw->col_group_pos,
                0,
                edw->col_group_pos,
                wattr.height
            );
            OSWDrawLine(edw->file_heading_win_buf,
                edw->col_date_pos,
                0,
                edw->col_date_pos,
                wattr.height
            );

            OSWSetFgPix(widget_global.surface_highlight_pix);
            OSWDrawLine(edw->file_heading_win_buf,
                edw->col_name_pos + 1,
                0,
                edw->col_name_pos + 1,
                wattr.height
            );
            OSWDrawLine(edw->file_heading_win_buf,
                edw->col_size_pos + 1,
                0,
                edw->col_size_pos + 1,
                wattr.height
            );
            OSWDrawLine(edw->file_heading_win_buf,
                edw->col_permission_pos + 1,
                0,
                edw->col_permission_pos + 1,
                wattr.height
            );
            OSWDrawLine(edw->file_heading_win_buf,
                edw->col_owner_pos + 1,
                0,
                edw->col_owner_pos + 1,
                wattr.height
            );
            OSWDrawLine(edw->file_heading_win_buf,
                edw->col_group_pos + 1,
                0,
                edw->col_group_pos + 1,
                wattr.height
            );
            OSWDrawLine(edw->file_heading_win_buf,
                edw->col_date_pos + 1,
                0,
                edw->col_date_pos + 1,
                wattr.height
            );
            
            OSWSetFgPix(widget_global.normal_text_pix);
	    OSWDrawStringLimited(edw->file_heading_win_buf,
		edw->col_name_pos + 8,
		(wattr.height / 2) + 5,
		"Name",
		MIN(
                    (edw->col_size_pos - edw->col_name_pos) / CHAR_WIDTH,
		    (int)strlen("Name")
		)
	    );
            OSWDrawStringLimited(edw->file_heading_win_buf,
                edw->col_size_pos + 8,
                (wattr.height / 2) + 5,
                "Size",
                MIN(
                    (edw->col_permission_pos - edw->col_size_pos) / CHAR_WIDTH,
                    (int)strlen("Size")
                )
            );
            OSWDrawStringLimited(edw->file_heading_win_buf,
                edw->col_permission_pos + 8, 
                (wattr.height / 2) + 5,
                "Permissions",
                MIN(
                    (edw->col_owner_pos - edw->col_permission_pos) / CHAR_WIDTH,
                    (int)strlen("Permissions")
                )
            );
            OSWDrawStringLimited(edw->file_heading_win_buf,
                edw->col_owner_pos + 8,
                (wattr.height / 2) + 5,
                "Owner",
                MIN(
                    (edw->col_group_pos - edw->col_owner_pos) / CHAR_WIDTH,
                    (int)strlen("Owner")
                )
            );
            OSWDrawStringLimited(edw->file_heading_win_buf,
                edw->col_group_pos + 8, 
                (wattr.height / 2) + 5,
                "Group",
                MIN(
                    (edw->col_date_pos - edw->col_group_pos) / CHAR_WIDTH,
                    (int)strlen("Group")
                )
            );
            OSWDrawStringLimited(edw->file_heading_win_buf,
                edw->col_date_pos + 8,
                (wattr.height / 2) + 5,
                "Last Modified",
                MIN(
                    ((int)wattr.width - edw->col_date_pos) / CHAR_WIDTH,
                    (int)strlen("Last Modified")
                )
            );


	    WidgetFrameButtonPixmap(edw->file_heading_win_buf,
		False,
		wattr.width, wattr.height,
		widget_global.surface_highlight_pix,
		widget_global.surface_shadow_pix
	    );

	    OSWPutBufferToWindow(
		edw->file_heading_win,
		edw->file_heading_win_buf
	    );


	    /* ********************************************** */

            /* Get window attributes (file window). */
            OSWGetWindowAttributes(edw->file_win, &wattr);

            /* Clear buffer. */
	    OSWClearPixmap(edw->file_win_buf,
		wattr.width, wattr.height,
		widget_global.surface_editable_pix
	    );

        
            /* Check if there are any files. */
            if( (edw->file != NULL) &&
                ((int)edw->total_files > 0)
            )
            {
		/* Calculate starting position for gradiant scrolling. */
                flist_y_pos = 5 - ((int)edw->file_win_sb.y_win_pos %
                    ROW_SPACING);

                /* Calculate starting file_num. */
                file_num = edw->file_win_sb.y_win_pos / ROW_SPACING;
                if(file_num < 0)
                    file_num = 0;
            
                /* Calculate items_visiable. */
                items_visiable = ((int)wattr.height / ROW_SPACING) + 1;
                     
                items_drawn = 0;
                        
                while(file_num < (int)edw->total_files)
                {
                    /* Don't draw too many then visable. */
                    if(items_drawn > items_visiable)
                        break;

                    /* Skip unallocated entries. */
                    if(edw->file[file_num] == NULL)   
                    {
                        /* Skip unallocated file entries. */
                        file_num++; continue;
                    }
                     
                    /* Allocate name for items with no name. */
                    if(edw->file[file_num]->name == NULL)
                    {
                        edw->file[file_num]->name = StringCopyAlloc(NO_NAME_STR);
                        if(edw->file[file_num]->name == NULL)
                        {
                            file_num++; continue;
                        }
                    }


                    /* Check if item is selected. */
                    if(EDW_ISFILESELECTED(edw, file_num))
                    {   
                        /* Selected file entry. */
                        OSWSetFgPix(widget_global.surface_selected_pix);
                        OSWDrawSolidRectangle(edw->file_win_buf,
                            edw->col_name_pos + 24 + 4,
                            flist_y_pos,
                            MIN(
                                edw->col_size_pos -
                                edw->col_name_pos - 24 - 4,
                                (int)(strlen(edw->file[file_num]->name) + 1)
				* CHAR_WIDTH
			    ),
                            ROW_SPACING
                        );

                        /* See which ximage to use. */
                        if(S_ISLNK(edw->file[file_num]->mode))
                            ximage = widget_global.linkicon_selected_img;
                        else if( (S_IXUSR & edw->file[file_num]->mode) ||
                                 (S_IXGRP & edw->file[file_num]->mode) ||
                                 (S_IXOTH & edw->file[file_num]->mode)
                        )
                            ximage = widget_global.execicon_selected_img;
                        else if(S_ISREG(edw->file[file_num]->mode))
                            ximage = widget_global.fileicon_selected_img;
                        else if(S_ISCHR(edw->file[file_num]->mode) ||
                                S_ISBLK(edw->file[file_num]->mode) ||
                                S_ISFIFO(edw->file[file_num]->mode) ||
                                S_ISSOCK(edw->file[file_num]->mode)
                        )
                            ximage = widget_global.pipeicon_selected_img;
                        else
                            ximage = widget_global.fileicon_selected_img;

                        OSWSetFgPix(widget_global.selected_text_pix);
                    }
                    else
                    {
                        /* See which ximage to use. */
                        if(S_ISLNK(edw->file[file_num]->mode))
                            ximage = widget_global.linkicon_normal_img;
                        else if( (S_IXUSR & edw->file[file_num]->mode) ||
                                 (S_IXGRP & edw->file[file_num]->mode) ||
                                 (S_IXOTH & edw->file[file_num]->mode)
                        )
                            ximage = widget_global.execicon_normal_img;
                        else if(S_ISREG(edw->file[file_num]->mode))
                            ximage = widget_global.fileicon_normal_img;
                        else if(S_ISCHR(edw->file[file_num]->mode) ||
                                S_ISBLK(edw->file[file_num]->mode) ||
                                S_ISFIFO(edw->file[file_num]->mode) ||
                                S_ISSOCK(edw->file[file_num]->mode)
                        )
                            ximage = widget_global.pipeicon_normal_img;
                        else
                            ximage = widget_global.fileicon_normal_img;

                        OSWSetFgPix(widget_global.editable_text_pix);
                    }


		    /* File name. */
		    if(edw->file_headers_flag & EDW_COL_NAME_MASK)
		    {
                        /* Put icon image. */
                        OSWPutImageToDrawablePos(
                            ximage,
                            (drawable_t)edw->file_win_buf,
                            edw->col_name_pos + 4,
                            flist_y_pos + 2
                        );
                        OSWDrawStringLimited(edw->file_win_buf,
                            edw->col_name_pos + 24 + 8,
                            flist_y_pos + 18,
			    edw->file[file_num]->name,
			    MIN(
                                (edw->col_size_pos -
                                edw->col_name_pos - 24) / CHAR_WIDTH,
                                (int)strlen(edw->file[file_num]->name)
			    )
                        );
		    }
                    /* File size. */
		    if(edw->file_headers_flag & EDW_COL_SIZE_MASK)
		    {
                        OSWSetFgPix(widget_global.editable_text_pix);
                        sprintf(stringa, "%ld", edw->file[file_num]->size);
                        OSWDrawStringLimited(edw->file_win_buf,
                            edw->col_size_pos + 4,
                            flist_y_pos + 18,
                            stringa,
			    MIN(
                                (edw->col_permission_pos -
                                edw->col_size_pos) / CHAR_WIDTH,
                                (int)strlen(stringa)
			    )
                        );
		    }
		    /* Permissions. */
		    if(edw->file_headers_flag & EDW_COL_PERMISSION_MASK)
		    {
                    stringa[0] = (edw->file[file_num]->mode & S_IRUSR) ? 'r' : '-';
                    stringa[1] = (edw->file[file_num]->mode & S_IWUSR) ? 'w' : '-';
                    stringa[2] = (edw->file[file_num]->mode & S_IXUSR) ? 'x' : '-';
		    if(edw->file[file_num]->mode & S_ISUID)
			stringa[2] = 's';

                    stringa[3] = (edw->file[file_num]->mode & S_IRGRP) ? 'r' : '-';
                    stringa[4] = (edw->file[file_num]->mode & S_IWGRP) ? 'w' : '-';
                    stringa[5] = (edw->file[file_num]->mode & S_IXGRP) ? 'x' : '-';
                    if(edw->file[file_num]->mode & S_ISGID)
                        stringa[5] = 's';

                    stringa[6] = (edw->file[file_num]->mode & S_IROTH) ? 'r' : '-';
                    stringa[7] = (edw->file[file_num]->mode & S_IWOTH) ? 'w' : '-';
                    stringa[8] = (edw->file[file_num]->mode & S_IXOTH) ? 'x' : '-';
		    if(edw->file[file_num]->mode & S_ISVTX)
			stringa[8] = 't';
                    stringa[9] = '\0';
                        OSWDrawStringLimited(edw->file_win_buf,
                            edw->col_permission_pos + 4,
                            flist_y_pos + 18,
                            stringa,
                            MIN(
                                (edw->col_owner_pos -
                                edw->col_permission_pos) / CHAR_WIDTH,
                                (int)strlen(stringa)
                            )
                        );
		    }
		    /* Owner. */
		    if(edw->file_headers_flag & EDW_COL_OWNER_MASK)
		    {
			strptr = EDW_GETOWNERNAME(edw->file[file_num]->uid);
                        if(strptr != NULL)
                        {
                            OSWDrawStringLimited(edw->file_win_buf,
                                edw->col_owner_pos + 4,
                                flist_y_pos + 18,
                                strptr,
                                MIN(
                                    (edw->col_group_pos -
                                    edw->col_owner_pos) / CHAR_WIDTH,
                                    (int)strlen(strptr)
                                )
                            );   
                        }
		    }
                    /* Group. */
                    if(edw->file_headers_flag & EDW_COL_GROUP_MASK)
                    {
                        strptr = EDW_GETGROUPNAME(edw->file[file_num]->gid);
                        if(strptr != NULL)
                        {
                            OSWDrawStringLimited(edw->file_win_buf,
                                edw->col_group_pos + 4,
                                flist_y_pos + 18,
                                strptr,
                                MIN(
                                    (edw->col_date_pos -
                                    edw->col_group_pos) / CHAR_WIDTH,
                                    (int)strlen(strptr)
                                )
                            );
                        }
                    }
		    /* Date (last modified). */
		    if(edw->file_headers_flag & EDW_COL_DATE_MASK)
                    {
			strptr = EDW_GETTIMESTR(
			    edw->file[file_num]->last_modified
			);

			if(strptr != NULL)
                        {
			    OSWDrawStringLimited(edw->file_win_buf,
                                edw->col_date_pos + 4,
                                flist_y_pos + 18,
                                strptr,
                                MIN(
                                    ((int)wattr.width -
                                    edw->col_date_pos) / CHAR_WIDTH,
                                    (int)strlen(strptr)
                                )
                            );
			}
		    }

		    /* Increment positions. */                          
                    flist_y_pos += ROW_SPACING;
                    items_drawn++;
                    file_num++;
                }
            }

	    /* Draw frame around file window. */        
            WidgetFrameButtonPixmap(edw->file_win_buf, True,
                wattr.width, wattr.height,
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix
            );

	    /* Put buffer. */
            OSWPutBufferToWindow(edw->file_win, edw->file_win_buf);
	}


	/* *********************************************************** */
	/* Status bar. */
	if((amount == DRAW_AMOUNT_COMPLETE) ||
           (amount == DRAW_AMOUNT_STATUS)
	)
	{
	    OSWGetWindowAttributes(edw->status_win, &wattr);
	    WidgetPutImageTile(
		edw->status_win_buf, widget_global.std_bkg_img,
		wattr.width, wattr.height
	    );

	    /* Padlock icon. */
	    if(option.write_protect)
	    {
		ximage = endeavour_image.padlock1;
	    }
	    else
	    {
		ximage = endeavour_image.padlock_tmp;
	    }
	    if(ximage != NULL)
	    {
                OSWPutImageToDrawablePos(
                    ximage,
                    (drawable_t)edw->status_win_buf,
                    (int)wattr.width - (int)ximage->width - 15,
                    ((int)wattr.height / 2) -
                        ((int)ximage->height / 2)
                );
	    }

	    /* Status message. */
	    if((edw->status_mesg != NULL) && (ximage != NULL))
	    {
	        OSWSetFgPix(widget_global.normal_text_pix);
		OSWDrawStringLimited(
		    edw->status_win_buf,
		    10,
		    ((int)wattr.height / 2) + 5,
		    edw->status_mesg,
		    MIN((int)strlen(edw->status_mesg),
			((int)wattr.width - (int)ximage->width - 25) /
			CHAR_WIDTH
		    )
		);
	    }

            /* Raised frame. */
            WidgetFrameButtonPixmap(edw->status_win_buf, False,
                wattr.width, wattr.height,
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix
            );

            /* Put buffer. */
            OSWPutBufferToWindow(edw->status_win, edw->status_win_buf);
	}



	return(0);
}



int EDWManage(endeavour_win_struct *edw, event_t *event)
{
	int i, n, x, y, z, root_x, root_y;
	int events_handled = 0;
	long l;
	win_attr_t wattr;
	static long last_buttonpress;
	image_t *image;
	char *strptr;

	char stringa[512];
	char tmp_path[PATH_MAX + NAME_MAX];
	char parent[PATH_MAX + NAME_MAX];


	if((edw == NULL) ||
           (event == NULL)
	)
	    return(events_handled);

	if(!edw->map_state &&
           (event->type != MapNotify)
	)
	    return(events_handled);

	switch(event->type)
	{
	  /* ********************************************************* */
	  case KeyPress:
	    /* Skip KeyPress if edw not in focus. */
	    if(!edw->is_in_focus)
		return(events_handled);

            /* Change values prompt. */
            if(edw->cv_prompt.is_in_focus)
            {
                /* Process contents of cv prompt buffer. */

                /* Change values prompt: Enter key */  
                if((event->xkey.keycode == osw_keycode.enter) ||
                   (event->xkey.keycode == osw_keycode.np_enter)
                )
                {
		    /* Turn off yes_to_all for upcomming prompting. */
		    yes_to_all = False;

		    /* Handle by cv prompt mode. */
                    switch(edw->cv_prompt_mode)
		    {
		      case CV_PROMPT_MODE_RENAME:
			if(strlen(edw->cv_prompt.buf) > 0)
			{
			    /* Get parent. */
			    strptr = GetParentDir(edw->cv_src_path);
			    if(strptr == NULL) break;
			    strncpy(parent, strptr, PATH_MAX + NAME_MAX);
			    parent[PATH_MAX + NAME_MAX - 1] = '\0';

			    /* Get target path. */
			    strptr = PrefixPaths(
				parent,
				edw->cv_prompt.buf
			    );
                            if(strptr == NULL) break;
                            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
                            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

			    if(
			        EDWPerformFileOp(
			            OP_CODE_RENAME,
			            tmp_path,		/* Target. */
			            edw->cv_src_path	/* Source. */
			        )
			    )
			    {
                                /* Set failed status message. */
                                sprintf(stringa,
                     "Error encountered when attempting to rename object"
                                );
                            }
                            else
			    {
				if(ISPATHDIR(tmp_path))
				{
				    /* Update dir item name on the list. */
				    EDWUpdateDirItemName(
				        edw,
				        edw->cv_src_path,	/* Old name. */
                                        tmp_path		/* New name. */
				    );
				}
				else
				{
				    /* Update file listing. */
				    EDWRefreshFileList(
					edw,
					edw->loc_prompt.buf
				    );
				}
				/* Set status message. */
			        sprintf(stringa,
                                "Renamed object to `%s'",
				    edw->cv_prompt.buf
			        );
			    }
/*
			    EDWRefreshFileList(edw);
*/
                            free(edw->status_mesg);
                            edw->status_mesg = StringCopyAlloc(stringa);
			}
			else
			{
			    sprintf(stringa,
                       "Insufficent characters given, name not changed");
			    free(edw->status_mesg);
                            edw->status_mesg = StringCopyAlloc(stringa);
			}
                        break;

                      case CV_PROMPT_MODE_CHMOD:
                        if(strlen(edw->cv_prompt.buf) > 0)
                        {
			    /* Check write protect since this op is direct. */
			    if(ISPATHDIR(edw->cv_src_path))
			        FileOpChmodDir(
                                    edw->cv_src_path,	/* Target. */
                                    edw->cv_prompt.buf  /* Argument. */
                                );
			    else
                                FileOpChmodFile(
				    edw->cv_src_path,	/* Target. */
				    edw->cv_prompt.buf  /* Argument. */
                                );

			    if(!ISPATHDIR(edw->cv_src_path))
                                EDWRefreshFileList(edw, edw->loc_prompt.buf);

                            /* Set status message. */
                            sprintf(stringa,
                                "Set new permissions on object `%s'",
                                edw->cv_src_path
                            );
                            free(edw->status_mesg);
                            edw->status_mesg = StringCopyAlloc(stringa);
			}
                        break;

                      case CV_PROMPT_MODE_MOVE:
                        if(!ISPATHDIR(edw->cv_src_path))
                            EDWRefreshFileList(edw, edw->loc_prompt.buf);
                        break;

                      case CV_PROMPT_MODE_COPY:
                        if(!ISPATHDIR(edw->cv_src_path))
                            EDWRefreshFileList(edw, edw->loc_prompt.buf);
                        break;

                      case CV_PROMPT_MODE_CHOWN_OWNER:
                        if(strlen(edw->cv_prompt.buf) > 0)
                        {
                            EDWPerformFileOp(
                                OP_CODE_CHOWN_OWNER,
				edw->cv_prompt.buf,
                                edw->cv_src_path
                            );

                            if(!ISPATHDIR(edw->cv_src_path))
                                EDWRefreshFileList(edw, edw->loc_prompt.buf);

                            /* Set status message. */
                            sprintf(stringa,
                                "Chowned object `%s' to owner `%s'",   
                                edw->cv_src_path,
				edw->cv_prompt.buf
                            );
                            free(edw->status_mesg);
                            edw->status_mesg = StringCopyAlloc(stringa);
			}
                        break;

		      case CV_PROMPT_MODE_CHOWN_GROUP:
                        if(strlen(edw->cv_prompt.buf) > 0)
                        {
                            EDWPerformFileOp(
                                OP_CODE_CHOWN_GROUP,
				edw->cv_prompt.buf,
                                edw->cv_src_path
                            );

                            if(!ISPATHDIR(edw->cv_src_path))
                                EDWRefreshFileList(edw, edw->loc_prompt.buf);

                            /* Set status message. */
                            sprintf(stringa,
                                "Chowned object `%s' to group `%s'",
				edw->cv_src_path,
                                edw->cv_prompt.buf
                            );
                            free(edw->status_mesg);
                            edw->status_mesg = StringCopyAlloc(stringa);
                        }
                        break;
                    }

                    /* Need to turn off global yes_to_all. */
		    yes_to_all = False;

                    /* Switch change values prompt mode back to none. */
                    edw->cv_prompt_mode = CV_PROMPT_MODE_NONE;
                 
                    PromptUnmap(&edw->cv_prompt);
                    edw->cv_prompt.is_in_focus = 0;
                    
                    events_handled++;
                }
		/* Change values prompt: Escape key */
		else if(event->xkey.keycode == osw_keycode.esc)
		{
                    /* Switch change values prompt mode back to none. */
                    edw->cv_prompt_mode = CV_PROMPT_MODE_NONE;

                    PromptUnmap(&edw->cv_prompt); 
                    edw->cv_prompt.is_in_focus = 0;

                    /* Set status message. */
                    free(edw->status_mesg);
                    edw->status_mesg = StringCopyAlloc(
			"Change of values operation aborted."
		    );

		    events_handled++;
		}
            }
	    /* Escape key. */
	    else if(event->xkey.keycode == osw_keycode.esc)
	    {
		/*   Note: if the change values prompt if active would
		 *   have already handled the escape key press.
		 */
		EDW_UNSELECTALLDIRS(edw);
		EDW_UNSELECTALLFILES(edw);

		/* Clear command prompt. */
		memset(edw->cmd_prompt.buf, 0x00, edw->cmd_prompt.buf_len);
		edw->cmd_prompt.buf_pos = 0;

                events_handled++;  
	    }
	    /* Refresh: F5. */
	    else if(event->xkey.keycode == osw_keycode.f5)
	    {
		EDWDoRefresh(edw);

		events_handled++;
		return(events_handled);
	    }
            /* Bring up move dialog: F7. */
            else if(event->xkey.keycode == osw_keycode.f7)
            {
		EDWDoMove(edw);

                events_handled++;
		return(events_handled);
            }
            /* Bring up copy dialog: F8. */
            else if(event->xkey.keycode == osw_keycode.f8)
            {
                EDWDoCopy(edw);

                events_handled++;
                return(events_handled);
            }
	    /* Recycle object (make sure no prompts are active). */
	    else if((edw->loc_prompt.is_in_focus == 0) &&
                    (edw->cmd_prompt.is_in_focus == 0) &&
                    (edw->cv_prompt.is_in_focus == 0) &&
                    (event->xkey.keycode == osw_keycode.ddelete)
	    )
	    {
		EDWDoRecycle(edw);

                events_handled++;
		return(events_handled);
	    }
	    /* Selected directory item Key handling. */
	    else if(edw->total_sel_dirs > 0)
	    {
                /* Enter key. */
                if((event->xkey.keycode == osw_keycode.enter) ||
                   (event->xkey.keycode == osw_keycode.np_enter)
                )
                {
		    /* Check if the alt key is pressed. */
		    if(drag_record[0].alt_key_state)
		    {
			/* Edit properties of selected dirs. */
			EDWDoEditProperties(edw);
		    }
		    else
		    {
			/* Change directory. */

                        /* Change all cursors to working. */
                        EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_WORKING);

		        /* Record old path temporarly. */
                        strncpy(tmp_path, edw->loc_prompt.buf, PATH_MAX + NAME_MAX);
                        tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

		        /* Let prompt manage so history buf gets recorded. */
                        PromptManage(&edw->loc_prompt, event);

		        strncpy(edw->loc_prompt.buf, tmp_path,
			    edw->loc_prompt.buf_len);
		        edw->loc_prompt.buf[(int)edw->loc_prompt.buf_len - 1] = '\0';


		        x = edw->sel_dir[0];
		        if((x >= 0) && (x < edw->total_dirs))
		        {
		            if(edw->dir[x] != NULL)
			    {
				if(edw->dir[x]->expanded)
				    EDWColapseDir(edw, x);
				else
				    EDWExpandDir(edw, x);
			    }
		        }
/*
			EDWRefreshAllLists(edw);
*/
                        /* Change all cursors. */
                        EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_NORMAL);
		    }

                    events_handled++;   
                }
		/* Minus/Underscore key. */
		else if((event->xkey.keycode == osw_keycode.minus) ||
                        (event->xkey.keycode == osw_keycode.comma) ||
                        (event->xkey.keycode == osw_keycode.np_minus)
		)
		{
		    x = edw->sel_dir[0];
		    EDWColapseDir(edw, x);

                    events_handled++;
		}
                /* Equal/Add key. */
                else if((event->xkey.keycode == osw_keycode.equal) || 
                        (event->xkey.keycode == osw_keycode.period) ||
                        (event->xkey.keycode == osw_keycode.np_add)
                )
                {
                    x = edw->sel_dir[0];
                    EDWExpandDir(edw, x);

                    events_handled++;
                }
	    }
            /* Selected file item key handling. */
            else if(edw->total_sel_files > 0)
            {
                /* Enter key. */
                if((event->xkey.keycode == osw_keycode.enter) ||
                   (event->xkey.keycode == osw_keycode.np_enter)
                )
                {
		    /* Check if the alt key is pressed. */
		    if(drag_record[0].alt_key_state)
		    {
			/* Edit properties of selected files. */
                        EDWDoEditProperties(edw);
		    }
		    else
		    {
		        /* Execute program for selected files on edw. */
		        EDWExecuteSelectedFiles(edw, True);
		    }

                    events_handled++;   
                }
            }
	    /* Location prompt. */
	    else if(edw->loc_prompt.is_in_focus == 1)
	    {
		/* Enter key. */
		if((event->xkey.keycode == osw_keycode.enter) ||
                   (event->xkey.keycode == osw_keycode.np_enter)
		)
		{
		    if(edw->loc_prompt.buf != NULL)
		    {
                        /* Record old path temporarly. */
                        strncpy(
			    tmp_path,
			    edw->loc_prompt.buf,
			    PATH_MAX + NAME_MAX
		        );
                        tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

                        /* Let prompt manage so history buf gets recorded. */
                        PromptManage(&edw->loc_prompt, event);

                        strncpy(
			    edw->loc_prompt.buf,
			    tmp_path,
                            edw->loc_prompt.buf_len
		        );
		        edw->loc_prompt.buf_pos = strlen(edw->loc_prompt.buf);

		        /* Seek new dir. */
                        x = EDWMatchDirFromPath(edw, edw->loc_prompt.buf);
		        EDWDoSelectDir(edw, x);
			EDWScrollToDirItem(edw, MAX(x - 1, 0));
		    }

		    events_handled++;
		}
	    }
	    /* Command prompt. */
	    else if(edw->cmd_prompt.is_in_focus)
	    {
		/* Execute command. */

                /* Command prompt: Enter key. */
                if((event->xkey.keycode == osw_keycode.enter) ||
                   (event->xkey.keycode == osw_keycode.np_enter)
                )
                {
		    EDWDoExecCmdPrompt(edw);

                    /* Manage prompt so that history buf gets recorded. */
                    PromptManage(&edw->cmd_prompt, event);   

/* Update prompt with kill command for last process. */
/*
sprintf(edw->cmd_prompt.buf, "kill %i", (int)ProcHistLastPid());
*/

                    PromptDraw(&edw->cmd_prompt, PROMPT_DRAW_AMOUNT_TEXTONLY);

		    events_handled++;
                    return(events_handled);
                }
	    }
	    break;

          /* ********************************************************* */
	  case KeyRelease:
            if(!edw->is_in_focus)
                return(events_handled);

            break;

          /* ********************************************************* */
          case ButtonPress:
            /* Toplevel (for list split drag. */
            if(event->xany.window == edw->toplevel)
            {
                OSWGetWindowAttributes(edw->toplevel, &wattr);
        
                x = event->xbutton.x;
                y = event->xbutton.y;
                z = (int)(edw->dir_win_width * (double)wattr.width);
                if((x >= z) && (x < (z + VERT_SEP_WIDTH)) &&
                   (y >= ((2 * PROMPT_HEIGHT) + TOOL_RIBBON_HEIGHT)) &&
                   (y < ((int)wattr.height - STATUS_HEIGHT))
                )
                {
                    /* Record drag for file heading window. */
                    drag_record[0].state = True;
                    drag_record[0].type = DRAG_TYPE_LIST_SPLIT;
                    drag_record[0].origin_ptr_type = POINTER_TYPE_EDW;
                    drag_record[0].origin_ptr = edw;

                    /* Set up split cursor window. */
                    OSWSetWindowCursor(edw->split_cur,
                        endeavour_cursor.list_heading_resize);
                    OSWReparentWindow(edw->split_cur, edw->toplevel);
                    OSWGetPointerCoords(
			edw->toplevel,
			&root_x, &root_y,
                        &x, &y
		    );
                    OSWMoveResizeWindow(edw->split_cur,
                        x - (VERT_SEP_WIDTH / 2),
                        (2 * PROMPT_HEIGHT) + TOOL_RIBBON_HEIGHT,
                        VERT_SEP_WIDTH,
                        MAX((int)edw->height - (2 * PROMPT_HEIGHT) -
                           TOOL_RIBBON_HEIGHT - STATUS_HEIGHT, 2)
                    );
                    OSWMapRaised(edw->split_cur);

                    /* Do not ungrab pointer, toplevel needs to track events. */
                }
                
                events_handled++;
                return(events_handled);
            }
	    /* Status window - padlock. */
	    else if((event->xany.window == edw->status_win) &&
                    (endeavour_image.padlock0 != NULL)
	    )
	    {
		image = endeavour_image.padlock0;

		OSWGetWindowAttributes(edw->status_win, &wattr);
		if((event->xbutton.x >= ((int)wattr.width - image->width - 15)) &&
                   (event->xbutton.x < ((int)wattr.width - 15)) &&
                   (event->xbutton.y >= 2) &&
                   (event->xbutton.y < ((int)wattr.height - 2))
		)
		{
		    if(option.write_protect)
			option.write_protect = 0;
		    else
			option.write_protect = 1;

                    /* Need to redraw padlocks on all windows that have. */
                    for(x = 0; x < (int)total_endeavour_wins; x++)
                    {   
                        if(endeavour_win[x] == NULL) continue;
                        if(endeavour_win[x]->map_state < 1) continue;
                        EDWDraw(endeavour_win[x], DRAW_AMOUNT_STATUS);
                    }
                    for(x = 0; x < (int)total_recbin_wins; x++)
                    {
                        if(recbin_win[x] == NULL) continue;
                        if(recbin_win[x]->map_state < 1) continue;
                        
                        RecBinDraw(recbin_win[x], DRAW_AMOUNT_STATUS);
                    }
#ifdef USE_IMLIB
                    for(x = 0; x < (int)total_img_browsers; x++)
                    {
                        if(img_browser[x] == NULL) continue;
                        if(img_browser[x]->map_state < 1) continue;
                        
                        IMBRDraw(img_browser[x], DRAW_AMOUNT_STATUS);
                    } 
#endif /* USE_IMLIB */

		    events_handled++;
		    return(events_handled);
		}
	    }
	    /* Tool ribbon: Mount/Unmount device. */
	    else if(event->xany.window == edw->tool_ribbon_win)
	    {
		x = DEVREC_X_POS_START;
		for(z = 0; z < total_devrecs; z++)
		{
		    if(!DevRecIsAllocated(z))
			continue;

		    if((event->xbutton.x >= x) &&
		       (event->xbutton.x < (x + DEVREC_ENTRY_SPACING))
		    )
		    {
			/* Change all cursors. */
			EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_WORKING);

			/* Button2 or Button3 mounts/unmounts. */
			if((event->xbutton.button == Button2) ||
                           (event->xbutton.button == Button3)
			)
			{
			    /* Mount or unmount? */
                            if(devrec[z]->is_mounted)
                            {
				/* Unmount device. */
                                if(!(devrec[z]->flags & DEV_TF_NOUNMOUNT) &&
                                   (devrec[z]->path != NULL) &&
                                   (event->xbutton.button == Button3)
				)
				{
				    /* Update status window message. */
                                    sprintf(stringa,
 "Device `%s' not mounted (right click to mount on directory `%s').",
                                        devrec[z]->label,
                                        devrec[z]->path
                                    );
				    free(edw->status_mesg);
                                    edw->status_mesg = StringCopyAlloc(stringa);

                                    /* Change working directory. */
                                    if(chdir("/") == 0)
                                        getcwd(dname.current_working, PATH_MAX);
                                    dname.current_working[PATH_MAX - 1] = '\0';
                                
                                    /* Unmount device. */
                                    DevRecDoUnmount(z);
 

				    /* Check all edw working directories to
				     * make sure they are not referancing
				     * this just unmounted directory.
				     */
                for(y = 0; y < (int)total_endeavour_wins; y++)
		{
		    if(endeavour_win[y] == NULL) continue;
		    if(endeavour_win[y]->loc_prompt.buf == NULL) continue;

		    /* Change path as needed. */
		    if(COMPARE_PARENT_PATHS(
                        endeavour_win[y]->loc_prompt.buf,
			devrec[z]->path
		       )
		    )
		    {
                        i = EDWMatchExistingDirFromPath(
			    endeavour_win[y],
			    devrec[z]->path
			);
                        EDWRefreshDirItem(endeavour_win[y], i);
                        EDWColapseDir(endeavour_win[y], i);
                        EDWDoSelectDir(endeavour_win[y], i);
			EDWScrollToDirItem(endeavour_win[y], MAX(i - 1, 0));
		    }

                    /* Skip current edw, it will be drawn later. */
                    if(endeavour_win[y] == edw) continue;

                    EDWDraw(endeavour_win[y], DRAW_AMOUNT_COMPLETE);
		}

				}
                            }
                            else
                            {
				/* Mount device. */
                                DevRecDoMount(z);

				/* Redraw all endeavour windows. */
                                for(y = 0; y < (int)total_endeavour_wins; y++)
                                {
                                  if(endeavour_win[y] == NULL) continue;

                                  /* Redraw file list as needed. */
				  if(endeavour_win[y]->loc_prompt.buf != NULL)
				  {
                                    if(!strcmp(endeavour_win[y]->loc_prompt.buf,
                                      devrec[z]->path
                                    ))
                                    {
                                      i = EDWMatchExistingDirFromPath(
                                        endeavour_win[y],
                                        devrec[z]->path
                                      );
				      EDWExpandDir(endeavour_win[y], i);

                                      endeavour_win[y]->file_win_sb.x_win_pos = 0;
                                      endeavour_win[y]->file_win_sb.x_win_pos = 0;

                                      EDWRefreshFileList(
					endeavour_win[y],
                                        devrec[z]->path
                                      );
                                    }
				  }

                                  /* Skip current edw, it will be drawn later. */
                                  if(endeavour_win[y] == edw) continue;

                                  if(endeavour_win[y]->map_state == 1)
                                    EDWDraw(endeavour_win[y], DRAW_AMOUNT_COMPLETE);
                                }

                                /* Update status window message. */
                                sprintf(stringa,
 "Device `%s' Total: %ld  Used: %ld  Available: %ld  Ratio: %.0f%%",
                                    devrec[z]->label,
                                    devrec[z]->capacity_total,
                                    devrec[z]->capacity_used,
                                    devrec[z]->capacity_available,
                                    devrec[z]->capacity_percent
                                );
                                free(edw->status_mesg);
                                edw->status_mesg = StringCopyAlloc(stringa);
                            }
			}
			/*   Button1 or Button2 changes directory.
                         *
                         *   (Do not use else if, independant condition).
			 */
			if((event->xbutton.button == Button1) ||
			   (event->xbutton.button == Button2)
                        )
			{
			    if(devrec[z]->is_mounted)
                            {
                                /* Change dir. */
                                if(devrec[z]->path != NULL) 
                                {
                                    i = EDWMatchDirFromPath(edw,
					devrec[z]->path);
				    EDWRefreshDirItem(edw, i);
                                    EDWExpandDir(edw, i);
                                    EDWDoSelectDir(edw, i);
				    EDWScrollToDirItem(edw, MAX(i - 1, 0));
				}
			    }
			    else if(event->xbutton.button == Button2)
			    {
				printdw(&err_dw,
"Cannot switch path to device:\n\n\
    Device is not mounted.\n\n\
To mount the device, press Button3 over the\n\
device icon.   Press Button3 again to unmount.\n\n"
				);
			    }
			}

                        /* Change back all cursors. */
                        EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_NORMAL);

			events_handled++;
			break;
		    }

		    /* Increment x position. */
		    x += DEVREC_ENTRY_SPACING;
		}

                /* Redraw complete. */
                if(events_handled > 0)
                {
                    EDWDraw(edw, DRAW_AMOUNT_COMPLETE);
                    return(events_handled);
                }
	    }
            /* Directory window: Button3 (Dir quick menu) */
            else if((event->xany.window == edw->dir_win) &&
                    (event->xbutton.button == Button3)
            )
            {
                OSWGetPointerCoords(edw->toplevel,
                    &root_x, &root_y,
                    &x, &y
                );
                OSWMoveWindow(edw->dir_qmenu.toplevel, root_x, root_y);
                edw->dir_qmenu.x = root_x;
                edw->dir_qmenu.y = root_y;

                MenuMap(&edw->dir_qmenu);

                events_handled++;
                return(events_handled);
            }
            /* File window: Button3 (File quick menu) */
            else if((event->xany.window == edw->file_win) &&
                    (event->xbutton.button == Button3)
            )
            {
                OSWGetPointerCoords(edw->toplevel,
                    &root_x, &root_y,
                    &x, &y
                );
                OSWMoveWindow(edw->file_qmenu.toplevel, root_x, root_y);
                edw->file_qmenu.x = root_x;
                edw->file_qmenu.y = root_y;

                MenuMap(&edw->file_qmenu);

                events_handled++;
                return(events_handled);
            }
	    /* Directory window: Button1: double click or click on
	     * expand cursor.
	     */
	    else if((event->xany.window == edw->dir_win) &&
                    (event->xbutton.button == Button1)
	    )
	    {
		/* Change scroll bar focuses. */
		edw->dir_win_sb.is_in_focus = 1;
		edw->file_win_sb.is_in_focus = 0;

		/* Get entry position number. */
                y = EDW_GETDIRNUMBYPOS(
                    edw,
                    event->xbutton.x,
                    event->xbutton.y
                );


		/* Get current millitime. */
		l = MilliTime();

		/* Fast double click. */
		if(((last_buttonpress + widget_global.double_click_int)
		   >= l) && EDW_ISDIRSELECTED(edw, y)
		)
		{
                    EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_WORKING);

                    if((y >= 0) &&
                       (y < (int)edw->total_dirs) &&
                       EDW_ISDIRSELECTED(edw, y)
                    )
		    {
			/* Change dir. */
			if(edw->dir[y] != NULL)
			{
                            /* Record old path temporarly. */
                            strncpy(
				tmp_path, edw->loc_prompt.buf,
				PATH_MAX + NAME_MAX
			    );
                            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

                            /* Let prompt manage so history buf gets recorded. */
                            PromptManage(&edw->loc_prompt, event);

			    PromptSetS(&edw->loc_prompt, tmp_path);

			    /* Clean up drag just in case. */
			    DragCleanup();

			    /* Expand or colapse dir? */
			    if(edw->dir[y]->expanded)
				EDWColapseDir(edw, y);
			    else
				EDWExpandDir(edw, y);
/*
			    EDWRefreshAllLists(edw);
*/
			    /* Redraw all and return. */
                            EDWDraw(edw, DRAW_AMOUNT_COMPLETE);
			}
		    }

                    EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_NORMAL);

                    last_buttonpress = 0;

		    events_handled++;
                    return(events_handled);
		}
                /* Slow double click (change name). */
                else if(((last_buttonpress + widget_global.relabel_item_delay)
                        >= l) && EDW_ISDIRSELECTED(edw, y) &&
                        (edw->total_sel_dirs == 1)
		)
                {
                    EDWDoMapCVPromptDir(
                        edw,
			y,
                        CV_PROMPT_MODE_RENAME
                    );

		    last_buttonpress = 0;

                    events_handled++;
                    return(events_handled);
		}
		/* Clicked on expand branch (this needs to be checked
		 * second to last).
		 */
                else if((y >= 0) && (y < edw->total_dirs))
                {
		    if(edw->dir[y] != NULL)
		    {
			if(((event->xbutton.x + edw->dir_win_sb.x_win_pos) >=
                            (edw->dir[y]->x - 10)) &&
                           ((event->xbutton.x + edw->dir_win_sb.x_win_pos) <
                            (edw->dir[y]->x))
			)
			{
                            EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_WORKING);

                            /* Record old path temporarly. */
                            strncpy(
                                tmp_path, edw->loc_prompt.buf,
                                PATH_MAX + NAME_MAX  
                            );
                            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
                                
                            /* Let prompt manage so history buf gets recorded. */
                            PromptManage(&edw->loc_prompt, event);
  
			    PromptSetS(&edw->loc_prompt, tmp_path);

                            /* Clean up drag just in case. */
                            DragCleanup();

                            /* Expand or colapse dir? */
                            if(edw->dir[y]->expanded)
                                EDWColapseDir(edw, y);
                            else
                                EDWExpandDir(edw, y);
/*
                            EDWRefreshAllLists(edw);
*/
                            /* Redraw all and return. */
                            EDWDraw(edw, DRAW_AMOUNT_COMPLETE);

                            EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_NORMAL);

			    last_buttonpress = 0;

                            events_handled++;
                            return(events_handled);
			}
		    }
                    /* Record last button press. */
                    last_buttonpress = l;
                }
		else
		{
                    /* Record last button press. */
                    last_buttonpress = l;
		}

		events_handled++;
	    }
	    /* File window: Button1: Double click. */
	    else if((event->xany.window == edw->file_win) &&
                    (event->xbutton.button == Button1)
            )
	    {
                /* Change scroll bar focuses. */
                edw->dir_win_sb.is_in_focus = 0;
                edw->file_win_sb.is_in_focus = 1;

                y = EDW_GETFILENUMBYPOS(
		    edw,
		    event->xbutton.x,
                    event->xbutton.y
                );

                /* Check timming for double click. */
                if(((last_buttonpress + widget_global.double_click_int) >=
                   MilliTime()) &&
                   EDW_ISFILESELECTED(edw, y)
                )
		{
                    EDWExecuteSelectedFiles(edw, True);
		    last_buttonpress = 0;
		}
                /* Slow double click (change values). */
                else if(((last_buttonpress + widget_global.relabel_item_delay)
                            >= MilliTime()) &&
                        EDW_ISFILESELECTED(edw, y) &&
                        (edw->total_sel_files == 1)
		)
                {
                    x = event->xbutton.x;

                    /* Name. */ 
                    if((x >= 0) && (x < edw->col_size_pos))
                    {
                        EDWDoMapCVPromptFile(   
                            edw,
                            y,
                            CV_PROMPT_MODE_RENAME
                        );
                    }
                    /* Permissions. */
                    else if((x >= edw->col_permission_pos) &&
                            (x < edw->col_owner_pos)
                    )
                    {
                        EDWDoMapCVPromptFile(   
                            edw,
                            y,
                            CV_PROMPT_MODE_CHMOD
                        );
                    }
                    /* Owner. */
                    else if((x >= edw->col_owner_pos) &&
                            (x < edw->col_group_pos)
                    )
                    {
                        EDWDoMapCVPromptFile(   
                            edw,
			    y,
                            CV_PROMPT_MODE_CHOWN_OWNER
                        );
                    }
                    /* Group. */
                    else if((x >= edw->col_group_pos) &&
                            (x < edw->col_date_pos)
                    )
                    {
                        EDWDoMapCVPromptFile(
                            edw,
			    y,
                            CV_PROMPT_MODE_CHOWN_GROUP
                        );
                    }

                    last_buttonpress = 0;

                    events_handled++;
                    return(events_handled);
		}
		else
		{
                    /* Record last button press. */
                    last_buttonpress = MilliTime();
		}

                events_handled++;
	    }

	    /* File heading window (start reposition colums). */
	    else if(event->xany.window == edw->file_heading_win)
	    {
		/* Record drag for file heading window. */
		drag_record[0].state = True;
                drag_record[0].type = DRAG_TYPE_FILE_HEADING;
		drag_record[0].origin_ptr_type = POINTER_TYPE_EDW;
		drag_record[0].origin_ptr = edw;

		/* Set up heading_col_flag. */
		drag_record[0].heading_col_flag = 0;

                if( (event->xbutton.x >= edw->col_date_pos - DRAG_TOLORANCE) &&
                         (event->xbutton.x < edw->col_date_pos + DRAG_TOLORANCE)
                )
                    drag_record[0].heading_col_flag |= EDW_COL_DATE_MASK;
                else if( (event->xbutton.x >= edw->col_group_pos - DRAG_TOLORANCE) &&
                         (event->xbutton.x < edw->col_group_pos + DRAG_TOLORANCE)
                )
                    drag_record[0].heading_col_flag |= EDW_COL_GROUP_MASK;
                else if( (event->xbutton.x >= edw->col_owner_pos - DRAG_TOLORANCE) &&
                         (event->xbutton.x < edw->col_owner_pos + DRAG_TOLORANCE)
                )
                    drag_record[0].heading_col_flag |= EDW_COL_OWNER_MASK;
                else if( (event->xbutton.x >= edw->col_permission_pos - DRAG_TOLORANCE) &&
                         (event->xbutton.x < edw->col_permission_pos + DRAG_TOLORANCE)
                )
                    drag_record[0].heading_col_flag |= EDW_COL_PERMISSION_MASK;
                else if( (event->xbutton.x >= edw->col_size_pos - DRAG_TOLORANCE) &&
                         (event->xbutton.x < edw->col_size_pos + DRAG_TOLORANCE)
                )
                    drag_record[0].heading_col_flag |= EDW_COL_SIZE_MASK;
                else if( (event->xbutton.x >= edw->col_name_pos - DRAG_TOLORANCE) &&
                         (event->xbutton.x < edw->col_name_pos + DRAG_TOLORANCE)
                )
                    drag_record[0].heading_col_flag |= EDW_COL_NAME_MASK;


                events_handled++;
		return(events_handled);
	    }

	    /* Redraw lists only as needed. */
	    if(events_handled > 0)
	    {
		EDWDraw(edw, DRAW_AMOUNT_LISTS);
		return(events_handled);
	    }

	    break;


          /* ********************************************************* */
          case ButtonRelease:
	    /* Dir window: Button1. */
	    if((event->xany.window == edw->dir_win) &&
               (event->xbutton.button == Button1) &&
               !drag_record[0].state &&
	       (edw->cv_prompt_mode == CV_PROMPT_MODE_NONE)
            )
	    {
		/* Select a directory. */
		OSWGetWindowAttributes(edw->dir_win, &wattr);

		/* Get directory number by position of pointer. */
		x = EDW_GETDIRNUMBYPOS(edw,
		    event->xbutton.x,
                    event->xbutton.y
		);
		EDWDoSelectDir(edw, x);

                /* Did we get a valid position? */
                if((x > -1) && (x < (int)edw->total_dirs))
                {
		    /* Scroll if near edge. */
		    OSWGetWindowAttributes(edw->dir_win, &wattr);
		    if(event->xbutton.y >
		       ((int)wattr.height - ROW_SPACING)
		    )
		    {
                        EDWDraw(edw, DRAW_AMOUNT_DIR_LIST);
			OSWGUISync(False);

			edw->dir_win_sb.y_win_pos += (int)wattr.height / 2;
			y = ((edw->total_dirs + 2) * ROW_SPACING) -
			    (int)wattr.height;
			if(edw->dir_win_sb.y_win_pos > y)
			    edw->dir_win_sb.y_win_pos = y;

			usleep(SELECT_SHIFT_DELAY * 1000);
                    }
		    else if(event->xbutton.y < ROW_SPACING)
		    {
                        EDWDraw(edw, DRAW_AMOUNT_DIR_LIST);
                        OSWGUISync(False);

		        edw->dir_win_sb.y_win_pos -= (int)wattr.height / 2;
                        usleep(SELECT_SHIFT_DELAY * 1000);
		    }
                    if(edw->dir_win_sb.y_win_pos < 0) 
                        edw->dir_win_sb.y_win_pos = 0;
                }

		/* Unselect all prompts. */
		edw->loc_prompt.is_in_focus = 0;
                edw->cmd_prompt.is_in_focus = 0;
                edw->cv_prompt.is_in_focus = 0;

                EDWDraw(edw, DRAW_AMOUNT_COMPLETE);

		events_handled++;
		return(events_handled);
	    }	/* Dir window. */

            /* File window: Button1. */
            else if((event->xany.window == edw->file_win) &&
                    (event->xbutton.button == Button1) &&
                    !drag_record[0].state &&
                    (edw->cv_prompt_mode == CV_PROMPT_MODE_NONE)
	    )
            {
		/* Select file(s). */
                OSWGetWindowAttributes(edw->file_win, &wattr);

	        /* Get file number by position of pointer. */
                x = EDW_GETFILENUMBYPOS(
                    edw,
		    event->xbutton.x,
		    event->xbutton.y
		);

                /* Did we get a valid position? */
                if((x > -1) && (x < (int)edw->total_files))
                {
		    /* Must unselect all dirs when a file is selected. */
		    EDW_UNSELECTALLDIRS(edw);

                    /* Shift key held down will select multiple files in a row. */
                    if(drag_record[0].shift_key_state)
                    {
			/* Only works if more than 1 file was selected. */
                        if((int)edw->total_sel_files > 0)
			{
			    z = 0;

			    /* Top to bottom select. */
			    if(x > (int)edw->sel_file[0])
			    {
				    y = edw->sel_file[0];
				    z = x + 1;
			    }
			    else if(x < (int)edw->sel_file[0])
			    {
				    y = x;
				    z = edw->sel_file[0] + 1;
			    }
			    EDW_UNSELECTALLFILES(edw);
			    while(y < z)
			    {
				    EDW_SELECTFILE(edw, y);
				    y++;
			    }
			}
                    }
		    /* Ctrl key held down will select multiple files. */
	            else if(drag_record[0].ctrl_key_state)
		    {
			if(EDW_ISFILESELECTED(edw, x))
				EDW_UNSELECTFILE(edw, x);
			else
				EDW_SELECTFILE(edw, x);
		    }
		    /* Single select. */
		    else
		    {
			    EDW_UNSELECTALLFILES(edw);
                            EDW_SELECTFILE(edw, x);
		    }

		    /* Update command prompt buffer. */
		    EDWExecuteSelectedFiles(edw, False);
		    PromptDraw(&edw->cmd_prompt,
			PROMPT_DRAW_AMOUNT_TEXTONLY);


		    /* Update status message. */
		    for(x = 0, z = 0; x < (int)edw->total_sel_files; x++)
		    {
			    y = edw->sel_file[x];
			    if((y < 0) || (y >= (int)edw->total_files)) continue;
			    if(edw->file[y] == NULL) continue;

			    z += edw->file[y]->size;
		    }

		    free(edw->status_mesg);
		    edw->status_mesg = NULL;
		    /* Use tmp_path for now. */
		    sprintf(tmp_path,
   "%i object%s selected, %i bytes (double click to view or execute).",
			    edw->total_sel_files,
			    ((int)edw->total_sel_files > 1) ? "s" : "",
			    z
		    );
		    edw->status_mesg = StringCopyAlloc(tmp_path);


                    /* Scroll if near edge. */
                    OSWGetWindowAttributes(edw->file_win, &wattr);
                    if(event->xbutton.y >
                       ((int)wattr.height - ROW_SPACING)
                    )
		    {
                            EDWDraw(edw, DRAW_AMOUNT_FILE_LIST);
                            OSWGUISync(False);

                            edw->file_win_sb.y_win_pos += (int)wattr.height / 2;
			    y = (((int)edw->total_files + 2) * ROW_SPACING) -
                                (int)wattr.height;
			    if(edw->file_win_sb.y_win_pos > y)
				edw->file_win_sb.y_win_pos = y;

			    usleep(SELECT_SHIFT_DELAY * 1000);
		    }
                    else if(event->xbutton.y < ROW_SPACING)
		    {
                            EDWDraw(edw, DRAW_AMOUNT_FILE_LIST);
                            OSWGUISync(False);

                            edw->file_win_sb.y_win_pos -= (int)wattr.height / 2;
                            usleep(SELECT_SHIFT_DELAY * 1000);
                    }
                    if(edw->file_win_sb.y_win_pos < 0)
                        edw->file_win_sb.y_win_pos = 0;
                }

                /* Unselect all prompts. */
                edw->loc_prompt.is_in_focus = 0;
                edw->cmd_prompt.is_in_focus = 0;
                edw->cv_prompt.is_in_focus = 0;
                
                EDWDraw(edw, DRAW_AMOUNT_COMPLETE);

                events_handled++;
		return(events_handled);
	    }	/* File window. */

            /* File heading window. */
            else if(event->xany.window == edw->file_heading_win)
	    {
                OSWGetWindowAttributes(edw->file_heading_win, &wattr);  

                /* Check if drag state was active. */
                if(drag_record[0].state)
                {
                    /* Drag records will be reset on exit of this function. */

		}

                /* Unselect all prompts. */
                edw->loc_prompt.is_in_focus = 0;
                edw->cmd_prompt.is_in_focus = 0;
                edw->cv_prompt.is_in_focus = 0;
                
                EDWDraw(edw, DRAW_AMOUNT_LISTS);
 
		events_handled++;
                return(events_handled); 
	    }	/* File heading window. */

            /* Unmap split cursor window on ButtonRelease. */
            else if(drag_record[0].state &&
                   (drag_record[0].type == DRAG_TYPE_LIST_SPLIT) &&
                   ((endeavour_win_struct *)drag_record[0].origin_ptr == edw)
            )
            {
                OSWGetWindowAttributes(edw->toplevel, &wattr);

                /* Unmap the split cursor window. */
                OSWUnmapWindow(edw->split_cur);

                x = event->xbutton.x;
                y = event->xbutton.y;

                if(x >= ((int)wattr.width - 10))
                    x = (int)wattr.width - 10;
                if(x < 10)
                    x = 10;

                /* Resize windows. */
                if(wattr.width != 0)
                {
                    edw->dir_win_width = (double)(x
                        - (int)(VERT_SEP_WIDTH / 2)) /
                        (double)wattr.width;
                }

                edw->width += 1;     /* Force EDWResize() to resize. */
                EDWResize(edw);
                EDWDraw(edw, DRAW_AMOUNT_LISTS);

                events_handled++;
                return(events_handled);
            }
            break;

          /* ********************************************************* */
          case MotionNotify:

	    /* Change drag state to active and record drag starts. */

	    /* Dir window. */
	    if(event->xany.window == edw->dir_win)
	    {
		/* Draw outline on pointer over item? */
		if(drag_record[0].state)
		{
		    EDWDraw(edw, DRAW_AMOUNT_DIR_LIST_OUTLINE);

                    /* Switch change values prompt mode back to none. */
		    if(edw->cv_prompt_mode != CV_PROMPT_MODE_NONE)
		    {
                        edw->cv_prompt_mode = CV_PROMPT_MODE_NONE;

                        PromptUnmap(&edw->cv_prompt);
                        edw->cv_prompt.is_in_focus = 0;
		    }

		    /* Do not report event for this. */
		}
	    }
            /* File window. */
            else if(event->xany.window == edw->file_win)
	    {
                /* Draw outline on pointer over item? */
                if(drag_record[0].state)
                {
                    /* Switch change values prompt mode back to none. */
                    if(edw->cv_prompt_mode != CV_PROMPT_MODE_NONE)
                    {
                        edw->cv_prompt_mode = CV_PROMPT_MODE_NONE;

                        PromptUnmap(&edw->cv_prompt);
                        edw->cv_prompt.is_in_focus = 0;
                    }

                    /* Do not report event for this. */

		}
	    }
	    /* File heading window. */
	    else if(event->xany.window == edw->file_heading_win)
	    {
		if(drag_record[0].state &&
                   (drag_record[0].type == DRAG_TYPE_FILE_HEADING)
		)
		{
		    /* Update file heading colum spacing. */
		    if(drag_record[0].heading_col_flag & EDW_COL_NAME_MASK)
		    {
			x = edw->col_name_pos;
		        edw->col_name_pos = event->xmotion.x;
			y = edw->col_name_pos - x;

			edw->col_size_pos += y;
			edw->col_permission_pos += y;
			edw->col_owner_pos += y;
			edw->col_group_pos += y;
			edw->col_date_pos += y;
		    }
                    else if(drag_record[0].heading_col_flag & EDW_COL_SIZE_MASK)
		    {
			x = edw->col_size_pos;
                        edw->col_size_pos = event->xmotion.x;
			y = edw->col_size_pos - x;

                        edw->col_permission_pos += y;
                        edw->col_owner_pos += y; 
                        edw->col_group_pos += y;
                        edw->col_date_pos += y;
		    }
                    else if(drag_record[0].heading_col_flag & EDW_COL_PERMISSION_MASK)
		    {
			x = edw->col_permission_pos;
                        edw->col_permission_pos = event->xmotion.x;
			y = edw->col_permission_pos - x;

                        edw->col_owner_pos += y; 
                        edw->col_group_pos += y;
                        edw->col_date_pos += y;
		    }
                    else if(drag_record[0].heading_col_flag & EDW_COL_OWNER_MASK)
		    {
			x = edw->col_owner_pos;
                        edw->col_owner_pos = event->xmotion.x;
			y = edw->col_owner_pos - x;

                        edw->col_group_pos += y;
                        edw->col_date_pos += y;
		    }
                    else if(drag_record[0].heading_col_flag & EDW_COL_GROUP_MASK)
		    {
			x = edw->col_group_pos;
                        edw->col_group_pos = event->xmotion.x;
			y = edw->col_group_pos - x;

                        edw->col_date_pos += y;
		    }
                    else if(drag_record[0].heading_col_flag & EDW_COL_DATE_MASK)
		    {
			x = edw->col_date_pos;
                        edw->col_date_pos = event->xmotion.x;
			y = edw->col_date_pos - x;
		    }

		    /* Sanitize colum spacing. */
		    EDWSANITIZEFILECOLSPACING(edw);

		    EDWDraw(edw, DRAW_AMOUNT_FILE_LIST);

		    events_handled++;
		    return(events_handled);
		}
		/* If not in any drag, check if cursor needs to change. */
		else if(!drag_record[0].state)
		{
		    x = event->xbutton.x;

		    if((x >= edw->col_date_pos - DRAG_TOLORANCE) &&
                       (x < edw->col_date_pos + DRAG_TOLORANCE)
                    )
                        WidgetSetWindowCursor(edw->file_heading_win,
                            widget_global.h_split_wcr);
		    else if((x >= edw->col_group_pos - DRAG_TOLORANCE) &&
                            (x < edw->col_group_pos + DRAG_TOLORANCE)
                    )
                        WidgetSetWindowCursor(edw->file_heading_win,
                            widget_global.h_split_wcr);
		    else if((x >= edw->col_owner_pos - DRAG_TOLORANCE) &&
                            (x < edw->col_owner_pos + DRAG_TOLORANCE)
                    )
                       WidgetSetWindowCursor(edw->file_heading_win,
                            widget_global.h_split_wcr);
		    else if((x >= edw->col_permission_pos - DRAG_TOLORANCE) &&
                            (x < edw->col_permission_pos + DRAG_TOLORANCE)
                    )
                        WidgetSetWindowCursor(edw->file_heading_win,
                            widget_global.h_split_wcr);
		    else if((x >= edw->col_size_pos - DRAG_TOLORANCE) &&
                            (x < edw->col_size_pos + DRAG_TOLORANCE)
                    )
                        WidgetSetWindowCursor(edw->file_heading_win,
                            widget_global.h_split_wcr);
                    else if((x >= edw->col_name_pos - DRAG_TOLORANCE) &&
                            (x < edw->col_name_pos + DRAG_TOLORANCE)
                    )
                        WidgetSetWindowCursor(edw->file_heading_win,
			    widget_global.h_split_wcr);
		    else
                        OSWUnsetWindowCursor(edw->file_heading_win);
		}
	    }

            /* Toplevel for dir and file window split. */
            else if(event->xany.window == edw->toplevel)
            {
                if(drag_record[0].state &&
                   (drag_record[0].type == DRAG_TYPE_LIST_SPLIT) &&
                   ((endeavour_win_struct *)drag_record[0].origin_ptr == edw)
                )
                {
                    /* Move the split cursor window. */
                    OSWGetPointerCoords(edw->toplevel,
			&root_x, &root_y,
                        &x, &y
		    );
                    if((x > 10) && (x <= ((int)edw->width - 10)))
                    {
                        OSWMoveResizeWindow(edw->split_cur,
                         x - (VERT_SEP_WIDTH / 2),
                         (2 * PROMPT_HEIGHT) + TOOL_RIBBON_HEIGHT,
                         VERT_SEP_WIDTH,
                         MAX((int)edw->height - (2 * PROMPT_HEIGHT) -
                           TOOL_RIBBON_HEIGHT - STATUS_HEIGHT, 2)
                        );
                    }
                
                    events_handled++;
                    return(events_handled);
                }
                else
                {
                    /* Change pointer cursor. */
                    OSWGetWindowAttributes(edw->toplevel, &wattr);

                    x = event->xmotion.x;
                    y = event->xmotion.y;
                    z = (int)(edw->dir_win_width * (double)wattr.width);
                    if((x >= z) && (x < (z + VERT_SEP_WIDTH + 10)) &&
                       (y >= ((2 * PROMPT_HEIGHT) + TOOL_RIBBON_HEIGHT)) &&
                       (y < ((int)wattr.height - STATUS_HEIGHT))
                    )
                        OSWSetWindowCursor(edw->toplevel,
                            endeavour_cursor.list_heading_resize);
                    else
                        OSWUnsetWindowCursor(edw->toplevel);

                    /* Do not report event for this. */
                }
	    }

	    /* Tool ribbon: Print device records information on status bar. */
            if(event->xany.window == edw->tool_ribbon_win)
	    {
                x = DEVREC_X_POS_START;
                for(z = 0; z < (int)total_devrecs; z++)
                {
                    if(!DevRecIsAllocated(z)) continue; 
                
                    if((event->xmotion.x >= x) &&
                       (event->xmotion.x < (x + DEVREC_ENTRY_SPACING))
		    )
		    {
			free(edw->status_mesg);
			edw->status_mesg = NULL;

			if(devrec[z]->is_mounted)
			{
			    sprintf(stringa,
 "Device `%s' Total: %ld  Used: %ld  Available: %ld  Ratio: %.0f%%",
			        devrec[z]->label,
			        devrec[z]->capacity_total,
			        devrec[z]->capacity_used,
			        devrec[z]->capacity_available,
			        devrec[z]->capacity_percent
			    );
			}
			else
			{
			    sprintf(stringa,
 "Device `%s' not mounted (right click to mount on directory `%s').",
				devrec[z]->label,
				devrec[z]->path
			    );
			}
			edw->status_mesg = StringCopyAlloc(stringa);
			EDWDraw(edw, DRAW_AMOUNT_STATUS);
			break;
		    }

		    x += DEVREC_ENTRY_SPACING;
		}

		events_handled++;
		return(events_handled);
	    }

	    /* Status window (padlock hints) */
            else if((event->xany.window == edw->status_win) &&
                    (endeavour_image.padlock0 != NULL)
            )
            {
                image = endeavour_image.padlock0;
		x = event->xmotion.x;
		y = event->xmotion.y;

                OSWGetWindowAttributes(edw->status_win, &wattr);
                if( (x >= ((int)wattr.width - (int)image->width - 15)) &&
                    (x < ((int)wattr.width - 15)) &&
                    (y >= 2) &&
                    (y < ((int)wattr.height - 2)) 
                )
		{
                    /* Schedual hints window map. */
                    HintWinSetSchedual(
                        widget_global.hintwin_map_delay,
                        edw->status_win
                    );

                    events_handled++;
		}
		else
		{
		    /* Unmap/unschedual hints window. */
                    HintWinUnmap();
		}
		return(events_handled);
	    }
          
            break;

	  /* ********************************************************* */
	  case Expose:
	    if(event->xany.window == edw->toplevel)
	    {
		EDWDraw(edw, DRAW_AMOUNT_COMPLETE);

                events_handled++;
		return(events_handled);
	    }
            else if(event->xany.window == edw->tool_ribbon_win)
	    {
                EDWDraw(edw, DRAW_AMOUNT_COMPLETE);

                events_handled++;
                return(events_handled);
	    }
            else if(event->xany.window == edw->dir_win)
            {
                EDWDraw(edw, DRAW_AMOUNT_DIR_LIST);

                events_handled++;
                return(events_handled);
            }   
            else if(event->xany.window == edw->file_win)
            {
                EDWDraw(edw, DRAW_AMOUNT_FILE_LIST);

                events_handled++;
                return(events_handled);
            }
            else if(event->xany.window == edw->status_win)
            {
                EDWDraw(edw, DRAW_AMOUNT_STATUS);

                events_handled++;
                return(events_handled);
            }   
	    break;

	  /* ********************************************************* */
	  case ConfigureNotify:
	    if((event->xany.window == edw->toplevel) ||
               (event->xany.window == edw->dir_win) ||
               (event->xany.window == edw->file_win)
            )
	    {
		EDWResize(edw);
                events_handled++;
	    }
	    break;

	  /* ********************************************************* */
	  case FocusIn:
	    if(event->xany.window == edw->toplevel)
	    {
		edw->is_in_focus = 1;
		events_handled++;
	    }
	    else
	    {
		edw->is_in_focus = 0;
		/* Do not report any event for focus out. */
	    }
	    return(events_handled);
	    break;

          /* ********************************************************* */
          case FocusOut:
            if(event->xany.window == edw->toplevel)
            {
                edw->is_in_focus = 0;
                events_handled++;
            }
            return(events_handled);
            break;

          /* ********************************************************* */
          case VisibilityNotify:
            if(event->xany.window == edw->toplevel)
            {
                edw->visibility_state = event->xvisibility.state;

                events_handled++;
		return(events_handled);
            }
            break;

	  /* ********************************************************* */
	  case LeaveNotify:
            if(event->xany.window == edw->toplevel)
            {
		/* Set back the cursor on toplevel if leaving toplevel
		 * so child windows get proper pointer when not in drag.
		 */
		if(!drag_record[0].state)
                    OSWUnsetWindowCursor(edw->toplevel);

                events_handled++;
                return(events_handled);
            }
	    else if(event->xany.window == edw->status_win)
	    {
                /* Unmap/unschedual hints window. */
                HintWinUnmap();

		events_handled++;		
		return(events_handled);
	    }
	    break;
	}


	/* Redraw as needed. */
	if(events_handled > 0)
	{
	    EDWDraw(edw, DRAW_AMOUNT_COMPLETE);
	    return(events_handled);
	}


	/* ************************************************************ */
	/* Manage widgets. */

	/* Change values prompt. */
	if(events_handled == 0)
	{
            events_handled += PromptManage(
                &edw->cv_prompt, event
	    );
	}
	/* Location prompt. */
	if(events_handled == 0)
	{
	    events_handled += PromptManage(
		&(edw->loc_prompt), event);

	    /* Must unselect files and dirs if prompt location changed. */
	    if((events_handled > 0) &&
               ( (event->type == KeyPress)
               ) &&
               ( ((int)edw->total_sel_dirs > 0) ||
                 ((int)edw->total_sel_files > 0)
               )
            )
	    {
		EDW_UNSELECTALLDIRS(edw);
		EDW_UNSELECTALLFILES(edw);
		EDWDraw(edw, DRAW_AMOUNT_LISTS);
	    }
	}
        /* Command prompt. */
        if(events_handled == 0)
        {
            events_handled += PromptManage(
                &edw->cmd_prompt, event);

            /* Must unselect files and dirs if prompt location changed. */
            if( (events_handled > 0) &&
                ( (event->type == KeyPress) ||
                  (event->type == KeyRelease) ||
                  (event->type == ButtonPress)
                ) &&
                ( ((int)edw->total_sel_dirs > 0) ||
                  ((int)edw->total_sel_files > 0)
                )
            )
            {
                EDW_UNSELECTALLDIRS(edw);
                EDW_UNSELECTALLFILES(edw);
                EDWDraw(edw, DRAW_AMOUNT_LISTS);
            }
        }


        /* Dir window scroll bars. */
        if(events_handled == 0)
        {
            OSWGetWindowAttributes(edw->dir_win, &wattr);
            events_handled += SBarManage(&edw->dir_win_sb,
                wattr.width,
                wattr.height,
                edw->max_dir_width,
                (edw->total_dirs + 2) * ROW_SPACING,
                event
            );
         
            if(events_handled > 0)
                EDWDraw(edw, DRAW_AMOUNT_DIR_LIST);
        }
 
        /* File window scroll bars. */
        if(events_handled == 0)
        {
            OSWGetWindowAttributes(edw->file_win, &wattr);
            events_handled += SBarManage(&edw->file_win_sb,
                wattr.width, 
                wattr.height,  
                wattr.width,
                (edw->total_files + 2) * ROW_SPACING,
                event
            );
             
            if(events_handled > 0)
                EDWDraw(edw, DRAW_AMOUNT_FILE_LIST);
        }


	/* *********************************************************** */
	/* Tool ribbon buttons. */

	/* Folder back. */
	if(events_handled == 0)
        {
	    events_handled += PBtnManage(
		&edw->folder_back_btn,
		event
	    );

	    if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
	    {
		EDWDoGoParentDir(edw);
	    }
	}
	/* Refresh. */
        if(events_handled == 0)
	{
            events_handled += PBtnManage(
                &edw->folder_refresh_btn,
                event 
            );

            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
            {
		EDWDoRefresh(edw);
	    }
	}

	/* Create new folder. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &edw->new_folder_btn,
                event
            );
            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
	    {
		EDWDoQCreateNewDir(edw);
	    }
        }
	/* Edit properties. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &edw->properties_btn,
                event
            );

            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
	    {
		EDWDoEditProperties(edw);
	    }
        }
	/* Move button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &edw->move_btn,   
                event
            );
            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
            {
		EDWDoMove(edw);
            }
        }
	/* Copy button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &edw->copy_btn,
                event
            );
            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
            {
		EDWDoCopy(edw);
	    }            
        }
	/* Recycle. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &edw->recycle_btn,
                event
            );

            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )       
            {
		EDWDoRecycle(edw);
	    }
        }

	/* New Endeavour browser. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &edw->new_browser_btn,
                event
            );

            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
            {
		EDWDoNewBrowser(edw);
	    }
        }
        /* Tile browser windows. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &edw->tile_btn,
                event
            );

            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
            {
		EDWDoTileBrowser(edw);
            }
        }
	/* Map recycle bin window. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &edw->recbin_map_btn,
                event
            );
            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
            {
		EDWDoMapRecycleBin(edw);
	    }
        }

        /* View output. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &edw->output_btn,
                event
            );
            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
            {
		/* Create an internal text viewer and view output. */
		DoCreateV(fname.prog_output);
            }
        }
        /* View button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &edw->view_btn,
                event
            );
            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
            {
		EDWExecuteSelectedFiles(edw, True);
            }
        }
	/* Edit button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &edw->edit_btn,
                event
            );
            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
            {
                /* Edit selected files. */
                for(i = 0; i < edw->total_sel_files; i++)
                {
                    n = edw->sel_file[i];
                    if((n < 0) || (n >= edw->total_files)) continue;
                    if(edw->file[n] == NULL) continue;

		    /* Run text editor. */
                    FileOpRunTextEditor(edw->file[n]->name);
                }
            }
        }
	/* Kill button. */
        if(events_handled == 0) 
        {
            events_handled += PBtnManage(
                &edw->kill_btn,
                event
            );
            if((event->type == ButtonRelease) &&
               (events_handled > 0)
            )
            {
/*   Kill last process, use process history item 0 for now since
 *   it is always last.
 */
		ProcHistKill(0);
            }   
        }     


	/* ********************************************************** */
	/* Quick menus. */
	if(events_handled == 0)
        {
            events_handled += MenuManage(
                &edw->dir_qmenu,
                event
            );
	}
        if(events_handled == 0)
        {
            events_handled += MenuManage(
                &edw->file_qmenu,  
                event
            );
        }






	return(events_handled);
}



int EDWManageAll(event_t *event)
{
	static int i, events_handled;


	/* Error checks. */
	if( !IDC() ||
            (event == NULL)
	)
	    return(0);


	/* ********************************************************** */
	events_handled = 0;

	for(i = 0; i < (int)total_endeavour_wins; i++)
	{
	    /* Skip unallocated ones. */
	    if(endeavour_win[i] == NULL)
		continue;

	    /* ClientMessage must be handled here. */
            if((event->type == ClientMessage) &&
               OSWIsEventDestroyWindow(endeavour_win[i]->toplevel, event)
	    )
            {
		/* Unmap and reset resources which may referance us. */
		ComfDialogUnmap();
		if((int)total_op_dialogs > 0)
		    OPDialogUnmap(op_dialog[0]);

		/* Delete then reclaim memory. */
		EDWDelete(i);
		EDWReclaimMemory();

		OSWPurgeAllEvents();

                events_handled++;
                break;	/* Break since events are now messed up due to delete. */
            }


	    events_handled += EDWManage(endeavour_win[i], event);
	}


	return(events_handled);
}


void EDWMap(endeavour_win_struct *edw)
{
	if(edw == NULL)
	    return;

	UnfocusAllTopLvlWins();

	edw->map_state = 0;
	EDWDraw(edw, DRAW_AMOUNT_COMPLETE);

	/* Make it in focus. */
	edw->is_in_focus = 1;

	return;
}



void EDWUnmap(endeavour_win_struct *edw)
{
        if(edw == NULL)
            return;


        PromptUnmap(&edw->cv_prompt);

	PBtnUnmap(&edw->output_btn);
        PBtnUnmap(&edw->kill_btn);
        PBtnUnmap(&edw->view_btn);
        PBtnUnmap(&edw->edit_btn);

        PromptUnmap(&edw->cmd_prompt);

        PromptUnmap(&edw->loc_prompt);


        PBtnUnmap(&edw->folder_back_btn);
        PBtnUnmap(&edw->folder_refresh_btn);
        PBtnUnmap(&edw->new_folder_btn);
        PBtnUnmap(&edw->properties_btn);
        PBtnUnmap(&edw->move_btn);
        PBtnUnmap(&edw->copy_btn);
        PBtnUnmap(&edw->recycle_btn);
        PBtnUnmap(&edw->new_browser_btn);
        PBtnUnmap(&edw->tile_btn);
        PBtnUnmap(&edw->recbin_map_btn);


	/* Unmap toplevel. */
	OSWUnmapWindow(edw->toplevel);        
        edw->map_state = 0;
        edw->is_in_focus = 0;


	return;
}



/*
 *	Destroys substuctures, widgets, windows, objects, etc.
 *	(But does NOT deallocate the edw itself!)
 */
void EDWDestroy(endeavour_win_struct *edw)
{
	if(edw == NULL)
	    return;


	/* Status message. */
	free(edw->status_mesg);
	edw->status_mesg = NULL;

        /* Free all allocated objects and selections. */
	EDWFreeAllListings(edw);

	/* Free change values target path. */
	free(edw->cv_src_path);
	edw->cv_src_path = NULL;


	if(IDC())
	{
	    PromptDestroy(&edw->cv_prompt);


	    PBtnDestroy(&edw->output_btn);
	    PBtnDestroy(&edw->kill_btn);
	    PBtnDestroy(&edw->view_btn);
	    PBtnDestroy(&edw->edit_btn);

	    PromptDestroy(&edw->cmd_prompt);
	    OSWDestroyWindow(&edw->cmd_tool_ribbon);
	    OSWDestroyPixmap(&edw->cmd_tool_ribbon_buf);

	    PromptDestroy(&edw->loc_prompt);
            OSWDestroyWindow(&edw->loc_tool_ribbon);
            OSWDestroyPixmap(&edw->loc_tool_ribbon_buf);

	    OSWDestroyWindow(&edw->split_cur);

	    /* Dir win. */
	    SBarDestroy(&edw->dir_win_sb);
            OSWDestroyWindow(&edw->dir_win);
	    OSWDestroyPixmap(&edw->dir_win_buf);
	    MenuDestroy(&edw->dir_qmenu);

	    /* File win. */
            OSWDestroyWindow(&edw->file_heading_win);
	    OSWDestroyPixmap(&edw->file_heading_win_buf);
	    SBarDestroy(&edw->file_win_sb);
            OSWDestroyWindow(&edw->file_win);
	    OSWDestroyPixmap(&edw->file_win_buf);
            MenuDestroy(&edw->file_qmenu);


	    /* Status win. */
	    HintWinDeleteMessage(edw->status_win);	/* Hints data. */
            OSWDestroyPixmap(&edw->status_win_buf);
            OSWDestroyWindow(&edw->status_win);


	    /* Tool ribbon (buttons and itself). */
            PBtnDestroy(&edw->folder_back_btn);   
            PBtnDestroy(&edw->folder_refresh_btn);
            PBtnDestroy(&edw->new_folder_btn);
            PBtnDestroy(&edw->properties_btn);
            PBtnDestroy(&edw->move_btn);
            PBtnDestroy(&edw->copy_btn);   
            PBtnDestroy(&edw->recycle_btn);
            PBtnDestroy(&edw->new_browser_btn);
            PBtnDestroy(&edw->tile_btn);
            PBtnDestroy(&edw->recbin_map_btn);
	    OSWDestroyPixmap(&edw->tool_ribbon_win_buf);
            OSWDestroyWindow(&edw->tool_ribbon_win);


	    /* Toplevel. */
	    OSWDestroyWindow(&edw->toplevel);
	}


        edw->map_state = 0;
        edw->visibility_state = VisibilityFullyObscured;
        edw->x = 0;
        edw->y = 0;
        edw->width = DEF_WIDTH;
        edw->height = DEF_HEIGHT;
        edw->is_in_focus = 0;
        edw->disabled = False;
        edw->dir_win_width = (double)DEF_DIR_WIN_WIDTH / 100;


	return;
}
