/*
              Endeavour Browser: File and Directory List Management

	Functions:

	image_t *EDW_GETDIRIMAGE(char *path, int sel)
	int EDWDoesDirHaveSubDirs(char *path)
	int EDWMatchDirFromPath(endeavour_win_struct *edw, char *path)
        int EDWMatchExistingDirFromPath(endeavour_win_struct *edw, char *path)
	int EDWScrollToDirItem(endeavour_win_struct *edw, int dir_num)
        int EDWScrollToFileItem(endeavour_win_struct *edw, int file_num)
	int EDWUpdateDirWinWidth(endeavour_win_struct *edw)

        int EDWInitToplevelDir(endeavour_win_struct *edw)
	int EDWUpdateDirItemName(endeavour_win_struct *edw,
	    char *old_name, char *new_name)

	void EDWDeleteDirItemPath(endeavour_win_struct *edw, char *path)

	void EDWFreeFileListItem(object_struct **obj)
	void EDWFreeDirListItem(edw_dir_struct **dir)

	void EDWFreeFileListing(endeavour_win_struct *edw)
	void EDWFreeDirListing(endeavour_win_struct *edw)
	void EDWFreeAllListings(endeavour_win_struct *edw)

	int EDWExpandDir(endeavour_win_struct *edw, int n)
	void EDWColapseDir(endeavour_win_struct *edw, int n)

	int EDWRefreshDirItem(endeavour_win_struct *edw, int dir_num)
        int EDWRefreshFileItem(endeavour_win_struct *edw, int file_num)
	void EDWRefreshDirList(endeavour_win_struct *edw)
	void EDWRefreshFileList(endeavour_win_struct *edw, char *path)
	void EDWRefreshAllLists(endeavour_win_struct *edw)


	---







 */

#include "endeavour.h"
#include "edwdefaults.h"

#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))



/* ******************************************************************* */

/*
 *	Returns a pointer to an image for a dir icon referanced in
 *	path.  The selected version will be returned if sel is 1 and
 *	the unselected version will be returned if sel is 0.
 */
image_t *EDW_GETDIRIMAGE(char *path, int sel)
{
	int i;
	struct stat stat_buf;


	if(path == NULL) return(NULL);


	/* ********************************************************* */
	/* Special Case: Toplevel */
	if(!strcmp(path, "/"))
	{
            if(lstat(path, &stat_buf)) return(NULL);

            if(S_ISLNK(stat_buf.st_mode))
            {
                if(sel)
                    return(widget_global.linkicon_selected_img);
                else
                    return(widget_global.linkicon_normal_img);
            }

            if(sel)
                return(endeavour_image.comp_s1);
            else
                return(endeavour_image.comp_s0);


	    return(NULL);
	}


	/* ********************************************************* */
	for(i = 0; i < total_devrecs; i++)
	{
	    if(devrec[i] == NULL) continue;
	    if(devrec[i]->path == NULL) continue;

	    if(!strcmp(devrec[i]->path, path)) break;
	}
	if(i < total_devrecs)
	{
	    switch(devrec[i]->image_code)
	    {
              case RECDEV_IMG_NFS:
		if(sel)
		     return(endeavour_image.dev_hd1);
		else
                     return(endeavour_image.dev_hd0);
		break;

              case RECDEV_IMG_MEMORY:
                if(sel)
                     return(endeavour_image.dev_hd1);
                else
                     return(endeavour_image.dev_hd0);
                break;

              case RECDEV_IMG_ZIP:
                if(sel)
                     return(endeavour_image.dev_zip1);
                else
                     return(endeavour_image.dev_zip0);
                break;

              case RECDEV_IMG_TAPE:
                if(sel)
                     return(endeavour_image.dev_fd1);
                else
                     return(endeavour_image.dev_fd0);
                break;

              case RECDEV_IMG_CDROM:
                if(sel)
                     return(endeavour_image.dev_cdrom1);
                else
                     return(endeavour_image.dev_cdrom0);
                break;

              case RECDEV_IMG_FD:
                if(sel)
                     return(endeavour_image.dev_fd1);
                else
                     return(endeavour_image.dev_fd0);
                break;

              default: /* Default to hard drive image. */
                if(sel)
                     return(endeavour_image.dev_hd1);
                else  
                     return(endeavour_image.dev_hd0);
                break;
	    }
	}
	else
	{
	    if(lstat(path, &stat_buf)) return(NULL);

	    if(S_ISLNK(stat_buf.st_mode))
	    {
                if(sel)
                    return(widget_global.linkicon_selected_img);
                else
                    return(widget_global.linkicon_normal_img);
	    }

	    if(sel)
	        return(widget_global.diricon_selected_img);
	    else
		return(widget_global.diricon_normal_img);
	}


	return(NULL);
}


/*
 *	Returns 1 if the givin path is a dir and has additional dirs
 *	within it.
 */
int EDWDoesDirHaveSubDirs(char *path)
{
	return(DirHasSubDirs(path));
}



/*
 *	Simple macro to create the toplevel dir (entry #0) on edw.
 */
int EDWInitToplevelDir(endeavour_win_struct *edw)
{
	struct stat stat_buf;


	/* Error checks. */
	if(edw == NULL) return(-1);

	/* Already has dirs? */
	if(edw->total_dirs > 0) return(0);


	/* ******************************************************** */

	/* Add toplevel to dir list. */
        edw->dir = NULL;
        edw->total_dirs = 1;
        
        edw->dir = (edw_dir_struct **)realloc(edw->dir,
            edw->total_dirs * sizeof(edw_dir_struct *)
        );
        if(edw->dir == NULL)
        {
            edw->total_dirs = 0;
            return(-1);
        }
        edw->dir[0] = (edw_dir_struct *)calloc(1, sizeof(edw_dir_struct));
        if(edw->dir[0] == NULL)
        {
            edw->total_dirs--;
            return(-1);
        }
        
        edw->dir[0]->name = StringCopyAlloc("/");
        edw->dir[0]->fullname = StringCopyAlloc("/"); 
          
        edw->dir[0]->link_dest = NULL;
        if(stat("/", &stat_buf)) return(-1);
            
        edw->dir[0]->level = 0;                 /* Starts on level 0. */
        edw->dir[0]->expandable = 1;            /* Should be expandable. */
        edw->dir[0]->expanded = 0;
        edw->dir[0]->x = 0;

        edw->dir[0]->sel_image = EDW_GETDIRIMAGE("/", 1);
        edw->dir[0]->unsel_image = EDW_GETDIRIMAGE("/", 0);

        edw->dir[0]->inode = stat_buf.st_ino;
        edw->dir[0]->mode = stat_buf.st_mode;
        edw->dir[0]->nlink = stat_buf.st_nlink;
        edw->dir[0]->uid = stat_buf.st_uid;
        edw->dir[0]->gid = stat_buf.st_gid;
        edw->dir[0]->rdevice = stat_buf.st_rdev;
        edw->dir[0]->size = stat_buf.st_size;
        edw->dir[0]->last_accessed = stat_buf.st_atime;
        edw->dir[0]->last_modified = stat_buf.st_mtime;
        edw->dir[0]->create_time = stat_buf.st_ctime; 


	return(0);
}



/*
 *	Searches and returns the dir item number matching
 *	path.    Returns -1 if no match can be made or error.
 */
int EDWMatchDirFromPath(endeavour_win_struct *edw, char *path)
{
	int i, strc;
	char **strv;
	int level;
	int dir_item = -1;


	/* Error checks. */
	if((edw == NULL) ||
           (path == NULL)
	)
	    return(-1);
	if(!ISPATHABSOLUTE(path))
	    return(-1);


	/* ************************************************************* */

	/* Explode names out of path. */
	strv = strchrexp(path + 1, '/', &strc);
	if((strv == NULL) || (strc <= 0))
	{
	    /* Return entry #0 if path contains no names. */
	    return(0);
	}


	/* Start on level 1, skipping toplevel. */
	level = 1;
	EDWExpandDir(edw, 0);

	for(i = 0; i < edw->total_dirs; i++)
	{
	    if(edw->dir[i] == NULL)
		continue;
	    if(edw->dir[i]->name == NULL)
		continue;

	    if(edw->dir[i]->level == level)
	    {	    
		/* Same level, check for match. */
		if((level > 0) && (level <= strc))
		{
		    if(!strcmp(edw->dir[i]->name, strv[level - 1]))
		    {
			EDWExpandDir(edw, i);
			dir_item = i;
		        level++;
		    }
		}
	    }

            /*   Level should equal number of strings gotten from
             *   exploding path.
             */
            if(level > strc)
	    {
		dir_item = i;
		break;
	    }
	}


	/* Free search strings. */
	for(i = 0; i < strc; i++)
	    free(strv[i]);
	free(strv);


	return(dir_item);
}



/*
 *	Similar to EDWMatchDirFromPath() except this function will not
 *	expand directories when searching for the path.
 *
 *	Also unlike EDWMatchDirFromPath() this function will return
 *	either a number for an exact match or -1 if it cannot be found,
 *	absolutely no approximations will be made.
 */
int EDWMatchExistingDirFromPath(endeavour_win_struct *edw, char *path)
{
        int i;

                
        /* Error checks. */
        if((edw == NULL) ||
           (path == NULL)
        )
            return(-1);
        if(!ISPATHABSOLUTE(path))
            return(-1);


        for(i = 0; i < edw->total_dirs; i++)
        {
            if(edw->dir[i] == NULL)
		continue;
            if(edw->dir[i]->fullname == NULL)
		continue;

	    if(!strcmp(path, edw->dir[i]->fullname))
		return(i);
	}


	return(-1);
}



/*
 *	Scrolls the dir window scroll bar to the position of dir_num.
 */
int EDWScrollToDirItem(endeavour_win_struct *edw, int dir_num)
{
	int y_max;
	win_attr_t wattr;


	/* Error checks. */	
	if(!IDC() ||
           (edw == NULL)
	)
	    return(-1);
	if((dir_num < 0) || (dir_num >= edw->total_dirs))
	    return(-1);
	if(edw->dir[dir_num] == NULL)
	    return(-1);


        OSWGetWindowAttributes(edw->dir_win, &wattr);

	/* Calculate x position. */
	if((int)edw->max_dir_width > (int)wattr.width)
	    edw->dir_win_sb.x_win_pos = edw->dir[dir_num]->x -
	        DIR_ROUTE_WIDTH - DIR_ICON_WIDTH + LIST_MARGIN;
	else
	    edw->dir_win_sb.x_win_pos = 0;
	if(edw->dir_win_sb.x_win_pos < 0)
	    edw->dir_win_sb.x_win_pos = 0;

        /* Calculate y position. */
	y_max = (edw->total_dirs * ROW_SPACING) - (int)wattr.height
	    + 30;	/* Lil extra tolorance for horiz scrollbar. */

        edw->dir_win_sb.y_win_pos = dir_num * ROW_SPACING;
	if(edw->dir_win_sb.y_win_pos > y_max)
	    edw->dir_win_sb.y_win_pos = y_max;
	if(edw->dir_win_sb.y_win_pos < 0)
	    edw->dir_win_sb.y_win_pos = 0;


	return(0);
}


/*
 *      Scrolls the file window scroll bar to the position of file_num.
 */
int EDWScrollToFileItem(endeavour_win_struct *edw, int file_num)
{
        if(edw == NULL) return(-1);
        if((file_num < 0) || (file_num >= edw->total_files)) return(-1);

        edw->file_win_sb.y_win_pos = file_num * ROW_SPACING;

        return(0);
}



/*
 *	Procedure to update the maximum dir window width
 *	on edw.
 *
 *	Note: This does NOT update the dir and file list split coeff.
 */
int EDWUpdateDirWinWidth(endeavour_win_struct *edw)
{
	int i, n;
	int highest_x_pos = 0;

	if(edw == NULL) return(-1);
	if(edw->total_dirs <= 0) return(0);


	/* Search through dirs and get highest item and position. */
	for(i = 0; i < edw->total_dirs; i++)
	{
	    if(edw->dir[i] == NULL) continue;
	    if(edw->dir[i]->name == NULL) continue;

	    /* Calculate length of this entry. */
	    n = edw->dir[i]->x + ((strlen(edw->dir[i]->name) + 2) *
                CHAR_WIDTH) +
                ((edw->dir[i]->link_dest == NULL) ? 0 :
                    (strlen(edw->dir[i]->link_dest) + 4) * CHAR_WIDTH
                )
	    ;

	    if(n > highest_x_pos)
		highest_x_pos = n;
	}


	/* Set new maximum dir window width. */
	edw->max_dir_width = highest_x_pos + (LIST_MARGIN * 2) +
            DIR_ROUTE_WIDTH + DIR_ICON_WIDTH;


	return(0);
}


/*
 *	Searches edw's dir items list for an item matching old_name,
 *	and changes that item's name to new_name.
 *
 *	old_name must be an absolute path.
 */
int EDWUpdateDirItemName(
	endeavour_win_struct *edw,
	char *old_name,
	char *new_name
)
{
	int x, level, expanded, expandable;
	int i;
	char tmp_old_name[PATH_MAX + NAME_MAX];
        char tmp_new_name[PATH_MAX + NAME_MAX];
	char tmp_new_child_name[PATH_MAX + NAME_MAX];
	char tmp_path[PATH_MAX + NAME_MAX];
	char *strptr;
	struct stat stat_buf;


	/* Error checks. */
	if((edw == NULL) ||
           (old_name == NULL) ||
           (new_name == NULL)
	)
	    return(-1);


	strncpy(tmp_old_name, old_name, PATH_MAX + NAME_MAX);
	tmp_old_name[PATH_MAX + NAME_MAX - 1] = '\0';
	StringStripSpaces(tmp_old_name);
	if(!ISPATHABSOLUTE(tmp_old_name))
	{
	    fprintf(stderr,
		"EDWUpdateDirItemName(): Old name `%s' not absolute path.\n",
		tmp_old_name
	    );
	    return(-2);
	}

	strncpy(tmp_new_name, new_name, PATH_MAX + NAME_MAX);
	tmp_new_name[PATH_MAX + NAME_MAX - 1] = '\0';
        StringStripSpaces(tmp_new_name);
	if(!ISPATHABSOLUTE(tmp_new_name))
	{
	    strptr = GetParentDir(tmp_old_name);
	    if(strptr == NULL)
		return(-1);

	    strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
	    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

	    strptr = PrefixPaths(tmp_path, tmp_new_name);
            if(strptr == NULL)
                return(-1);
	    strncpy(tmp_new_name, strptr, PATH_MAX + NAME_MAX);
	    tmp_new_name[PATH_MAX + NAME_MAX - 1] = '\0';
	}

	strncpy(tmp_new_child_name, tmp_new_name, PATH_MAX + NAME_MAX);
	tmp_new_child_name[PATH_MAX + NAME_MAX - 1] = '\0';
	StripAbsolutePath(tmp_new_child_name);


	/* ***************************************************** */

	for(i = 0; i < edw->total_dirs; i++)
	{
	    if(edw->dir[i] == NULL) continue;
	    if(edw->dir[i]->fullname == NULL) continue;

	    /* Skip if name does not match. */
	    if(strcmp(edw->dir[i]->fullname, tmp_old_name)) continue;


	    /* Name matches, free old dir item and get new one. */

	    /* Record old positions. */
	    level = edw->dir[i]->level;
	    expandable = edw->dir[i]->expandable;
            expanded = edw->dir[i]->expanded;
            x = edw->dir[i]->x;

	    /* Free listing. */
	    EDWFreeDirListItem(&edw->dir[i]);

	    edw->dir[i] = (edw_dir_struct *)calloc(1, sizeof(edw_dir_struct));
	    if(edw->dir[i] == NULL) break;

	    edw->dir[i]->name = StringCopyAlloc(tmp_new_child_name);
	    edw->dir[i]->fullname = StringCopyAlloc(tmp_new_name);
            edw->dir[i]->link_dest = GetAllocLinkDest(tmp_new_name);

            edw->dir[i]->level = level;
            edw->dir[i]->expandable = expandable;
            edw->dir[i]->expanded = expanded;
            edw->dir[i]->x = x;

            edw->dir[i]->sel_image = EDW_GETDIRIMAGE(tmp_new_name, 1);
            edw->dir[i]->unsel_image = EDW_GETDIRIMAGE(tmp_new_name, 0);

            if(!lstat(edw->dir[i]->fullname, &stat_buf))
	    {
                edw->dir[i]->inode = stat_buf.st_ino;
                edw->dir[i]->mode = stat_buf.st_mode;
                edw->dir[i]->nlink = stat_buf.st_nlink;
                edw->dir[i]->uid = stat_buf.st_uid;
                edw->dir[i]->gid = stat_buf.st_gid;
                edw->dir[i]->rdevice = stat_buf.st_rdev;
                edw->dir[i]->size = stat_buf.st_size;
                edw->dir[i]->last_accessed = stat_buf.st_atime; 
                edw->dir[i]->last_modified = stat_buf.st_mtime;
                edw->dir[i]->create_time = stat_buf.st_ctime;
	    }

	    break;
	}


	return(0);
}



/*
 *	Searches edw's dir list for item's fullname matching path.
 *	If match is found the item is deleted from the list,
 *	the actual dir is not removed ofcourse.
 *
 *	Note that path must be an absolute path as the member fullname
 *	in the dir list is also an absolute path.
 */
void EDWDeleteDirItemPath(endeavour_win_struct *edw, char *path)
{
	int i, n;

	if((edw == NULL) || (path == NULL)) return;
	if(!ISPATHABSOLUTE(path)) return;


	for(i = 0; i < edw->total_dirs; i++)
	{
	    if(edw->dir[i] == NULL) continue;
	    if(edw->dir[i]->fullname == NULL) continue;

	    if(!strcmp(path, edw->dir[i]->fullname))
	    {
		/* Dir item matches, delete it. */
		EDWFreeDirListItem(&edw->dir[i]);

		/* Realign pointers. */
		for(n = i; (n + 1) < edw->total_dirs; n++)
		{
		    edw->dir[n] = edw->dir[n + 1];
		}

		/* Reallocate pointers. */
		edw->total_dirs -= 1;
		edw->dir = (edw_dir_struct **)realloc(edw->dir,
		    edw->total_dirs * sizeof(edw_dir_struct *)
		);
		if(edw->dir == NULL)
		{
		    edw->total_dirs = 0;
		    break;
		}
		continue;
	    }
	}


	return;
}

/*
 *	Frees an object_struct structure and all its substructures,
 *	then sets the pointed location to NULL.
 */
void EDWFreeFileListItem(object_struct **obj)
{
	if(obj == NULL) return;
	if(*obj == NULL) return;


	free((*obj)->name);
	free((*obj)->link_dest);

	free(*obj);
	*obj = NULL;


	return;
}


/*
 *      Frees an edw_dir_struct structure and all its substructures,
 *      then sets the pointed location to NULL.
 */
void EDWFreeDirListItem(edw_dir_struct **dir)  
{
        if(dir == NULL) return;
        if(*dir == NULL) return;

 
        free((*dir)->name);
        free((*dir)->fullname);
        free((*dir)->link_dest);

        free(*dir);
        *dir = NULL;

        return;
}       



/*
 *	Macro to unselect all files and free all files listed
 *	on edw's file list.
 */
void EDWFreeFileListing(endeavour_win_struct *edw) 
{
	int i;


	/* Error checks. */
        if(edw == NULL) return;


	/* Unselect all files. */
	EDW_UNSELECTALLFILES(edw);

	/* Free each file object struct. */
	for(i = 0; i < edw->total_files; i++)
	{
	    EDWFreeFileListItem(&edw->file[i]);
	}

	free(edw->file);
	edw->file = NULL;

	edw->total_files = 0;


	return;
}


/*
 *      Macro to unselect all files and free all files listed
 *      on edw's file list.
 */
void EDWFreeDirListing(endeavour_win_struct *edw)
{
        int i;


        /* Error checks. */
        if(edw == NULL) return;


        /* Unselect all dirs. */
        EDW_UNSELECTALLDIRS(edw);


        /* Free each dir object struct. */
        for(i = 0; i < edw->total_dirs; i++)
        {
            EDWFreeDirListItem(&edw->dir[i]);
        }
        
        free(edw->dir);
        edw->dir = NULL;
  
        edw->total_dirs = 0;


        return;
}  


/*
 *      Frees both file and dir list items on edw.
 */
void EDWFreeAllListings(endeavour_win_struct *edw)
{
	if(edw == NULL) return;

	EDWFreeDirListing(edw);
	EDWFreeFileListing(edw);


        return;
} 



/*
 *	Expands the directory item n on edw by one level if
 *	it is expandable.  Updates the directory list, leaves file
 *	list alone.
 */
int EDWExpandDir(endeavour_win_struct *edw, int n)
{
	int i, j;
	char *strptr;
	struct stat stat_buf;
	int start_level;

        edw_dir_struct **sub_dir;
	int total_sub_dirs;		/* Including links to dirs. */
	edw_dir_struct *ds_ptr;
	char parent[PATH_MAX + NAME_MAX];
	char tmp_path[PATH_MAX + NAME_MAX];
	char **dent;
	int total_dents;


	if(edw == NULL) return(-1);
	if((n < 0) || (n >= edw->total_dirs)) return(-1);
	if(edw->dir == NULL) return(-1);
	ds_ptr = (edw_dir_struct *)edw->dir[n];

	if(ds_ptr == NULL) return(-1);
	if(ds_ptr->fullname == NULL) return(-1);

	/* Is it expandable? */
	if(!ds_ptr->expandable)
	{
	    /* It is set not expandable, but just in case, check if it is. */
	    if(EDWDoesDirHaveSubDirs(ds_ptr->fullname))
	        ds_ptr->expandable = 1;
	    else
		return(-2);
	}

	/* Is it already expanded? */
	if(ds_ptr->expanded) return(-2);


	/* ******************************************************** */

	/* Mark dir as expanded. */
	ds_ptr->expanded = 1;

	start_level = ds_ptr->level;

	strncpy(parent, ds_ptr->fullname, PATH_MAX + NAME_MAX);
	parent[PATH_MAX + NAME_MAX - 1] = '\0';

	/* Get directory listing. */
	dent = (char **)GetDirEntNames(parent);
	if(dent == NULL)
	{
	    /* Hmm, no dir entries? then set givin dir not expandable. */
	    ds_ptr->expandable = 0;
	    ds_ptr->expanded = 0;

	    return(-2);
	}

        /* Count dents. */
        total_dents = 0;
        while(dent[total_dents] != NULL)
        {
            total_dents++;
        }
        
        /* Sort dents. */
        dent = StringQSort(dent, total_dents);
        if(dent == NULL)
	    return(-1);


	/* Get subdirs. */
	total_dents = 0;
        sub_dir = NULL;
	total_sub_dirs = 0;
	while(dent[total_dents] != NULL)
	{
	    j = total_dents;
	    total_dents++;

	    /* Skip current and parent directory notations. */
	    if(!strcmp(dent[j], ".")) continue;
            if(!strcmp(dent[j], "..")) continue;

            strptr = PrefixPaths(parent, dent[j]);
            if(strptr == NULL) continue;
            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

	    /* Skip non dirs and links that do not point to dirs. */
	    if(!ISPATHDIR(tmp_path)) continue;


	    /* Is a dir or a link to a dir, allocate new subdir. */
	    i = total_sub_dirs;
	    total_sub_dirs++;
	    sub_dir = (edw_dir_struct **)realloc(sub_dir,
		total_sub_dirs * sizeof(edw_dir_struct *));
	    if(sub_dir == NULL)
	    {
		total_sub_dirs = 0;
		continue;
	    }
	    sub_dir[i] = (edw_dir_struct *)calloc(1, sizeof(edw_dir_struct));
	    if(sub_dir[i] == NULL)
	    {
		total_sub_dirs--;
		continue;
	    }

	    /* Get and set stats. */
	    lstat(tmp_path, &stat_buf);

	    sub_dir[i]->name = StringCopyAlloc(dent[j]);
	    sub_dir[i]->fullname = StringCopyAlloc(tmp_path);
	    sub_dir[i]->link_dest = GetAllocLinkDest(tmp_path);

	    sub_dir[i]->level = start_level + 1;
            sub_dir[i]->expandable = EDWDoesDirHaveSubDirs(tmp_path);
            sub_dir[i]->expanded = 0;
            sub_dir[i]->x = sub_dir[i]->level * DIR_LEVEL_INC;

	    sub_dir[i]->sel_image = EDW_GETDIRIMAGE(tmp_path, 1);
	    sub_dir[i]->unsel_image = EDW_GETDIRIMAGE(tmp_path, 0);

            sub_dir[i]->inode = stat_buf.st_ino;
            sub_dir[i]->mode = stat_buf.st_mode;
            sub_dir[i]->nlink = stat_buf.st_nlink;
            sub_dir[i]->uid = stat_buf.st_uid;
            sub_dir[i]->gid = stat_buf.st_gid;
            sub_dir[i]->rdevice = stat_buf.st_rdev;
            sub_dir[i]->size = stat_buf.st_size;
            sub_dir[i]->last_accessed = stat_buf.st_atime;
            sub_dir[i]->last_modified = stat_buf.st_mtime;
            sub_dir[i]->create_time = stat_buf.st_ctime;
	}


	/* Free directory entry names. */
	for(i = 0; i < total_dents; i++)
	    free(dent[i]);
	free(dent);


        /* ******************************************************** */
	/* Got subdir entries, now realign edw's dir list pointers. */

	/* Allocate more dir pointers on edw. */
	edw->total_dirs += total_sub_dirs;
	edw->dir = (edw_dir_struct **)realloc(edw->dir,
	    edw->total_dirs * sizeof(edw_dir_struct *));
	if(edw->dir == NULL)
	{
	    edw->total_dirs = 0;

	    for(i = 0; i < total_sub_dirs; i++)
	        EDWFreeDirListItem(&sub_dir[i]);
	    free(sub_dir);

	    return(-1);
	}


	/* Shift current dir pointers on edw. */
	for(i = edw->total_dirs - 1;
	    (i - total_sub_dirs) > n;
	    i--
	)
	{
	    if(i <= 0) break;

	    edw->dir[i] = edw->dir[i - total_sub_dirs];
	}

	/* Set new pointers to be the new subdirs. */
	for(i = 0; i < total_sub_dirs; i++)
	{
	    edw->dir[n + i + 1] = sub_dir[i];
	}

	/* Free sub dir pointers but not each sub dir structure. */
	free(sub_dir);
	sub_dir = NULL;
	total_sub_dirs = 0;


	/* Update the maximum dir window width. */
	EDWUpdateDirWinWidth(edw);


	return(0);
}


/*
 *	Colapses the directory item n and all subsiquent
 *	subdirectories on edw.
 */
void EDWColapseDir(endeavour_win_struct *edw, int n)
{
	int i;
	int end_level;
	int start_item, end_item, diff;


        if(edw == NULL) return;
        if((n < 0) || (n >= edw->total_dirs)) return;
	if(edw->dir[n] == NULL) return;

	/* Already colapsed? */
	if(!edw->dir[n]->expanded) return;

	/* Mark dir n as colapsed. */
	edw->dir[n]->expanded = 0;


	/* ************************************************************* */

	/* Get end level, the level dir n is on. */
	end_level = edw->dir[n]->level;

	start_item = n + 1;
	for(i = start_item; i < edw->total_dirs; i++)
	{
	    /* End of subdirectories? */
	    if(edw->dir[i]->level <= end_level) break;

	    EDWFreeDirListItem(&edw->dir[i]);
	}
	end_item = i;
	diff = end_item - start_item;
	if(diff < 0) return;

	for(i = start_item; (i + diff) < edw->total_dirs; i++)
	{
	    edw->dir[i] = edw->dir[i + diff];
	}

	/* Reallocate dir item pointers. */
	edw->total_dirs -= diff;
	if(edw->total_dirs <= 0)
	{
	    edw->total_dirs = 0;
	    free(edw->dir);
	    edw->dir = NULL;
	}
	else
	{
	    edw->dir = (edw_dir_struct **)realloc(edw->dir,
		edw->total_dirs * sizeof(edw_dir_struct *)
	    );
	    if(edw->dir == NULL)
	    {
		edw->total_dirs = 0;
	    }
	}

        /* Update the maximum dir window width. */
        EDWUpdateDirWinWidth(edw);


	return;
}



/*
 *	Refreshes a single dir item on edw.
 */
int EDWRefreshDirItem(endeavour_win_struct *edw, int dir_num)
{
	int i;


        /* Error checks. */
        if(edw == NULL) return(-1);
	if((dir_num < 0) || (dir_num >= edw->total_dirs)) return(-1);
	if(edw->dir[dir_num] == NULL) return(-1);
	if(edw->dir[dir_num]->fullname == NULL) return(-1);


	/* Check if it has sub dirs. */
	i = EDWDoesDirHaveSubDirs(edw->dir[dir_num]->fullname);
	if(i)
	{
	    edw->dir[dir_num]->expandable = 1;
	}
	else
	{
	    EDWColapseDir(edw, dir_num);
	    edw->dir[dir_num]->expandable = 0;
	    edw->dir[dir_num]->expanded = 0;
	}


	return(0);
}


/*
 *	Refreshes a single file item on edw.
 */
int EDWRefreshFileItem(endeavour_win_struct *edw, int file_num)
{
        /* Error checks. */
        if(edw == NULL) return(-1);
	if((file_num < 0) || (file_num >= edw->total_files)) return(-1);



        return(0);
}



/*
 *	Scans through edw's dir list and removes any non-existant
 *	dirs.
 */
void EDWRefreshDirList(endeavour_win_struct *edw)
{
	int i, n;
	int need_delete;
	struct stat stat_buf;


	if(edw == NULL)
	    return;


	for(i = 0; i < edw->total_dirs; i++)
	{
	    need_delete = 0;

	    if(edw->dir[i] == NULL)
		need_delete = 1;
	    else if(edw->dir[i]->name == NULL)
		need_delete = 1;
	    else if(edw->dir[i]->fullname == NULL)
		need_delete = 1;
	    /* Does it exist (local)? */
	    else if(lstat(edw->dir[i]->fullname, &stat_buf))
		need_delete = 1;
	    /* Does it ultimately go to a dir? */
	    else if(!ISPATHDIR(edw->dir[i]->fullname))
		need_delete = 1;

	    if(need_delete)
	    {
/*
fprintf(stderr, "Removing dir entry %i: (%s longer exists).\n",
 i, ((edw->dir[i]->fullname == NULL) ? "null" : edw->dir[i]->fullname)
);
 */

		/* Unselect dir as needed. */
		if(EDW_ISDIRSELECTED(edw, i))
		    EDW_UNSELECTDIR(edw, i);

		/* Delete dir entry i on edw. */
		EDWFreeDirListItem(&edw->dir[i]);

		/* Realign dir list pointers. */
		for(n = i; (n + 1) < edw->total_dirs; n++)
		    edw->dir[n] = edw->dir[n + 1];


		/* Reallocate dir pointers. */
		edw->total_dirs--;

		if(edw->total_dirs > 0)
		{
		    edw->dir = (edw_dir_struct **)realloc(edw->dir,
			edw->total_dirs * sizeof(edw_dir_struct *)
		    );
		    if(edw->dir == NULL)
		    {
			edw->total_dirs = 0;
		    }
		}
		else
		{
		    edw->total_dirs = 0;
		    free(edw->dir);
		    edw->dir = NULL;
		}

		/* Since numbers shifted, reduce i by 1 to check
		 * next entry that is now currently i.
		 */
		i--;
		if(i < -1)
		    i = -1;
	    }
	    /* Is valid and exists, do other tests. */
	    else
	    {
		/* Is it expandable? */
		edw->dir[i]->expandable = EDWDoesDirHaveSubDirs(
		    edw->dir[i]->fullname
		);

		if(!(edw->dir[i]->expandable) &&
                   (edw->dir[i]->expanded)
		)
		    edw->dir[i]->expanded = 0;
	    }

	    /* Do not access dir beyond this point, it might be NULL. */
	}


	return;
}



/*
 *	Unselects all files on edw, frees all file listings on edw
 *	and gets new listing from absolute path path.
 */
void EDWRefreshFileList(endeavour_win_struct *edw, char *path)
{
	int n, i;
	char *strptr;
	char tmp_path[PATH_MAX + NAME_MAX];
	char **dent;
	int total_dents;
	struct stat stat_buf;


	/* Error checks. */
        if(edw == NULL) return;
	if(path == NULL) return;


	/* Unselect all items in file list and free file list. */
	EDWFreeFileListing(edw);


	/* ********************************************************* */

	/* Get dents. */
        dent = (char **)GetDirEntNames(path);
        if(dent == NULL) return;

	/* Count dents. */
	total_dents = 0;
	while(dent[total_dents] != NULL)
	{
	    total_dents++;
	}

	/* Sort dents. */
	dent = StringQSort(dent, total_dents);
	if(dent == NULL) return;


	/* ******************************************************** */

	edw->file = NULL; edw->total_files = 0;

	for(i = 0; i < total_dents; i++)
	{
            /* Skip current and parent dir notations. */
            if(!strcmp(dent[i], ".")) continue;
            if(!strcmp(dent[i], "..")) continue;   

	    /* Get tmp_path. */
	    strptr = PrefixPaths(path, dent[i]);
	    if(strptr == NULL) continue;	    

	    strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
	    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

	    /* Skip directories. */
	    if(ISPATHDIR(tmp_path)) continue;


	    /* Get local stats. */
	    if(lstat(tmp_path, &stat_buf)) continue;

	    /* Allocate new file object structure on edw. */
	    n = edw->total_files;
	    edw->total_files++;
	    edw->file = (object_struct **)realloc(edw->file,
		edw->total_files * sizeof(object_struct *)
	    );
	    if(edw->file == NULL)
	    {
		edw->total_files = 0;
		continue;
	    }

	    edw->file[n] = (object_struct *)calloc(1, sizeof(object_struct));
	    if(edw->file[n] == NULL)
	    {
		edw->total_files--;
		continue;
	    }

	    /* Set stats. */
            edw->file[n]->name = StringCopyAlloc(dent[i]);
            edw->file[n]->link_dest = GetAllocLinkDest(tmp_path);

            edw->file[n]->inode = stat_buf.st_ino;
            edw->file[n]->mode = stat_buf.st_mode;
            edw->file[n]->nlink = stat_buf.st_nlink;
            edw->file[n]->uid = stat_buf.st_uid;
            edw->file[n]->gid = stat_buf.st_gid;  
            edw->file[n]->rdevice = stat_buf.st_rdev;
            edw->file[n]->size = stat_buf.st_size;
            edw->file[n]->last_accessed = stat_buf.st_atime;
            edw->file[n]->last_modified = stat_buf.st_mtime;
            edw->file[n]->create_time = stat_buf.st_ctime;
	}

        for(i = 0; i < total_dents; i++)
	    free(dent[i]);
	free(dent);


	/* Need to move file scroll position to upper left. */
/*
	edw->file_win_sb.x_win_pos = 0;
	edw->file_win_sb.y_win_pos = 0;
*/

	return;
}


/*
 *	Refreshes all listings on edw.
 */
void EDWRefreshAllLists(endeavour_win_struct *edw)
{
	char stringa[512];


	/* Refresh dir list. */
	EDWRefreshDirList(edw);


        /* Update the maximum dir window width. */
        EDWUpdateDirWinWidth(edw);

	/* Refresh file list. */
	EDWRefreshFileList(edw, edw->loc_prompt.buf);


        /* ********************************************************* */
        /* Change status message. */

        free(edw->status_mesg);
        edw->status_mesg = NULL;
  
        sprintf(stringa, "%i directories, %i objects.",
            edw->total_dirs, edw->total_files
        );
        edw->status_mesg = (char *)calloc(1,
            (strlen(stringa) + 1) * sizeof(char)
        ); 
        if(edw->status_mesg != NULL) 
        {
            strcpy(edw->status_mesg, stringa);
        }



        /* ********************************************************* */
        /* Update recycle bin button on the tool ribbon. */

        if((int)total_recbin_wins > 0)
        {  
            if(recbin_win[0] != NULL)
            {   
                if(recbin_win[0]->total_objects > 0)
                {
                    edw->recbin_map_btn.image =
                        endeavour_image.recycle_bin_small1;
                }
                else
                {
                    edw->recbin_map_btn.image =
                        endeavour_image.recycle_bin_small0;
                }
            }
            else
            {
                edw->recbin_map_btn.image =
                    endeavour_image.recycle_bin_small0;
            }
        }               


	return;
}
