/*
                         Endeavour Browser Window: Macros

	Functions:

        int EDW_GETDIRNUMBYPOS(endeavour_win_struct *edw, int x, int y)
        int EDW_GETFILENUMBYPOS(endeavour_win_struct *edw, int x, int y)

	char *EDW_GETTIMESTR(time_t seconds)
	char *EDW_GETOWNERNAME(uid_t uid)
	char *EDW_GETGROUPNAME(gid_t gid)

	void EDWSETALLCURSORS(endeavour_win_struct *edw, int code)
	void EDWSANITIZEFILECOLSPACING(endeavour_win_struct *edw)

	int EDWDoMapCVPromptDir(endeavour_win_struct *edw, int mode)
        int EDWDoMapCVPromptFile(endeavour_win_struct *edw, int mode)

	int EDWDoGoParentDir(endeavour_win_struct *edw);
	int EDWDoRefresh(endeavour_win_struct *edw);

	int EDWDoQCreateNewDir(endeavour_win_struct *edw);
	int EDWDoQCreateNewLink(endeavour_win_struct *edw);
	int EDWDoEditProperties(endeavour_win_struct *edw);
	int EDWDoMove(endeavour_win_struct *edw);
	int EDWDoCopy(endeavour_win_struct *edw);
	int EDWDoRecycle(endeavour_win_struct *edw);

	int EDWDoNewBrowser(endeavour_win_struct *edw);
	int EDWDoTileBrowser(endeavour_win_struct *edw);
	int EDWDoMapRecycleBin(endeavour_win_struct *edw);

	void EDWDoExecCmdPrompt(endeavour_win_struct *edw)



 */


#include "endeavour.h"
#include "edwdefaults.h"


#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))

/* ******************************************************************** */

/*
 *	Returns the dir item number by givin coordinates.
 */
int EDW_GETDIRNUMBYPOS(endeavour_win_struct *edw, int x, int y)
{
        int dir_num;

        if(edw == NULL)
            return(-1);


        dir_num = ((int)edw->dir_win_sb.y_win_pos / ROW_SPACING) +
            ((y + ((int)edw->dir_win_sb.y_win_pos % ROW_SPACING)) /
            ROW_SPACING);

        if(dir_num >= (int)edw->total_dirs)
            dir_num = -1;

        return(dir_num);
}


/*
 *      Returns the file item number by givin coordinates.
 */
int EDW_GETFILENUMBYPOS(endeavour_win_struct *edw, int x, int y)
{
        int file_num;

        if(edw == NULL)
            return(-1);


        file_num = ((int)edw->file_win_sb.y_win_pos / ROW_SPACING) +
            ((y + ((int)edw->file_win_sb.y_win_pos % ROW_SPACING)) /
            ROW_SPACING);

        if(file_num >= (int)edw->total_files)
            file_num = -1;


        return(file_num);
}

/*
 *	Returns a statically allocated string formatted to a date and time
 *      string relative to seconds.
 */
char *EDW_GETTIMESTR(time_t seconds)
{
	size_t len;
        struct tm *tm_ptr;
	static char s[MAX_TIME_STR];


        if(seconds <= 0)
            return(LISTING_TIME_UNKNOWN_STR);

	tm_ptr = localtime(&seconds);

	len = strftime(
	    s,
            MAX_TIME_STR,
            LISTING_MTIME_FORMAT_STR,
            tm_ptr
        );
	if(len >= MAX_TIME_STR)
	    len = MAX_TIME_STR - 1;
	if(len < 0)
	    len = 0;

        s[len] = '\0';


	return(s);
}

/*
 *	Returns a statically allocated string containing the owner
 *	name of the given uid.
 */
char *EDW_GETOWNERNAME(uid_t uid)
{
	char *strptr;
	static char rtn_str[NAME_MAX];

	strptr = IDGetNameByUID(uid);
	if(strptr == NULL)
	{
            sprintf(rtn_str, "%i", uid);
	}
	else
	{
	    strncpy(rtn_str, strptr, NAME_MAX);
	}
        rtn_str[NAME_MAX - 1] = '\0';

	return(rtn_str);
}


/*
 *      Returns a statically allocated string containing the group
 *      name of the given gid.
 */
char *EDW_GETGROUPNAME(gid_t gid)
{
	char *strptr;
        static char rtn_str[NAME_MAX];

        strptr = IDGetGroupByGID(gid);
	if(strptr == NULL)
	{
	    sprintf(rtn_str, "%i", gid);
	}
	else
	{
	    strncpy(rtn_str, strptr, NAME_MAX);
	}
	rtn_str[NAME_MAX - 1] = '\0';

        return(rtn_str);
}


/*
 *      Sets all cursors on the endeavour window to display to
 *      the scheme givin by code.
 */             
void EDWSETALLCURSORS(endeavour_win_struct *edw, int code)
{
        if(edw == NULL)
            return;


        switch(code)
        {
          case EDW_CURSORS_CODE_WORKING:
            OSWSetWindowCursor(edw->toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(edw->tool_ribbon_win, endeavour_cursor.working);
   
            OSWSetWindowCursor(edw->folder_back_btn.toplevel,
                endeavour_cursor.working);
            OSWSetWindowCursor(edw->folder_refresh_btn.toplevel,
                endeavour_cursor.working);
         
            OSWSetWindowCursor(edw->new_folder_btn.toplevel,
                endeavour_cursor.working);
            OSWSetWindowCursor(edw->properties_btn.toplevel,
                endeavour_cursor.working);
            OSWSetWindowCursor(edw->move_btn.toplevel,
                endeavour_cursor.working);
            OSWSetWindowCursor(edw->copy_btn.toplevel,
                endeavour_cursor.working);
            OSWSetWindowCursor(edw->recycle_btn.toplevel,
                endeavour_cursor.working);
  
            OSWSetWindowCursor(edw->new_browser_btn.toplevel,
                endeavour_cursor.working);
            OSWSetWindowCursor(edw->recbin_map_btn.toplevel,  
                endeavour_cursor.working);
                
            OSWSetWindowCursor(edw->dir_win, endeavour_cursor.working);
            OSWSetWindowCursor(edw->file_win, endeavour_cursor.working);
            OSWSetWindowCursor(edw->file_heading_win, endeavour_cursor.working);
            OSWSetWindowCursor(edw->status_win, endeavour_cursor.working);
        
            break; 
            
          default: /* EDW_CURSORS_CODE_NORMAL */
            OSWUnsetWindowCursor(edw->toplevel);
            OSWUnsetWindowCursor(edw->tool_ribbon_win);

            OSWUnsetWindowCursor(edw->folder_back_btn.toplevel);
            OSWUnsetWindowCursor(edw->folder_refresh_btn.toplevel);

            OSWUnsetWindowCursor(edw->new_folder_btn.toplevel);
            OSWUnsetWindowCursor(edw->properties_btn.toplevel);
            OSWUnsetWindowCursor(edw->move_btn.toplevel);
            OSWUnsetWindowCursor(edw->copy_btn.toplevel);
            OSWUnsetWindowCursor(edw->recycle_btn.toplevel);

            OSWUnsetWindowCursor(edw->new_browser_btn.toplevel);
            OSWUnsetWindowCursor(edw->recbin_map_btn.toplevel);

            OSWUnsetWindowCursor(edw->dir_win);
            OSWUnsetWindowCursor(edw->file_win);
            OSWUnsetWindowCursor(edw->file_heading_win);
            OSWUnsetWindowCursor(edw->status_win);

            break;
        }
                
        OSWGUISync(False);


        return;
}


/*
 *      Sanitizes file heading colum spacing on edw.
 */         
void EDWSANITIZEFILECOLSPACING(endeavour_win_struct *edw)
{
        static win_attr_t wattr;
        
        if(edw == NULL)
            return;
   
        /* Name. */
        if(edw->col_name_pos >= edw->col_size_pos)
            edw->col_name_pos = edw->col_size_pos - 1;
        if(edw->col_name_pos < 0)  
            edw->col_name_pos = 0;
        
        /* Size. */
        if(edw->col_size_pos >= edw->col_permission_pos)
            edw->col_size_pos = edw->col_permission_pos - 1;
        if(edw->col_size_pos < 0)
            edw->col_size_pos = 0;
            
        /* Permission. */
        if(edw->col_permission_pos >= edw->col_owner_pos)
            edw->col_permission_pos = edw->col_owner_pos - 1;
        if(edw->col_permission_pos < 0)
            edw->col_permission_pos = 0;
            
        /* Owner. */
        if(edw->col_owner_pos >= edw->col_group_pos)
            edw->col_owner_pos = edw->col_group_pos - 1;
        if(edw->col_owner_pos < 0)
            edw->col_owner_pos = 0;
            
        /* Group. */
        if(edw->col_group_pos >= edw->col_date_pos)
            edw->col_group_pos = edw->col_date_pos - 1;
        if(edw->col_group_pos < 0)
            edw->col_group_pos = 0;
        
        /* Date. */
        OSWGetWindowAttributes(edw->file_heading_win, &wattr);
        if(edw->col_date_pos >= wattr.width)
            edw->col_date_pos = wattr.width - 1;
        if(edw->col_date_pos < 0)
            edw->col_date_pos = 0;
            

        return;
}       



/*
 *	Maps the change values prompt on edw's dir window
 *	at the appropriate position.
 */
int EDWDoMapCVPromptDir(endeavour_win_struct *edw, int dir_num, int mode)
{
	int x, y;
	int width, height;
        win_attr_t wattr;
	char stringa[PATH_MAX + NAME_MAX + 128];


	if(edw == NULL)
	    return(-1);
	if((dir_num < 0) || (dir_num >= edw->total_dirs))
	    return(-1);
        if(edw->dir[dir_num] == NULL)
	    return(-1);
        if(edw->dir[dir_num]->name == NULL)
	    return(-1);
        if(edw->dir[dir_num]->fullname == NULL)
	    return(-1);

        /* Skip special notations. */
        if(!strcmp(edw->dir[dir_num]->name, "..") ||
           !strcmp(edw->dir[dir_num]->name, ".") ||
           !strcmp(edw->dir[dir_num]->name, "/")
        )
            return(-2);

	/* Write protect check. */
	if(option.write_protect)
	    return(-3);

        /* Select dir. */
        EDWDoSelectDir(edw, dir_num);

        /* Reparent the cv_prompt. */
        OSWReparentWindow(edw->cv_prompt.toplevel, edw->dir_win);
        OSWGetWindowAttributes(edw->dir_win, &wattr);

        /* Record cv_src_path. */
        free(edw->cv_src_path);
        edw->cv_src_path = StringCopyAlloc(
            edw->dir[dir_num]->fullname
        );


        /* ********************************************************** */
	/* Mode: Rename */
	if(mode == CV_PROMPT_MODE_RENAME)
	{
	    /* Calculate x and y coordinate position. */
	    x = LIST_MARGIN + edw->dir[dir_num]->x +
                DIR_ICON_WIDTH + 4 - edw->dir_win_sb.x_win_pos;

            y = (ROW_SPACING * (dir_num -
                ((int)edw->dir_win_sb.y_win_pos / ROW_SPACING)))
                - ((int)edw->dir_win_sb.y_win_pos % ROW_SPACING);

	    /* Calculate width and height. */
	    width = MIN(
	        (int)wattr.width - SCROLLBAR_YBAR_WIDTH - x,
	        (int)wattr.width - SCROLLBAR_YBAR_WIDTH
	    );
	    if(width < 2) return(-2);

	    height = ROW_SPACING + 2;
	    if(height < 2) return(-2);

	    /* Reposition cv_prompt window. */
            OSWMoveResizeWindow(
                edw->cv_prompt.toplevel,
	        x, y + 3, width, height
	    );

	    /* Put current name into cv prompt buffer. */
            strncpy(
                edw->cv_prompt.buf,
                edw->dir[dir_num]->name,
                edw->cv_prompt.buf_len
            );
            edw->cv_prompt_mode = CV_PROMPT_MODE_RENAME;

            /* Set status message. */ 
            sprintf(stringa,
 "Changing name for `%s' (press escape key to cancel).",
                edw->dir[dir_num]->name
            );
            free(edw->status_mesg);
            edw->status_mesg = StringCopyAlloc(stringa);
	}
	/* Unsupported mode. */
	else
	{
            /* Set change values prompt type and map. */ 
            edw->cv_prompt_mode = CV_PROMPT_MODE_NONE;

	    memset(edw->cv_prompt.buf, 0x00, edw->cv_prompt.buf_len);
	    edw->cv_prompt.buf_pos = 0;
	    PromptUnmap(&edw->cv_prompt);

	    sprintf(stringa,
"Internal non-critical error:\n\n\
    %i: Unsupported change values operation mode.\n\n",
		mode
	    );
	    printdw(&err_dw, stringa);
	}

        /* Mark entire cv prompt buffer. */
        PromptMarkBuffer(
            &edw->cv_prompt,
            PROMPT_POS_END 
        );
        /* Map the cv prompt in focus. */
        PromptMap(&edw->cv_prompt);
        edw->cv_prompt.is_in_focus = 1;


	return(0);
}


/*
 *	Procedure to map the change values prompt on the file window on
 *	edw over item file_num.
 */
int EDWDoMapCVPromptFile(endeavour_win_struct *edw, int file_num, int mode)
{
        int x, y;
        int width, height;
	char *strptr;
	win_attr_t wattr;
        char stringa[PATH_MAX + NAME_MAX + 128];


	/* Error checks. */
        if(!IDC() || (edw == NULL)) return(-1);
        if((file_num < 0) || (file_num >= edw->total_files)) return(-1);
        if(edw->file[file_num] == NULL) return(-1);
        if(edw->file[file_num]->name == NULL) return(-1);

	/* Skip special notations. */
        if(!strcmp(edw->file[file_num]->name, "..") ||   
           !strcmp(edw->file[file_num]->name, ".") ||
           !strcmp(edw->file[file_num]->name, "/")
        )
	    return(-2);

        /* Write protect check. */
        if(option.write_protect)
        {
            return(-3);
        }


	/* Unselect everything and select just the file. */
        EDW_UNSELECTALLDIRS(edw);
        EDW_UNSELECTALLFILES(edw);

        EDW_SELECTFILE(edw, file_num);
	EDWDraw(edw, DRAW_AMOUNT_FILE_LIST);

        /* Reparent the cv_prompt. */
        OSWReparentWindow(edw->cv_prompt.toplevel, edw->file_win);
        OSWGetWindowAttributes(edw->file_win, &wattr);

        /* Record cv_src_path. */
        free(edw->cv_src_path);  
        edw->cv_src_path = StringCopyAlloc(
            PrefixPaths(
		edw->loc_prompt.buf,
		edw->file[file_num]->name
	    )
        );


        /* ********************************************************** */
        /* Mode: Rename */
        if(mode == CV_PROMPT_MODE_RENAME)
        {
            /* Calculate x and y coordinate position. */
            x = edw->col_name_pos + 24 + 4;

            y = (ROW_SPACING * (file_num -
                ((int)edw->file_win_sb.y_win_pos / ROW_SPACING)))
                - ((int)edw->file_win_sb.y_win_pos % ROW_SPACING);

            /* Calculate width and height. */
	    width = edw->col_size_pos - edw->col_name_pos - 24 - 4;
	    if(width < 2) return(-2);
	    height = ROW_SPACING + 2;
	    if(height < 2) return(-2);

            /* Reposition cv_prompt window. */
            OSWMoveResizeWindow(
		edw->cv_prompt.toplevel,
                x, y + 3, width, height
            );

            /* Put current name into cv prompt buffer. */
            strncpy(
                edw->cv_prompt.buf,
                edw->file[file_num]->name,
                edw->cv_prompt.buf_len
            );
            edw->cv_prompt_mode = CV_PROMPT_MODE_RENAME;


            /* Set status message. */
            sprintf(stringa,
 "Changing name for `%s' (press escape key to cancel).",
                edw->file[file_num]->name
            );
            free(edw->status_mesg);
            edw->status_mesg = StringCopyAlloc(stringa);
        }
        /* Mode: CHMOD */
        else if(mode == CV_PROMPT_MODE_CHMOD)
        {
            /* Calculate x and y coordinate position. */
            x = edw->col_permission_pos;

            y = (ROW_SPACING * (file_num -
                ((int)edw->file_win_sb.y_win_pos / ROW_SPACING)))
                - ((int)edw->file_win_sb.y_win_pos % ROW_SPACING);
                
            /* Calculate width and height. */
            width = edw->col_owner_pos -
                edw->col_permission_pos + 4;
            if(width < 2) return(-2);
            height = ROW_SPACING + 2;
            if(height < 2) return(-2);

            /* Reposition cv_prompt window. */
            OSWMoveResizeWindow(
                edw->cv_prompt.toplevel,
                x, y + 3, width, height
            );

            edw->cv_prompt.buf[0] =
                (edw->file[file_num]->mode & S_IRUSR) ? 'r' : '-';
            edw->cv_prompt.buf[1] =
                (edw->file[file_num]->mode & S_IWUSR) ? 'w' : '-';
            edw->cv_prompt.buf[2] =
                (edw->file[file_num]->mode & S_IXUSR) ? 'x' : '-';
            if(edw->file[file_num]->mode & S_ISUID)
                edw->cv_prompt.buf[2] = 's';

            edw->cv_prompt.buf[3] =
                (edw->file[file_num]->mode & S_IRGRP) ? 'r' : '-';
            edw->cv_prompt.buf[4] =
                (edw->file[file_num]->mode & S_IWGRP) ? 'w' : '-';
            edw->cv_prompt.buf[5] =
                (edw->file[file_num]->mode & S_IXGRP) ? 'x' : '-';
            if(edw->file[file_num]->mode & S_ISGID)
                edw->cv_prompt.buf[5] = 's';

            edw->cv_prompt.buf[6] =
                (edw->file[file_num]->mode & S_IROTH) ? 'r' : '-';
            edw->cv_prompt.buf[7] =
                (edw->file[file_num]->mode & S_IWOTH) ? 'w' : '-';
            edw->cv_prompt.buf[8] =
                (edw->file[file_num]->mode & S_IXOTH) ? 'x' : '-';
            if(edw->file[file_num]->mode & S_ISVTX)
                 edw->cv_prompt.buf[8] = 't';
            edw->cv_prompt.buf[9] = '\0';

            edw->cv_prompt_mode = CV_PROMPT_MODE_CHMOD;


            /* Set status message. */
            sprintf(stringa,
 "Changing permissions for `%s' (press escape key to cancel).",
                edw->file[file_num]->name 
            );
            free(edw->status_mesg);
            edw->status_mesg = StringCopyAlloc(stringa);
 	}
        /* Mode: Chown Owner */
        else if(mode == CV_PROMPT_MODE_CHOWN_OWNER)
        {
            /* Calculate x and y coordinate position. */
            x = edw->col_owner_pos;

            y = (ROW_SPACING * (file_num -
                ((int)edw->file_win_sb.y_win_pos / ROW_SPACING))) 
                - ((int)edw->file_win_sb.y_win_pos % ROW_SPACING);

            /* Calculate width and height. */
            width = edw->col_group_pos - edw->col_owner_pos - 2;
            if(width < 2) return(-2);
            height = ROW_SPACING + 2;
            if(height < 2) return(-2);

            /* Reposition cv_prompt window. */
            OSWMoveResizeWindow(
                edw->cv_prompt.toplevel,
                x, y + 3, width, height
            );

            strptr = IDGetNameByUID(edw->file[file_num]->uid); 
            if(strptr != NULL)
                strncpy(
                    edw->cv_prompt.buf,
                    strptr,
                    edw->cv_prompt.buf_len
                );
	    else
                strncpy(
                    edw->cv_prompt.buf,
                    "?",
                    edw->cv_prompt.buf_len
                );
            edw->cv_prompt_mode = CV_PROMPT_MODE_CHOWN_OWNER;


            /* Set status message. */
            sprintf(stringa,
 "Changing owner for `%s' (press escape key to cancel).",
                edw->file[file_num]->name
            );
            free(edw->status_mesg); 
            edw->status_mesg = StringCopyAlloc(stringa);
	}
        /* Mode: Chown Group */
        else if(mode == CV_PROMPT_MODE_CHOWN_GROUP)
        {
            /* Calculate x and y coordinate position. */
            x = edw->col_group_pos;

            y = (ROW_SPACING * (file_num -
                ((int)edw->file_win_sb.y_win_pos / ROW_SPACING)))
                - ((int)edw->file_win_sb.y_win_pos % ROW_SPACING);

            /* Calculate width and height. */
            width = edw->col_date_pos - edw->col_group_pos - 2;
            if(width < 2) return(-2);
            height = ROW_SPACING + 2;
            if(height < 2) return(-2);

            /* Reposition cv_prompt window. */
            OSWMoveResizeWindow(
                edw->cv_prompt.toplevel,
                x, y + 3, width, height
            );

            strptr = IDGetGroupByGID(edw->file[file_num]->gid);
            if(strptr != NULL)
                strncpy(
                    edw->cv_prompt.buf,
                    strptr,
                    edw->cv_prompt.buf_len
                );
 	    else
                strncpy(
                    edw->cv_prompt.buf,
                    "?",
                    edw->cv_prompt.buf_len
                );
            edw->cv_prompt_mode = CV_PROMPT_MODE_CHOWN_GROUP;


            /* Set status message. */ 
            sprintf(stringa,
 "Changing group for `%s' (press escape key to cancel).",
                edw->file[file_num]->name
            );
            free(edw->status_mesg);
            edw->status_mesg = StringCopyAlloc(stringa);
        }
        /* Unsupported mode. */
        else
        {
            /* Set change values prompt type and map. */
            edw->cv_prompt_mode = CV_PROMPT_MODE_NONE;

            memset(edw->cv_prompt.buf, 0x00, edw->cv_prompt.buf_len);
            edw->cv_prompt.buf_pos = 0;
            PromptUnmap(&edw->cv_prompt);

            sprintf(stringa,
"Internal non-critical error:\n\n\
    %i: Unsupported change values operation mode.\n\n",
                mode
            );
            printdw(&err_dw, stringa);
        }
        /* Mark entire cv prompt buffer. */
        PromptMarkBuffer(
            &edw->cv_prompt,
            PROMPT_POS_END
        );
        /* Map the cv prompt in focus. */
        PromptMap(&edw->cv_prompt);
        edw->cv_prompt.is_in_focus = 1;


        return(0);
}


/*
 *	Selects the parent directory of the currently
 *	selected dir item on edw.
 */
int EDWDoGoParentDir(endeavour_win_struct *edw)
{
	int i, sel_dir;
	int cur_level;


	if(edw == NULL)
	    return(-1);
	if(edw->total_sel_dirs <= 0)
	    return(-1);

	sel_dir = edw->sel_dir[0];
	if((sel_dir < 0) || (sel_dir >= edw->total_dirs))
	    return(-1);

	cur_level = edw->dir[sel_dir]->level;


        EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_WORKING);

	for(i = sel_dir - 1; i >= 0; i--)
	{
	    if(edw->dir[i] == NULL) continue;

	    if(edw->dir[i]->level < cur_level)
	    {
		EDWDoSelectDir(edw, i);
		break;
	    }
	}

	EDWDraw(edw, DRAW_AMOUNT_LISTS);
	PromptDraw(&edw->loc_prompt, PROMPT_DRAW_AMOUNT_TEXTONLY);

        EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_NORMAL);


	return(0);
}


/*
 *	Performs a complete refresh procedure on edw, refreshing
 *	lists and device records.
 */
int EDWDoRefresh(endeavour_win_struct *edw)
{
        if(edw == NULL) return(-1);

        EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_WORKING);
             
        EDWRefreshAllLists(edw);
        DevRecRefreshList();

        EDWDraw(edw, DRAW_AMOUNT_COMPLETE);

        EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_NORMAL);

	return(0);
}



/*
 *	Creates a new directory `quick style' on the edw.
 */
int EDWDoQCreateNewDir(endeavour_win_struct *edw)
{
	int i, sel_dir, status;
	struct stat stat_buf;
	mode_t mode;
	char dirname[PATH_MAX + NAME_MAX];
	char parent[PATH_MAX + NAME_MAX];
	char *strptr;
	char stringa[512];


	/* Error checks. */
	if(edw == NULL)
	    return(-1);

	/* Atleast one dir must be selected. */
	if(edw->total_sel_dirs <= 0)
	    return(-2);

	/* Get sel_dir number and check if it is valid. */
	sel_dir = edw->sel_dir[0];
	if((sel_dir < 0) || (sel_dir >= edw->total_dirs)) return(-1);
	if(edw->dir[sel_dir] == NULL) return(-1);
	if(edw->dir[sel_dir]->fullname == NULL) return(-1);

	strncpy(parent, edw->dir[sel_dir]->fullname,
	    PATH_MAX + NAME_MAX);
	parent[PATH_MAX + NAME_MAX - 1] = '\0';


        /* Check if internal write-protect is on. */
        if(option.write_protect)
        {
	    printdw(&err_dw,
"Cannot create new directory:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
            );
            return(-3);
        }

        /* Need to change current directory. */
        if(chdir(parent) == 0)
        {
            getcwd(dname.current_working, PATH_MAX);
            dname.current_working[PATH_MAX - 1] = '\0';
        }
	else
	{
	    /* Cannot change working dir then cannot create dir. */
            printdw(&err_dw,
"Cannot create new directory:\n\n\
    Current working directory invalid.\n\n\
Make sure that the directory you have currently\n\
selected is valid and allows creation of additional\n\
directories within.\n"
            );

            return(-1);
	}


	/* ************************************************************* */

	/* Search for a name not in use. */
	status = -1;
	for(i = 0; i < 100; i++)	/* 100 retires */
	{
	    sprintf(dirname, "newdir%i", i + 1);
	    strptr = PrefixPaths(parent, dirname);
	    if(strptr == NULL)
		continue;
	    strncpy(dirname, strptr, PATH_MAX + NAME_MAX);
	    dirname[PATH_MAX + NAME_MAX - 1] = '\0';

	    /* Check if the name already exists. */
	    if(lstat(dirname, &stat_buf) == 0)
	    {
		continue;
	    }
	    else
	    {
		status = 0;
	        break;
	    }
	}
	/* Check if we were not able to generate a name. */
	if(status)
	{
            printdw(&err_dw,
"Cannot create new directory:\n\n\
    Unable to generate a new tempory name.\n\n\
You may have an existing object(s) with the name\n\
`newdir#', where # is a number.\n"
            );
	    return(-4);
	}


	/* Set permissions for new directory. */
	mode = 0;
	mode |= S_IRUSR;
        mode |= S_IWUSR;
        mode |= S_IXUSR;

        mode |= S_IRGRP;
	/* No write permission for group. */
        mode |= S_IXGRP;

        mode |= S_IROTH;
        /* No write permissions for others. */
        mode |= S_IXOTH;

	/* Create a new directory based on dirname. */
	status = mkdir(dirname, mode);
	if(status == -1)
	{
	    switch(errno)
	    {
	      case EEXIST:
		strptr = "Pathname already exists (not necessarily as a directory).";
		break;

              case EFAULT:
                strptr = "Pathname points outside your accessible address space.";
                break;

	      case EACCES:
		strptr = "The parent directory does not allow write permission.";
		break;

	      case ENAMETOOLONG:
                strptr = "Pathname was too long.";
		break;

	      case ENOENT:
                strptr = "A directory component in pathname does not exist.";
                break;

	      case ENOTDIR:
                strptr = "A component used as a directory in pathname is not a directory.";
                break;

	      case ENOMEM:
                strptr = "Insufficient kernel memory was available.";
                break;

              case EROFS:
                strptr = "Pathname refers to a file on a read-only filesystem.";
                break;

              case ELOOP:
                strptr = "Pathname contains a reference to a circular symbolic link.";
		break;

              case ENOSPC:
                strptr = "The device containing pathname has no more room.";
		break;

	      default:
                strptr = "Unknown error.";
		break;
	    }
	    sprintf(stringa,
"Cannot create a new directory:\n\n\
    mkdir(): %s\n\n",
		strptr
	    );
            printdw(&err_dw, stringa);

	    return(-1);
	}


	/* *************************************************************** */
	/* Directory create was successful. */

	/* Refresh list on endeavour window. */
	EDWColapseDir(edw, sel_dir);
	EDWExpandDir(edw, sel_dir);
/*
	EDWRefreshAllLists(edw);
*/
	/* Select new dir. */
	i = EDWMatchExistingDirFromPath(edw, dirname);
	EDWDoSelectDir(edw, i);


	StripAbsolutePath(dirname);

	/* Update status message. */
	free(edw->status_mesg);
	edw->status_mesg = NULL;
	sprintf(stringa,
            "New directory `%s' created and selected.",
	    dirname
	);
	edw->status_mesg = StringCopyAlloc(stringa);

	/* Redraw edw. */
        EDWDraw(edw, DRAW_AMOUNT_COMPLETE);


	return(i);
}



/*
 *	Creates a new link by Endeavour create `style'.
 */
int EDWDoQCreateNewLink(endeavour_win_struct *edw)
{






	return(0);
}



/*
 *	Performs a map edit properties dialogs procedure on edw.
 */
int EDWDoEditProperties(endeavour_win_struct *edw)
{
	int i, n;
	char tmp_path[PATH_MAX + NAME_MAX];

        if(edw == NULL) return(-1);

        /* Directories. */
        for(i = 0; i < (int)edw->total_sel_dirs; i++)
        {
            n = edw->sel_dir[i];
            if((n < 0) || (n >= (int)edw->total_dirs)) continue;
            if(edw->dir[n] == NULL) continue;
            if(edw->dir[n]->fullname == NULL) continue;

            DoCreatePropDialog(edw->dir[n]->fullname);
        }

	/* Files. */
        for(i = 0; i < (int)edw->total_sel_files; i++)
        {
            n = edw->sel_file[i];
            if((n < 0) || (n >= (int)edw->total_files)) continue;
            if(edw->file[n] == NULL) continue;
            if(edw->file[n]->name == NULL) continue;

	    if(edw->loc_prompt.buf != NULL)
                strncpy(tmp_path, PrefixPaths(
                    edw->loc_prompt.buf,
                    edw->file[n]->name),
                    PATH_MAX + NAME_MAX
                );
            else
                continue;

            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

            DoCreatePropDialog(tmp_path);
        }


	return(0);
}


/*
 *      Performs a map op dialog for move procedure on edw.
 */
int EDWDoMove(endeavour_win_struct *edw)
{
        int i, n;           
        char tmp_path[PATH_MAX + NAME_MAX];
        char *strptr;


        /* Error checks. */
        if(edw == NULL) return(-1);
	if(total_op_dialogs <= 0) return(-1);
	if((edw->total_sel_dirs + edw->total_sel_files) <= 0) return(-1);
	if(edw->loc_prompt.buf == NULL) return(-1);


        /* Clear previous op dialog queue of objects. */
	OPDialogClearObjects(op_dialog[0]);	

	/* Directories. */
        for(i = 0; i < edw->total_sel_dirs; i++)
        {
            n = edw->sel_dir[i];
            if((n < 0) || (n >= (int)edw->total_dirs)) continue;
            if(edw->dir[n] == NULL) continue;
            if(edw->dir[n]->fullname == NULL) continue;

	    strptr = GetParentDir(edw->dir[n]->fullname);
	    if(strptr == NULL) continue;

	    strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
	    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
 
            OPDialogAddObject(
                op_dialog[0],
                OP_CODE_MOVE,
                tmp_path,		/* Current dir. */
                edw->dir[n]->fullname	/* Target. */
            );
	}

	/* Files. */
        for(i = 0; i < edw->total_sel_files; i++)
        {
            n = edw->sel_file[i];
            if((n < 0) || (n >= (int)edw->total_files)) continue;
            if(edw->file[n] == NULL) continue;
            if(edw->file[n]->name == NULL) continue;

            strptr = PrefixPaths(edw->loc_prompt.buf,
                edw->file[n]->name);
            if(strptr == NULL) continue;
            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

            OPDialogAddObject(
                op_dialog[0],
                OP_CODE_MOVE,
                edw->loc_prompt.buf,
                tmp_path
            );
        }
        OPDialogMap(op_dialog[0]);


	return(0);
}


/*      
 *      Performs a map op dialog for copy procedure on edw.
 */
int EDWDoCopy(endeavour_win_struct *edw)
{
        int i, n;
        char tmp_path[PATH_MAX + NAME_MAX];   
        char *strptr;


	/* Error checks. */
        if(edw == NULL) return(-1);
        if(total_op_dialogs <= 0) return(-1);
        if((edw->total_sel_dirs + edw->total_sel_files) <= 0) return(-1);
        if(edw->loc_prompt.buf == NULL) return(-1);
            
  
	/* Clear previous op dialog queue of objects. */
        OPDialogClearObjects(op_dialog[0]);

	/* Directories. */  
        for(i = 0; i < edw->total_sel_dirs; i++)
        {
            n = edw->sel_dir[i];
            if((n < 0) || (n >= (int)edw->total_dirs)) continue;
            if(edw->dir[n] == NULL) continue;
            if(edw->dir[n]->fullname == NULL) continue;
         
            strptr = GetParentDir(edw->dir[n]->fullname);
            if(strptr == NULL) continue;
            
            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

            OPDialogAddObject(
                op_dialog[0],
                OP_CODE_COPY,
                tmp_path,               /* Current dir. */
                edw->dir[n]->fullname   /* Target. */
            );
        }

        /* Files. */
        for(i = 0; i < edw->total_sel_files; i++)
        {
            n = edw->sel_file[i];
            if((n < 0) || (n >= (int)edw->total_files)) continue;
            if(edw->file[n] == NULL) continue;
            if(edw->file[n]->name == NULL) continue;

            strptr = PrefixPaths(edw->loc_prompt.buf,
                edw->file[n]->name);
            if(strptr == NULL) continue;
            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

            OPDialogAddObject(
                op_dialog[0],
                OP_CODE_COPY,
                edw->loc_prompt.buf,
                tmp_path
            );
        }
        OPDialogMap(op_dialog[0]);


        return(0);
}


/*
 *	Performs a recycle selected objects procedure on edw.
 */
int EDWDoRecycle(endeavour_win_struct *edw)
{
        int i, n;
        char tmp_path[PATH_MAX + NAME_MAX];   
	char *strptr;


        /* Error checks. */            
        if(edw == NULL) return(-1);
        if(edw->loc_prompt.buf == NULL) return(-1);



        EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_WORKING);
               
        /* Need to change current directory. */
        if(chdir(edw->loc_prompt.buf) == 0)
        {
            getcwd(dname.current_working, PATH_MAX);
            dname.current_working[PATH_MAX - 1] = '\0';
        }
                
        /* Need to turn off yes_to_all just in case. */
        yes_to_all = False;

	/* Move comfermation dialog to position of pointer. */
        WidgetCenterWindow(comf_dialog.toplevel, WidgetCenterWindowToPointer);

        for(i = 0; i < edw->total_sel_dirs; i++)
        {
            n = edw->sel_dir[i];
            if((n < 0) || (n >= edw->total_dirs)) continue;
            if(edw->dir[n] == NULL) continue;
            if(edw->dir[n]->fullname == NULL) continue;

            /* Need to change current to its parent. */
            if(chdir(GetParentDir(edw->dir[n]->fullname)) == 0)
            {
                getcwd(dname.current_working, PATH_MAX);
                dname.current_working[PATH_MAX - 1] = '\0';
            }

            EDWPerformFileOp(
                OP_CODE_RECYCLE,
                NULL,
                edw->dir[n]->fullname
            );
        }

        for(i = 0; i < edw->total_sel_files; i++)
        {
            n = edw->sel_file[i];
            if((n < 0) || (n >= edw->total_files)) continue;
            if(edw->file[n] == NULL) continue;
            if(edw->file[n]->name == NULL) continue;

	    strptr = PrefixPaths(edw->loc_prompt.buf, edw->file[n]->name);
	    if(strptr == NULL) continue;

            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

            EDWPerformFileOp(
                OP_CODE_RECYCLE,
                NULL,
                tmp_path   
            );
	}

        /* Need to turn off yes_to_all. */
        yes_to_all = False;
                
        /* Need to refresh list. */
        EDWRefreshAllLists(edw);
                        
        EDWDraw(edw, DRAW_AMOUNT_COMPLETE);

        EDWSETALLCURSORS(edw, EDW_CURSORS_CODE_NORMAL);

	return(0);
}


/*
 *	Performs a create new browser window procedure on edw.
 */
int EDWDoNewBrowser(endeavour_win_struct *edw)
{
	int status;


        /* Error checks. */
        if(edw == NULL) return(-1);
        if(edw->loc_prompt.buf == NULL) return(-1);
 
        status = DoCreateEDW(edw->loc_prompt.buf);


	return(status);
}


/*
 *	Tiles all endeavour windows.
 */
int EDWDoTileBrowser(endeavour_win_struct *rel_edw)
{
	int i, x, y;
	int width, height;
	int total_height;
	win_attr_t wattr;
	char height_too_big = 0;


	/* Error checks. */
	if(!IDC())
	    return(-1);


	/* Sanitize global total_endeavour_wins. */
	if((int)total_endeavour_wins < 0)
	    total_endeavour_wins = 0;

	/* Must have atleast two endeavour windows to tile. */
	if((int)total_endeavour_wins < 2)
	{
	    printdw(&err_dw,
"Cannot tile browsers:\n\n\
    There is one or less browser in existance.\n\n\
(Press the `New Browser Button' to create a new browser.)\n"
	    );
	    return(-2);
	}

	/* Use first edw if givin relative is NULL. */
	if(rel_edw == NULL)
	    rel_edw = endeavour_win[0];

	/* Relative must be allocated and initialized. */
	if(rel_edw == NULL)
	    return(-1);
	else if(rel_edw->toplevel == 0)
	    return(-1);


	/* ************************************************************ */

	/* All tiling is relative from the relative edw. */
	OSWGetWindowAttributes(rel_edw->toplevel, &wattr);
	OSWGetWindowRootPos(rel_edw->toplevel, &x, &y);


        /* Calculate new width. */
        width = wattr.width;
	/* Sanitize width. */
        if((int)width > (int)osw_gui[0].display_width)
            width = osw_gui[0].display_width;
        if((int)width < MIN_WIDTH)
            width = MIN_WIDTH;


	/* Calculate new height. */
	total_height = ((int)wattr.height * (int)total_endeavour_wins) +
	    (WM_TITLE_HEIGHT * (int)total_endeavour_wins);
	if((total_height + y) > (int)osw_gui[0].display_height)
	{
	    height_too_big = 1;

	    height = ((int)osw_gui[0].display_height /
                (int)total_endeavour_wins) -
		(WM_TITLE_HEIGHT * (int)total_endeavour_wins);
	    y = (int)height + (2 * WM_TITLE_HEIGHT);
	}
	else
	{
	    height = wattr.height;
	    y += (int)height + WM_TITLE_HEIGHT;
	}


	if((int)height < MIN_HEIGHT)
	    height = MIN_HEIGHT;



	/* ******************************************************* */
	/* Begin tiling. */
	for(i = 0; i < (int)total_endeavour_wins; i++)
	{
	    if(endeavour_win[i] == NULL) continue;
	    if(endeavour_win[i]->map_state == 0) continue;
	    if(endeavour_win[i]->toplevel == 0) continue;

	    if(endeavour_win[i] == rel_edw)
	    {
		if(height_too_big)
		{
		    OSWMoveResizeWindow(
			endeavour_win[i]->toplevel,
			x, WM_TITLE_HEIGHT, width, height
		    );
		    OSWMapRaised(endeavour_win[i]->toplevel);
		    EDWResize(endeavour_win[i]);
		}
	    }
	    else
	    {
		OSWMoveResizeWindow(
		    endeavour_win[i]->toplevel,
		    x, y, width, height
		);
		OSWMapRaised(endeavour_win[i]->toplevel);
		EDWResize(endeavour_win[i]);

                /* Increment y position. */
                y += (int)height + WM_TITLE_HEIGHT;
	    }
	}


	return(0);
}


/*
 *      Performs a map recycle bin procedure on edw.
 */
int EDWDoMapRecycleBin(endeavour_win_struct *edw)
{
	DoCreateRB(NULL);

 	return(0);
}



/*
 *	Executes contents in edw's command prompt.
 */
int EDWDoExecCmdPrompt(endeavour_win_struct *edw)
{
	char *cmd_prompt_buf, *loc_prompt_buf;
	int cmd_len;
	char **strv;
	int strc;
	char do_not_execute = 0;
	pid_t p = 0;

#ifndef CMD_ARG_DELIMITER
# define CMD_ARG_DELIMITER	';'
#endif


	if(edw == NULL)
	    return(-1);

	cmd_prompt_buf = PromptGetS(&edw->cmd_prompt);
	loc_prompt_buf = PromptGetS(&edw->loc_prompt);
	if((cmd_prompt_buf == NULL) ||
           (loc_prompt_buf == NULL)
	)
	    return(-1);


	/* Get length of command. */
	cmd_len = strlen(cmd_prompt_buf);
	if(cmd_len < 1)
	    return(-1);


	/* If write protect is on, check if there is a rm command. */
	if(option.write_protect)
        {
	    char *tmp_cmd = StringCopyAlloc(cmd_prompt_buf);

	    /* Will this command possibly execute the remove command? */
	    do_not_execute = (ISCMDREMOVE(tmp_cmd) ? 1 : 0);

	    free(tmp_cmd);
	}
	if(do_not_execute)
	{
	    printdw(&err_dw,
"The command you have typed may cause a deliberate object\n\
removal.   If you still want to execute the command, turn\n\
off the master write-protect.\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
	    );
	    return(-1);
	}


        /* Change working directory. */
        if(chdir(loc_prompt_buf) == 0)
        {
            /* Update global working dir. */
            getcwd(dname.current_working, PATH_MAX);   
            dname.current_working[PATH_MAX - 1] = '\0';
	}


	/* Parse initial command. */
	strv = strexp(cmd_prompt_buf, &strc);
	if((strv == NULL) || (strc < 1))
	    return(-1);


	/* ************************************************************** */
	/* Load to internal (regular) viewer? */
	if(strcasepfx(strv[0], "%v"))
	{
	    /* Must have atleast two arguments. */
	    if(strc > 1)
	    {
		/* Create Internal (not image) Viewer and map. */
		DoCreateV(strv[1]);
	    }
	}
#ifdef USE_IMLIB
	/* Run image browser. */
        else if(strcasepfx(strv[0], "%imageb") ||
                strcasepfx(strv[0], "%imgb") ||
                strcasepfx(strv[0], "%imbr") ||
                strcasepfx(strv[0], "%ib")
        )
	{
            /* Create Image Browser and map. */
	    DoCreateIB(edw->loc_prompt.buf);
	}
#endif /* USE_IMLIB */

#ifdef USE_IMLIB
        /* Load to internal image viewer? */
        else if(strcasepfx(strv[0], "%imagev") ||
                strcasepfx(strv[0], "%imgv") ||
                strcasepfx(strv[0], "%iv")
        )
        {
            /* Must have atleast two arguments. */
            if(strc > 1)
            {
                /* Create Internal Image Viewer and map. */
                DoCreateIV(strv[1]);
            }
        }
#endif /* USE_IMLIB */
	/* Run text editor. */
        else if(strcasepfx(strv[0], "%ed"))
	{
            /* Must have atleast two arguments. */
            if(strc > 1)
            {
		/* Run text editor. */
		FileOpRunTextEditor(strv[1]);
	    }
	}
	/* Execute command literally. */
	else
	{
	    char *tmp_cmd;


	    /* Remove previous program stdout spool file if it exists. */
	    SFRemoveProgStdout();

	    /* Format command. */
	    tmp_cmd = (char *)malloc((CMD_MAX + 1) * sizeof(char));
	    if(tmp_cmd != NULL)
	    {
		strncpy(tmp_cmd, cmd_prompt_buf, CMD_MAX);
		tmp_cmd[CMD_MAX - 1] = '\0';
		p = ExecAOE(tmp_cmd, fname.prog_output, fname.prog_output);

		/* Record process history. */
		ProcHistAdd(p, time(NULL));
	    }

	    free(tmp_cmd);
	}

	/* Free allocated strings. */
	StringFreeArray(strv, strc);

	return(-1);
}
