/*
                             E N D E A V O R

               Copyright (C) 1997, 1998, 1999, 2000 WolfPack
 */

#ifndef ENDEAVOUR_H
#define ENDEAVOUR_H



#include <stdio.h>
#include <errno.h>
extern int errno;
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <time.h>
extern char *tzname[2];
#include <signal.h>
#include <asm/types.h>

#include <sys/types.h>
#include <sys/stat.h>

#ifdef USE_IMLIB
# include "../include/Imlibosw.h"
#endif

#include "../include/os.h"
#include "../include/cfgfmt.h"
#include "../include/graphics.h"
#include "../include/strexp.h"
#include "../include/string.h"
#include "../include/prochandle.h"
#include "../include/fio.h"
#include "../include/disk.h"
#include "../include/osw-x.h"
#include "../include/widget.h"
#include "../include/tga.h"

#include "id.h"
#include "object.h"
#include "recbin.h"
#include "fass.h"
#include "config.h"



/*
 *	Program name and version:
 */
#define PROG_NAME	"Endeavour"
#define PROG_VERSION	"1.13.0"

#define VERSION_MAJOR	1
#define VERSION_MINOR	13

/*
 *	Program help message:
 */
#define PROG_HELP_MESG	"\
Usage: endeavour [options] [GUI_options] [startup_directory]\n\
\n\
    [options] can be any of the following:\n\n\
        --help                     Shows this message.\n\
        --version                  Shows version number.\n\
        --recyclebin               Startup in recycle bin.\n\
        --imagebrowser             Startup in image browser.\n\
\n\
    [startup_directory] is an absolute directory path to start up\n\
    in.\n\
\n\
    [GUI_options] can be any options standard to your GUI, consult\n\
    your GUI's manual for available options.\n\
\n"


/*
 *	Program copyright message:
 */
#define PROG_COPYRIGHT	"\
Copyright (C) 1997, 1998, 1999, 2000 WolfPack Development Organization."



/*
 *	Draw amount codes:
 *
 *	To be passed to *Draw() functions.
 */
#define DRAW_AMOUNT_COMPLETE		0
#define DRAW_AMOUNT_LISTS		1
#define DRAW_AMOUNT_DIR_LIST		2
#define DRAW_AMOUNT_FILE_LIST		3
#define DRAW_AMOUNT_FILE_HEADING	4
#define DRAW_AMOUNT_STATUS		5
#define DRAW_AMOUNT_MOUNT_BAR		6
#define DRAW_AMOUNT_DIR_LIST_OUTLINE	7
#define DRAW_AMOUNT_FILE_LIST_OUTLINE	8



/*
 *                      Drag information
 */
#define DRAG_TYPE_NONE		0
#define DRAG_TYPE_FILE  	1
#define DRAG_TYPE_DIR   	2
#define DRAG_TYPE_RECYCLE_BIN	3
#define DRAG_TYPE_FILE_HEADING	4
#define DRAG_TYPE_DIR_HEADING	5	/* Not supported. */
#define DRAG_TYPE_LIST_SPLIT	6	/* File and dir window split. */
#define DRAG_TYPE_THUMBS_LIST	7

#define POINTER_TYPE_UNDEFINED	0
#define POINTER_TYPE_EDW	1	/* Endeavour browser. */
#define POINTER_TYPE_RB		2	/* Recycle bin. */
#define POINTER_TYPE_IV		3	/* Image viewer. */
#define POINTER_TYPE_IB		4	/* Image browser. */


/*
 *      Operations codes:
 *
 *	These are used in fileop.c, endeavour pointer motion handling
 *	and the comfermation dialog.
 *
 */
#define OP_CODE_NONE		0
#define OP_CODE_TEST            1

#define OP_CODE_COPY            2
#define OP_CODE_MOVE            3
#define OP_CODE_RENAME		4
#define OP_CODE_CHMOD		5
#define OP_CODE_CHOWN_OWNER	6
#define OP_CODE_CHOWN_GROUP	7
#define OP_CODE_RECYCLE		8
#define OP_CODE_RECOVER		9

#define OP_CODE_RECBIN_PURGE		10
#define OP_CODE_RECBIN_PURGE_ALL	11

/*
 *	Comfermation button codes:
 */
#define YES_PBTN_NUM            0
#define NO_PBTN_NUM             1
#define ALL_PBTN_NUM            2


/* ***************************************************************** */

/*
 *	Program runlevel:
 */
extern int runlevel;

/*
 *	Current time since midnight in milliseconds:
 */
extern long cur_millitime;


/*
 *	Parameters and options:
 */
typedef struct {

	int rc_version_major,
	    rc_version_minor;

	char write_protect;	/* 0 = off, 1 = on. */

	/* Comfermations; 0 = off, 1 = on. */
	char	comferm_pointer_motion_move,
		comferm_pointer_motion_copy,
		comferm_overwrites,
		comferm_recycle,
		comferm_purge,
		comferm_purge_all,
		comferm_unmount_onexit;

	char show_prog_output;	/*   0 = no, 1 = if any
                                 *   2 = always
				 */

	char show_expandable_dirs;	/* Default setting for edws. */

	char animated_padlock;

	int file_exe_max_burst;		/*   Max external program executions
                                         *   per 'enter' press.
					 */

	time_t mount_timeout;		/* In seconds. */

	/* Recbin items warn. */
	int recbin_items_warn;
	off_t recbin_size_warn;

} endeavour_options_struct;
extern endeavour_options_struct option;


/*
 *	User and current process information:
 */
typedef struct {

	char name[NAME_MAX];	/* Guess NAME_MAX is the max for this. */
	char group[NAME_MAX];

	pid_t pid;	/* Our pid. */

	uid_t uid;
	uid_t euid;	/* Effective uid. */

	gid_t gid;
	gid_t egid;	/* Effective gid. */

} userinfo_struct;
extern userinfo_struct userinfo;


/*
 *	Directory names:
 */
typedef struct {

	char current_working[PATH_MAX];
	char home[PATH_MAX];			/* Local. */
	char prog_home[PATH_MAX];		/* Local. */
	char recycle_bin[PATH_MAX];

} dname_struct;
extern dname_struct dname;

/*
 *	File names:
 */
typedef struct {

	char rcfile[PATH_MAX + NAME_MAX];

	/* Do not confused this one with the one in struct dname. */
	char recycle_bin[PATH_MAX + NAME_MAX];

	/* Runned program stdout and stderr are redirected to this file. */
	char prog_output[PATH_MAX + NAME_MAX];
        char prog_output_ori[PATH_MAX + NAME_MAX];

	/* Archive program list output is redirected to this file. */
	char list_output[PATH_MAX + NAME_MAX];
        char list_output_ori[PATH_MAX + NAME_MAX];

	/* Communications FIFO pipe, for multiple processes to communicate. */
	char com_pipe[PATH_MAX + NAME_MAX];
        char com_pipe_ori[PATH_MAX + NAME_MAX];

} fname_struct;
extern fname_struct fname;


/*
 *	Operation commands:
 *
 *	This program uses external program calls to do various
 *	operations to ensure portability.
 *
 *	Note: These are NOT the command lines for file associations.
 *
 *	Redirects are *not* supported, stdout and stderr are automatically
 *	redirected to the file specified in the fname structure, member
 *	prog_output.
 */
typedef struct {

	char *move_file;
	char *move_dir;

	char *copy_file;
	char *copy_dir;

	char *rename_file;
	char *rename_dir;

	char *chmod_file;
	char *chmod_dir;

	char *chown_file;
	char *chown_dir;

	char *touch_file;
	char *touch_dir;

	char *text_editor;
	char *df;		/* Device find. */

} command_struct;
extern command_struct command;


/*
 *	Timmed event scheduals:
 */
typedef struct {

	/* Redraws. */
	long rd_padlock_glow;

	/* Checks. */
	long sanity;

} next_struct;
extern next_struct next;


/*
 *	Timmers:
 */
#define T_SGLOW_CODE	0
#define T_MGLOW_CODE    1
#define T_LGLOW_CODE    2
#define MAX_TIMMERS	3	/* 1 higher than highest timmer code. */

#define T_SGLOW_MAX	8	/* Higher numbers mean smoother gradiance. */
#define T_MGLOW_MAX	8
#define T_LGLOW_MAX	8

#define T_SGLOW_INT	90	/* Milliseconds. */
#define T_MGLOW_INT	190
#define T_LGLOW_INT	300

typedef struct {

	int val;
	int max;

	long i;		/* In milliseconds. */
	long next;	/* In milliseconds. */

} timmer_struct;
extern timmer_struct timmer[MAX_TIMMERS];



/*
 *      Fonts:
 */
typedef struct {

        font_t  *std,
                *bold;

} font_struct;
extern font_struct font;


/*
 *	Pointer cursors:
 */
typedef struct {

	/* GUI standard cursors. */
	cursor_t	list_heading_resize,
			cross_hairs,
			working;

	/* Widget cursors. */
	WCursor		*translate,
			*zoom;

} endeavour_cursor_struct;
extern endeavour_cursor_struct endeavour_cursor;

/*
 *	Endeavour images:
 */
typedef struct {

	image_t		*endeavour_icon,	/* Browser desktop icon. */

                        *recycle_bin_large0,	/* Recbin desktop icons. */
                        *recycle_bin_large1,

#ifdef USE_IMLIB
			*imbr_icon,	/* Image browser desktop icon. */
			*imbr_err_img,  /*   Images with errors are displayed
                                         *   as this (including their thumbs).
                                         */
#endif /* USE_IMLIB */

			*error_icon,
			*info_icon,
                        *purge_icon,
			*question_icon,
			*warning_icon,

			*padlock0,
			*padlock1,
			*padlock_tmp,		/* For glowing blitting. */

			*copy,			/* Tool ribbon btn images. */
			*folder_back,
			*folder_new,
			*move,
			*properties,
			*recycle,
			*refresh,

                        *new_browser,
			*tile,
			*recycle_bin_small0,
			*recycle_bin_small1;

	/* Device icons. */
	image_t		*dev_cdrom0,
			*dev_cdrom1,
			*dev_fd0,
			*dev_fd1,
			*dev_hd0,
			*dev_hd1,
			*dev_tape0,
			*dev_tape1,
			*dev_zip0,
			*dev_zip1;

	/* Computer icon. */
	image_t		*comp_l0,
			*comp_l1,
			*comp_s0,	/* Used for dir list. */
			*comp_s1;

} endeavour_image_struct;
extern endeavour_image_struct endeavour_image;


/*
 *	Yes to all global:
 *
 *	Used for multiple comfermation loops.
 */
extern bool_t yes_to_all;


/*
 *	Process execution history.
 */
typedef struct
{
	char is_running;	/* True if running. */

	pid_t pid;		/* Process's ID. */
	time_t s_time;		/* Start time, in systime seconds. */

} proc_hist_struct;

extern proc_hist_struct **proc_hist;
extern int total_proc_hists;



/* ***************************************************************** */

/*
 *	Directory item structure for Endeavour Browser's directory
 *	window.
 */
typedef struct {

	char *name;
	char *fullname;		/* Absolute path. */
	char *link_dest;

	int level;		/* Level starting from 0. */
	char expandable;	/* 0 = not expandable, 1 = expandable. */
	char expanded;		/* 0 = not expanded, 1 = expanded. */

	int x;			/* x coordinate on list window. */

	image_t *unsel_image, *sel_image;	/* Shared, do not free. */

	/* Stats. */
        ino_t inode;        /* Inode. */
        umode_t mode;       /* Mode/permissions/type. */
 
        nlink_t nlink;      /* Number of hard links. */
        uid_t uid;          /* User ID of owner */
        gid_t gid;          /* Group ID of owner */
        dev_t rdevice;      /* Device type (if inode device) */
        off_t size;         /* Total size, in blocks. */
        time_t last_accessed;       /* time of last access */
        time_t last_modified;       /* time of last modification */
        time_t create_time;         /* time of last change */

} edw_dir_struct;

/*
 *	Endeavour Browser window structure:
 */
#define CV_PROMPT_MODE_NONE		0
#define CV_PROMPT_MODE_RENAME		1
#define CV_PROMPT_MODE_CHMOD		2
#define CV_PROMPT_MODE_MOVE		3
#define CV_PROMPT_MODE_COPY		4
#define CV_PROMPT_MODE_CHOWN_OWNER	5
#define CV_PROMPT_MODE_CHOWN_GROUP	6

#define EDW_COL_NAME_MASK		(1 << 0)
#define EDW_COL_SIZE_MASK		(1 << 1)
#define EDW_COL_PERMISSION_MASK		(1 << 2)
#define EDW_COL_OWNER_MASK		(1 << 3)
#define EDW_COL_GROUP_MASK		(1 << 4)
#define EDW_COL_DATE_MASK		(1 << 5)

#define EDW_SHOW_EXPANDABLE_MASK	(1 << 0)
#define EDW_SHOW_BRANCH_MASK		(1 << 1)


typedef struct {

        int map_state;
        int visibility_state;
        char is_in_focus;
        int x, y;
        unsigned width, height;
        bool_t disabled;

	double dir_win_width;	/* In percent! */
	u_int32_t show_options;


	/* Toplevel */
        win_t toplevel;

	/* Tool ribbon window and buttons. */
	win_t tool_ribbon_win;
	pixmap_t tool_ribbon_win_buf;

        push_button_struct folder_back_btn;
        push_button_struct folder_refresh_btn;

        push_button_struct new_folder_btn;
        push_button_struct properties_btn;
        push_button_struct move_btn;   
        push_button_struct copy_btn;   
        push_button_struct recycle_btn;

	push_button_struct new_browser_btn;
	push_button_struct tile_btn;
        push_button_struct recbin_map_btn;

	/* Buttons on command bar. */
	push_button_struct	output_btn,	/* View output. */
				view_btn,
				edit_btn,
				kill_btn;


	/* Directory window. */
	win_t dir_win;
	pixmap_t dir_win_buf;
	scroll_bar_struct dir_win_sb;

	/* File window. */
	win_t file_win;
	pixmap_t file_win_buf;
	scroll_bar_struct file_win_sb;
	win_t file_heading_win;			/* File list heading. */
        pixmap_t file_heading_win_buf;

        win_t split_cur;        /* Vertical split cursor shown when splitting
                                 * the list and viewer windows.
                                 */

	/* Status bar (horizontal bar at the bottom). */
	win_t status_win;
	pixmap_t status_win_buf;
        progress_bar_struct pbar;


	u_int32_t	file_headers_flag;	/* Items shown in file window. */

	int	col_name_pos,	/* All colum positions are in pixels. */
		col_size_pos,
		col_permission_pos,
		col_owner_pos,
		col_group_pos,
		col_date_pos;


	/* Directory objects. */
	edw_dir_struct **dir;
        int total_dirs;
	int *sel_dir;
	int total_sel_dirs;
        menu_struct dir_qmenu;
	unsigned int max_dir_width; /* Width info for scroll bar. */

	/* File objects. */
	object_struct **file;
        int total_files;
	int *sel_file;
	int total_sel_files;
	menu_struct file_qmenu;


	/* Location ribbon. */
	prompt_window_struct loc_prompt;
        win_t loc_tool_ribbon;
        pixmap_t loc_tool_ribbon_buf;

	/* Command ribbon. */
	prompt_window_struct cmd_prompt;
        win_t cmd_tool_ribbon;    
        pixmap_t cmd_tool_ribbon_buf;


	/* Change values prompt. */
	prompt_window_struct cv_prompt;
	int cv_prompt_mode;
	char *cv_src_path;		/* Path and name. */

	/* Status message. */
	char *status_mesg;

} endeavour_win_struct;

extern endeavour_win_struct **endeavour_win;
extern int total_endeavour_wins;



/*
 *      Primary file viewer:
 *
 *	These are also known as `Internal viewers', using the widget set's
 *	viewer widget.
 */
extern file_viewer_struct **pri_fviewer;
extern int total_pri_fviewers;


/*
 *	Error dialog:
 */
extern dialog_win_struct err_dw;


/*
 *	Operations dialog:
 *
 *	This dialog is used for prompting copy, move, rename, etc.
 */
typedef struct {

        int map_state;
        int visibility_state;
        char is_in_focus;
        int x, y;
        unsigned width, height;
        bool_t disabled;

	int op_code;

        /* Toplevel */
        win_t toplevel;
	pixmap_t toplevel_buf;

	win_t status_win;
	pixmap_t status_win_buf;
	char *status_mesg;

	prompt_window_struct tar_prompt;
	prompt_window_struct src_prompt;

	push_button_struct copy_btn;
	push_button_struct move_btn;
	push_button_struct rename_btn;
	push_button_struct ok_btn;
	push_button_struct cancel_btn;

	progress_bar_struct pbar;

	/* List of objects to operate on. */
	char **object;
	int total_objects;

	int last_object;	/* Last object operated on. */

	char *current_dir;

} op_dialog_struct;
extern op_dialog_struct **op_dialog;
extern int total_op_dialogs;


/*
 *	Comfermation dialog:
 */
#define CDIALOG_QUERY_NONE		0
#define CDIALOG_QUERY_CONFIRM_MOTION	1
#define CDIALOG_QUERY_OVERWRITE		2
#define CDIALOG_QUERY_RECBIN_PURGE	3	/* Recycle bin purge. */
#define CDIALOG_QUERY_RECBIN_PURGE_ALL	4

#define CDIALOG_QUERY_DEVICE_MOUNT	5
#define CDIALOG_QUERY_DEVICE_UNMOUNT	6 


typedef struct {

        int map_state;
        int visibility_state;
        char is_in_focus;
        int x, y;
        unsigned width, height;
        bool_t disabled;

	win_t toplevel;
	pixmap_t toplevel_buf;

	push_button_struct yes_btn;
	push_button_struct no_btn;
        push_button_struct all_btn;


	/* Query type. */
	int query_type;		/* CDIALOG_QUERY_* */
	int op_code;		/* Operation code. */

	/* To comferm overwrite. */
	char *tar_path;
	char *src_path;

	/* To comferm motion. */
	void *tar_ptr, *src_ptr;
	int tar_ptr_type, src_ptr_type;
	int tar_drag_type, src_drag_type;

        int tar_item_num;

} comfermation_dialog_struct;

extern comfermation_dialog_struct comf_dialog;



/*
 *	Properties dialog structure:
 */
typedef struct {

        int map_state;
        int visibility_state;
        char is_in_focus;
        int x, y;
        unsigned width, height;
        bool_t disabled;

        win_t toplevel;
        pixmap_t toplevel_buf;
  
        push_button_struct ok_btn;
        push_button_struct cancel_btn;
        push_button_struct apply_btn;

	push_button_struct touch_btn;


	prompt_window_struct name_prompt;
	prompt_window_struct link_dest_prompt;

	prompt_window_struct owner_prompt;
	prompt_window_struct group_prompt;

	/* Permissions toggle buttons. */
	toggle_button_struct	perm_uid_tb,
				perm_gid_tb,
				perm_sticky_tb,
				perm_usrr_tb,
				perm_usrw_tb,
				perm_usrx_tb,
				perm_grpr_tb,
				perm_grpw_tb,
				perm_grpx_tb,
				perm_othr_tb,
				perm_othw_tb,
				perm_othx_tb;


	/* Original object information. */
	object_struct ori_info;


} prop_dialog_struct;

extern prop_dialog_struct **prop_dialog;
extern int total_prop_dialogs;



#ifdef USE_IMLIB
/*
 *	Internal image viewer:
 *
 *	Available only when USE_IMLIB is defined.
 */
#define IV_ZOOM_MIN     0.10
#define IV_ZOOM_MAX 16384.00
#define IV_ZOOM_NORMAL	1.00
typedef struct {

        int map_state;
        int visibility_state;
        char is_in_focus;
        int x, y;
        unsigned width, height;
        bool_t disabled;

        win_t toplevel;
        pixmap_t toplevel_buf;

        win_t view_screen;
	scroll_bar_struct view_screen_sb; /*   Only used for storing
                                           *   coordinates.
					   */
	shared_image_t *view_screen_data;

	win_t console;
	pixmap_t console_buf;

	push_button_struct close_btn;
	push_button_struct reload_btn;

	menu_struct qmenu;

	image_t *image_data;

	double zoom;	/* IV_ZOOM_MIN to IV_ZOOM_MAX */

	char *filename;

} img_viewer_struct;
extern img_viewer_struct **img_viewer;
extern int total_img_viewers;

#endif /* USE_IMLIB */


#ifdef USE_IMLIB
/*
 *	Image browser
 *
 *      Available only when USE_IMLIB is defined.
 */
typedef struct {

	char *filename;		/* Not absolute path. */
	image_t *image;

	size_t ori_size;	/* In bytes. */
	unsigned int ori_width, ori_height;

} img_browser_thumb_struct;

typedef struct {

        int map_state;
        int visibility_state;
        char is_in_focus;
        int x, y;
        unsigned width, height;
        bool_t disabled;
  
        win_t toplevel;
        pixmap_t toplevel_buf;

	double list_win_width;		/* Coefficient. */
	win_t list;
	pixmap_t list_buf;
	scroll_bar_struct list_sb;

	win_t viewer;
	pixmap_t viewer_buf;
	/* No scrolling for viewer. */

	win_t split_cur;	/* Vertical split cursor shown when splitting
                                 * the list and viewer windows.
				 */

	win_t loc_ribbon;
	pixmap_t loc_ribbon_buf;

	win_t status_win;
	pixmap_t status_win_buf;
	char *status_mesg;
	progress_bar_struct pbar;

	push_button_struct	prev_btn,
				next_btn,
				view_btn,
				reload_btn,
				stop_btn;

	prompt_window_struct cv_prompt;
	int cv_prompt_mode;
	prompt_window_struct loc_prompt;
	char *current_dir;

	menu_struct qmenu;

	/* Thumbs. */
	img_browser_thumb_struct **thumb;
	int total_thumbs;

	int total_sel_thumbs;
	int *sel_thumb;

	unsigned int thumbs_per_row;
	unsigned int thumb_width, thumb_height;
	unsigned int thumb_label_height;

	int thumb_col_margin, thumb_row_margin;

	unsigned int	list_row_spacing,
			list_col_spacing;

	/* File load queue list. */
	char **fq_name;
	int total_fqs;
	int last_fq;
	bool_t stop_load;	/* Stop loading images if True. */

} img_browser_struct;
extern img_browser_struct **img_browser;
extern int total_img_browsers;

#endif /* USE_IMLIB */


/*
 *	Device record structure:
 *
 *	Records for mounted devices.
 */
#define DEV_TF_PHYSICAL		(1 << 1)	/* False = memory. */
#define DEV_TF_FIXED		(1 << 2)	/* False = removeable. */
#define DEV_TF_READONLY		(1 << 3)	/* False = read/write. */
#define DEV_TF_LOCAL		(1 << 10)	/* False = remote. */
#define DEV_TF_NOUNMOUNT	(1 << 11)	/* False = unmountable. */

#define RECDEV_IMG_HD		0
#define RECDEV_IMG_FD		1
#define RECDEV_IMG_CDROM	2
#define RECDEV_IMG_TAPE		3
#define RECDEV_IMG_ZIP		4
#define RECDEV_IMG_MEMORY	5
#define RECDEV_IMG_NFS		6

typedef struct {

	char *label;		/* Friendly name. */
        u_int64_t flags;        /* Flags (not complaint to any standard). */

	char *devname;		/* Device path. */
	char *path;		/* Where it is/would be mounted too. */
	char *filesystem;	/* File system name. */
	char is_mounted;

	int image_code;		/* Icon that should be used for it. */

	char *cmd_mount;
	char *cmd_unmount;
	char *cmd_status;	/* Not used. */


	/* Statistical information (units depend on platform). */
	off_t capacity_total;
	off_t capacity_used;
	off_t capacity_available;
	double capacity_percent;

} devrec_struct;

extern devrec_struct **devrec;
extern int total_devrecs;




/*
 *	Drag record:
 *
 *	For recording drag states on the endeavour windows.
 */
typedef struct {
        
        bool_t  state;          /* True if drag is active. */
        
        bool_t  button1_state,
                button2_state,
                button3_state;
                                           
        bool_t  shift_key_state,
                alt_key_state,
                ctrl_key_state;
   
        /* position drag started. */
        int start_x, start_y;

	int type;		/* Source, one of DRAG_TYPE_*. */

        int origin_ptr_type;	/* Source, one of POINTER_TYPE_* */
	void *origin_ptr;	/* Pointer to source window struct. */

        /* For dragging in file heading. */
        u_int32_t heading_col_flag;
 
} drag_record_struct;
extern drag_record_struct drag_record[1];



/* ***************************************************************** */

/* In comfdialog.c */
extern int ComfDialogYesPBCB(void *ptr);
extern int ComfDialogNoPBCB(void *ptr);
extern int ComfDialogAllPBCB(void *ptr);

extern int ComfDialogInit();
extern int ComfDialogDraw();
extern int ComfDialogManage(event_t *event);
extern void ComfDialogMapQueryMovement(
	void *tar_ptr,
        int tar_ptr_type,
        void *src_ptr,
        int src_ptr_type,
        int tar_drag_type,
        int src_drag_type,
        int tar_item_num,
        int op_code
);
extern int ComfDialogMapQueryOverWrite(
        char *tar_path,
        char *src_path,
        int op_code
);
extern int ComfDialogMapQueryPurge(
        recbin_win_struct *rb,
        char *src_path
);
extern int ComfDialogMapQueryPurgeAll(
        recbin_win_struct *rb
);
extern int ComfDialogMapQueryMount(
        char *src_path
);
extern int ComfDialogMapQueryUnmount(
        char *src_path
);
extern void ComfDialogResetQuery();
extern void ComfDialogUnmap();
extern void ComfDialogDestroy();


/* In cursor.c */
extern int LoadAllCursors();
extern void DestroyAllCursors();


/* In devrec.c */
extern int DevRecIsReallyMounted(char *dev);
extern int DevRecIsAllocated(int dev_num);

extern int DevRecDoMount(int dev_num);
extern int DevRecDoUnmount(int dev_num);

extern int DevRecRefreshList();
extern int DevRecAllocate();
extern void DevRecDelete(int rec_num);
extern void DevRecDeleteAllRecords();
extern void DevRecDoComfUnmountAll();


/* In drag.c */
extern void DragDoAllWindowsActive();
extern void DragCleanup();
extern int DragManage(event_t *event);


/* In edwalloc.c */
extern int EDWIsAllocated(int edw_num);
extern int EDWIsInit(int edw_num);
extern int EDWAllocate();
extern void EDWReclaimMemory();
extern void EDWDelete(int edw_num);
extern void EDWDeleteAll();

/* In edwcb.c */
extern int EDWPerformFileOp(
        int op_code,
        char *tar_path,
        char *src_path 
);
extern int EDWProcessMovement(
        void *tar_ptr,
        int tar_ptr_type,       /* Always POINTER_TYPE_EDW. */
        void *src_ptr,
        int src_ptr_type,

        int tar_drag_type,
        int src_drag_type, 
        int tar_item_num,
        int op_code,
        bool_t comfermed
);
extern int EDWExecuteSelectedFiles(
        endeavour_win_struct *edw, 
        bool_t execute
);
extern int EDWMenuCB(void *client_data, int id_code);


/* In edwevent.c */
extern int EDWInit( 
        endeavour_win_struct *edw,
        int x, int y, 
        unsigned int width, unsigned int height,
        char *start_path
);
extern int EDWResize(endeavour_win_struct *edw);
extern int EDWDraw(endeavour_win_struct *edw, int amount);
extern int EDWManage(endeavour_win_struct *edw, event_t *event);
extern int EDWManageAll(event_t *event);
extern void EDWMap(endeavour_win_struct *edw);
extern void EDWUnmap(endeavour_win_struct *edw);
extern void EDWDestroy(endeavour_win_struct *edw);


/* In edwlist.c */
extern int EDWDoesDirHaveSubDirs(char *path);
extern int EDWMatchDirFromPath(endeavour_win_struct *edw, char *path);
extern int EDWMatchExistingDirFromPath(endeavour_win_struct *edw, char *path);
extern int EDWScrollToDirItem(endeavour_win_struct *edw, int dir_num);
extern int EDWScrollToFileItem(endeavour_win_struct *edw, int dir_num);
extern int EDWUpdateDirWinWidth(endeavour_win_struct *edw);

extern int EDWInitToplevelDir(endeavour_win_struct *edw);
extern int EDWUpdateDirItemName(
        endeavour_win_struct *edw,
        char *old_name,
        char *new_name
);

extern void EDWDeleteDirItemPath(endeavour_win_struct *edw, char *path);

extern void EDWFreeFileListItem(object_struct **obj);
extern void EDWFreeDirListItem(edw_dir_struct **dir);

extern void EDWFreeFileListing(endeavour_win_struct *edw);
extern void EDWFreeDirListing(endeavour_win_struct *edw);
extern void EDWFreeAllListings(endeavour_win_struct *edw);

extern int EDWExpandDir(endeavour_win_struct *edw, int n);
extern void EDWColapseDir(endeavour_win_struct *edw, int n);

extern int EDWRefreshDirItem(endeavour_win_struct *edw, int dir_num);
extern int EDWRefreshFileItem(endeavour_win_struct *edw, int file_num);
extern void EDWRefreshDirList(endeavour_win_struct *edw);
extern void EDWRefreshFileList(endeavour_win_struct *edw, char *path);
extern void EDWRefreshAllLists(endeavour_win_struct *edw);


/* In edwmacros.c */
extern int EDW_GETDIRNUMBYPOS(endeavour_win_struct *edw, int x, int y);
extern int EDW_GETFILENUMBYPOS(endeavour_win_struct *edw, int x, int y);
extern char *EDW_GETTIMESTR(time_t seconds);
extern char *EDW_GETOWNERNAME(uid_t uid);
extern char *EDW_GETGROUPNAME(gid_t gid);

#define EDW_CURSORS_CODE_NORMAL		0
#define EDW_CURSORS_CODE_WORKING	1
extern void EDWSETALLCURSORS(endeavour_win_struct *edw, int code);
extern void EDWSANITIZEFILECOLSPACING(endeavour_win_struct *edw);

extern int EDWDoMapCVPromptDir(endeavour_win_struct *edw,
	int dir_num, int mode);
extern int EDWDoMapCVPromptFile(endeavour_win_struct *edw,
        int file_num, int mode);

extern int EDWDoGoParentDir(endeavour_win_struct *edw);
extern int EDWDoRefresh(endeavour_win_struct *edw);

extern int EDWDoQCreateNewDir(endeavour_win_struct *edw);
extern int EDWDoQCreateNewLink(endeavour_win_struct *edw);
extern int EDWDoEditProperties(endeavour_win_struct *edw);
extern int EDWDoMove(endeavour_win_struct *edw);
extern int EDWDoCopy(endeavour_win_struct *edw);
extern int EDWDoRecycle(endeavour_win_struct *edw);

extern int EDWDoNewBrowser(endeavour_win_struct *edw);
extern int EDWDoTileBrowser(endeavour_win_struct *edw);
extern int EDWDoMapRecycleBin(endeavour_win_struct *edw);

extern int EDWDoExecCmdPrompt(endeavour_win_struct *edw);


/* In edwselect.c */
extern int EDW_ISDIRSELECTED(endeavour_win_struct *edw, int dir_num);
extern int EDW_ISFILESELECTED(endeavour_win_struct *edw, int file_num);
extern int EDW_SELECTDIR(endeavour_win_struct *edw, int dir_num);
extern int EDW_SELECTFILE(endeavour_win_struct *edw, int file_num);
extern void EDW_UNSELECTDIR(endeavour_win_struct *edw, int dir_num);
extern void EDW_UNSELECTFILE(endeavour_win_struct *edw, int file_num);
extern void EDW_UNSELECTALLDIRS(endeavour_win_struct *edw);
extern void EDW_UNSELECTALLFILES(endeavour_win_struct *edw);

extern int EDWDoSelectDir(endeavour_win_struct *edw, int dir_num);


/* In fileop.c */
extern void STRINGSUBTARGET(char *string, char *sub);
extern void STRINGSUBSOURCE(char *string, char *sub);
extern void STRINGSUBARG(char *string, char *sub);
extern int DESTEXISTS(char *tar_path, char *src_path);
extern int ISCMDREMOVE(char *cmd);

extern int FileOpMoveFile(char *tar_path, char *src_path);
extern int FileOpMoveDir(char *tar_path, char *src_path);

extern int FileOpCopyFile(char *tar_path, char *src_path);
extern int FileOpCopyDir(char *tar_path, char *src_path);

extern int FileOpRenameFile(char *tar_path, char *src_path);
extern int FileOpRenameDir(char *tar_path, char *src_path);

extern int FileOpChmodFile(char *tar_path, char *mode_str);
extern int FileOpChmodDir(char *tar_path, char *mode_str);

extern int FileOpChownOwner(char *tar_path, char *owner);
extern int FileOpChownGroup(char *tar_path, char *group);

extern int FileOpTouchFile(char *tar_path);
extern int FileOpTouchDir(char *tar_path);

extern int FileOpRunTextEditor(char *tar_path);


/* In images.c */
extern int LoadEndeavourImages();
extern void FreeEndeavourImages();



#ifdef USE_IMLIB
/* In imbralloc.c */
extern int IMBRIsAllocated(int n);
extern int IMBRIsInit(int n);

extern int IMBRAllocate();
extern void IMBRReclaimMemory();
extern void IMBRDelete(int n);
extern void IMBRDeleteAll();

/* In imbrcb.c */
extern int IMBRRepeatCB();
extern int IMBRProcessMovement(
        void *tar_ptr, int tar_ptr_type,
        void *src_ptr, int src_ptr_type,
        int tar_drag_type, int src_drag_type,
        int tar_item_num, int op_code,
        bool_t comfermed
);
extern int IMBRMenuCB(void *client_data, int id_code);


/* In imbrevent.c */
extern int IMBRInit(img_browser_struct *ib);
extern int IMBRResize(img_browser_struct *ib);
extern int IMBRDraw(img_browser_struct *ib, int amount);
extern int IMBRManage(img_browser_struct *ib, event_t *event);
extern int IMBRManageAll(event_t *event);
extern void IMBRMap(img_browser_struct *ib);
extern void IMBRUnmap(img_browser_struct *ib);
extern void IMBRDestroy(img_browser_struct *ib);


/* In imbrmacros.c */
extern int IMBR_GETTHUMBFROMPOS(img_browser_struct *ib, int x, int y);
extern void IMBR_SETALLCURSORS(img_browser_struct *ib, int code);

extern int IMBRSetStatusMesg(img_browser_struct *ib, char *mesg);

extern int IMBRDisplayOnRoot(img_browser_struct *ib, int n);
extern int IMBRDoChangeDir(img_browser_struct *ib, char *path);
extern int IMBRLoad(img_browser_struct *ib, char *path);
extern int IMBRLoadImageFromThumb(img_browser_struct *ib, int n);
extern int IMBRLoadThumbFromQueue(img_browser_struct *ib, int fq_num);

extern int IMBRDoPrev(img_browser_struct *ib);
extern int IMBRDoNext(img_browser_struct *ib);
extern int IMBRDoView(img_browser_struct *ib);
extern int IMBRDoReload(img_browser_struct *ib);
extern int IMBRDoStop(img_browser_struct *ib);

extern int IMBRDoQuickRename(img_browser_struct *ib);
extern int IMBRDoMove(img_browser_struct *ib);
extern int IMBRDoCopy(img_browser_struct *ib);
extern int IMBRDoRecycle(img_browser_struct *ib);
extern int IMBRDoPropDialog(img_browser_struct *ib);

extern int IMBRSelect(img_browser_struct *ib, int n);
extern void IMBRUnselect(img_browser_struct *ib, int n);
extern int IMBRIsSelected(img_browser_struct *ib, int n);
extern void IMBRDeleteFileQueue(img_browser_struct *ib);
extern void IMBRUnselectAll(img_browser_struct *ib);

extern void IMBRDeleteAllThumbs(img_browser_struct *ib);

#endif /* USE_IMLIB */


#ifdef USE_IMLIB
/* In imgviewer.c */
extern int ImgViewerMenuCB(void *client_data, int id_code);
extern void ImgViewerRepeatCB();

extern int ImgViewerIsAllocated(int n);
extern int ImgViewerIsInit(int n);
extern int ImgViewerAllocate();
extern void ImgViewerReclaimMemory();
extern void ImgViewerDelete(int n);
extern void ImgViewerDeleteAll();

extern int ImgViewerLoadFile(img_viewer_struct *iv, char *filename);
extern void ImgViewerUnload(img_viewer_struct *iv);
extern int ImgViewerInit(img_viewer_struct *iv);
extern int ImgViewerDraw(img_viewer_struct *iv, int amount);
extern int ImgViewerResize(img_viewer_struct *iv);
extern int ImgViewerManage(img_viewer_struct *iv, event_t *event);
extern int ImgViewerManageAll(event_t *event);
extern void ImgViewerMap(img_viewer_struct *iv);
extern void ImgViewerUnmap(img_viewer_struct *iv);
extern void ImgViewerDestroy(img_viewer_struct *iv);


#endif /* USE_IMLIB */



/* In install.c */
extern int InstallDefaultFiles();


/* In intviewer.c (internal viewers, not image viewers) */
extern int IntViewerIsAllocated(int viewer_num);
extern int IntViewerIsInit(int viewer_num);

extern int IntViewerAllocate();
extern int IntViewerReclaimMemory();
extern int IntViewerManageAll(event_t *event);
extern void IntViewerDelete(int viewer_num);
extern void IntViewerDeleteAll();


/* In main.c */
extern int ScrollBarNotifyHandler(scroll_bar_struct *sb);

extern int DoCreateEDW(char *path);
extern int DoCreateRB(char *path);
extern int DoCreateV(char *path);
#ifdef USE_IMLIB
extern int DoCreateIV(char *path);
extern int DoCreateIB(char *path);
#endif /* USE_IMLIB */
extern int DoCreatePropDialog(char *path);

extern void UnfocusAllTopLvlWins();
extern void ManagePadlockDraw();
extern void ResetGlobalTimming();
extern int SanityCheck();
extern void FreeGlobalMemory();
extern int ManageEvents();


/* In opdialog.c */
extern int OPDialogAddObject(
	op_dialog_struct *od,
	int op_code,
        char *current_dir,
	char *object
);
extern void OPDialogClearObjects(op_dialog_struct *od);

extern void OPDialogClearObjects(op_dialog_struct *od);
extern int OPDialogApply(op_dialog_struct *od);

extern int OPDialogInit(op_dialog_struct *od);
extern int OPDialogResize(op_dialog_struct *od);
extern int OPDialogDraw(op_dialog_struct *od);
extern int OPDialogManage(op_dialog_struct *od, event_t *event);
extern void OPDialogMap(op_dialog_struct *od);
extern void OPDialogUnmap(op_dialog_struct *od);
extern void OPDialogDestroy(op_dialog_struct *od);


/* In prochist.c */
extern pid_t ProcHistLastPid();
extern int ProcHistIsAllocated(int n);
extern int ProcHistAdd(
	pid_t pid,
        time_t s_time
);
extern void ProcHistDelete(int n);
extern void ProcHistDeleteAll();
extern int ProcHistKill(int n);


/* In propdialog.c */
extern int PropDialogIsAllocated(int dialog_num);
extern int PropDialogIsInit(int dialog_num);

extern int PropDialogLoad(prop_dialog_struct *pd, char *path);
extern int PropDialogApply(prop_dialog_struct *pd);

extern int PropDialogReclaimMemory(); 
extern int PropDialogAllocate();
extern void PropDialogDelete(int pd_num);
extern void PropDialogDeleteAll();
extern int PropDialogManageAll(event_t *event);

extern int PropDialogInit(prop_dialog_struct *pd);
extern int PropDialogResize(prop_dialog_struct *pd);
extern int PropDialogDraw(prop_dialog_struct *pd);
extern int PropDialogManage(prop_dialog_struct *pd, event_t *event);
extern void PropDialogMap(prop_dialog_struct *pd);
extern void PropDialogUnmap(prop_dialog_struct *pd);
extern void PropDialogDestroy(prop_dialog_struct *pd);
 

/* In rcfile.c */
extern int RCLoadFromFile(char *filename);


/* In sighandle.c */
extern void HandleSignal(int da_sig);


/* In spoolfiles.c */
extern void SFRemoveProgStdout();
extern void SFRemoveList();
extern void SFRemoveAll();
extern int SFDoStartupWarn();
extern int SFDoShutdownWarn();
extern int SFViewerProgStdout();
extern int SFShowOutputOnDialog(dialog_win_struct *dw);


/* In timming.c */
extern long MilliTime(void);   


/* In timmer.c */
extern int TimmerInitAll();
extern void TimmerResetAll();
extern void TimmerManageAll();


/* In xerror.c */
extern int EDWXErrorHandler(Display *da_display, XErrorEvent *da_error_event);


#endif /* ENDEAVOUR_H */
