/*
                          Endeavour Native Image Loading

	Functions:

	int LoadEndeavourImages()
	void FreeEndeavourImages()



 */

#include "endeavour.h"

#include "images/endeavour_icon.h"

#include "images/recycle_bin_large0.h"
#include "images/recycle_bin_large1.h"

#ifdef USE_IMLIB
#include "images/imbr_icon.h"
#include "images/imbr_err_img.h"
#endif /* USE_IMLIB */

#include "../include/icons/error.h"
#include "../include/icons/info.h"
#include "../include/icons/purge.h"
#include "../include/icons/question.h"
#include "../include/icons/warning.h"

#include "images/padlock0.h"
#include "images/padlock1.h"

#include "images/copy.h"
#include "images/new_browser.h"
#include "images/folder_back.h"
#include "images/folder_new.h"
#include "images/move.h"
#include "images/properties.h"
#include "images/recycle.h"
#include "images/recycle_bin_small0.h"
#include "images/recycle_bin_small1.h"  
#include "images/refresh.h"
#include "images/tile.h"

#include "images/dev_cd0.h"
#include "images/dev_cd1.h"
#include "images/dev_fd0.h"
#include "images/dev_fd1.h"
#include "images/dev_hd0.h"
#include "images/dev_hd1.h"
#include "images/dev_zip0.h"
#include "images/dev_zip1.h"

#include "images/comp_l0.h"
#include "images/comp_l1.h"
#include "images/comp_s0.h"
#include "images/comp_s1.h"



int LoadEndeavourImages()
{
        image_t *image;


        /* Error checks. */
        if(!IDC())
        {
            fprintf(stderr, "LoadAllImages(): Display not connected.\n");
            return(-1);
        }


        /* Endeavour desktop icon. */
        image = WidgetLoadImageFromTgaData(endeavour_icon_tga);
        if(image == NULL) return(-1);
        endeavour_image.endeavour_icon = image;

        /* Recycle bin desktop icons. */
        image = WidgetLoadImageFromTgaData(recycle_bin_large0_tga);
        if(image == NULL) return(-1);
        endeavour_image.recycle_bin_large0 = image;
        
        image = WidgetLoadImageFromTgaData(recycle_bin_large1_tga);
        if(image == NULL) return(-1);
        endeavour_image.recycle_bin_large1 = image;

#ifdef USE_IMLIB
	/* Image browser desktop icon. */
        image = WidgetLoadImageFromTgaData(imbr_icon_tga);
        if(image == NULL) return(-1);
        endeavour_image.imbr_icon = image;

	/* Image browser failed load image. */
        image = WidgetLoadImageFromTgaData(imbr_err_img_tga);
        if(image == NULL) return(-1);
        endeavour_image.imbr_err_img = image;
#endif /* USE_IMLIB */
        
        
        /* Error icon. */
        image = WidgetLoadImageFromTgaData(error_tga);
        if(image == NULL) return(-1);
        endeavour_image.error_icon = image;
        
        image = WidgetLoadImageFromTgaData(info_tga);
        if(image == NULL) return(-1);
        endeavour_image.info_icon = image;

        image = WidgetLoadImageFromTgaData(purge_tga);
        if(image == NULL) return(-1);
        endeavour_image.purge_icon = image;

        image = WidgetLoadImageFromTgaData(question_tga);
        if(image == NULL) return(-1);
        endeavour_image.question_icon = image;

        image = WidgetLoadImageFromTgaData(warning_tga);
        if(image == NULL) return(-1);
        endeavour_image.warning_icon = image;


        /* Padlock images. */
        image = WidgetLoadImageFromTgaData(padlock0_tga);
        if(image == NULL) return(-1);
        endeavour_image.padlock_tmp = image;
        
        image = WidgetLoadImageFromTgaData(padlock0_tga);
        if(image == NULL) return(-1);
        endeavour_image.padlock0 = image;

        image = WidgetLoadImageFromTgaData(padlock1_tga);
        if(image == NULL) return(-1);
        endeavour_image.padlock1 = image;
        
        
        /* Tool bar images. */
        image = WidgetLoadImageFromTgaData(copy_tga);
        if(image == NULL) return(-1);
        endeavour_image.copy = image;
 
        image = WidgetLoadImageFromTgaData(folder_back_tga);
        if(image == NULL) return(-1);
        endeavour_image.folder_back = image;
        
        image = WidgetLoadImageFromTgaData(folder_new_tga);
        if(image == NULL) return(-1);
        endeavour_image.folder_new = image;
        
        image = WidgetLoadImageFromTgaData(move_tga);
        if(image == NULL) return(-1);
        endeavour_image.move = image;
         
        image = WidgetLoadImageFromTgaData(properties_tga);
        if(image == NULL) return(-1);
        endeavour_image.properties = image;

        image = WidgetLoadImageFromTgaData(recycle_tga);
        if(image == NULL) return(-1);
        endeavour_image.recycle = image;

        image = WidgetLoadImageFromTgaData(refresh_tga);
        if(image == NULL) return(-1);
        endeavour_image.refresh = image;
        
        
        image = WidgetLoadImageFromTgaData(new_browser_tga);
        if(image == NULL) return(-1);
        endeavour_image.new_browser = image;
        
        image = WidgetLoadImageFromTgaData(tile_tga);
        if(image == NULL) return(-1);
        endeavour_image.tile = image;

        image = WidgetLoadImageFromTgaData(recycle_bin_small0_tga);
        if(image == NULL) return(-1);
        endeavour_image.recycle_bin_small0 = image;
        
        image = WidgetLoadImageFromTgaData(recycle_bin_small1_tga);
        if(image == NULL) return(-1);
        endeavour_image.recycle_bin_small1 = image;

        
        /* Device icons. */
        image = WidgetLoadImageFromTgaData(dev_cd0_tga);
        if(image == NULL) return(-1);
        endeavour_image.dev_cdrom0 = image;

        image = WidgetLoadImageFromTgaData(dev_cd1_tga);
        if(image == NULL) return(-1);
        endeavour_image.dev_cdrom1 = image;
        
        
        image = WidgetLoadImageFromTgaData(dev_fd0_tga);
        if(image == NULL) return(-1);
        endeavour_image.dev_fd0 = image;
        
        image = WidgetLoadImageFromTgaData(dev_fd1_tga);
        if(image == NULL) return(-1);
        endeavour_image.dev_fd1 = image;
        
 
        image = WidgetLoadImageFromTgaData(dev_hd0_tga);
        if(image == NULL) return(-1);
        endeavour_image.dev_hd0 = image;
        
        image = WidgetLoadImageFromTgaData(dev_hd1_tga);
        if(image == NULL) return(-1);
        endeavour_image.dev_hd1 = image;

        
/*
        image = WidgetLoadImageFromTgaData(dev_tape0_tga);
        if(image == NULL) return(-1);
        endeavour_image.dev_tape0 = image;

        image = WidgetLoadImageFromTgaData(dev_tape1_tga);
        if(image == NULL) return(-1);
        endeavour_image.dev_tape1 = image;
*/

        image = WidgetLoadImageFromTgaData(dev_zip0_tga);
        if(image == NULL) return(-1);
        endeavour_image.dev_zip0 = image;
        
        image = WidgetLoadImageFromTgaData(dev_zip1_tga);
        if(image == NULL) return(-1);
        endeavour_image.dev_zip1 = image;


	/* Computer icon. */
        image = WidgetLoadImageFromTgaData(comp_l0_tga);
        if(image == NULL) return(-1);
        endeavour_image.comp_l0 = image;

        image = WidgetLoadImageFromTgaData(comp_l1_tga);
        if(image == NULL) return(-1);
        endeavour_image.comp_l1 = image;


        image = WidgetLoadImageFromTgaData(comp_s0_tga);
        if(image == NULL) return(-1);
        endeavour_image.comp_s0 = image;

        image = WidgetLoadImageFromTgaData(comp_s1_tga);
        if(image == NULL) return(-1);
        endeavour_image.comp_s1 = image;



	return(0);
}



void FreeEndeavourImages()
{
	if(!IDC())
	{
	    return;
	}



        OSWDestroyImage(&endeavour_image.endeavour_icon);

        OSWDestroyImage(&endeavour_image.recycle_bin_large0);
        OSWDestroyImage(&endeavour_image.recycle_bin_large1);

#ifdef USE_IMLIB
        OSWDestroyImage(&endeavour_image.imbr_icon);
        OSWDestroyImage(&endeavour_image.imbr_err_img);
#endif /* USE_IMLIB */

        OSWDestroyImage(&endeavour_image.error_icon);
        OSWDestroyImage(&endeavour_image.info_icon);
        OSWDestroyImage(&endeavour_image.purge_icon);
        OSWDestroyImage(&endeavour_image.question_icon);
        OSWDestroyImage(&endeavour_image.warning_icon);

        OSWDestroyImage(&endeavour_image.padlock0); 
        OSWDestroyImage(&endeavour_image.padlock1);
        OSWDestroyImage(&endeavour_image.padlock_tmp);

        OSWDestroyImage(&endeavour_image.copy);   
        OSWDestroyImage(&endeavour_image.folder_back);
        OSWDestroyImage(&endeavour_image.folder_new);
        OSWDestroyImage(&endeavour_image.move);
        OSWDestroyImage(&endeavour_image.properties);
        OSWDestroyImage(&endeavour_image.recycle);
        OSWDestroyImage(&endeavour_image.refresh);

        OSWDestroyImage(&endeavour_image.new_browser);
        OSWDestroyImage(&endeavour_image.tile);
        OSWDestroyImage(&endeavour_image.recycle_bin_small0);
        OSWDestroyImage(&endeavour_image.recycle_bin_small1);

        OSWDestroyImage(&endeavour_image.dev_cdrom0);
        OSWDestroyImage(&endeavour_image.dev_cdrom1);   
        OSWDestroyImage(&endeavour_image.dev_fd0);    
        OSWDestroyImage(&endeavour_image.dev_fd1);
        OSWDestroyImage(&endeavour_image.dev_hd0);
        OSWDestroyImage(&endeavour_image.dev_hd1);
        OSWDestroyImage(&endeavour_image.dev_tape0);
        OSWDestroyImage(&endeavour_image.dev_tape1);
        OSWDestroyImage(&endeavour_image.dev_zip0);
        OSWDestroyImage(&endeavour_image.dev_zip1);

        OSWDestroyImage(&endeavour_image.comp_l0);
        OSWDestroyImage(&endeavour_image.comp_l1); 
        OSWDestroyImage(&endeavour_image.comp_s0);
        OSWDestroyImage(&endeavour_image.comp_s1); 


	return;
}
