/*

               Image Browser: Memory Management and Allocation

	Functions:

	int IMBRIsAllocated(int n)
	int IMBRIsInit(int n)

	int IMBRAllocate()
	void IMBRReclaimMemory()
	void IMBRDelete(int n)
	void IMBRDeleteAll()

 */

#ifdef USE_IMLIB

#include "endeavour.h"
#include "imbrdefaults.h"

#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))


/*
 *	Checks if Image Browser n is allocated.
 */
int IMBRIsAllocated(int n)
{
	if((img_browser == NULL) ||
           (n < 0) ||
           (n >= (int)total_img_browsers)
	)
	{
	    return(0);
	}
	else if(img_browser[n] == NULL)
	{
	    return(0);
	}
	else
	{
	    return(1);
	}
}



int IMBRIsInit(int n)
{
	if(!IMBRIsAllocated(n))
	{
            return(0);
        }
        else if(img_browser[n]->toplevel == 0)
        {
            return(0);
        }
        else
        {
            return(1);
        }
}



/*
 *	Allocate a new image browser.
 */
int IMBRAllocate()
{
        int i, n;


        /* Sanitize global total_img_browsers. */
        if((int)total_img_browsers < 0)
            total_img_browsers = 0;


        /* Check for available entry. */
        for(i = 0; i < (int)total_img_browsers; i++)
        {   
            if(img_browser[i] == NULL)
                break;
        }
        if(i < (int)total_img_browsers)
        {
            n = i;
        }
        else
        {
            n = total_img_browsers;
            total_img_browsers++;

            img_browser = (img_browser_struct **)realloc(
                img_browser, total_img_browsers * sizeof(img_browser_struct *)
            );
            if(img_browser == NULL)
            {
                total_img_browsers = 0;
                return(-1);
            }
        }


        /* Allocate a new image viewer structure. */
        img_browser[n] = (img_browser_struct *)calloc(1,
            sizeof(img_browser_struct));
        if(img_browser[n] == NULL)
        {
            total_img_browsers--;
            return(-1);
        }


        return(n);
}



/*
 *      Reclaims image browser pointers for unallocated pointers.
 */
void IMBRReclaimMemory()
{
        int i, h;


        /* Sanitize global total_img_browsers. */
        if((int)total_img_browsers < 0)
            total_img_browsers = 0;


        /* Get highest number h. */
        for(i = 0, h = -1; i < (int)total_img_browsers; i++)
        {
            if(IMBRIsAllocated(i))
                h = i;
        }

	/* Adjust global variable total_img_browsers. */
        total_img_browsers = h + 1;

	/* No more allocated? */
	if((int)total_img_browsers <= 0)
	{
	    total_img_browsers = 0;
	    free(img_browser);
	    img_browser = NULL;
	}
	/* Need to reclaim memory. */
	else
        {
            img_browser = (img_browser_struct **)realloc(
                img_browser,
                (int)total_img_browsers * sizeof(img_browser_struct *)
            );
            if(img_browser == NULL)
            {
                total_img_browsers = 0;
            }
        }


        return;
}



/*
 *	Destroy resources of and free memory for given image browser
 *	number n.
 */
void IMBRDelete(int n)
{
	if(!IMBRIsAllocated(n))
	    return;


	/* Deallocate all substructures. */
	IMBRDestroy(img_browser[n]);

	/* Free structure itself. */
	free(img_browser[n]);
	img_browser[n] = NULL;


	/* Clean up drag referances, just in case. */
	DragCleanup();


	return;
}



/*
 *	Deletes all image browsers allocated.
 */
void IMBRDeleteAll() 
{
	int i;

	for(i = 0; i < (int)total_img_browsers; i++)
	{
	    IMBRDelete(i);
	}

	free(img_browser);
	img_browser = NULL;

	total_img_browsers = 0;

	return;
}



#endif /* USE_IMLIB */
