/*

                    Image Browser: Call Backs

	Functions:

	int IMBRRepeatCB()
	int IMBRProcessMovement(
	        void *tar_ptr, int tar_ptr_type,
	        void *src_ptr, int src_ptr_type,  
	        int tar_drag_type, int src_drag_type,
	        int tar_item_num, int op_code,
	        bool_t comfermed
	)
	int IMBRMenuCB(void *client_data, int id_code)

 */

#ifdef USE_IMLIB

#include "endeavour.h"
#include "imbrdefaults.h"

#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))


/*
 *	Repeat callback.
 *
 *	Handles load of thumbs on the file queue
 *	and other repeatable things of each ib.
 */
int IMBRRepeatCB()
{
	int x;


	/* ********************************************************* */
	/* Handle each Image Browser. */

	for(x = 0; x < (int)total_img_browsers; x++)
	{
	    /* Must be initialized. */
	    if(!IMBRIsInit(x))
		continue;


            /* ***************************************************** */
            /* Load next image? */

	    /* Are there any items left on the file queue to be loaded? */
	    if(((int)img_browser[x]->last_fq <
                ((int)img_browser[x]->total_fqs - 1)
	       ) &&
               !img_browser[x]->stop_load	/* Interupted? */
	    )
	    {
                /* Increment last File Queue number (starts at -1) */
                img_browser[x]->last_fq++;

		/* Load thumb from queue. */
		IMBRLoadThumbFromQueue(
		    img_browser[x],
		    img_browser[x]->last_fq
		);
	    }
	}


	return(0);
}


/*
 *	Process drag release on Image Browser.
 */
int IMBRProcessMovement(
	void *tar_ptr, int tar_ptr_type,	/* Always image browser. */
	void *src_ptr, int src_ptr_type,
	int tar_drag_type, int src_drag_type,
	int tar_item_num, int op_code,
        bool_t comfermed
)
{
        int i, n;
	char *strptr;

        char tmp_path[PATH_MAX + NAME_MAX];
        char tar_path[PATH_MAX + NAME_MAX];
        char src_path[PATH_MAX + NAME_MAX];

        char need_comferm;

        endeavour_win_struct *src_edw;
        recbin_win_struct *src_rb;
        img_viewer_struct *src_iv;
        img_browser_struct *src_ib, *tar_ib;


        /* Error checks. */
        if((tar_ptr == NULL) ||         /* Target must exist. */
           (tar_ptr_type == POINTER_TYPE_UNDEFINED)
        )
            return(-1);


        /* Target pointer must be of type image browser. */
        if(tar_ptr_type != POINTER_TYPE_IB)
        {
            fprintf(stderr,
        "IMBRProcessMovement(): tar_ptr_type must be POINTER_TYPE_IB.\n"
            );
            return(-1);
        }

        /* ************************************************************ */
        /* Get target and source pointers. */
        src_edw = NULL;
        src_rb = NULL;
        src_iv = NULL;
        src_ib = NULL;
	tar_ib = (img_browser_struct *)tar_ptr;	/* Target always ib. */


        /* Check source pointer type. */
        switch(src_ptr_type)
        {
          case POINTER_TYPE_EDW:
            src_edw = (endeavour_win_struct *)src_ptr;
            break;

          case POINTER_TYPE_RB:
            src_rb = (recbin_win_struct *)src_ptr;
            break;

          case POINTER_TYPE_IV:
            src_iv = (img_viewer_struct *)src_ptr;
            break;
            
          case POINTER_TYPE_IB:
            src_ib = (img_browser_struct *)src_ptr;
            break;

          default:
            fprintf(stderr,
        "EDWProcessMovement(): Unknown source pointer type %i.\n",
                src_ptr_type
            );
            return(-1);
            break;
        }


        /* ************************************************************ */
        /* Endeavour Browser to Image Browser? */
        if((src_edw != NULL) &&
           (tar_ib != NULL)
        )
        {
	    /* Location information must be available. */
	    if((src_edw->loc_prompt.buf == NULL) ||
               (tar_ib->current_dir == NULL)
	    )
		return(-1);


            /* Check and get tar_path. */
            if((tar_item_num >= 0) &&
               (tar_item_num < tar_ib->total_thumbs)
            )
	    {
                if(tar_ib->thumb[tar_item_num] != NULL)
		{
		    strptr = PrefixPaths(tar_ib->current_dir,
			tar_ib->thumb[tar_item_num]->filename);
		    if(strptr != NULL)
		    {
			strncpy(tar_path, strptr, PATH_MAX + NAME_MAX);
			if(!ISPATHDIR(tar_path))
			{
			    strncpy(tar_path, tar_ib->current_dir,
				PATH_MAX + NAME_MAX);
			}
		    }
		    else
		    {
			strncpy(tar_path, tar_ib->current_dir,
			    PATH_MAX + NAME_MAX);
		    }
		}
		else
		{
		    strncpy(tar_path, tar_ib->current_dir, PATH_MAX + NAME_MAX);
		}
	    }
	    else
	    {
		/* No valid selection, use current_dir as target. */
                strncpy(tar_path, tar_ib->current_dir, PATH_MAX + NAME_MAX);
	    }
	    tar_path[PATH_MAX + NAME_MAX - 1] = '\0';
            SimplifyPath(tar_path);


            /* Check if comfermation is needed. */
            if(!comfermed)
            {
                if((op_code == OP_CODE_COPY) &&
                   option.comferm_pointer_motion_copy
                )
                    need_comferm = 1; 
                else if((op_code == OP_CODE_MOVE) &&
                        option.comferm_pointer_motion_move
                )
                    need_comferm = 1;
                else
                    need_comferm = 0;
            
                if(need_comferm)
                {
                    /* Move comfermation dialog to pointer's center. */
                    WidgetCenterWindow(comf_dialog.toplevel,
                        WidgetCenterWindowToPointer);

                    ComfDialogMapQueryMovement(
                        tar_ib, POINTER_TYPE_IB,	/* Target. */
                        src_edw, POINTER_TYPE_EDW,	/* Source. */
                        tar_drag_type,
                        src_drag_type,
                        tar_item_num,
                        op_code
                    );

                    /*   Return, do not continue.
                     *   Comfermation will handle query and call this function
                     *   again as needed.
                     */
                    return(0);
                }
            }


            /* Move comfermation dialog to pointer's center. */
            WidgetCenterWindow(comf_dialog.toplevel,
                WidgetCenterWindowToPointer);

            /* Source drag type file? */
            if(src_drag_type == DRAG_TYPE_FILE)
            {
                for(i = 0; i < src_edw->total_sel_files; i++)
                {
                    n = src_edw->sel_file[i];
                    if((n < 0) || (n >= (int)src_edw->total_files))
                        continue;
                    if(src_edw->file[n] == NULL) continue;
                    if(src_edw->file[n]->name == NULL) continue;  
                    strptr = PrefixPaths(src_edw->loc_prompt.buf,
                        src_edw->file[n]->name);
                    if(strptr == NULL) continue;

                    /* Get source path. */
                    strncpy(src_path, strptr, PATH_MAX + NAME_MAX);
                    src_path[PATH_MAX + NAME_MAX - 1] = '\0';
                    SimplifyPath(src_path);

                    /* Skip if paths are the same. */
                    if(!strcmp(tar_path, src_path))
                        continue;
            
                    /* Perform file operation. */
                    EDWPerformFileOp(op_code, tar_path, src_path);
                }
            }
            /* Source drag type directory? */
            else if(src_drag_type == DRAG_TYPE_DIR)
            {
                for(i = 0; i < (int)src_edw->total_sel_dirs; i++)
                {
                    n = src_edw->sel_dir[i];
                    if((n < 0) || (n >= (int)src_edw->total_dirs))
                        continue;
                    if(src_edw->dir[n] == NULL) continue;
                    if(src_edw->dir[n]->fullname == NULL) continue;
                    
                    /* Get source path. */
                    strncpy(src_path, src_edw->dir[n]->fullname,
			PATH_MAX + NAME_MAX);
                    src_path[PATH_MAX + NAME_MAX - 1] = '\0';  
                    SimplifyPath(src_path); 

                    /* Skip if paths are the same. */
                    if(!strcmp(tar_path, src_path))
                        continue;

                    /* Perform file operation. */
                    EDWPerformFileOp(op_code, tar_path, src_path);
                }
	    }

            /* Refresh target Image Browser. */
            strncpy(tmp_path, tar_ib->current_dir, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
            IMBRLoad(tar_ib, tmp_path);
            IMBRDraw(tar_ib, DRAW_AMOUNT_COMPLETE);
 
            /* Refresh source Endeavour Browser. */
            EDWRefreshAllLists(src_edw);
            EDWDraw(src_edw, DRAW_AMOUNT_COMPLETE);
	}
        /* ************************************************************ */
        /* Image Browser to Image Browser? */
        else if((src_ib != NULL) &&
                (tar_ib != NULL)
        )
        {
	    /* Current dirs on both Image Browsers must be valid. */
	    if((tar_ib->current_dir == NULL) ||
               (src_ib->current_dir == NULL)
	    )
	        return(-1);

            /* Check and get tar_path. */
            if((tar_item_num >= 0) &&
               (tar_item_num < (int)tar_ib->total_thumbs)
            )
            {
                if(tar_ib->thumb[tar_item_num] != NULL)
                {
                    strptr = PrefixPaths(tar_ib->current_dir,
                        tar_ib->thumb[tar_item_num]->filename);
                    if(strptr != NULL)
                    {
                        strncpy(tar_path, strptr, PATH_MAX + NAME_MAX);
                        if(!ISPATHDIR(tar_path))
                        {
                            strncpy(tar_path, tar_ib->current_dir,  
                                PATH_MAX + NAME_MAX);
                        }
                    }
                    else
                    {     
                        strncpy(tar_path, tar_ib->current_dir,
                            PATH_MAX + NAME_MAX); 
                    }
                }
                else
                {
                    strncpy(tar_path, tar_ib->current_dir, PATH_MAX + NAME_MAX);
                }
            }
            else
            {
                /* No valid selection, use current_dir as target. */
                strncpy(tar_path, tar_ib->current_dir, PATH_MAX + NAME_MAX);
            }
            tar_path[PATH_MAX + NAME_MAX - 1] = '\0';
            SimplifyPath(tar_path);


	    /* Duplicity check: Same ib, source and object same? */
	    if(tar_ib == src_ib)
	    {
		/* Only one or less seleted thumbs? */
		if(src_ib->total_sel_thumbs <= 1)
		{
		    /* Target item is same as seleted thumb #0. */
		    if(tar_item_num == src_ib->sel_thumb[0])
			return(-2);
		}
	    }

	    /* Duplicity check: Target path same as current dir on target? */
	    strncpy(tmp_path, src_ib->current_dir, PATH_MAX + NAME_MAX);
	    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
            SimplifyPath(tmp_path);
	    if(!strcmp(tar_path, tmp_path))
	    {
		/* Target path is the same as the current dir on target. */
		return(-2);
	    }


            /* Check if comfermation is needed. */
            if(!comfermed)
            {
                if((op_code == OP_CODE_COPY) &&
                   option.comferm_pointer_motion_copy
                )
                    need_comferm = 1;
                else if((op_code == OP_CODE_MOVE) &&
                        option.comferm_pointer_motion_move
                )
                    need_comferm = 1;
                else
                    need_comferm = 0;   

		if(need_comferm)            
                {
                    /* Move comfermation dialog to pointer's center. */
                    WidgetCenterWindow(comf_dialog.toplevel,
                        WidgetCenterWindowToPointer);
                    
                    ComfDialogMapQueryMovement(
                        tar_ib, POINTER_TYPE_IB,	/* Target. */
                        src_ib, POINTER_TYPE_IB,	/* Source. */
                        tar_drag_type,
                        src_drag_type,
                        tar_item_num,   
                        op_code
                    );

                    /*   Return, do not continue.
                     *   Comfermation will handle query and call this function
                     *   again as needed.
                     */
                    return(0);
                }
            }

	    /* Process objects. */
	    for(i = 0; i < src_ib->total_sel_thumbs; i++)
	    {
		n = src_ib->sel_thumb[i];
		if((n < 0) || (n >= src_ib->total_thumbs)) continue;
		if(src_ib->thumb[n] == NULL) continue;
		if(src_ib->thumb[n]->filename == NULL) continue;

		strptr = PrefixPaths(src_ib->current_dir,
		    src_ib->thumb[n]->filename);
		if(strptr == NULL) continue;

		strncpy(src_path, strptr, PATH_MAX + NAME_MAX);
		src_path[PATH_MAX + NAME_MAX - 1] = '\0';
                SimplifyPath(src_path);

                /* Skip if source is the target. */
                if(!strcmp(tar_path, src_path))
                    continue;

                /* Perform file operation. */
                EDWPerformFileOp(op_code, tar_path, src_path);
	    }


            /* Refresh target Image Browser. */
            strncpy(tmp_path, tar_ib->current_dir, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
            IMBRLoad(tar_ib, tmp_path);
            IMBRDraw(tar_ib, DRAW_AMOUNT_COMPLETE);

            /* Refresh souce Image Browser. */
            strncpy(tmp_path, src_ib->current_dir, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
            IMBRLoad(src_ib, tmp_path);
            IMBRDraw(src_ib, DRAW_AMOUNT_COMPLETE);
        }   




        /* Turn yes to all off. */
        yes_to_all = False;


	return(0);
}             



/*
 *	Image browser quick menu callback handler.
 */
int IMBRMenuCB(void *client_data, int id_code)
{
	int i;
	char stringa[PATH_MAX + NAME_MAX + 128];
	img_browser_struct *ib;


	/* Error checks. */
        if(!IDC() ||
           (client_data == NULL)
        )
            return(-1);


        /* ************************************************************** */
  
        /* Search for valid img_browser by client_data pointer. */
        for(i = 0; i < total_img_browsers; i++)
        {
            if(img_browser[i] == client_data)
                break;
        }
        /* Could not find match? */
        if(i >= total_img_browsers)
            return(-2);

	/* Is initialized? */
	if(!IMBRIsInit(i))
	    return(-1);
	/* Is mapped? */
	if(img_browser[i]->map_state == 0)
	    return(-1);

	/* Get pointer to image browser ib. */
	ib = (img_browser_struct *)img_browser[i];



        /* Handle menu function. */
        switch(id_code)
        {
          /* ************************************************************ */
	  case IB_MC_NONE:
	    break;

	  /* ************************************************************ */
	  case IB_MC_PREV:
	    IMBRDoPrev(ib);
	    break;

          /* ************************************************************ */
	  case IB_MC_NEXT:
	    IMBRDoNext(ib);
	    break;

          /* ************************************************************ */
          case IB_MC_VIEW:
	    IMBRDoView(ib);
	    break;

          /* ************************************************************ */
          case IB_MC_STOP:
	    IMBRDoStop(ib);
            break;

          /* ************************************************************ */
          case IB_MC_RELOAD:
	    IMBRDoReload(ib);
	    break;

          /* ************************************************************ */
          case IB_MC_RENAME:
	    IMBRDoQuickRename(ib);
            break;

          /* ************************************************************ */
          case IB_MC_MOVE:
	    IMBRDoMove(ib);
	    break;

          /* ************************************************************ */
          case IB_MC_COPY:
	    IMBRDoCopy(ib);
            break;

          /* ************************************************************ */
          case IB_MC_RECYCLE:
	    IMBRDoRecycle(ib);
            break;

          /* ************************************************************ */
          case IB_MC_PROPS:
	    IMBRDoPropDialog(ib);
            break;

          /* ************************************************************ */
	  case IB_MC_DROOT:
	    if(ib->total_sel_thumbs > 0)
	    {
	        IMBRDisplayOnRoot(ib, ib->sel_thumb[0]);
	    }
	    break;

          /* ************************************************************ */
	  case IB_MC_NEWEDW:
	    DoCreateEDW(ib->current_dir);
            break;

          /* ************************************************************ */
          case IB_MC_NEWRB:
	    DoCreateRB(ib->current_dir);
            break;

#ifdef USE_IMLIB
          /* ************************************************************ */
          case IB_MC_NEWIB:
            DoCreateIB(ib->current_dir);
            break;
#endif /* IB_MC_NEWIB */

          /* ************************************************************ */
	  default:
	    sprintf(stringa,
"Internal non-critical error:\n\n\
    Menu function code %i not supported.\n\n",
		id_code
	    );
	    printdw(&err_dw, stringa);
	    break;
	}


	return(0);
}


#endif /* USE_IMLIB */
