/*

                 Image Browser: GUI Event Handling


	Functions:

	int IMBRInit(img_browser_struct *ib)
	int IMBRResize(img_browser_struct *ib)
	int IMBRDraw(img_browser_struct *ib, int amount)
	int IMBRManage(img_browser_struct *ib, event_t *event)
	int IMBRManageAll(event_t *event)
        void IMBRMap(img_browser_struct *ib)
        void IMBRUnmap(img_browser_struct *ib)
	void IMBRDestroy(img_browser_struct *ib)

 */

#ifdef USE_IMLIB

#include "endeavour.h"
#include "imbrdefaults.h"

#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))

static long last_button_press;


/*
 *	Initializes resources of an already allocated ib.
 */
int IMBRInit(img_browser_struct *ib)
{
        int btn_x, btn_y;

        pixmap_t pixmap;
        win_attr_t wattr;

	char *strptr;
	char title[256];


        if(ib == NULL)
            return(-1);

        ib->map_state = 0;
        ib->is_in_focus = 0;
        ib->visibility_state = VisibilityFullyObscured;
        ib->x = osw_gui[0].def_toplevel_x;
        ib->y = osw_gui[0].def_toplevel_y;
        ib->width = IB_DEF_WIDTH;
        ib->height = IB_DEF_HEIGHT;
        ib->disabled = False;

	ib->cv_prompt_mode = CV_PROMPT_MODE_NONE;

	ib->list_win_width = IB_DEF_LIST_WIDTH_COEFF;
        ib->stop_load = True;		/* Just for safety. */

	/* These are assumed not allocated. */
	ib->total_thumbs = 0;
	ib->total_fqs = 0;
	ib->last_fq = -1;


        /* ************************************************************** */

        /* Create toplevel. */
        if(
            OSWCreateWindow(
                &ib->toplevel,
                osw_gui[0].root_win,
                ib->x, ib->y,
                ib->width, ib->height
            )
        )
            return(-1);
        OSWSetWindowInput(ib->toplevel, ButtonPressMask | ButtonReleaseMask |
            PointerMotionMask | KeyPressMask |
            KeyReleaseMask | ExposureMask | StructureNotifyMask |
            FocusChangeMask | LeaveWindowMask
        );
	OSWSetWindowBkg(ib->toplevel, 0, widget_global.std_bkg_pm);
        WidgetCenterWindow(ib->toplevel, WidgetCenterWindowToRoot);
        OSWGetWindowAttributes(ib->toplevel, &wattr);
        ib->x = wattr.x;
        ib->y = wattr.y;
        ib->width = wattr.width;
        ib->height = wattr.height;

	/* Get icon. */
        pixmap = OSWCreatePixmapFromImage(endeavour_image.imbr_icon);


        /* Set title. */
        if(userinfo.uid != userinfo.euid)
        {
            char username[NAME_MAX], eusername[NAME_MAX], hostname[NAME_MAX];


            strptr = IDGetNameByUID(userinfo.uid);
            strncpy(
                username, ((strptr == NULL) ? "" : strptr), NAME_MAX
            );
            username[NAME_MAX - 1] = '\0';
         
            strptr = IDGetNameByUID(userinfo.euid);
            strncpy(
                eusername, ((strptr == NULL) ? "" : strptr), NAME_MAX
            );
            eusername[NAME_MAX - 1] = '\0';
              
            strptr = getenv("HOSTNAME");  
            if(strptr == NULL)
                strptr = getenv("HOST");
            strncpy(hostname, ((strptr == NULL) ? "" : strptr), NAME_MAX);
            hostname[NAME_MAX - 1] = '\0';

            sprintf(title, "Endeavour Image Browser: %s(%s)@%s",
                username, eusername, hostname
            );
        }
        else
        {
            char username[NAME_MAX], hostname[NAME_MAX];

            strptr = IDGetNameByUID(userinfo.uid);
            strncpy(
                username, ((strptr == NULL) ? "" : strptr), NAME_MAX
            );
            username[NAME_MAX - 1] = '\0';

            strptr = getenv("HOSTNAME");
            if(strptr == NULL)
                strptr = getenv("HOST");
            strncpy(hostname, ((strptr == NULL) ? "" : strptr), NAME_MAX);
            hostname[NAME_MAX - 1] = '\0';

            sprintf(title, "Endeavour Image Browser: %s@%s",
                username, hostname
            );
        }

	/* WM Properties. */
	OSWSetWindowWMProperties(
	    ib->toplevel,
	    title,
            "Image Browser",
            ((pixmap == 0) ? widget_global.std_icon_pm : pixmap),
            !osw_gui[0].def_geometry_set,
            ib->x, ib->y,
	    IB_MIN_WIDTH, IB_MIN_HEIGHT,
            osw_gui[0].display_width, osw_gui[0].display_height,
            WindowFrameStyleStandard,
            NULL, 0
	);
        pixmap = 0;     /* Do not destroy pixmap. */


	/* Location ribbon. */
	if(
            OSWCreateWindow( 
                &ib->loc_ribbon,
		ib->toplevel,
		0, 0,
		ib->width,
		IB_LOC_RIBBON_HEIGHT
	    )
	)
            return(-1);
        OSWSetWindowInput(ib->loc_ribbon, ExposureMask);
        OSWGetWindowAttributes(ib->loc_ribbon, &wattr);
        if(OSWCreatePixmap(&ib->loc_ribbon_buf, wattr.width, wattr.height))
            return(-1);

	/* Location prompt. */
	if(
	    PromptInit(
		&ib->loc_prompt,
		ib->loc_ribbon,
		2, 2,
		((int)wattr.width / 2),
		MAX(IB_PROMPT_HEIGHT - 2, 2),
		PROMPT_STYLE_FLUSHED,
		"Location:",
		PATH_MAX + NAME_MAX,
                25,
                NULL
	    )
        )
            return(-1);

	/* Previous button. */
	btn_x = ((int)wattr.width / 2) + IB_DEF_MARGIN;
	btn_y = 4;

	if(
	    PBtnInit(
		&ib->prev_btn,
		ib->loc_ribbon,
		btn_x, btn_y,
		IB_BTN_WIDTH, IB_BTN_HEIGHT,
		"Prev",
		PBTN_TALIGN_CENTER,
		NULL,
		(void *)&ib->prev_btn,
		NULL
	    )
	)
	    return(-1);
	PBtnSetHotKeys(
	    &ib->prev_btn,
	    "pb,["
	);
        PBtnSetHintMessage(
            &ib->prev_btn,
            HINT_MESG_PREV_BTN
        );


	/* Next button. */
	btn_x += IB_BTN_WIDTH + IB_DEF_MARGIN;
        if(
            PBtnInit(
                &ib->next_btn,
                ib->loc_ribbon,
                btn_x, btn_y,
                IB_BTN_WIDTH, IB_BTN_HEIGHT,
                "Next",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&ib->next_btn,
                NULL
            )
        )
            return(-1);
        PBtnSetHotKeys(
            &ib->next_btn,
            "n .]"
        );
        PBtnSetHintMessage(
            &ib->next_btn,
            HINT_MESG_NEXT_BTN 
        );

	/* View button. */
        btn_x += IB_BTN_WIDTH + IB_DEF_MARGIN;
        if(
            PBtnInit(
                &ib->view_btn,
                ib->loc_ribbon,
                btn_x, btn_y,
                IB_BTN_WIDTH, IB_BTN_HEIGHT,
                "View",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&ib->view_btn,
                NULL
            )
        )
            return(-1);
        PBtnSetHotKeys(
            &ib->view_btn,
            "v\n\r"
        );
        PBtnSetHintMessage(
            &ib->view_btn,
            HINT_MESG_VIEW_BTN 
        );

        /* Reload button. */
        btn_x += IB_BTN_WIDTH + IB_DEF_MARGIN;
        if(
            PBtnInit(
                &ib->reload_btn,
                ib->loc_ribbon,
                btn_x, btn_y,
                IB_BTN_WIDTH, IB_BTN_HEIGHT,
                "Reload",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&ib->reload_btn,
                NULL   
            )
        )
            return(-1);
        PBtnSetHotKeys(
            &ib->reload_btn,
            "r"
        );
        PBtnSetHintMessage(
            &ib->reload_btn,
            HINT_MESG_RELOAD_BTN
        );

        /* Stop button. */
        btn_x += IB_BTN_WIDTH + IB_DEF_MARGIN;
        if(
            PBtnInit(
                &ib->stop_btn,
                ib->loc_ribbon, 
                btn_x, btn_y,   
                IB_BTN_WIDTH, IB_BTN_HEIGHT,
                "Stop",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&ib->stop_btn,
                NULL 
            )
        )
            return(-1);
        PBtnSetHotKeys(
            &ib->stop_btn,
            "s"
        );
        PBtnSetHintMessage(
            &ib->stop_btn,
            HINT_MESG_STOP_BTN 
        );


        /* List window. */
        if(
            OSWCreateWindow( 
                &ib->list,
                ib->toplevel,
                0, IB_LOC_RIBBON_HEIGHT,
                (int)((double)ib->width * ib->list_win_width),
		(int)ib->height - IB_LOC_RIBBON_HEIGHT - IB_STATUS_HEIGHT
            )
        )
            return(-1);
        OSWSetWindowInput(ib->list, ButtonPressMask |
            ButtonReleaseMask | PointerMotionMask | ExposureMask);
        OSWGetWindowAttributes(ib->list, &wattr);
        if(OSWCreatePixmap(&ib->list_buf, wattr.width, wattr.height))
            return(-1);

	if(
	    SBarInit(
		&ib->list_sb, ib->list,
		wattr.width, wattr.height
	    )
	)
	    return(-1);

	/* Viewer window. */
        if(
            OSWCreateWindow(
                &ib->viewer,
                ib->toplevel,
                (int)(((double)ib->width * ib->list_win_width) +
                    IB_DEF_SEPERATOR_WIDTH),
                IB_LOC_RIBBON_HEIGHT,
                (int)(((double)ib->width * (double)(1 - ib->list_win_width)) -
		    IB_DEF_SEPERATOR_WIDTH),
                (int)ib->height - IB_LOC_RIBBON_HEIGHT - IB_STATUS_HEIGHT
            )
        )
            return(-1);
        OSWSetWindowInput(ib->viewer, ButtonPressMask |
            ButtonReleaseMask | PointerMotionMask | ExposureMask);
        OSWGetWindowAttributes(ib->viewer, &wattr);
        if(OSWCreatePixmap(&ib->viewer_buf, wattr.width, wattr.height))
            return(-1);

	/* List and viewer split cursor window. */
	if(
            OSWCreateWindow(
                &ib->split_cur,
		ib->toplevel,
		0, 0, 5, 5
	    )
	)
	    return(-1);
	OSWSetWindowBkg(ib->split_cur, osw_gui[0].white_pix, 0);


	/* Status window. */
	if(
            OSWCreateWindow(
                &ib->status_win,
		ib->toplevel,
		0, (int)ib->height - IB_STATUS_HEIGHT,
		ib->width, IB_STATUS_HEIGHT
	    )
	)
	    return(-1);
        OSWSetWindowInput(ib->status_win, ButtonPressMask |
            ButtonReleaseMask | PointerMotionMask | ExposureMask |
	    EnterWindowMask | LeaveWindowMask
	);
        OSWGetWindowAttributes(ib->status_win, &wattr);
        if(OSWCreatePixmap(&ib->status_win_buf, wattr.width, wattr.height))
            return(-1);
	/* Progress bar. */
        if(
            PBarInit(
                &ib->pbar,
                ib->status_win,
                (int)wattr.width - IB_PBAR_WIDTH - 90,
                ((int)wattr.height / 2) - (IB_PBAR_HEIGHT / 2),
                IB_PBAR_WIDTH,
                IB_PBAR_HEIGHT,
                0,              /* Start value. */
                0, 1,           /* Min and max. */
                NULL,
                PBAR_COMPLETION_HOLD
            )
        )
            return(-1);

	/* Padlock hint. */
        OSWGetWindowAttributes(ib->toplevel, &wattr);
        if(
            HintWinAddMessage( 
                ib->status_win,
                ib->toplevel,
                (int)wattr.width - 275,
                (int)wattr.height - 50,
                HINT_MESG_PADLOCK
            ) < 0
        )
            return(-1);


        /* Change values prompt. */
	if(
            PromptInit(
                &ib->cv_prompt,
                ib->toplevel,
                0, 0,
                200,			/* Arbatary numbers for now. */
                IB_PROMPT_HEIGHT,
                PROMPT_STYLE_NOBORDER,	/* Height must be size of row. */
                NULL,			/* No name. */
                PATH_MAX + NAME_MAX,
                0,
                NULL
	    )
        )
            return(-1);

	/* Quick menu. */
	if(
            MenuInit(
                &ib->qmenu,
                osw_gui[0].root_win,
                IMBRMenuCB,
                (void *)ib
            )
	)
	    return(-1);
        MenuAddItem(
            &ib->qmenu,
            "Image Browser Commands:",
            MENU_ITEM_TYPE_COMMENT,
            NULL,
            IB_MC_NONE,
            -1
        );
        MenuAddItem(
            &ib->qmenu,
            "",
            MENU_ITEM_TYPE_HR,
            NULL,
            IB_MC_NONE,
            -1
        );
        MenuAddItem(
            &ib->qmenu,
            "Previous               P",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IB_MC_PREV,
            -1   
        );
        MenuAddItem(
            &ib->qmenu,
            "Next                   N",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IB_MC_NEXT,
            -1
        );
        MenuAddItem(
            &ib->qmenu,
            "View                   V",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IB_MC_VIEW,
            -1
        );
        MenuAddItem(
            &ib->qmenu,
            "",
            MENU_ITEM_TYPE_HR,
            NULL,
            IB_MC_NONE,
            -1
        );
        MenuAddItem(
            &ib->qmenu,
            "Rename           Button2",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IB_MC_RENAME,
            -1
        );
        MenuAddItem(
            &ib->qmenu,
            "Move                  F7",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IB_MC_MOVE,
            -1
	);
        MenuAddItem(
            &ib->qmenu,
            "Copy                  F8",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IB_MC_COPY,
            -1
        );
        MenuAddItem(   
            &ib->qmenu,
            "Recycle           Delete",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IB_MC_RECYCLE,
            -1
        );
        MenuAddItem(
            &ib->qmenu,
            "Properties     Alt+Enter",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IB_MC_PROPS,
            -1
        );
        MenuAddItem(   
            &ib->qmenu,
            "",
            MENU_ITEM_TYPE_HR,
            NULL,
            IB_MC_NONE,
            -1
        );
        MenuAddItem(
            &ib->qmenu,
            "Display on Root",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IB_MC_DROOT,
            -1
        );
        MenuAddItem(
            &ib->qmenu,
            "",
            MENU_ITEM_TYPE_HR,
            NULL,
            IB_MC_NONE,
            -1
        );
        MenuAddItem(   
            &ib->qmenu,
            "Browser",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IB_MC_NEWEDW,
            -1
        );
        MenuAddItem(
            &ib->qmenu,
            "Recycle Bin", 
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IB_MC_NEWRB,
            -1
        );
        MenuAddItem(
            &ib->qmenu,
            "Image Browser",
            MENU_ITEM_TYPE_ENTRY,
            NULL,
            IB_MC_NEWIB,
            -1
        );



	/* Calculate thumb sizing. */
	ib->thumb_width = IB_DEF_THUMB_WIDTH;
	ib->thumb_height = IB_DEF_THUMB_HEIGHT;
	ib->thumb_label_height = IB_DEF_THUMB_LABEL_HEIGHT;
	ib->thumb_col_margin = IB_DEF_THUMB_COL_MARGIN;
	ib->thumb_row_margin = IB_DEF_THUMB_ROW_MARGIN;

	OSWGetWindowAttributes(ib->list, &wattr);
	ib->thumbs_per_row = (int)MAX(
            ((int)wattr.width - (int)ib->thumb_col_margin), 1)
	    / (int)MAX((int)ib->thumb_col_margin + (int)ib->thumb_width, 1)
	;

	ib->list_col_spacing = (int)ib->thumb_col_margin +
	    (int)ib->thumb_width;
	ib->list_row_spacing = (int)ib->thumb_height +
	    (int)ib->thumb_label_height + (int)ib->thumb_row_margin;


	return(0);
}



int IMBRResize(img_browser_struct *ib)
{
	int btn_x, btn_y;
        win_attr_t wattr;


        /* Error checks. */
        if(!IDC() || (ib == NULL))
            return(-1);

        if(ib->toplevel == 0)
            return(-1);


        /* ************************************************************** */

        /* Get toplevel attributes. */
        OSWGetWindowAttributes(ib->toplevel, &wattr); 

        /* Skip if no change. */
        if(((int)ib->width == wattr.width) &&
           ((int)ib->height == wattr.height)
        )
            return(0); 


	/* Record new size values. */
	ib->x = wattr.x;
	ib->y = wattr.y;
	ib->width = wattr.width;
	ib->height = wattr.height;


        /* Location ribbon. */
        OSWMoveResizeWindow(
            ib->loc_ribbon,
            0, 0,
            ib->width,
            IB_LOC_RIBBON_HEIGHT
	);
        OSWGetWindowAttributes(ib->loc_ribbon, &wattr);
        OSWDestroyPixmap(&ib->loc_ribbon_buf);
        if(OSWCreatePixmap(&ib->loc_ribbon_buf, wattr.width, wattr.height))
            return(-1);

	/* Location prompt. */
	OSWMoveResizeWindow(
	    ib->loc_prompt.toplevel,
            2, 2,
            ((int)wattr.width / 2),
            MAX(IB_PROMPT_HEIGHT - 2, 2)
	);
	ib->loc_prompt.map_state = 0;
	PromptDraw(&ib->loc_prompt, PROMPT_DRAW_AMOUNT_COMPLETE);

	/* Prev button. */
        btn_x = ((int)wattr.width / 2) + IB_DEF_MARGIN;
        btn_y = 4;

	OSWMoveWindow(ib->prev_btn.toplevel, btn_x, btn_y);
	btn_x += IB_BTN_WIDTH + IB_DEF_MARGIN;

        OSWMoveWindow(ib->next_btn.toplevel, btn_x, btn_y);

        btn_x += IB_BTN_WIDTH + IB_DEF_MARGIN;
        OSWMoveWindow(ib->view_btn.toplevel, btn_x, btn_y);

        btn_x += IB_BTN_WIDTH + IB_DEF_MARGIN;
        OSWMoveWindow(ib->reload_btn.toplevel, btn_x, btn_y);

        btn_x += IB_BTN_WIDTH + IB_DEF_MARGIN;
        OSWMoveWindow(ib->stop_btn.toplevel, btn_x, btn_y);


	/* List. */
	OSWMoveResizeWindow(
            ib->list,
            0, IB_LOC_RIBBON_HEIGHT,
            (int)((double)ib->width * ib->list_win_width),
            (int)ib->height - IB_LOC_RIBBON_HEIGHT - IB_STATUS_HEIGHT
        );
        OSWGetWindowAttributes(ib->list, &wattr);
	OSWDestroyPixmap(&ib->list_buf);
        if(OSWCreatePixmap(&ib->list_buf, wattr.width, wattr.height))
            return(-1);
	SBarResize(&ib->list_sb, wattr.width, wattr.height);


        /* Viewer. */
        OSWMoveResizeWindow(
            ib->viewer,
            (int)(((double)ib->width * ib->list_win_width) +
                IB_DEF_SEPERATOR_WIDTH),
            IB_LOC_RIBBON_HEIGHT,
            (int)(((double)ib->width * (double)(1 - ib->list_win_width)) -
		IB_DEF_SEPERATOR_WIDTH),
            (int)ib->height - IB_LOC_RIBBON_HEIGHT - IB_STATUS_HEIGHT
        );
        OSWGetWindowAttributes(ib->viewer, &wattr);
        OSWDestroyPixmap(&ib->viewer_buf);
        if(OSWCreatePixmap(&ib->viewer_buf, wattr.width, wattr.height))
            return(-1);

	/* Reload image to viewer. */
	if((int)ib->total_sel_thumbs > 0)
	{
	    IMBRLoadImageFromThumb(ib, ib->sel_thumb[0]);
	}


	/* Status window. */
	OSWMoveResizeWindow(
            ib->status_win,
            0, (int)ib->height - IB_STATUS_HEIGHT,
            ib->width, IB_STATUS_HEIGHT
	);
        OSWGetWindowAttributes(ib->status_win, &wattr);
        OSWDestroyPixmap(&ib->status_win_buf);
        if(OSWCreatePixmap(&ib->status_win_buf, wattr.width, wattr.height))
            return(-1);
        /* Progress bar. */
        OSWMoveResizeWindow(
            ib->pbar.toplevel,
            (int)wattr.width - IB_PBAR_WIDTH - 90,
            ((int)wattr.height / 2) - (IB_PBAR_HEIGHT / 2),
            IB_PBAR_WIDTH,
            IB_PBAR_HEIGHT
        );
	OSWGetWindowAttributes(ib->pbar.toplevel, &wattr);
	ib->pbar.x = wattr.x;
	ib->pbar.y = wattr.y;
	ib->pbar.width = wattr.width;
	ib->pbar.height = wattr.height;

        OSWGetWindowAttributes(ib->toplevel, &wattr);


        /* Calculate thumb sizing. */
        ib->thumb_width = IB_DEF_THUMB_WIDTH;
        ib->thumb_height = IB_DEF_THUMB_HEIGHT;
        ib->thumb_label_height = IB_DEF_THUMB_LABEL_HEIGHT;
        ib->thumb_col_margin = IB_DEF_THUMB_COL_MARGIN;
        ib->thumb_row_margin = IB_DEF_THUMB_ROW_MARGIN;
 
        OSWGetWindowAttributes(ib->list, &wattr);
        ib->thumbs_per_row = (int)MAX(
            ((int)wattr.width - (int)ib->thumb_col_margin), 1)
            / (int)MAX((int)ib->thumb_col_margin + (int)ib->thumb_width, 1)
        ;
            
        ib->list_col_spacing = (int)ib->thumb_col_margin +
            (int)ib->thumb_width;
        ib->list_row_spacing = (int)ib->thumb_height +
            (int)ib->thumb_label_height + (int)ib->thumb_row_margin;



	return(0);
}



int IMBRDraw(img_browser_struct *ib, int amount)
{
	int x, y, z;
	int root_x, root_y, start_x, start_y;
	int i, sl, len;
	int drag_outline_num;
	int thumb_row_count, rows_visible;
	int rows_drawn;
	win_attr_t wattr;
	image_t *ximage;


        if(!IDC() || (ib == NULL))
	    return(-1);

	if(ib->toplevel == 0)
	    return(-1);


	/* Map as needed. */
	if(ib->map_state == 0)
	{
	    OSWMapRaised(ib->toplevel);
            OSWMapWindow(ib->loc_ribbon);
	    PromptMap(&ib->loc_prompt);
	    OSWMapWindow(ib->list);
	    OSWMapWindow(ib->viewer);

            OSWMapWindow(ib->status_win);
	    PBarMap(&ib->pbar);

	    PBtnMap(&ib->prev_btn);
            PBtnMap(&ib->next_btn);
            PBtnMap(&ib->view_btn);
            PBtnMap(&ib->reload_btn);
            PBtnMap(&ib->stop_btn);

	    if(ib->cv_prompt_mode == CV_PROMPT_MODE_NONE)
		PromptUnmap(&ib->cv_prompt);
	    else
		PromptMap(&ib->cv_prompt);

	    ib->map_state = 1;
	}


        /* ********************************************************** */
        /* Draw location ribbon. */
        if(amount == DRAW_AMOUNT_COMPLETE)
	{
            /* Get window attributes. */
            OSWGetWindowAttributes(ib->loc_ribbon, &wattr);

            /* Redraw background. */
            WidgetPutImageTile(
                (drawable_t)ib->loc_ribbon_buf,
                widget_global.std_bkg_img,
                wattr.width, wattr.height
            );

            /* Draw outer frame. */
            WidgetFrameButtonPixmap(
                ib->loc_ribbon_buf,
                False,
                wattr.width, wattr.height,
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix
            );


            OSWPutBufferToWindow(ib->loc_ribbon, ib->loc_ribbon_buf);
	}


        /* ********************************************************** */
        /* Draw thumbs list. */
        if((amount == DRAW_AMOUNT_COMPLETE) ||      
           (amount == DRAW_AMOUNT_LISTS) ||
           (amount == DRAW_AMOUNT_DIR_LIST) ||
           (amount == DRAW_AMOUNT_DIR_LIST_OUTLINE)
	)
	{
            /* Get window attributes. */
            OSWGetWindowAttributes(ib->list, &wattr);

            /* Clear buffer. */
            OSWClearPixmap(ib->list_buf,
                wattr.width, wattr.height,
                widget_global.surface_editable_pix
            );

            /* Check if there are any thumbs. */
            if((ib->thumb != NULL) &&
               ((int)ib->total_thumbs > 0) &&
               (ib->list_row_spacing != 0)
            )
	    {
                /* Calculate starting position for gradiant scrolling. */
		start_x = IB_DEF_MARGIN;
                start_y = IB_DEF_MARGIN - ((int)ib->list_sb.y_win_pos %
                    (int)ib->list_row_spacing);

                /* Calculate starting thumb_row_count. */
                thumb_row_count = ib->list_sb.y_win_pos /
		    (int)ib->list_row_spacing;
                if(thumb_row_count < 0)
                    thumb_row_count = 0;

                /* Calculate rows_visible. */
                rows_visible = ((int)wattr.height /
		    (int)ib->list_row_spacing) + 3;

		/* Calculate maximum file name length. */
		len = MAX(
		    (int)ib->thumb_width / IB_CHAR_WIDTH,
		    0
		);

		/* Calculate drag_outline_num. */
		if(amount == DRAW_AMOUNT_DIR_LIST_OUTLINE)
		{
		    OSWGetPointerCoords(ib->list, &root_x, &root_y, &x, &y);
		    drag_outline_num = IMBR_GETTHUMBFROMPOS(ib, x, y);
		}
		else
		{
		    drag_outline_num = -1;
		}

		/* Begin drawing thumbs. */
		y = start_y;
		rows_drawn = 0;
		while(rows_drawn < rows_visible)
		{
		    z = MAX(thumb_row_count * (int)ib->thumbs_per_row, 0);

		    /* Draw each colum of this row. */
		    for(i = 0, x = start_x;
                        i < (int)ib->thumbs_per_row;
			i++, x += ib->list_col_spacing, z++
		    )
		    {
			if(z >= (int)ib->total_thumbs) break;
			if(ib->thumb[z] == NULL) continue;

			/* Draw image. */
			if(ib->thumb[z]->image != NULL)
			{
			    OSWPutImageToDrawablePos(
				ib->thumb[z]->image,
				(drawable_t)ib->list_buf,
				x, y
			    );
			}
			/* Draw outline if entry is selected. */
			if(IMBRIsSelected(ib, z) &&
                           (ib->thumb[z]->filename != NULL)
			)
			{
                            OSWSetFgPix(widget_global.editable_text_pix);
                            OSWDrawRectangle((drawable_t)ib->list_buf,
                                x, y,
                                ib->thumb_width, ib->thumb_height
			    );

			    sl = strlen(ib->thumb[z]->filename);
                            OSWSetFgPix(widget_global.surface_selected_pix);
                            OSWDrawSolidRectangle((drawable_t)ib->list_buf,
                                (sl < len) ? x + (len * IB_CHAR_WIDTH / 2) -
                                    (sl * IB_CHAR_WIDTH / 2) : x,
				y + (int)ib->thumb_height + 1,
				(MIN(sl, len) * IB_CHAR_WIDTH) + 5,
				ib->thumb_label_height
                            );
                            OSWSetFgPix(widget_global.selected_text_pix); 
                            OSWDrawStringLimited((drawable_t)ib->list_buf,
                                (sl < len) ? x + (len * IB_CHAR_WIDTH / 2) -
                                    (sl * IB_CHAR_WIDTH / 2) + 2 : x + 2,
				y + ib->thumb_height + 15,
                                ib->thumb[z]->filename,
                                MIN(sl, len)
                            );
			}
			/* Draw file name. */
			else if(ib->thumb[z]->filename != NULL)
			{
			    sl = strlen(ib->thumb[z]->filename);
			    OSWSetFgPix(widget_global.editable_text_pix);
                            OSWDrawStringLimited((drawable_t)ib->list_buf,
                                (sl < len) ? x + (len * IB_CHAR_WIDTH / 2) -
                                    (sl * IB_CHAR_WIDTH / 2) + 2 : x + 2,
                                y + ib->thumb_height + 15,
                                ib->thumb[z]->filename,
                                MIN(sl, len)
                            );
			}

			/* Draw drag outline. */
			if((amount == DRAW_AMOUNT_DIR_LIST_OUTLINE) &&
                           (drag_outline_num == z)
			)
			{
                            OSWSetFgPix(widget_global.editable_text_pix);
                            OSWDrawRectangle((drawable_t)ib->list_buf,
                                x, y,
                                ib->thumb_width, ib->thumb_height
                            );
			}

		    }


		    /* Increment row positions. */
		    y += ib->list_row_spacing;
		    thumb_row_count++;
		    rows_drawn++;
		}

	    }

	    /* Draw frame. */
            WidgetFrameButtonPixmap(ib->list_buf, True,
                wattr.width, wattr.height,
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix
            );

	    OSWPutBufferToWindow(ib->list, ib->list_buf);
	}

        /* ********************************************************** */
        /* Draw viewer. */
        if((amount == DRAW_AMOUNT_COMPLETE) ||
           (amount == DRAW_AMOUNT_LISTS) ||   
           (amount == DRAW_AMOUNT_FILE_LIST)
        )
        {
	    OSWGetWindowAttributes(ib->viewer, &wattr);

	    if((int)ib->total_sel_thumbs <= 0)
	    {
                /* Redraw background of viewer. */
                WidgetPutImageTile(
                    (drawable_t)ib->viewer_buf,
                    widget_global.std_bkg_img,
                    wattr.width, wattr.height  
                );

                /* Draw outer frame on viewer. */
                WidgetFrameButtonPixmap(
                    ib->viewer_buf,
                    True,
                    wattr.width, wattr.height,
                    widget_global.surface_highlight_pix,
                    widget_global.surface_shadow_pix 
                );
	    }

            OSWPutBufferToWindow(ib->viewer, ib->viewer_buf);
	}

	/* *********************************************************** */
	/* Status window. */
        if((amount == DRAW_AMOUNT_COMPLETE) ||
           (amount == DRAW_AMOUNT_STATUS)
	)
	{
            OSWGetWindowAttributes(ib->status_win, &wattr);

            /* Redraw background. */
            WidgetPutImageTile(
                (drawable_t)ib->status_win_buf,
                widget_global.std_bkg_img,
                wattr.width, wattr.height
            );

            /* Draw outer frame. */
            WidgetFrameButtonPixmap(
                ib->status_win_buf,
                False,
                wattr.width, wattr.height,
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix
            );

            /* Padlock icon. */
            if(option.write_protect)
            {
                ximage = endeavour_image.padlock1;
            }            
            else        
            {
                ximage = endeavour_image.padlock_tmp;
            }
            if(ximage != NULL) 
            {
                OSWPutImageToDrawablePos(
                    ximage,
                    (drawable_t)ib->status_win_buf,
                    (int)wattr.width - (int)ximage->width - 15,
                    ((int)wattr.height / 2) -
                        ((int)ximage->height / 2)   
                );
            }

            /* Status message. */   
            if((ib->status_mesg != NULL) && (ximage != NULL))
            {
                OSWSetFgPix(widget_global.normal_text_pix);
                OSWDrawStringLimited(
                    ib->status_win_buf,
                    10,
                    ((int)wattr.height / 2) + 5,
                    ib->status_mesg,
                    MIN((int)strlen(ib->status_mesg),
                        ((int)wattr.width - (int)ximage->width - 25) /
                        IB_CHAR_WIDTH
                    )
                );
            }


            OSWPutBufferToWindow(ib->status_win, ib->status_win_buf);
	}


	/* *********************************************************** */

	/* Draw scroll bars. */
        if(amount == DRAW_AMOUNT_COMPLETE)
	{
            /* Get window attributes. */
            OSWGetWindowAttributes(ib->list, &wattr);

	    if(ib->thumbs_per_row != 0)
	    {
                SBarDraw(
		    &ib->list_sb,
                    wattr.width,
                    wattr.height,
                    wattr.width,
                    (((int)ib->total_thumbs / (int)ib->thumbs_per_row) + 2) *
                    ib->list_row_spacing
                );
	    }
	}

	/* Draw progress bar. */
/*
        if(amount == DRAW_AMOUNT_COMPLETE)
	{
	    PBarDraw(&ib->pbar);
	}
*/


	return(0);
}



int IMBRManage(img_browser_struct *ib, event_t *event)
{
	int x, y, z, root_x, root_y;
	int events_handled = 0;
	win_attr_t wattr;
	image_t *image;
	char *strptr;
        struct stat stat_buf;
	long t;
	char stringa[PATH_MAX + NAME_MAX + 80];
	char tmp_path[PATH_MAX + NAME_MAX];
        char tar_path[PATH_MAX + NAME_MAX];


        if((ib == NULL) ||
           (event == NULL)
	)
	    return(events_handled);


	switch(event->type)
	{
          /* ************************************************************ */
          case ButtonPress:
            /* Toplevel (for thumbs list and viewer split drag. */
            if(event->xany.window == ib->toplevel)
            {
                OSWGetWindowAttributes(ib->toplevel, &wattr);

                x = event->xbutton.x;
                y = event->xbutton.y;
                z = (int)(ib->list_win_width * (double)wattr.width);
                if((x >= z) && (x < (z + IB_DEF_SEPERATOR_WIDTH)) &&
                   (y >= IB_LOC_RIBBON_HEIGHT) &&
                   (y < ((int)wattr.height - IB_STATUS_HEIGHT))
                )
                {
                    /* Record drag for file heading window. */
                    drag_record[0].state = True;
                    drag_record[0].type = DRAG_TYPE_LIST_SPLIT;
                    drag_record[0].origin_ptr_type = POINTER_TYPE_IB;
                    drag_record[0].origin_ptr = ib;

		    /* Set up split cursor window. */
                    OSWSetWindowCursor(ib->split_cur,
                        endeavour_cursor.list_heading_resize);
		    OSWReparentWindow(ib->split_cur, ib->toplevel);
		    OSWGetPointerCoords(ib->toplevel, &root_x, &root_y,
			&x, &y);
		    OSWMoveResizeWindow(ib->split_cur,
			x - (IB_DEF_SEPERATOR_WIDTH / 2),
			IB_LOC_RIBBON_HEIGHT,
			IB_DEF_SEPERATOR_WIDTH,
			MAX((int)ib->height - IB_LOC_RIBBON_HEIGHT -
                           IB_STATUS_HEIGHT, 2)
		    );
		    OSWMapRaised(ib->split_cur);

		    /* Do not ungrab pointer, toplevel needs to track events. */
                }

                events_handled++;
                return(events_handled);
	    }
	    /* List. */
	    else if(event->xany.window == ib->list)
	    {
		/* Set list scroll bars into focus. */
		ib->list_sb.is_in_focus = 1;


		/* Sanitize last_button_press. */
		t = MilliTime();
		if(last_button_press > t)
		    last_button_press = 0;

		/* Check for double click. */
		if((last_button_press + widget_global.double_click_int) >= t)
		{
		    if((event->xbutton.button == Button1) &&
                       (ib->current_dir != NULL)
		    )
		    {
		        /*   On double click, change to selected dir or
                         *   spawn viewer if image is selected.
		         */
		        for(x = 0; x < (int)ib->total_sel_thumbs; x++)
		        {
			    y = ib->sel_thumb[x];
			    if((y < 0) || (y >= (int)ib->total_thumbs))
				continue;
			    if(ib->thumb[y] == NULL) continue;
			    if(ib->thumb[y]->filename == NULL) continue;

			    /* Get full path as tmp_path. */
			    strptr = PrefixPaths(ib->current_dir,
				ib->thumb[y]->filename);
			    if(strptr == NULL) continue;

			    strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
			    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

			    /* Is directory? */
			    if(ISPATHDIR(tmp_path))
			    {
				/* Change directory. */
                                IMBRDoChangeDir(ib, ib->thumb[y]->filename);

				/* Must break after dir change. */
				break;
			    }
			    /* Spawn image viewer. */
			    else
			    {
                                /* Allocate and initialize a new internal viewer. */
                                y = ImgViewerAllocate();
                                if(ImgViewerIsAllocated(y))  
                                {
                                    ImgViewerInit(img_viewer[y]);

                                    ImgViewerLoadFile(img_viewer[y], tmp_path);
                                    ImgViewerMap(img_viewer[y]);
                                }
			    }
			}
		    }

		    last_button_press = 0;
		}
		else
                {
		    last_button_press = t;             
                }	/* Check for double click. */

                /* Map the quick menu. */
                if(event->xbutton.button == Button3)
                {
                    OSWGetPointerCoords(ib->toplevel,
                        &root_x, &root_y,
                        &x, &y
                    );
                    OSWMoveWindow(ib->qmenu.toplevel, root_x, root_y);
                    ib->qmenu.x = root_x;
                    ib->qmenu.y = root_y;

                    MenuMap(&ib->qmenu);

                    events_handled++;
                    return(events_handled);
                }

		events_handled++;
	    }
            /* Viewer. */
            else if(event->xany.window == ib->viewer)
            {
                /* Sanitize last_button_press. */  
                t = MilliTime();
                if(last_button_press > t)
                    last_button_press = 0;

                /* Check for double click. */
                if((last_button_press + widget_global.double_click_int) >= t)
                {
                    if((event->xbutton.button == Button1) &&
                       (ib->current_dir != NULL)
                    )
                    {
			/* Put selected thumb into viewer. */
			if(ib->total_sel_thumbs > 0)
			{
			    x = ib->sel_thumb[0];
			    if((x >= 0) && (x < (int)ib->total_thumbs))
			    {
			      if(ib->thumb[x] != NULL)
			      {
				strptr = PrefixPaths(ib->current_dir,
				    ib->thumb[x]->filename);
				strncpy(tmp_path, (strptr == NULL) ? "/" :
                                    strptr, PATH_MAX + NAME_MAX);
				tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

                                y = ImgViewerAllocate();
                                if(ImgViewerIsAllocated(y))
                                {
                                    ImgViewerInit(img_viewer[y]);

                                    ImgViewerLoadFile(img_viewer[y], tmp_path);
                                    ImgViewerMap(img_viewer[y]);
				}
                              }
                            }
			}
		    }


                    last_button_press = 0;
                    events_handled++;
                }
                else
                {
                    last_button_press = t;
                }       /* Check for double click. */


                /* Map the quick menu. */
                if(event->xbutton.button == Button3)
                {
                    OSWGetPointerCoords(ib->toplevel,
                        &root_x, &root_y,
                        &x, &y
                    );
                    OSWMoveWindow(ib->qmenu.toplevel, root_x, root_y);
                    ib->qmenu.x = root_x;
                    ib->qmenu.y = root_y;

                    MenuMap(&ib->qmenu);

                    events_handled++;
                    return(events_handled);
                }
	    }
            /* Status window - padlock. */
            else if((event->xany.window == ib->status_win) &&
                    (endeavour_image.padlock0 != NULL)
            )
            {
                image = endeavour_image.padlock0;

                OSWGetWindowAttributes(ib->status_win, &wattr);
                if((event->xbutton.x >= ((int)wattr.width - image->width - 15)) &&
                   (event->xbutton.x < ((int)wattr.width - 15)) &&
                   (event->xbutton.y >= 2) &&
                   (event->xbutton.y < ((int)wattr.height - 2))
                )
                {
                    if(option.write_protect)
                        option.write_protect = 0;
                    else
                        option.write_protect = 1;

                    /* Need to redraw padlocks on all windows that have. */
                    for(x = 0; x < (int)total_endeavour_wins; x++)
                    {
                        if(endeavour_win[x] == NULL) continue;
                        if(endeavour_win[x]->map_state < 1) continue;
                        EDWDraw(endeavour_win[x], DRAW_AMOUNT_STATUS);  
                    }
                    for(x = 0; x < (int)total_recbin_wins; x++)
                    {
                        if(recbin_win[x] == NULL) continue;
                        if(recbin_win[x]->map_state < 1) continue;

                        RecBinDraw(recbin_win[x], DRAW_AMOUNT_STATUS);
                    }
#ifdef USE_IMLIB
                    for(x = 0; x < (int)total_img_browsers; x++)
                    {
                        if(img_browser[x] == NULL) continue;
                        if(img_browser[x]->map_state < 1) continue;
               
                        IMBRDraw(img_browser[x], DRAW_AMOUNT_STATUS);
                    }
#endif /* USE_IMLIB */

                    events_handled++;
                    return(events_handled);
                }
            }


	    break;

	  /* ************************************************************ */
	  case ButtonRelease:
	    /* List. */
            if(event->xany.window == ib->list)
            {
		/* Button1 or Button2: Select thumb. */
		if(((event->xbutton.button == Button1) ||
                    (event->xbutton.button == Button2)
                   ) && !drag_record[0].state	/* Must not be in drag. */
		)
		{
		    /* Take prompts out of focus. */
		    ib->loc_prompt.is_in_focus = 0;
		    ib->cv_prompt.is_in_focus = 0;
		    PromptDraw(&ib->loc_prompt, PROMPT_DRAW_AMOUNT_TEXTONLY);

		    /* Calculate thumb number from coordinates. */
		    x = IMBR_GETTHUMBFROMPOS(ib,
			event->xbutton.x, event->xbutton.y);

		    /* Shift key down: Select range? */
		    if(drag_record[0].shift_key_state)
		    {
			z = 0;

                        /* Check if atleast one thumb is selected. */
                        if((int)ib->total_sel_thumbs > 0)
                        {
                            /* Top to bottom select. */
                            if(x > (int)ib->sel_thumb[0])
                            {
                                y = ib->sel_thumb[0];
                                z = x + 1;
                            }
                            else if(x < (int)ib->sel_thumb[0])
                            {
                                y = x;
                                z = ib->sel_thumb[0] + 1;
                            }
                            IMBRUnselectAll(ib);
                            while(y < z)
                            {
                                IMBRSelect(ib, y);
                                y++;
                            }
                        }
		    }
		    /* Ctrl key down: Multiple select? */
		    else if(drag_record[0].ctrl_key_state)
		    {
                        if(IMBRIsSelected(ib, x))
                                IMBRUnselect(ib, x);
                        else
                                IMBRSelect(ib, x);
		    }
		    /* Single item select and display in viewer? */
		    else if(!IMBRIsSelected(ib, x))
		    {
		      /* Unselect all. */
		      IMBRUnselectAll(ib);
		      /* Select new item. */
		      IMBRSelect(ib, x);

		      /* Check if image or directory. */
		      if((x >= 0) && (x < (int)ib->total_thumbs) &&
                         (event->xbutton.button == Button1)
                      )
		      {
			if(ib->thumb[x] != NULL)
			{
			    IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_WORKING);

			    strptr = PrefixPaths(ib->current_dir,
				ib->thumb[x]->filename);
			    if(strptr == NULL)
				tmp_path[0] = '\0';
			    else
				strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
			    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';


			    /* Is it a directory? */
                            if(ISPATHDIR(tmp_path) &&
			       (ib->thumb[x]->filename != NULL)
			    )
			    {
				if(!strcmp(ib->thumb[x]->filename, ".."))
				    sprintf(stringa,
 "Parent directory selected (double click to go to that directory)."
				    );
				else
				    sprintf(stringa,
 "Directory `%s' selected (double click to go to that directory).",
				        ib->thumb[x]->filename
				    );
				IMBRSetStatusMesg(ib, stringa);
			    }
			    /* Is it an image? */
			    else if(ISEXTIMAGE(ib->thumb[x]->filename))
			    {
		                sprintf(stringa, "Loading: %s...",
			            ib->thumb[x]->filename
			        );
				IMBRSetStatusMesg(ib, stringa);
		                IMBRDraw(ib, DRAW_AMOUNT_DIR_LIST);

				/* Load thumb to viewer. */
				PBarUpdate(&ib->pbar, 0.0, NULL);
				OSWGUISync(False);
			        IMBRLoadImageFromThumb(ib, x);			
				PBarUpdate(&ib->pbar, 1.0, NULL);

			        sprintf(stringa,
				    "Displaying: %s (%ix%i) %i Bytes",
				    ib->thumb[x]->filename,
				    ib->thumb[x]->ori_width,
				    ib->thumb[x]->ori_height,
				    ib->thumb[x]->ori_size
				);
				IMBRSetStatusMesg(ib, stringa);
			    }

                            IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_NORMAL);
                        }
		      }
                    }	/* Check if its already selected. */


                    /* Scroll if near edge. */
                    OSWGetWindowAttributes(ib->list, &wattr);
		    y = ((((int)ib->total_thumbs / (int)ib->thumbs_per_row)
			+ 2) * ib->list_row_spacing) - (int)wattr.height;

                    if((event->xbutton.y >
                       ((int)wattr.height - (int)ib->list_row_spacing)) &&
                       (ib->list_sb.y_win_pos < y)
                    )
                    {
			/* Scroll down half a window unit. */

                        IMBRDraw(ib, DRAW_AMOUNT_DIR_LIST);
                        OSWGUISync(False);

                        ib->list_sb.y_win_pos += ((int)wattr.height / 2);
                        if(ib->list_sb.y_win_pos > y)
                            ib->list_sb.y_win_pos = y;

                        usleep(SELECT_SHIFT_DELAY * 1000);
                    }
                    else if((event->xbutton.y < (int)ib->list_row_spacing) &&
                            (ib->list_sb.y_win_pos > 0)
                    )
                    {
			/* Scroll up half a window unit. */

			IMBRDraw(ib, DRAW_AMOUNT_DIR_LIST);
                        OSWGUISync(False);

                        ib->list_sb.y_win_pos -= ((int)wattr.height / 2);

                        usleep(SELECT_SHIFT_DELAY * 1000);
                    }
                    if(ib->list_sb.y_win_pos < 0)
                        ib->list_sb.y_win_pos = 0;
		}
		/* Button2: Map change values prompt. */
		if((event->xbutton.button == Button2) &&
                   ((int)ib->total_sel_thumbs > 0) &&
                   !option.write_protect
                )
		{
		    IMBRDoQuickRename(ib);
		}


		/* Report event only for Button1 or Button2. */
		if((event->xbutton.button == Button1) ||
		   (event->xbutton.button == Button2)
		)
		    events_handled++;
	    }

	    /* Unmap split cursor window on ButtonRelease. */
	    else if(drag_record[0].state &&
                   (drag_record[0].type == DRAG_TYPE_LIST_SPLIT) &&
                   ((img_browser_struct *)drag_record[0].origin_ptr == ib)
            )
            {
                OSWGetWindowAttributes(ib->toplevel, &wattr);

		/* Unmap the split cursor window. */
                OSWUnmapWindow(ib->split_cur);

		x = event->xbutton.x;
		y = event->xbutton.y;

		if(x >= ((int)wattr.width - 10))
		    x = (int)wattr.width - 10;
		if(x < 10)
		    x = 10;

		/* Resize windows. */
                if(wattr.width != 0)
                {
                    ib->list_win_width = (double)(x
                        - (int)(IB_DEF_SEPERATOR_WIDTH / 2)) /
                        (double)wattr.width;
                }

                ib->width += 1;     /* Force IMBRResize() to resize. */
                IMBRResize(ib);
                IMBRDraw(ib, DRAW_AMOUNT_LISTS);

                events_handled++;
                return(events_handled);
	    }

	    break;

          /* ************************************************************ */
	  case MotionNotify:

            /* Status window (padlock hints) */ 
            if((event->xany.window == ib->status_win) &&
               (endeavour_image.padlock0 != NULL)
            )
            {
                image = endeavour_image.padlock0;
                x = event->xmotion.x;
                y = event->xmotion.y;
                
                OSWGetWindowAttributes(ib->status_win, &wattr);
                if((x >= ((int)wattr.width - (int)image->width - 15)) &&
                   (x < ((int)wattr.width - 15)) &&
                   (y >= 2) &&
                   (y < ((int)wattr.height - 2))
                )
                {
                    /* Schedual hints window map. */
                    HintWinSetSchedual(
                        widget_global.hintwin_map_delay,
                        ib->status_win
                    );   
                        
                    events_handled++;
                }
                else 
                {
                    /* Unmap/unschedual hints window. */
                    HintWinUnmap();
                }
                return(events_handled);
            }
            /* Toplevel for thumbs list and viewer split. */
            else if(event->xany.window == ib->toplevel)
            {
                if(drag_record[0].state &&
                   (drag_record[0].type == DRAG_TYPE_LIST_SPLIT) &&
                   ((img_browser_struct *)drag_record[0].origin_ptr == ib)
                )
                {
                    /* Move the split cursor window. */
                    OSWGetPointerCoords(ib->toplevel, &root_x, &root_y,
                        &x, &y);
		    if((x > 10) && x <= ((int)ib->width - 10))
		    {
                        OSWMoveResizeWindow(ib->split_cur,
                         x - (IB_DEF_SEPERATOR_WIDTH / 2),
                         IB_LOC_RIBBON_HEIGHT,
                         IB_DEF_SEPERATOR_WIDTH,
                         MAX((int)ib->height - IB_LOC_RIBBON_HEIGHT -
                           IB_STATUS_HEIGHT, 2)
                        );
		    }

                    events_handled++;
                    return(events_handled);
		}
		else
		{
		    /* Change pointer cursor. */
                    OSWGetWindowAttributes(ib->toplevel, &wattr);

                    x = event->xmotion.x;
                    y = event->xmotion.y;
                    z = (int)(ib->list_win_width * (double)wattr.width);
                    if((x >= z) && (x < (z + IB_DEF_SEPERATOR_WIDTH)) &&
                       (y >= IB_LOC_RIBBON_HEIGHT) &&
                       (y < ((int)wattr.height - IB_STATUS_HEIGHT))
                    )
                        OSWSetWindowCursor(ib->toplevel,
                            endeavour_cursor.list_heading_resize);
                    else
                        OSWUnsetWindowCursor(ib->toplevel);

                    /* Do not report event for this. */
                }
	    }
	    /* Drag for thumbs list. */
	    else if(event->xany.window == ib->list)
	    {
		if(drag_record[0].state)
		{
		    IMBRDraw(ib, DRAW_AMOUNT_DIR_LIST_OUTLINE);
		}

		events_handled++;
		return(events_handled);
	    }

            break;

          /* ************************************************************ */
          case KeyPress:
	    if(ib->is_in_focus == 0)
		return(events_handled);

	    /* Escape key. */
	    if(event->xkey.keycode == osw_keycode.esc)
	    {
		if(ib->cv_prompt_mode != CV_PROMPT_MODE_NONE)
		{
		    PromptUnmap(&ib->cv_prompt);
                    ib->cv_prompt.is_in_focus = 0;
		    ib->cv_prompt_mode = CV_PROMPT_MODE_NONE;

                    /* Set status message. */
		    IMBRSetStatusMesg(ib,
			"Change of values operation aborted");
		}
		events_handled++;
	    }
	    /* Enter key. */
	    else if((event->xkey.keycode == osw_keycode.enter) ||
                    (event->xkey.keycode == osw_keycode.np_enter)
	    )
	    {
		/* Change values prompt apply. */
		if(ib->cv_prompt_mode != CV_PROMPT_MODE_NONE)
	        {
		    /* Ignore, will be handled when KeyRelease is recieved. */

		}
		/* Location prompt. */
		else if(ib->loc_prompt.is_in_focus == 1)
		{
		    strptr = StringCopyAlloc(ib->loc_prompt.buf);

		    /* Allow loc_prompt.buf to be recorded into history. */
		    PromptManage(&ib->loc_prompt, event);

		    strncpy(ib->loc_prompt.buf, strptr,
			ib->loc_prompt.buf_len);
		    ib->loc_prompt.buf[(int)ib->loc_prompt.buf_len - 1] = '\0';

		    IMBRDoChangeDir(ib, ib->loc_prompt.buf);

		    free(strptr); strptr = NULL;
		}
		/* If alt key is held down, map prop dialogs. */
		else if(drag_record[0].alt_key_state)
		{
		    IMBRDoPropDialog(ib);
		}
		/* No prompts selected then handle selected thumbs. */
		else
		{
                    for(x = 0; x < (int)ib->total_sel_thumbs; x++)
                    {
                        y = ib->sel_thumb[x];
                        if((y < 0) || (y >= (int)ib->total_thumbs))
                            continue;
                        if(ib->thumb[y] == NULL) continue;
                        if(ib->thumb[y]->filename == NULL) continue;

                        /* Get full path as tmp_path. */
                        strptr = PrefixPaths(ib->current_dir,
                            ib->thumb[y]->filename);
                        if(strptr == NULL) continue;   

                        strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
                        tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

                        /* Is directory? */
                        if(ISPATHDIR(tmp_path))
                        {
                            /* Change directory. */
                            IMBRDoChangeDir(ib, ib->thumb[y]->filename);

                            /* Must break after dir change. */
                            break;
                        }
                        /* Spawn image viewer. */
                        else
                        {
                            /* Allocate and initialize a new internal viewer. */
                            y = ImgViewerAllocate();
                            if(ImgViewerIsAllocated(y))
                            {
                                ImgViewerInit(img_viewer[y]);
                                ImgViewerLoadFile(img_viewer[y], tmp_path);
                                ImgViewerMap(img_viewer[y]);
                            }
                        }
                    }
		}	/* No prompts selected then handle selected thumbs. */

		events_handled++;
	    }
            /* F5: Reload */
            else if(event->xkey.keycode == osw_keycode.f5)
            {
		OSWKBAutoRepeatOff();

		ib->reload_btn.state = PBTN_ARMED;
		PBtnDraw(&ib->reload_btn);

                events_handled++;
	    }
            /* Bring up move dialog: F7. */
            else if(event->xkey.keycode == osw_keycode.f7)
            {
		IMBRDoMove(ib);

                events_handled++;
                return(events_handled);
            }
            /* Bring up copy dialog: F8. */
            else if(event->xkey.keycode == osw_keycode.f8)
            {
                IMBRDoCopy(ib);

                events_handled++;
                return(events_handled);
            }
            /* Recycle object (make sure no prompts are active). */
            else if((ib->loc_prompt.is_in_focus == 0) &&
                    (ib->cv_prompt.is_in_focus == 0) &&
                    (event->xkey.keycode == osw_keycode.ddelete)
            )
            {
		IMBRDoRecycle(ib);

                events_handled++;
            }

            break;


          /* ************************************************************ */
          case KeyRelease:
            if(ib->is_in_focus == 0)
                return(events_handled);


            /* Enter key. */
            if((event->xkey.keycode == osw_keycode.enter) ||  
               (event->xkey.keycode == osw_keycode.np_enter)
            )
            {
                /* Change values prompt apply. */
                if(ib->cv_prompt_mode != CV_PROMPT_MODE_NONE)
                {
                    if((ib->current_dir != NULL) &&
                       (ib->total_sel_thumbs > 0)
                    )
                    {
                        z = ib->sel_thumb[0];
                        if((z >= 0) && (z < (int)ib->total_thumbs))
                        {
			  /* Rename. */
                          if(ib->thumb[z] != NULL)
                          {
			    /* Get tar_path. */
                            strptr = PrefixPaths(ib->current_dir,
                                ib->cv_prompt.buf);
                            strncpy(tar_path, (strptr == NULL) ? "/badname" :
                                strptr, PATH_MAX + NAME_MAX);
                            tar_path[PATH_MAX + NAME_MAX - 1] = '\0';
                            SimplifyPath(tar_path);

                            strptr = PrefixPaths(ib->current_dir,
                                ib->thumb[z]->filename);
                            if((strptr != NULL) &&
                               (lstat(tar_path, &stat_buf) == -1)
                            )
                            {
                              strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
                              tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
                              if(ISPATHDIR(tmp_path))
                              {
                                FileOpRenameDir(tar_path, tmp_path);
                              }
                              else
                              {
                                FileOpRenameFile(tar_path, tmp_path);
                              }

                              /* Update thumb's name. */
                              free(ib->thumb[z]->filename);
                              ib->thumb[z]->filename = StringCopyAlloc(
                                  ib->cv_prompt.buf);

                              /* Set status message. */
			      sprintf(stringa,
                                "Renamed object to `%s'",
				ib->thumb[z]->filename
			      );
			      IMBRSetStatusMesg(ib, stringa);
			    }
			    else
			    {
			      /* File already exists, cannot rename. */
                              /* Set status message. */
                              sprintf(stringa,
                                "Cannot rename, object `%s' already exists.",
                                tar_path
                              );
			      IMBRSetStatusMesg(ib, stringa);
			    }
                          }
                        }
                    }
                     
                    /* Unmap change values prompt. */
                    PromptUnmap(&ib->cv_prompt);
                    ib->cv_prompt.is_in_focus = 0;
                    ib->cv_prompt_mode = CV_PROMPT_MODE_NONE;
                }

		events_handled++;
	    }
            /* F5: Reload */
            else if(event->xkey.keycode == osw_keycode.f5)
            {
		OSWKBAutoRepeatOn();

                ib->reload_btn.state = PBTN_UNARMED;
                PBtnDraw(&ib->reload_btn);

		IMBRDoReload(ib);

		events_handled++;
            }
            break;


	  /* ************************************************************ */
	  case Expose:
            if(event->xany.window == ib->toplevel)
            {
                IMBRDraw(ib, DRAW_AMOUNT_COMPLETE);
                OSWPurgeWindowTypedEvent(event->xany.window, Expose);

                events_handled++;
                return(events_handled);
            }
            else if(event->xany.window == ib->list)
            {
                IMBRDraw(ib, DRAW_AMOUNT_DIR_LIST);
                OSWPurgeWindowTypedEvent(event->xany.window, Expose);

                events_handled++;
                return(events_handled);
            }
            else if(event->xany.window == ib->viewer)
            {
                IMBRDraw(ib, DRAW_AMOUNT_FILE_LIST);
                OSWPurgeWindowTypedEvent(event->xany.window, Expose);

                events_handled++;
                return(events_handled);
            }
            else if(event->xany.window == ib->status_win)
            {
                IMBRDraw(ib, DRAW_AMOUNT_STATUS);
                OSWPurgeWindowTypedEvent(event->xany.window, Expose);

                events_handled++;
                return(events_handled);
            }
	    break;

          /* ********************************************************* */
          case ConfigureNotify:
            if(event->xany.window == ib->toplevel)
            {
                IMBRResize(ib);
                events_handled++;
            }
            break;

          /* ********************************************************* */
          case FocusIn:
            if(event->xany.window == ib->toplevel)
            {
                ib->is_in_focus = 1;
                events_handled++;
            }
            else
            {
                ib->is_in_focus = 0;
                /* Do not report any event for focus out. */
            }
            return(events_handled);
            break;
             
          /* ********************************************************* */
          case FocusOut:
            if(event->xany.window == ib->toplevel)
            {
                ib->is_in_focus = 0;
                events_handled++;
            }
            return(events_handled);
            break;

          /* ********************************************************* */
          case VisibilityNotify:
            if(event->xany.window == ib->toplevel)
            {
                ib->visibility_state = event->xvisibility.state;
             
                events_handled++;
                return(events_handled);
            }
            break;
          
          /* ********************************************************* */
          case LeaveNotify:
	    if(event->xany.window == ib->toplevel)
	    {
		if(!drag_record[0].state)
		    OSWUnsetWindowCursor(ib->toplevel);

                events_handled++;
                return(events_handled);
	    }
	    else if(event->xany.window == ib->status_win)
	    {
                /* Unmap/unschedual hints window. */
                HintWinUnmap();

		events_handled++;
                return(events_handled);
	    }
            break;
	}

        /* Redraw as needed. */
        if(events_handled > 0)
        {
            IMBRDraw(ib, DRAW_AMOUNT_COMPLETE);
            return(events_handled);
        }

                
        /* ************************************************************ */
        /* Manage widgets. */

        /* if event is button press, bring all prompts out of focus. */
        if(event->type == ButtonPress)
        {
            ib->loc_prompt.is_in_focus = 0;   
            ib->cv_prompt.is_in_focus = 0;
        }

	/* Location prompt. */
	if(events_handled == 0)
        {
	    events_handled += PromptManage(&ib->loc_prompt, event);

	}

	/* Change values prompt. */
        if((events_handled == 0) && (ib->cv_prompt_mode != CV_PROMPT_MODE_NONE))
        {
            events_handled += PromptManage(&ib->cv_prompt, event);
         
        }


        /* List scroll bars. */
        if((events_handled == 0) && (ib->thumbs_per_row != 0))
        {
            OSWGetWindowAttributes(ib->list, &wattr);
            events_handled += SBarManage(&ib->list_sb,
                wattr.width,
                wattr.height,
                wattr.width,
                (((int)ib->total_thumbs / (int)ib->thumbs_per_row) + 2) *
		ib->list_row_spacing,
                event
            );

            if(events_handled > 0)
                IMBRDraw(ib, DRAW_AMOUNT_DIR_LIST);
        }


        /* Progress bar. */
        if(events_handled == 0)
	{
	    events_handled += PBarManage(&ib->pbar, event);
	}


	/* Prev button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(&ib->prev_btn, event);
	    if((events_handled > 0) &&
               ((event->type == ButtonRelease) ||
                (event->type == KeyRelease)
	       )
	    )
	    {
		IMBRDoPrev(ib);
	    }
        }
        /* Next button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(&ib->next_btn, event);
            if((events_handled > 0) &&
               ((event->type == ButtonRelease) ||
                (event->type == KeyRelease)
               )
            )
            {
		IMBRDoNext(ib);
            }
        }
        /* View button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(&ib->view_btn, event);
            if((events_handled > 0) &&
               ((event->type == ButtonRelease) ||
                (event->type == KeyRelease)
               )
            )
            {
		IMBRDoView(ib);
            }
        }
        /* Reload button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(&ib->reload_btn, event);
            if((events_handled > 0) &&
               ((event->type == ButtonRelease) ||
                (event->type == KeyRelease)
               )
            )
            {
		IMBRDoReload(ib);
	    }
        }
        /* Stop button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(&ib->stop_btn, event);
            if((events_handled > 0) &&
               ((event->type == ButtonRelease) ||
                (event->type == KeyRelease)
               )
            )
            {
		IMBRDoStop(ib);
            }
        }
                
        /* Quick menu. */
        if(events_handled == 0)
        {
            x = ib->qmenu.map_state;

            events_handled += MenuManage(&ib->qmenu, event);

            /* Unmap as needed. */
            if((events_handled <= 0) &&
               (event->type == ButtonRelease)
            )
            {
                MenuUnmap(&ib->qmenu);
            }
            
            /* Check if it just unmapped. */
            if((x == 1) &&
               (ib->qmenu.map_state == 0)
            )
            {
                OSWPurgeAllEvents();
                IMBRDraw(ib, DRAW_AMOUNT_COMPLETE);
            }
        }



	return(events_handled);
}



int IMBRManageAll(event_t *event)
{
        static int i, events_handled;
         
         
        /* Error checks. */
        if(!IDC() ||
           (event == NULL)
        )
            return(0);

        /* ********************************************************** */
        events_handled = 0;
             
        for(i = 0; i < (int)total_img_browsers; i++)
        {
            if(!IMBRIsInit(i))
                continue;

            /* ClientMessage must be handled here. */
            if((event->type == ClientMessage) &&
               OSWIsEventDestroyWindow(img_browser[i]->toplevel, event)
            )
            {
                /* Unmap and reset resources which may referance us. */
                ComfDialogUnmap();
                if((int)total_op_dialogs > 0)
                    OPDialogUnmap(op_dialog[0]);

                /* Delete then reclaim memory. */
                IMBRDelete(i);
		IMBRReclaimMemory();

                OSWPurgeAllEvents();
 
                events_handled++;
                break;  /* Break since events are now messed up due to delete. */
            }
 

            events_handled += IMBRManage(img_browser[i], event);
        }


        return(events_handled);
}



void IMBRMap(img_browser_struct *ib)  
{
        if(!IDC() || (ib == NULL)) return;

        UnfocusAllTopLvlWins();

	ib->map_state = 0;
	IMBRDraw(ib, DRAW_AMOUNT_COMPLETE);
 
        ib->is_in_focus = 1;
        ib->visibility_state = VisibilityUnobscured;

        return;

}



void IMBRUnmap(img_browser_struct *ib)                
{
	if(!IDC() || (ib == NULL)) return;


	OSWUnmapWindow(ib->toplevel);
	ib->map_state = 0;

	ib->is_in_focus = 0;
	ib->visibility_state = VisibilityFullyObscured;

	return;
}



/*
 *	Deallocates substructures of ib but does not deallocate
 *	ib itself.
 */
void IMBRDestroy(img_browser_struct *ib)
{
	if(ib == NULL) return;

	ib->map_state = 0;
        ib->visibility_state = VisibilityFullyObscured;
        ib->is_in_focus = 0;
        ib->x = 0;
        ib->y = 0;
        ib->width = 0;
        ib->height = 0;
        ib->disabled = False;

	ib->cv_prompt_mode = CV_PROMPT_MODE_NONE;

	/* Free current dir. */
	free(ib->current_dir);
	ib->current_dir = NULL;

	/* Free status message. */
	free(ib->status_mesg);
	ib->status_mesg = NULL;

	/* Free file queue. */
	IMBRDeleteFileQueue(ib);

	/* Unselect and delete all loaded thumbs. */
	IMBRDeleteAllThumbs(ib);

	if(IDC())
	{
	    MenuDestroy(&ib->qmenu);

	    PromptDestroy(&ib->cv_prompt);
	    PromptDestroy(&ib->loc_prompt);

	    PBtnDestroy(&ib->prev_btn);
            PBtnDestroy(&ib->next_btn);
            PBtnDestroy(&ib->view_btn);
            PBtnDestroy(&ib->reload_btn);
            PBtnDestroy(&ib->stop_btn);

	    OSWDestroyPixmap(&ib->loc_ribbon_buf);
            OSWDestroyWindow(&ib->loc_ribbon);

	    OSWDestroyWindow(&ib->split_cur);

	    HintWinDeleteMessage(ib->status_win);          /* Padlock hint. */
	    PBarDestroy(&ib->pbar);
	    OSWDestroyPixmap(&ib->status_win_buf);
	    OSWDestroyWindow(&ib->status_win);

	    SBarDestroy(&ib->list_sb);
	    OSWDestroyPixmap(&ib->list_buf);
            OSWDestroyWindow(&ib->list);

            OSWDestroyPixmap(&ib->viewer_buf);
            OSWDestroyWindow(&ib->viewer);

	    OSWDestroyPixmap(&ib->toplevel_buf);
	    OSWDestroyWindow(&ib->toplevel);

	}

	/* Stop loading (just for safety). */
	ib->stop_load = True;


	return;
}


#endif /* USE_IMLIB */
