/*

                        Image Browser: Macros


	Functions:

        void IMBR_SETALLCURSORS(img_browser_struct *ib, int code)
	int IMBR_GETTHUMBFROMPOS(img_browser_struct *ib, int x, int y)

	int IMBRSetStatusMesg(img_browser_struct *ib, char *mesg)

	int IMBRDisplayOnRoot(img_browser_struct *ib, int n)
	int IMBRDoChangeDir(img_browser_struct *ib, char *path)
        int IMBRLoad(img_browser_struct *ib, char *path)
	int IMBRLoadImageFromThumb(img_browser_struct *ib, int n)
	int IMBRLoadThumbFromQueue(img_browser_struct *ib, int fq_num)
	void IMBRDeleteFileQueue(img_browser_struct *ib)

	int IMBRDoPrev(img_browser_struct *ib)
	int IMBRDoNext(img_browser_struct *ib)
	int IMBRDoView(img_browser_struct *ib)
        int IMBRDoReload(img_browser_struct *ib)
        int IMBRDoStop(img_browser_struct *ib)
        int IMBRDoRecycle(img_browser_struct *ib)

        int IMBRSelect(img_browser_struct *ib, int n)
        void IMBRUnselect(img_browser_struct *ib, int n)
	int IMBRIsSelected(img_browser_struct *ib, int n)
	void IMBRUnselectAll(img_browser_struct *ib)

	void IMBRDeleteAllThumbs(img_browser_struct *ib)

 */

#ifdef USE_IMLIB

#include "endeavour.h"
#include "imbrdefaults.h"

#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))



/*
 *	Sets all pointer cursors for ib's windows
 *	depending on the code.
 */
void IMBR_SETALLCURSORS(img_browser_struct *ib, int code)
{
	if(!IDC() || (ib == NULL))
	    return;


        switch(code)
        {
          case EDW_CURSORS_CODE_WORKING:  
            OSWSetWindowCursor(ib->toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(ib->loc_ribbon, endeavour_cursor.working);
            OSWSetWindowCursor(ib->list, endeavour_cursor.working);
            OSWSetWindowCursor(ib->viewer, endeavour_cursor.working);
            OSWSetWindowCursor(ib->status_win, endeavour_cursor.working);
	    break;

          default: /* EDW_CURSORS_CODE_NORMAL */
	    OSWUnsetWindowCursor(ib->toplevel);
            OSWUnsetWindowCursor(ib->loc_ribbon);
            OSWUnsetWindowCursor(ib->list);
            OSWUnsetWindowCursor(ib->viewer);
            OSWUnsetWindowCursor(ib->status_win);
            break;
	}

	OSWGUISync(False);

	return;
}



/*
 *	Calculate thumb position from x and y coordinates.
 */
int IMBR_GETTHUMBFROMPOS(img_browser_struct *ib, int x, int y)
{
	int row, col, z;


	if(ib == NULL)
	    return(0);

        /* Calculate thumb that is selected. */
        row = (int)(y - IB_DEF_MARGIN +
            ib->list_sb.y_win_pos) / ib->list_row_spacing;
        col = (int)(x - IB_DEF_MARGIN +
            ib->list_sb.x_win_pos) / ib->list_col_spacing;

        z = (row * (int)ib->thumbs_per_row) + col;


	return(z);
}



/*
 *	Updates the status message on ib.
 */
int IMBRSetStatusMesg(img_browser_struct *ib, char *mesg)
{
	if(ib == NULL) return(-1);


	/* If mesg is pointed to NULL then unset message. */
	if(mesg == NULL)
	{
	    free(ib->status_mesg);
	    ib->status_mesg = NULL;
	}
	/* Else set new message. */
	else
	{
            free(ib->status_mesg); 
            ib->status_mesg = StringCopyAlloc(mesg);
	}


	/* Redraw if display is connected. */
	if(IDC())
	{
	    IMBRDraw(ib, DRAW_AMOUNT_STATUS);
	    OSWGUISync(False);
	}


	return(0);
}



/*
 *	Displays thumb n of ib on root window.
 */
int IMBRDisplayOnRoot(img_browser_struct *ib, int n)
{
	char *strptr;
	char stringa[PATH_MAX + NAME_MAX + 128];
	char filename[PATH_MAX + NAME_MAX];
	image_t *ori_img, *scaled_img;
	pixmap_t pixmap;
	win_attr_t wattr;


	/* Error checks. */
	if(!IDC() ||
           (ib == NULL)
	)
	    return(-1);

	if((n < 0) || (n >= ib->total_thumbs))
	    return(-1);
	if(ib->thumb[n] == NULL)
	    return(-1);
	if(ib->thumb[n]->filename == NULL)
	    return(-1);
	if(ib->current_dir == NULL)
	    return(-1);


	/* ************************************************************* */

	/* Get filename. */
	strptr = PrefixPaths(ib->current_dir, ib->thumb[n]->filename);
	if(strptr == NULL) return(-1);
	strncpy(filename, strptr, PATH_MAX + NAME_MAX);
	filename[PATH_MAX + NAME_MAX - 1] = '\0';

	/* Set message. */
	sprintf(stringa, "Loading: %s...", ib->thumb[n]->filename);
	IMBRSetStatusMesg(ib, stringa);

	/* Set cursors to working. */
	IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_WORKING);
        PBarUpdate(&ib->pbar, 0.00, NULL); OSWGUISync(False);

	/* Load image. */
	if(IMLIBOSWLoadFileToImage(filename, &ori_img))
	{
            PBarUpdate(&ib->pbar, 0.25, NULL); OSWGUISync(False);

	    /* Error loading image. */
            sprintf(stringa, "Loading: Failed");
            IMBRSetStatusMesg(ib, stringa);

	    sprintf(stringa,
"Error loading image:\n\n\
    %s\n\n",
		filename
	    );
	    printdw(&err_dw, stringa);
	}
	else
	{
            PBarUpdate(&ib->pbar, 0.25, NULL); OSWGUISync(False);

	    /* Rescale image to fit desktop. */
	    OSWGetWindowAttributes(osw_gui[0].root_win, &wattr);

	    sprintf(stringa, "Loading: Resizing buffer...");
            IMBRSetStatusMesg(ib, stringa);

	    /* Create scaled image. */
	    if(OSWCreateImage(&scaled_img, wattr.width, wattr.height))
	    {
		/* Free up resources on fail. */
		OSWDestroyImage(&ori_img);
                IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_NORMAL);
		PBarUpdate(&ib->pbar, 0.00, NULL); OSWGUISync(False);
		return(-1);
	    }
	    /* Resize buffer. */
	    WidgetResizeImageBuffer(
		osw_gui[0].depth,
		(u_int8_t *)scaled_img->data,		/* Target. */
		(u_int8_t *)ori_img->data,		/* Source. */
		scaled_img->width, scaled_img->height,
		ori_img->width, ori_img->height
	    );
            PBarUpdate(&ib->pbar, 0.50, NULL); OSWGUISync(False);
	    /* Destroy original image. */
	    OSWDestroyImage(&ori_img);

	    /* Create pixmap to recieve scaled_img. */
	    if(OSWCreatePixmap(&pixmap, scaled_img->width, scaled_img->height))
	    {
		OSWDestroyImage(&scaled_img);
                IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_NORMAL);
                PBarUpdate(&ib->pbar, 0.00, NULL); OSWGUISync(False);
		return(-1);
	    }

	    sprintf(stringa, "Loading: Blitting buffer...");
	    IMBRSetStatusMesg(ib, stringa);

	    /* Put image to pixmap. */
            OSWPutImageToDrawable(scaled_img, (drawable_t)pixmap);
            /* Destroy scaled image. */  
            OSWDestroyImage(&scaled_img);
            PBarUpdate(&ib->pbar, 0.75, NULL); OSWGUISync(False);

	    /* Set root window background to be the pixmap. */
	    OSWSetWindowBkg(osw_gui[0].root_win, 0, pixmap);
	    OSWClearWindow(osw_gui[0].root_win);
	    OSWCopyDrawables(osw_gui[0].root_win, pixmap,
		wattr.width, wattr.height);

	    OSWDestroyPixmap(&pixmap);
	}


        /* Set cursors to normal. */
        IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_NORMAL);
	PBarUpdate(&ib->pbar, 1.00, NULL);

        sprintf(stringa, "Loading: Done");
	IMBRSetStatusMesg(ib, stringa);


	return(0);
}



/*
 *	Handles change of directory, adjusts all resources on the ib
 *	and sets up its file queue to load to thumbs.
 */
int IMBRDoChangeDir(img_browser_struct *ib, char *path)
{
	char *strptr;
	char tmp_path[PATH_MAX + NAME_MAX];


	/* Error checks. */
	if(!IDC() || (ib == NULL)) return(-1);
	if(path == NULL) return(-2);


	/* Parse paths. */
	strptr = PrefixPaths(ib->current_dir, path);
	if(strptr == NULL) return(-1);

	strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
	tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

	/* Simply new path. */
	SimplifyPath(tmp_path);


	/*   Note: current_dir and loc_prompt.buf will be set in
         *   IMBRLoad().
	 */
  
	/* ************************************************************* */

        IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_WORKING);


	/* Free thumbs and previous file queue, then set new file queue. */
	IMBRLoad(ib, tmp_path);

	/* Move scroll bars to top. */
	ib->list_sb.x_win_pos = 0;
        ib->list_sb.y_win_pos = 0;

	/* Redraw ib completely including its loc_prompt. */
	IMBRDraw(ib, DRAW_AMOUNT_COMPLETE);
	PromptDraw(&ib->loc_prompt, PROMPT_DRAW_AMOUNT_COMPLETE);

        /* Select entire buffer of the target prompt. */
        PromptMarkBuffer(&ib->loc_prompt, PROMPT_POS_END);


        IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_NORMAL);
	

	return(0);
}



/*
 *	Sets up ib's file queue list to load all images and dirs
 *	in path.  Does not actually load to each thumb, that is handled
 *	in IMBRRepeatCB()
 */
int IMBRLoad(img_browser_struct *ib, char *path)
{
	int i, n;
	char **strv;
	int strc;
	char *strptr;
        char tmp_path[PATH_MAX + NAME_MAX];


	/* Error checks. */
	if((ib == NULL) || (path == NULL))
	    return(-1);


	/* Attempt to get directory listing. */
	strv = GetDirEntNames(path);
	if(strv == NULL)
	    return(-1);
	for(i = 0; strv[i] != NULL; i++)
	{
	}
	strc = i;

	StringQSort(strv, strc);


	/* Unload (if any) previous selection, thumbs, and file queue. */
	IMBRDeleteAllThumbs(ib);	/* This will unselect as well. */
	IMBRDeleteFileQueue(ib);

	/* Reset stop_load. */
	ib->stop_load = False;

	/* Set new current_dir. */
	free(ib->current_dir);
	ib->current_dir = StringCopyAlloc(path);

        /* Set new current directory to location prompt. */
        if(ib->loc_prompt.buf != NULL)
        {
            strncpy(ib->loc_prompt.buf, ib->current_dir, ib->loc_prompt.buf_len);
            ib->loc_prompt.buf[(int)ib->loc_prompt.buf_len - 1] = '\0';

	    /* Select entire buffer of the target prompt. */
            PromptMarkBuffer(&ib->loc_prompt, PROMPT_POS_END);
        }
 

	/* Set only image files and dirs to file queue. */
	ib->total_fqs = 0;
	ib->last_fq = -1;

	/* Get directories first. */
	for(i = 0; i < strc; i++)
	{
	    if(strv[i] == NULL) continue;

	    strptr = PrefixPaths(ib->current_dir, strv[i]);
	    if(strptr == NULL) continue;

	    strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
	    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

	    /* Do not count special dir notations. */
	    if(!strcmp(strv[i], "."))
	    {
                free(strv[i]);
                strv[i] = NULL; 
		continue;
	    }

	    /* Check if it is a directory. */
	    if(ISPATHDIR(tmp_path))
	    {
		/* Path is a directory. */

		/* Allocate a new fq name entry. */
		n = ib->total_fqs;
		ib->total_fqs++;
		ib->fq_name = (char **)realloc(ib->fq_name,
		    ib->total_fqs * sizeof(char *)
		);
		if(ib->fq_name == NULL)
		{
		    ib->total_fqs = 0;
		    continue;
		}

		/* Set file queue name pointer to entry. */
		ib->fq_name[n] = strv[i];
	    }
	}

        /* Get images second. */
        for(i = 0; i < strc; i++)
        {
            if(strv[i] == NULL) continue;

            strptr = PrefixPaths(ib->current_dir, strv[i]);
            if(strptr == NULL) continue;
        
            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

            /* Skip directories, we already got them. */
            if(ISPATHDIR(tmp_path)) continue;

	    /* Check extension of name to see if it is an image. */
	    if(ISEXTIMAGE(strv[i]))
	    {
		/* Path is an image. */

                /* Allocate a new fq name entry. */
                n = ib->total_fqs;
                ib->total_fqs++;
                ib->fq_name = (char **)realloc(ib->fq_name,
                    ib->total_fqs * sizeof(char *)
                );
                if(ib->fq_name == NULL)
                {
                    ib->total_fqs = 0;
                    continue;
                }

                /* Set file queue name pointer to entry. */
                ib->fq_name[n] = strv[i];
	    }
	    /* Entry is not an image or a dir, free it. */
	    else
	    {
		free(strv[i]);
		strv[i] = NULL;
	    }
	}


	/* Free pointers to dir entries but not each dir entry. */
	free(strv);
	strv = NULL;


	/* Scroll thumbs scroll bar to top. */
	ib->list_sb.x_win_pos = 0;
	ib->list_sb.y_win_pos = 0;


	return(0);
}



/*
 *	Loads an image with information referanced on an allocated thumb
 *	structure on ib.
 */
int IMBRLoadImageFromThumb(img_browser_struct *ib, int n)
{
        int x, y;
        char tmp_path[PATH_MAX + NAME_MAX];
        char *strptr;

	double coeff;
        image_t *image;
        unsigned int width, height;	/* Original image size. */

	image_t *s_image;
	unsigned int s_width, s_height;	/* Scaled image size. */

	win_attr_t wattr;		/* Attributes of viewer window. */


	/* Error checks. */
	if(!IDC() || (ib == NULL)) return(-1);
	if(ib->current_dir == NULL) return(-1);
	if((n < 0) || (n >= (int)ib->total_thumbs)) return(-1);
	if(ib->thumb[n] == NULL) return(-1);
        if(ib->thumb[n]->filename == NULL) return(-1);


	/* Must be an image. */
	if(!ISEXTIMAGE(ib->thumb[n]->filename)) return(-1);


	/* Set tmp_path as absolute path to image. */
	strptr = PrefixPaths(ib->current_dir, ib->thumb[n]->filename);
	if(strptr == NULL) return(-1);

	strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
	tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

	/* Make sure tmp_path is not a directory. */
	if(ISPATHDIR(tmp_path)) return(-1);


	/* ************************************************************* */

	/* Load actual image. */
	if(IMLIBOSWLoadFileToImage(tmp_path, &image))
	{
	    /* Use `failed load' image. */
	    if(endeavour_image.imbr_err_img != NULL)
	    {
		width = endeavour_image.imbr_err_img->width;
		height = endeavour_image.imbr_err_img->height;

                /* Create GUI native image buffer. */
                if(OSWCreateImage(&image, width, height))
                    return(-1);

		memcpy(
		    image->data,	/* Target. */
		    endeavour_image.imbr_err_img->data,	/* Source. */
		    image->width * image->height * (osw_gui[0].depth >> 3)
		);
	    }
	    else
	    {
	        width = 48;
	        height = 48;

                /* Create GUI native image buffer. */
                if(OSWCreateImage(&image, width, height))
                    return(-1);
	    }
        }
        else
        {
            /* Successful image load. */

            /* Get image values. */
            width = image->width;
            height = image->height;
	}

	/* ******************************************************* */
	/* Modify image size to fit on viewer_buf pixmap. */

	/* Get attributes for viewer window. */
	OSWGetWindowAttributes(ib->viewer, &wattr);

	/* Redraw background of viewer. */
	WidgetPutImageTile(
	    (drawable_t)ib->viewer_buf,
	    widget_global.std_bkg_img,
	    wattr.width, wattr.height
	);


	/* Calculate scaled down width and height. */
	if(((int)image->width > (int)wattr.width) &&
           (image->width != 0)
	)
	{
	    coeff = (double)wattr.width / (double)image->width;
	    s_width = (unsigned int)(coeff * (double)image->width);
            s_height = (unsigned int)(coeff * (double)image->height);
	}
	else if(((int)image->height > (int)wattr.height) &&
                (image->height != 0)
	)
        {
            coeff = (double)wattr.height / (double)image->height;
            s_width = (unsigned int)(coeff * (double)image->width);
            s_height = (unsigned int)(coeff * (double)image->height);
        }
	else
        {
            s_width = image->width;
            s_height = image->height;
        }

	if(((int)s_width > (int)wattr.width) &&
	   (s_width != 0)
	)
	{
	    coeff = (double)wattr.width / (double)s_width;
	    s_width = (unsigned int)(coeff * (double)s_width);
            s_height = (unsigned int)(coeff * (double)s_height);
	}
        if(((int)s_height > (int)wattr.height) &&
           (s_height != 0)
	)
        {
            coeff = (double)wattr.height / (double)s_height;
            s_width = (unsigned int)(coeff * (double)s_width);
            s_height = (unsigned int)(coeff * (double)s_height);
        }

	/* Create scaled GUI native image. */
	if(OSWCreateImage(&s_image, s_width, s_height))
	{
            OSWDestroyImage(&image);
	    return(-1);
	}

	WidgetResizeImageBuffer(
	    osw_gui[0].depth,
	    (u_int8_t *)s_image->data,		/* Target. */
	    (u_int8_t *)image->data,		/* Source. */
	    s_image->width, s_image->height,
	    image->width, image->height
	);

	/* Put scaled image to viewer_buf. */
	x = MAX(((int)wattr.width / 2) - ((int)s_image->width / 2), 0);
	y = MAX(((int)wattr.height / 2) - ((int)s_image->height / 2), 0);
	OSWPutImageToDrawablePos(
	    s_image, (drawable_t)ib->viewer_buf, x, y
	);


        /* Draw inner frame on viewer. */
	OSWSetFgPix(widget_global.surface_shadow_pix);
	OSWDrawLine((drawable_t)ib->viewer_buf,
	    x - 1, y + (int)s_image->height,
	    x - 1, y - 1
	);
        OSWDrawLine((drawable_t)ib->viewer_buf,
            x - 1, y - 1,
            x + (int)s_image->width, y - 1
        );
        OSWSetFgPix(widget_global.surface_highlight_pix);
        OSWDrawLine((drawable_t)ib->viewer_buf,
            x - 1, y + (int)s_image->height,
            x + (int)s_image->width, y + (int)s_image->height
        );
        OSWDrawLine((drawable_t)ib->viewer_buf,
            x + (int)s_image->width, y + (int)s_image->height,
            x + (int)s_image->width, y - 1
        );

	/* Draw outer frame on viewer. */
        WidgetFrameButtonPixmap(
            ib->viewer_buf,
            True,
            wattr.width, wattr.height,
            widget_global.surface_highlight_pix,
            widget_global.surface_shadow_pix
        );

	

	/* Destroy GUI native images. */
	OSWDestroyImage(&image);
        OSWDestroyImage(&s_image);


	return(0);
}



/*
 *	Loads a Thumb from the File Queue.
 */
int IMBRLoadThumbFromQueue(img_browser_struct *ib, int fq_num)
{
        int thumb_num;
        unsigned int width, height;
        char filename[PATH_MAX + NAME_MAX];
        char stringa[PATH_MAX + NAME_MAX + 128];
        char *strptr;
        image_t *image;
        struct stat stat_buf;


	/* Error checks. */
	if(!IDC() || (ib == NULL)) return(-1);
	if(ib->current_dir == NULL) return(-1);

	/* Check if fq_num is valid. */
	if((fq_num < 0) || (fq_num >= (int)ib->total_fqs)) return(-1);
	if(ib->fq_name[fq_num] == NULL) return(-1);


	/* ************************************************************* */

        /* Sanitize total_thumbs values. */
        if((int)ib->total_thumbs < 0)
            ib->total_thumbs = 0;

        strptr = PrefixPaths(ib->current_dir, ib->fq_name[fq_num]);
        if(strptr == NULL) return(-1);

        strncpy(filename, strptr, PATH_MAX + NAME_MAX);
        filename[PATH_MAX + NAME_MAX - 1] = '\0';

        /* Update status message. */
        sprintf(stringa,
            "Loading: Object %i of %i (%.0f%% Complete)",
            (int)fq_num + 1,
            (int)ib->total_fqs,
            (double)fq_num /
                (double)ib->total_fqs * 100
        );
        IMBRSetStatusMesg(ib, stringa);


        /* Allocate a new thumb structure. */
        thumb_num = ib->total_thumbs;
        ib->total_thumbs++;
        ib->thumb = (img_browser_thumb_struct **)realloc(
            ib->thumb,
            ib->total_thumbs * sizeof(img_browser_thumb_struct *)
        );
        if(ib->thumb == NULL)
        {
            ib->total_thumbs = 0;
            return(-1);
        }
        ib->thumb[thumb_num] = (img_browser_thumb_struct *)
            calloc(1, sizeof(img_browser_thumb_struct)
        );
        if(ib->thumb[thumb_num] == NULL)
        {
            ib->total_thumbs--;
            return(-1);
        }


        /* Set file name. */
        ib->thumb[thumb_num]->filename = StringCopyAlloc(ib->fq_name[fq_num]);

        /* Get file size (follow links). */
        stat(filename, &stat_buf);
        ib->thumb[thumb_num]->ori_size = stat_buf.st_size;


        /* Create thumb image. */
        if(
            OSWCreateImage(
                &ib->thumb[thumb_num]->image,
                ib->thumb_width,
                ib->thumb_height
            )
        )
	    return(-1);


        /* Is directory? */
        if(ISPATHDIR(filename))
        { 
	    /* Load directory thumb image. */

            /* Set image values. */
            ib->thumb[thumb_num]->ori_width = 0;
            ib->thumb[thumb_num]->ori_height = 0;

            /* Copy dir icon image to thumb image. */
            if(widget_global.diricon_normal_img != NULL)
            {
                width = widget_global.diricon_normal_img->width;
                height = widget_global.diricon_normal_img->height;

                if(OSWCreateImage(&image, width, height))
                    return(-1);

                memcpy(
                    image->data, 
                    widget_global.diricon_normal_img->data,
                    image->width * image->height * (osw_gui[0].depth >> 3)
                );
            }

            /* Resize original image to thumb image. */
            WidgetResizeImageBuffer(
                osw_gui[0].depth,
                (u_int8_t *)ib->thumb[thumb_num]->image->data,	/* Tar. */
                (u_int8_t *)image->data,			/* Src. */
                ib->thumb[thumb_num]->image->width,
                ib->thumb[thumb_num]->image->height,
                image->width, image->height
            );
                    
            OSWDestroyImage(&image);
	}
        /* Load actual image from disk. */
        else 
        {
            if(IMLIBOSWLoadFileToImage(filename, &image))
            {
                /* Use `failed load' image. */
                if(endeavour_image.imbr_err_img != NULL) 
                {
                    ib->thumb[thumb_num]->ori_width =
                        endeavour_image.imbr_err_img->width;
                    ib->thumb[thumb_num]->ori_height =
                        endeavour_image.imbr_err_img->height;

                    /* Resize image to thumb image size. */
                    WidgetResizeImageBuffer(
                        osw_gui[0].depth,
                        (u_int8_t *)ib->thumb[thumb_num]->image->data,	/* Tar. */
			(u_int8_t *)endeavour_image.imbr_err_img->data,	/* Src. */
                        ib->thumb[thumb_num]->image->width,
                        ib->thumb[thumb_num]->image->height,
                        endeavour_image.imbr_err_img->width,
                        endeavour_image.imbr_err_img->height
                    );
                }
            }
            else
            {
                /* Record original image values. */  
                ib->thumb[thumb_num]->ori_width = image->width;
                ib->thumb[thumb_num]->ori_height = image->height;

                /* Resize image to thumb image size. */
                WidgetResizeImageBuffer(
                    osw_gui[0].depth,
                    (u_int8_t *)ib->thumb[thumb_num]->image->data,	/* Tar. */
                    (u_int8_t *)image->data,				/* Src. */
		    ib->thumb[thumb_num]->image->width,
		    ib->thumb[thumb_num]->image->height,
		    image->width, image->height
                );

                /* Destroy original GUI native image. */
                OSWDestroyImage(&image);
            }
        }

        /* Update progress bar. */
        PBarUpdate(
            &ib->pbar,
            (double)(fq_num + 1) / (double)ib->total_fqs,
            NULL
        );

        /* Draw `Loading: Done' message? */
        if((fq_num + 1) >= (int)ib->total_fqs)
        {
	    IMBRSetStatusMesg(ib, "Loading: Done");
        }

        /* Redraw ib completely. */
        IMBRDraw(ib, DRAW_AMOUNT_COMPLETE);


	return(0);
}


/*
 *	Free file queue on ib and current_dir.
 */
void IMBRDeleteFileQueue(img_browser_struct *ib)
{
	int i;

	if(ib == NULL) return;


	/* Free file load queue. */
	for(i = 0; i < ib->total_fqs; i++)
	{
	    free(ib->fq_name[i]);
	}
	free(ib->fq_name);
	ib->fq_name = NULL;

	ib->total_fqs = 0;
	ib->last_fq = -1;


	/* Free current_dir. */
	free(ib->current_dir);
	ib->current_dir = NULL;


	return;
}



/*
 *	Select previous thumb procedure.
 */
int IMBRDoPrev(img_browser_struct *ib)
{
	int y, z;
        char stringa[PATH_MAX + NAME_MAX + 128];
        win_attr_t wattr;


	/* Error checks. */
        if(!IDC() || (ib == NULL)) return(-1);


        OSWGetWindowAttributes(ib->list, &wattr);
                        
        /* If no thumb is selected, go to last. */
        if(((int)ib->total_sel_thumbs <= 0) &&
           ((int)ib->total_thumbs > 0)
        )
        {
            z = ib->total_thumbs - 1;
            IMBRSelect(ib, z);
        }
        /* Select previous thumb. */
        else if(((int)ib->total_sel_thumbs > 0) &&
                ((int)ib->total_thumbs > 0)
        )
        {
            z = ib->sel_thumb[0] - 1;
            if(z < 0)
            {
                ib->list_sb.y_win_pos =
                    ((((int)ib->total_thumbs / (int)ib->thumbs_per_row)
                    + 2) * ib->list_row_spacing) - (int)wattr.height; 
                z = ib->total_thumbs - 1;
            }

            IMBRUnselectAll(ib);
            IMBRSelect(ib, z); 
        }
        else
        {
            return(-2);
        }

        /* Scroll if near edge. */
        y = (z - (z % (int)ib->thumbs_per_row)) /
            (int)ib->thumbs_per_row;
        y *= (int)ib->list_row_spacing;

        if(y < ib->list_sb.y_win_pos)
            ib->list_sb.y_win_pos -= (int)wattr.height / 2;
        if(y > ((int)ib->list_sb.y_win_pos + ((int)wattr.height / 2)))
            ib->list_sb.y_win_pos += (int)wattr.height / 2;

        /* Sanitize scroll bars. */
        y = ((((int)ib->total_thumbs / (int)ib->thumbs_per_row)
            + 2) * ib->list_row_spacing) - (int)wattr.height;
        if(ib->list_sb.y_win_pos > y)
            ib->list_sb.y_win_pos = y;
        if(ib->list_sb.y_win_pos < 0)
            ib->list_sb.y_win_pos = 0;

        /* Load newly selected thumb to viewer. */
        if(ib->thumb[z] != NULL)
        {
            IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_WORKING);

            if(ISEXTIMAGE(ib->thumb[z]->filename))
            {
                sprintf(stringa, "Loading: %s...",
                    ib->thumb[z]->filename
                );
                IMBRSetStatusMesg(ib, stringa);

                /* Load thumb to viewer. */
                PBarUpdate(&ib->pbar, 0.0, NULL);
                OSWGUISync(False);
                IMBRLoadImageFromThumb(ib, z);
                PBarUpdate(&ib->pbar, 1.0, NULL);
                IMBRDraw(ib, DRAW_AMOUNT_FILE_LIST);

                sprintf(stringa,
                    "Displaying: %s (%ix%i) %i Bytes",
                    ib->thumb[z]->filename,
                    ib->thumb[z]->ori_width,
                    ib->thumb[z]->ori_height,
                    ib->thumb[z]->ori_size
                );
                IMBRSetStatusMesg(ib, stringa);
            }
            else
            {
                IMBRSetStatusMesg(ib, NULL);
            }

            IMBRDraw(ib, DRAW_AMOUNT_COMPLETE);
            IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_NORMAL);
        }


	return(0);
}


/*
 *      Select next thumb procedure.
 */
int IMBRDoNext(img_browser_struct *ib)
{
        int y, z;
        char stringa[PATH_MAX + NAME_MAX + 128]; 
        win_attr_t wattr;


        /* Error checks. */
        if(!IDC() || (ib == NULL)) return(-1);


        OSWGetWindowAttributes(ib->list, &wattr);

        /* If no thumb is selected, go to first. */
        if(((int)ib->total_sel_thumbs <= 0) &&
           ((int)ib->total_thumbs > 0)
        )
        {
            z = 0;
            IMBRSelect(ib, z);
        }
        /* Select next thumb. */
        else if(((int)ib->total_sel_thumbs > 0) &&
                ((int)ib->total_thumbs > 0)
        )
        {
            z = ib->sel_thumb[0] + 1;
            if(z >= (int)ib->total_thumbs)
            {
                ib->list_sb.y_win_pos = 0;
                z = 0;
            }
            IMBRUnselectAll(ib);
            IMBRSelect(ib, z);
        }
        else
        {
            return(-2);
        }

        /* Scroll if near edge. */
        y = (z - (z % (int)ib->thumbs_per_row)) /
            (int)ib->thumbs_per_row;
        y *= (int)ib->list_row_spacing;

        if(y < ib->list_sb.y_win_pos)
            ib->list_sb.y_win_pos -= (int)wattr.height / 2;
        if(y > ((int)ib->list_sb.y_win_pos + ((int)wattr.height / 2)))
            ib->list_sb.y_win_pos += (int)wattr.height / 2;

        /* Sanitize scroll bars. */
        y = ((((int)ib->total_thumbs / (int)ib->thumbs_per_row)
            + 2) * ib->list_row_spacing) - (int)wattr.height;
        if(ib->list_sb.y_win_pos > y)
            ib->list_sb.y_win_pos = y;
        if(ib->list_sb.y_win_pos < 0)
            ib->list_sb.y_win_pos = 0;


        /* Load newly selected thumb to viewer. */
        if(ib->thumb[z] != NULL)
        {
            IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_WORKING);

            if(ISEXTIMAGE(ib->thumb[z]->filename))
            {
                sprintf(stringa, "Loading: %s...",
                    ib->thumb[z]->filename
                );
                IMBRSetStatusMesg(ib, stringa);

                /* Load thumb to viewer. */
                PBarUpdate(&ib->pbar, 0.0, NULL);
                OSWGUISync(False);
                IMBRLoadImageFromThumb(ib, z);
                PBarUpdate(&ib->pbar, 1.0, NULL);
                IMBRDraw(ib, DRAW_AMOUNT_FILE_LIST);

                sprintf(stringa,
                    "Displaying: %s (%ix%i) %i Bytes",
                    ib->thumb[z]->filename,
                    ib->thumb[z]->ori_width,
                    ib->thumb[z]->ori_height,
                    ib->thumb[z]->ori_size
                );
                IMBRSetStatusMesg(ib, stringa);
            }
            else
            {
                IMBRSetStatusMesg(ib, NULL);
            }

            IMBRDraw(ib, DRAW_AMOUNT_COMPLETE);
            IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_NORMAL);
        }


	return(0);
}



/*
 *      View thumb in Image Viewer procedure.
 */
int IMBRDoView(img_browser_struct *ib)
{
        int x, y;
        char *strptr;
	char tmp_path[PATH_MAX + NAME_MAX];


        /* Error checks. */
        if(!IDC() || (ib == NULL)) return(-1);


        /* Go through each selected thumb. */
        for(x = 0; x < (int)ib->total_sel_thumbs; x++)
        {
            y = ib->sel_thumb[x];
            if((y < 0) || (y >= (int)ib->total_thumbs))
                continue;
            if(ib->thumb[y] == NULL) continue;
            if(ib->thumb[y]->filename == NULL) continue;

            /* Get full path as tmp_path. */
            strptr = PrefixPaths(ib->current_dir,
                ib->thumb[y]->filename);
            if(strptr == NULL) continue;

            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

            /* Is directory? */
            if(ISPATHDIR(tmp_path))
            {
                /* Ignore directories. */
            }
            /* Spawn image viewer. */
            else
            {
                /* Allocate and initialize a new internal viewer. */
                DoCreateIV(tmp_path);
            }
        }       /* Go through each selected thumb. */

	return(0);
}



/*
 *	Reload thumbs listing procedure.
 */
int IMBRDoReload(img_browser_struct *ib)
{
        char tmp_path[PATH_MAX + NAME_MAX];


	/* Error checks. */
        if(!IDC() || (ib == NULL)) return(-1);
        if(ib->current_dir == NULL) return(-1);


	/* Change dir will cause reload. */
        strncpy(tmp_path, ib->current_dir, PATH_MAX + NAME_MAX);
        IMBRDoChangeDir(ib, tmp_path);


	return(0);
}



/*
 *	Stop procedure procedure.
 */
int IMBRDoStop(img_browser_struct *ib)
{
	/* Error checks. */
	if(!IDC() || (ib == NULL)) return(-1);


        /* Stop loading? */
        if(!ib->stop_load &&
           ((int)ib->last_fq < ((int)ib->total_fqs - 1))
        )
        {
            ib->stop_load = True;

            /* Update status message. */
            IMBRSetStatusMesg(ib, "Loading: Interupted");

            IMBRDraw(ib, DRAW_AMOUNT_COMPLETE);
        }


	return(0);
}



/*
 *      Rename first selected thumb procedure.
 */
int IMBRDoQuickRename(img_browser_struct *ib)
{
	int x, y, z;
	char stringa[PATH_MAX + NAME_MAX + 128];


        /* Error checks. */
        if(!IDC() || (ib == NULL)) return(-1);

	/* Check global write-protect. */
	if(option.write_protect == 1)
	{
            printdw(&err_dw,
"Cannot rename object:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
            );

	    return(-3);
	}

	/* Must have atleast one thumb selected. */
	if((int)ib->total_sel_thumbs <= 0) return(-2);


        /* Get first selected thumb. */
        z = ib->sel_thumb[0];
        if((z >= 0) && (z < (int)ib->total_thumbs) &&
           (ib->thumbs_per_row != 0)
        )
        {
            /* Reparent and position change values prompt. */
            OSWReparentWindow(ib->cv_prompt.toplevel, ib->list);
            x = IB_DEF_MARGIN + ((z % (int)ib->thumbs_per_row)
                * (int)ib->list_col_spacing);
            y = IB_DEF_MARGIN + (((int)(z - (z %
                (int)ib->thumbs_per_row)) / (int)ib->thumbs_per_row)
                * (int)ib->list_row_spacing) - ib->list_sb.y_win_pos;

            OSWMoveResizeWindow(ib->cv_prompt.toplevel,
                x + 1,
                y + (int)ib->thumb_height,
                ib->thumb_width,
                (int)ib->thumb_label_height + 2
            );

            if(ib->thumb[z] != NULL) 
            {
                if(ib->thumb[z]->filename != NULL)
                {   
                    strncpy(
                        ib->cv_prompt.buf,
                        ib->thumb[z]->filename,
                        ib->cv_prompt.buf_len
                    );
                    PromptMarkBuffer(&ib->cv_prompt, PROMPT_POS_END);
                }
            }

            PromptMap(&ib->cv_prompt);
            ib->cv_prompt_mode = CV_PROMPT_MODE_RENAME;
            ib->cv_prompt.is_in_focus = 1;

            /* Set status message. */
            sprintf(stringa,
                "Changing name for `%s' (press escape key to cancel)", 
                ib->cv_prompt.buf
            );
            IMBRSetStatusMesg(ib, stringa);
        }



        return(0);
}



/*
 *	Move set up procedure.
 */
int IMBRDoMove(img_browser_struct *ib)
{
        int x, y;
	char *strptr;
	char tmp_path[PATH_MAX + NAME_MAX];
         
                
        /* Error checks. */
        if(!IDC() || (ib == NULL)) return(-1);
	if((int)total_op_dialogs <= 0) return(-1);
	if((int)ib->total_sel_thumbs <= 0) return(-1);
        if(ib->current_dir == NULL) return(-1);


        for(x = 0; x < (int)ib->total_sel_thumbs; x++)
        {
            y = ib->sel_thumb[x];
            if((y < 0) || (y >= (int)ib->total_thumbs)) continue;
            if(ib->thumb[y] == NULL) continue;
            if(ib->thumb[y]->filename == NULL) continue;

            strptr = PrefixPaths(ib->current_dir,
                ib->thumb[y]->filename);
            if(strptr == NULL) continue;
            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

            OPDialogAddObject(
                op_dialog[0],
                OP_CODE_MOVE,		/* Operation code. */
                ib->current_dir,	/* Current dir. */
                tmp_path		/* Source (absolute). */
            );
        }

        OPDialogMap(op_dialog[0]);


	return(0);
}



/*
 *	Copy set up procedure.
 */
int IMBRDoCopy(img_browser_struct *ib)
{
        int x, y;
	char *strptr;
        char tmp_path[PATH_MAX + NAME_MAX];
        
        
        /* Error checks. */
        if(!IDC() || (ib == NULL)) return(-1);
        if((int)total_op_dialogs <= 0) return(-1);
        if((int)ib->total_sel_thumbs <= 0) return(-1);
        if(ib->current_dir == NULL) return(-1);
  
  
        for(x = 0; x < (int)ib->total_sel_thumbs; x++)
        {
            y = ib->sel_thumb[x];
            if((y < 0) || (y >= (int)ib->total_thumbs)) continue;
            if(ib->thumb[y] == NULL) continue;
            if(ib->thumb[y]->filename == NULL) continue;
        
            strptr = PrefixPaths(ib->current_dir,
                ib->thumb[y]->filename);
            if(strptr == NULL) continue;
            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';    

            OPDialogAddObject(
                op_dialog[0],
                OP_CODE_COPY,           /* Operation code. */
                ib->current_dir,        /* Current dir. */
                tmp_path                /* Source (absolute). */
            );
        }

        OPDialogMap(op_dialog[0]);


        return(0);
}



/*
 *	Recycle selected thumbs procedure.
 */
int IMBRDoRecycle(img_browser_struct *ib)
{
	int x, y, total_recycled;
	char src_path[PATH_MAX + NAME_MAX];


        /* Error checks. */
        if(!IDC() || (ib == NULL)) return(-1);
        if(ib->current_dir == NULL) return(-1);


	/* ********************************************************* */

        /* Change all cursors. */
        IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_WORKING);

        /* Need to change current working directory. */
        if(chdir(ib->current_dir) == 0)
        {
            getcwd(dname.current_working, PATH_MAX);
            dname.current_working[PATH_MAX - 1] = '\0';
        }

        /* Need to turn off yes_to_all just in case. */
        yes_to_all = False;

        /* Move comfermation dialog to pointer's center. */
        WidgetCenterWindow(comf_dialog.toplevel,
            WidgetCenterWindowToPointer);


        total_recycled = 0;
        for(x = 0; x < (int)ib->total_sel_thumbs; x++)
        {
            y = ib->sel_thumb[x];
            if((y < 0) || (y >= (int)ib->total_thumbs)) continue;
            if(ib->thumb[y] == NULL) continue;
            if(ib->thumb[y]->filename == NULL) continue;

            strncpy(src_path, PrefixPaths(ib->current_dir,
                ib->thumb[y]->filename), PATH_MAX + NAME_MAX);
            src_path[PATH_MAX + NAME_MAX - 1] = '\0';

            EDWPerformFileOp(
                OP_CODE_RECYCLE,
                NULL,		/* Target (none). */
                src_path	/* Source. */
            );
	    total_recycled++;
        }

        /* Need to turn off yes_to_all. */
        yes_to_all = False;

        /* Need to refresh list. */
	if(total_recycled > 0)
	{
            strncpy(src_path, ib->current_dir, PATH_MAX + NAME_MAX);
            IMBRLoad(ib, src_path);
	}

        /* Change all cursors. */
        IMBR_SETALLCURSORS(ib, EDW_CURSORS_CODE_NORMAL);


        return(0);
}



/*
 *	Sets up properties dialog map for selected thumb on ib.
 */
int IMBRDoPropDialog(img_browser_struct *ib)
{
        int x, y;
	char *strptr;
	char tmp_path[PATH_MAX + NAME_MAX];


        /* Error checks. */
        if(!IDC() || (ib == NULL)) return(-1);
        if(ib->current_dir == NULL) return(-1);


        for(x = 0; x < (int)ib->total_sel_thumbs; x++)
        {
            y = ib->sel_thumb[x];
            if((y < 0) || (y >= (int)ib->total_thumbs)) continue;
            if(ib->thumb[y] == NULL) continue;
            if(ib->thumb[y]->filename == NULL) continue;

	    strptr = PrefixPaths(ib->current_dir,
		ib->thumb[y]->filename);
	    if(strptr == NULL) continue;

            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
            tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

            DoCreatePropDialog(tmp_path);
        }


	return(0);
}



/*
 *	Selects thumb n on ib.
 */
int IMBRSelect(img_browser_struct *ib, int n)
{
	int i;

	if(ib == NULL) return(-1);
	if((n < 0) || (n >= (int)ib->total_thumbs)) return(-1);


	/* Sanitize total_sel_thumbs. */
	if(ib->total_sel_thumbs < 0)
	    ib->total_sel_thumbs = 0;


	/* See if we can add to already allocated selection. */
	for(i = 0; i < (int)ib->total_sel_thumbs; i++)
	{
	    if(ib->sel_thumb[i] < 0)
	    {
		ib->sel_thumb[i] = n;
		return(0);
	    }
	}


	/* Allocate new selection. */
	ib->total_sel_thumbs++;
	ib->sel_thumb = (int *)realloc(ib->sel_thumb,
	    ib->total_sel_thumbs * sizeof(int)
	);
	if(ib->sel_thumb == NULL)
	{
	    ib->total_sel_thumbs = 0;
	    return(-1);
	}

	/* Set selection. */
	ib->sel_thumb[ib->total_sel_thumbs - 1] = n;


	return(0);
}



/*
 *	Unselect thumb on image browser.
 */
void IMBRUnselect(img_browser_struct *ib, int n)
{
	int i;

	if(ib == NULL)
	    return;

	if((n < 0) || (n >= ib->total_thumbs))
	    return;

	for(i = 0; i < ib->total_sel_thumbs; i++)
	{
	    if(ib->sel_thumb[i] == n)
		ib->sel_thumb[i] = -1;
	}

	return;
}



/*
 *	Returns 1 if n is selected on ib.
 */
int IMBRIsSelected(img_browser_struct *ib, int n)
{
	static int i;


	/* Error checks. */
        if((ib == NULL) || (n < 0)) return(0);

        /* See if we can add to already allocated selection. */
        for(i = 0; i < (int)ib->total_sel_thumbs; i++)
        {
            if(ib->sel_thumb[i] == n)
                return(1);
        }


	return(0);
}


/*
 *	Unselect all thumbs on ib.
 */
void IMBRUnselectAll(img_browser_struct *ib)
{
        if(ib == NULL) return;

	free(ib->sel_thumb);
	ib->sel_thumb = NULL;

	ib->total_sel_thumbs = 0;   

	return; 
}        



/*
 *	Unselect and delete all thumbs on ib.
 */
void IMBRDeleteAllThumbs(img_browser_struct *ib)
{
	int i;

        if(ib == NULL) return;


	/* Unselect all thumbs. */
	IMBRUnselectAll(ib);

	/* Free all thumbs and their substructures. */
	for(i = 0; i < ib->total_thumbs; i++)
	{
	    if(ib->thumb[i] == NULL) continue;

	    free(ib->thumb[i]->filename);
	    OSWDestroyImage(&ib->thumb[i]->image);

	    free(ib->thumb[i]);
	    ib->thumb[i] = NULL;
	}

	ib->total_thumbs = 0;

	free(ib->thumb);
	ib->thumb = NULL;

	return;
}


#endif /* USE_IMLIB */
