/*
                        Installation Functions

	Functions:

	int InstallDefaultFiles()



*/

#include "endeavour.h"

#include "endeavourrc.h"


int CREATE_PROG_DIR_ASNEEDED()
{
	char tmp_path[PATH_MAX + NAME_MAX];
	int status;


	/* Make sure names are set. */
	if(dname.home == NULL)
	    return(-2);
	if(strlen(dname.home) < 1)
	    return(-2);


	/* Create local program directory. */
	strncpy(tmp_path,
	    PrefixPaths(dname.home, LOCAL_ENDEAVOUR_DIR),
	    PATH_MAX + NAME_MAX
	);
	tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
	SimplifyPath(tmp_path);


	/* Check if local program directory exists. */
	if(access(tmp_path, F_OK) == -1)
	{
	    status = mkdir(
		tmp_path,
		S_IRUSR | S_IWUSR | S_IXUSR |
		S_IRWXG | S_IXGRP |
		S_IROTH | S_IXOTH
	    );
	    if(status != 0)
	    {
		return(-1);
	    }
	}
	/* Path must be a directory. */
	else if(!ISPATHDIR(tmp_path))
	{
	    return(-3);
	}


	return(0);
}



int InstallDefaultFiles()
{
	FILE *fp;
	int len;
        int status;


	/* Create program local directory as needed. */
	if(CREATE_PROG_DIR_ASNEEDED())
	    return(-1);


	len = sizeof(endeavourrc);

	if(fname.rcfile == NULL)
	    return(-2);
	if(strlen(fname.rcfile) < 1)
	    return(-2);


	/* Write default rcfile as needed. */
	if(access(fname.rcfile, F_OK) == -1)
	{
	    fp = fopen(fname.rcfile, "wb");
	    if(fp == NULL)
		return(-1);

	    status = fwrite(endeavourrc, sizeof(char), len, fp);
	    if(status < 0)
	    {
		fclose(fp); fp = NULL;
		return(-1);
	    }

	    fclose(fp); fp = NULL;
	}


	return(0);
}
