/*
                        Internal Viewer Management


	Functions:

	int IntViewerIsAllocated(int viewer_num)
	int IntViewerIsInit(int viewer_num)

	int IntViewerAllocate()
	int IntViewerReclaimMemory()
	int IntViewerManageAll(event_t event)
	void IntViewerDelete(int viewer_num)
	void IntViewerDeleteAll()

	---

	Note that the internal viewer uses the viewer widget
	for each viewer instance, to make things easier.

*/


#include "endeavour.h"



int IntViewerIsAllocated(int viewer_num)
{
	if( (pri_fviewer == NULL) ||
            (viewer_num < 0) ||
            (viewer_num >= (int)total_pri_fviewers)
	)
	{
	    return(0);
	}
	else if(pri_fviewer[viewer_num] == NULL)
	{
	    return(0);
	}
	else
	{
	    return(1);
	}
}



int IntViewerIsInit(int viewer_num)
{
        if(!IntViewerIsAllocated(viewer_num))
        {
            return(0);
        }
        else if(pri_fviewer[viewer_num]->toplevel == 0)
        {
            return(0);
        }
        else
        {
            return(1);
        }
}



int IntViewerAllocate()
{
        int i;
        int new_viewer_num;


        /* Sanitize global total_pri_fviewers. */
        if((int)total_pri_fviewers < 0)
            total_pri_fviewers = 0;


        /* Check for available entry. */
        for(i = 0; i < (int)total_pri_fviewers; i++)
        {
            if(pri_fviewer[i] == NULL)
                break;
        }
        if(i < (int)total_pri_fviewers)
        {
            /* Found an unallocated pointer. */
            new_viewer_num = i;

            /* Allocate a new viewer structure for that pointer. */
            pri_fviewer[new_viewer_num] = (file_viewer_struct *)calloc(1,
                sizeof(file_viewer_struct));
            if(pri_fviewer[new_viewer_num] == NULL)
            {
                return(-1);
            }

            return(new_viewer_num);
        }


        /* ************************************************************** */

        /* Calculate new viewer number and increment total. */
        new_viewer_num = total_pri_fviewers;
	total_pri_fviewers++;


        /* Allocate more viewer window pointers. */
        pri_fviewer = (file_viewer_struct **)realloc(
            pri_fviewer, total_pri_fviewers * sizeof(file_viewer_struct *)
        );
        if(pri_fviewer == NULL)
        {
            total_pri_fviewers = 0;
            return(-1);
        }

        /* Allocate a new viewer structure. */
        pri_fviewer[new_viewer_num] = (file_viewer_struct *)calloc(1,
            sizeof(file_viewer_struct));
        if(pri_fviewer[new_viewer_num] == NULL)
        {
            total_pri_fviewers--;
            return(-1);
        }


        return(new_viewer_num);
}



int IntViewerReclaimMemory()
{
        int i, highest_viewer_num;


        /* Skip if nothing is allocated. */
        if((pri_fviewer == NULL) ||
           ((int)total_pri_fviewers <= 0)  
        )
            return(-1);

        /* Reclaim memory. */
        highest_viewer_num = -1;
        for(i = 0; i < (int)total_pri_fviewers; i++)
        {
            if(pri_fviewer[i] == NULL)
                continue;

            highest_viewer_num = i;  
        }
        if(highest_viewer_num < ((int)total_pri_fviewers - 1))
        {
            /* Adjust global variable total_pri_fviewers. */
            total_pri_fviewers = highest_viewer_num + 1;
             
            /* Reallocate or free pointers. */
            if(total_pri_fviewers <= 0)
            {
                /* total_pri_fviewers has dropped to 0, free all pointers. */
                free(pri_fviewer);
                pri_fviewer = NULL;
                total_pri_fviewers = 0;
            }
            else
            {
                pri_fviewer = (file_viewer_struct **)realloc(
                    pri_fviewer,
                    (int)total_pri_fviewers * sizeof(file_viewer_struct *)
                );
                if(pri_fviewer == NULL)
                {  
                    total_pri_fviewers = 0;
                }
            }
        }

        return(0);
}




int IntViewerManageAll(event_t *event)
{
	static int i, events_handled;


        /* Error checks. */
        if( !IDC() ||
            (event == NULL)
        )
            return(0);


        /* ********************************************************** */
	events_handled = 0;

	for(i = 0; i < (int)total_pri_fviewers; i++)
        {
            /* Skip unallocated ones. */
            if(pri_fviewer[i] == NULL)
                continue;

	    /* ClientMessage must be handled here. */
            if((event->type == ClientMessage) &&
               OSWIsEventDestroyWindow(pri_fviewer[i]->toplevel, event)
            )
            {
		/* Delete viewer and reclaim memory. */
                IntViewerDelete(i);
                IntViewerReclaimMemory();

		OSWPurgeAllEvents();

                events_handled++;
                break;
            }

	    /* Send event to viewer widget management for handling. */
            events_handled += ViewerManage(pri_fviewer[i], event);


	    /*   If viewer unmapped while it was being managed, we should
	     *   just delete it.
	     */
	    if(pri_fviewer[i]->map_state == 0)
	    {
                IntViewerDelete(i);
                IntViewerReclaimMemory();     /* Reclaim mem after each delete. */
	    }
        }


        return(events_handled);
}




void IntViewerDelete(int viewer_num)
{
	if(!IntViewerIsAllocated(viewer_num))
	    return;


	/* Have widget management deallocated substructures. */
	ViewerDestroy(pri_fviewer[viewer_num]);


	/* Free viewer structure itself. */
	free(pri_fviewer[viewer_num]);
	pri_fviewer[viewer_num] = NULL;

	/* Do not adjust global total (reclaim should be called after this). */


	return;
}



void IntViewerDeleteAll()
{
	int i;


	for(i = 0; i < (int)total_pri_fviewers; i++)
	{
	    IntViewerDelete(i);
	}
	free(pri_fviewer);
	pri_fviewer = NULL;

	total_pri_fviewers = 0;


	return;
}
