/*
                               Main Functions

	Functions:

	int ScrollBarNotifyHandler(scroll_bar_struct *sb)   

	int DoCreateEDW(char *path)
        int DoCreateRB(char *path)
	int DoCreateV(char *path)
        int DoCreateIV(char *path)
        int DoCreateIB(char *path)
	int DoCreatePropDialog(char *path)

	void UnfocusAllTopLvlWins()
	void ResetGlobalTimming()
	int SanityCheck()
	void FreeGlobalMemory()
	void ManagePadlockDraw()
	int ManageEvents()

	int main(int argc, char *argv[])

	---




 */

#include "endeavour.h"

#define MIN(a,b)        ((a) < (b) ? (a) : (b))
#define MAX(a,b)        ((a) > (b) ? (a) : (b))


int runlevel;
long cur_millitime;

endeavour_options_struct option;
userinfo_struct userinfo;

dname_struct dname;
fname_struct fname;

command_struct command;
next_struct next;

timmer_struct timmer[MAX_TIMMERS];

font_struct font;
endeavour_cursor_struct endeavour_cursor;
endeavour_image_struct endeavour_image;
bool_t yes_to_all;

proc_hist_struct **proc_hist;
int total_proc_hists;

endeavour_win_struct **endeavour_win;
int total_endeavour_wins;

file_viewer_struct **pri_fviewer;
int total_pri_fviewers;

dialog_win_struct err_dw;

op_dialog_struct **op_dialog;
int total_op_dialogs; 

comfermation_dialog_struct comf_dialog;

prop_dialog_struct **prop_dialog;
int total_prop_dialogs;

#ifdef USE_IMLIB
img_viewer_struct **img_viewer;
int total_img_viewers;
#endif	/* USE_IMLIB */

#ifdef USE_IMLIB
img_browser_struct **img_browser;
int total_img_browsers;
#endif /* USE_IMLIB */

devrec_struct **devrec;
int total_devrecs;

drag_record_struct drag_record[1];



recbin_win_struct **recbin_win;
int total_recbin_wins;

file_ass_struct **file_ass;
int total_file_asses;   

id_record_struct **id_record;
int total_id_records;




/*
 *	Handles the widget scroll bar repeat notification.
 *
 *	Redraws lists being controlled by that scroll bar.
 */
int ScrollBarNotifyHandler(scroll_bar_struct *sb)
{
	int i;

	if(sb == NULL)
	    return(-1);


	/* Search through endeavour browsers. */
	for(i = 0; i < (int)total_endeavour_wins; i++)
	{
	    if(endeavour_win[i] == NULL) continue;

	    if(&endeavour_win[i]->dir_win_sb == sb)
	    {
		EDWDraw(endeavour_win[i], DRAW_AMOUNT_DIR_LIST);
		break;
	    }
	    if(&endeavour_win[i]->file_win_sb == sb)  
            {
                EDWDraw(endeavour_win[i], DRAW_AMOUNT_FILE_LIST);
                break;
            }
	}

        /* Search through recycle bins. */
        for(i = 0; i < (int)total_recbin_wins; i++)
        {
            if(recbin_win[i] == NULL) continue;
            
            if(&recbin_win[i]->list_sb == sb)  
            {
                RecBinDraw(recbin_win[i], DRAW_AMOUNT_LISTS);
                break;
            }
        }

#ifdef USE_IMLIB
        /* Search through image browsers. */
        for(i = 0; i < (int)total_img_browsers; i++)
        {
            if(img_browser[i] == NULL) continue;
	    if(img_browser[i]->map_state < 1) continue;
             
            if(&img_browser[i]->list_sb == sb)
            {
                IMBRDraw(img_browser[i], DRAW_AMOUNT_DIR_LIST);
                break;
            }
        }
#endif /* USE_IMLIB */

	return(0);
}



/*
 *	Procedure to create an Endeavour Browser.
 */
int DoCreateEDW(char *path)
{
	int i;

	if(!IDC())
	    return(-1);


        i = EDWAllocate();
        if(EDWIsAllocated(i))
        {
            EDWInit(endeavour_win[i], 0, 0, 0, 0,
                (path == NULL) ? dname.current_working : path
            );
            EDWMap(endeavour_win[i]);
        }

	return(0);
}



/*
 *	Procedure to create an Recycle Bin.
 */
int DoCreateRB(char *path)
{
        if(!IDC())
	    return(-1);


	if((int)total_recbin_wins > 0)
	{
            RecBinMap(recbin_win[0]);
	}

	return(0);
}



/*
 *      Procedure to create an Internal (not image) Viewer.
 */
int DoCreateV(char *path)
{
        int i;
        
        if(!IDC()) return(-1);
        
        i = IntViewerAllocate();
        if(IntViewerIsAllocated(i))
        {
            ViewerInit(pri_fviewer[i], 0, 0, 0, 0);
   
            ViewerLoadFile(pri_fviewer[i],
                (path == NULL) ? dname.current_working : path
            );
            ViewerMap(pri_fviewer[i]);
        }

        return(0);
}


#ifdef USE_IMLIB
/*
 *	Procedure to create an Internal Image Viewer.
 */
int DoCreateIV(char *path)
{
        int i;

        if(!IDC()) return(-1);

        i = ImgViewerAllocate();
        if(ImgViewerIsAllocated(i)) 
        {
            ImgViewerInit(img_viewer[i]);

            ImgViewerLoadFile(img_viewer[i],
                (path == NULL) ? dname.current_working : path
	    );
            ImgViewerMap(img_viewer[i]);
        }

	return(0);
}


/*
 *      Procedure to create an Image Browser.
 */
int DoCreateIB(char *path)
{
        int i;

        if(!IDC()) return(-1);

        i = IMBRAllocate();
        if(IMBRIsAllocated(i))
        {
            IMBRInit(img_browser[i]);

            IMBRLoad(img_browser[i],
		(path == NULL) ? dname.current_working : path
	    );
            IMBRMap(img_browser[i]);
        }   

	return(0);
}
#endif /* USE_IMLIB */


/*
 *	Procedure to create an Prop Dialog.
 */
int DoCreatePropDialog(char *path)
{
        int i;
        
        if(!IDC()) return(-1);
 
        i = PropDialogAllocate();
	if(PropDialogIsAllocated(i))
	{
	    PropDialogInit(prop_dialog[i]);
	    PropDialogLoad(prop_dialog[i],
		(path == NULL) ? dname.current_working : path
	    );
	    PropDialogMap(prop_dialog[i]);
	}

	return(0);
}


/*
 *	Sets all toplevel windows (created by this program) to be
 *	out of focus.
 */
void UnfocusAllTopLvlWins()
{
	int i;


	/* Endeavour windows. */
	for(i = 0; i < (int)total_endeavour_wins; i++)
	{
	    if(endeavour_win[i] == NULL) continue;

	    endeavour_win[i]->is_in_focus = 0;
	}

        /* Internal viewers. */
        for(i = 0; i < (int)total_pri_fviewers; i++)
        {
            if(pri_fviewer[i] == NULL) continue;
            
            pri_fviewer[i]->is_in_focus = 0;
        }

        /* Error dialog. */
	err_dw.is_in_focus = 0;

	/* Operations dialog. */
        for(i = 0; i < (int)total_op_dialogs; i++)
        {
            if(op_dialog[i] == NULL) continue;
   
            op_dialog[i]->is_in_focus = 0;
        }

	/* Comfermation dialog. */
	comf_dialog.is_in_focus = 0;

        /* Properties dialog. */
        for(i = 0; i < (int)total_prop_dialogs; i++)
        {
            if(prop_dialog[i] == NULL) continue;

            prop_dialog[i]->is_in_focus = 0;
        }

#ifdef USE_IMLIB
        /* Internal image viewers. */
        for(i = 0; i < (int)total_img_viewers; i++)
        {
            if(img_viewer[i] == NULL) continue;

            img_viewer[i]->is_in_focus = 0;
        }
#endif /* USE_IMLIB */

#ifdef USE_IMLIB
        /* Image browsers. */
        for(i = 0; i < (int)total_img_browsers; i++)
        {
            if(img_browser[i] == NULL) continue;
            
            img_browser[i]->is_in_focus = 0;
        }
#endif /* USE_IMLIB */


	return;
}



/*
 *	Resets all global timmers.
 */
void ResetGlobalTimming()
{
	/* Next event scheduals. */
	next.rd_padlock_glow = 0;
	next.sanity = 0;

	/* Timmers. */
	TimmerResetAll();


	return;
}



/*
 *	Makes sure events, timmings, etc are sane.
 */
int SanityCheck()
{
	/* Is it time to check sanity? */
        if(next.sanity > cur_millitime)
            return(0);


	/* Toplevel focus check. */


        /* When all primary windows are gone, switch to runlevel 1. */
        if(((int)total_endeavour_wins <= 0)
           &&
           (((int)total_recbin_wins > 0) ? (recbin_win[0]->map_state == 0) : 1)
	   &&
           ((int)total_pri_fviewers <= 0)
#ifdef USE_IMLIB
           &&
           ((int)total_img_viewers <= 0)
           &&
           ((int)total_img_browsers <= 0)
#endif /* USE_IMLIB */
        )
            runlevel = 1;



	/* Schedual next sanity check. */
	next.sanity = cur_millitime + 1500;


	return(0);
}



/*
 *	Frees all globally allocated resources including GUI resources.
 */
void FreeGlobalMemory()
{
	/* Free GUI dependant allocated memory. */
	if(IDC())
	{
	    /* Property dialogs. */
	    PropDialogDeleteAll();

            /* All internal viewers. */
            IntViewerDeleteAll();

#ifdef USE_IMLIB
	    /* Image viewers. */
	    ImgViewerDeleteAll();
#endif /* USE_IMLIB */

#ifdef USE_IMLIB
            /* Image browsers. */
            IMBRDeleteAll();
#endif /* USE_IMLIB */

            /* All endeavour windows. */
            EDWDeleteAll();

	    /* Comfermation dialog. */
	    ComfDialogDestroy();

	    /* Error dialog. */
	    DialogWinDestroy(&err_dw);

	    /* Operations dialog. */
	    if(total_op_dialogs > 0)
            {
                OPDialogDestroy(op_dialog[0]);
                free(op_dialog[0]);
                free(op_dialog); 
                op_dialog = NULL;

                total_op_dialogs = 0;
            }

	    /* Recycle bin. */
	    if(total_recbin_wins > 0)
	    {
	        RecBinDestroy(recbin_win[0]);
	        free(recbin_win[0]);
	        free(recbin_win);
	        recbin_win = NULL;

	        total_recbin_wins = 0;
            }

	    /* Endeavour allocated pointer cursors. */
	    DestroyAllCursors();

	    /* Endeavour images. */
	    FreeEndeavourImages();

#ifdef USE_IMLIB
	    /* Shutdown imlib. */
	    IMLIBOSWShutdown();
#endif /* USE_IMLIB */


	    /* Fonts. */
	    OSWUnloadFont(&font.std);
	    OSWUnloadFont(&font.bold);


	    OSWKBAutoRepeatOn();
	}

        /* File associations. */
        FAssDeleteAll();

	/* Device records (for mounting). */
	DevRecDeleteAllRecords();

	/* Remove tempory output spool files. */
	SFRemoveAll();

	/* Command lines. */
	free(command.move_file);
	command.move_file = NULL;
        free(command.move_dir);
        command.move_dir = NULL;

        free(command.copy_file);
        command.copy_file = NULL;
        free(command.copy_dir);
        command.copy_dir = NULL;

        free(command.rename_file);
        command.rename_file = NULL;
        free(command.rename_dir);
        command.rename_dir = NULL;

        free(command.chmod_file);
        command.chmod_file = NULL;
        free(command.chmod_dir);
        command.chmod_dir = NULL;

        free(command.chown_file);
        command.chown_file = NULL;
        free(command.chown_dir);
        command.chown_dir = NULL;

        free(command.touch_file);
        command.touch_file = NULL;
        free(command.touch_dir);
        command.touch_dir = NULL;

	free(command.text_editor);
	command.text_editor = NULL;

	free(command.df);
	command.df = NULL;

	/* Process history. */
	ProcHistDeleteAll();

	/* Owner and group ID records. */
	IDDeleteAll();


	return;
}



/*
 *	Redraws padlock animation on all windows that show it.
 */
void ManagePadlockDraw()
{
	int i;
	double gam;


	if(!IDC() ||
           (option.write_protect == 1) ||
           (next.rd_padlock_glow > cur_millitime) ||
           !option.animated_padlock
	)
	    return;

	if((endeavour_image.padlock_tmp == NULL) ||
           (endeavour_image.padlock0 == NULL)
	)
	    return;

        if((endeavour_image.padlock_tmp->data == NULL) ||
           (endeavour_image.padlock0->data == NULL)
        )
	    return;


	/* ***************************************************************** */
	/* Modify padlock_tmp. */

	/* Copy padlock0 buffer to padlock_tmp. */
	/* 8 bits. */
	if(osw_gui[0].depth == 8)
	{
            memcpy(
                endeavour_image.padlock_tmp->data,
		endeavour_image.padlock0->data,
                MIN((int)endeavour_image.padlock_tmp->width *
                    (int)endeavour_image.padlock_tmp->height *
                    BYTES_PER_PIXEL8,
                    (int)endeavour_image.padlock0->width *
                    (int)endeavour_image.padlock0->height *
                    BYTES_PER_PIXEL8
                )
            );
	}
	/* 16 bits. */
        else if(osw_gui[0].depth == 16)
        {
            memcpy(  
                endeavour_image.padlock_tmp->data,
                endeavour_image.padlock0->data,   
                MIN((int)endeavour_image.padlock_tmp->width * 
                    (int)endeavour_image.padlock_tmp->height *
                    BYTES_PER_PIXEL16,
                    (int)endeavour_image.padlock0->width * 
                    (int)endeavour_image.padlock0->height *
                    BYTES_PER_PIXEL16
                )
            );
        }
	/* 24 or 32 bits. */
        else if((osw_gui[0].depth == 24) ||
                (osw_gui[0].depth == 32)
	)
        {
            memcpy(  
                endeavour_image.padlock_tmp->data,
                endeavour_image.padlock0->data,   
                MIN((int)endeavour_image.padlock_tmp->width *
                    (int)endeavour_image.padlock_tmp->height *
                    BYTES_PER_PIXEL32,
                    (int)endeavour_image.padlock0->width *
                    (int)endeavour_image.padlock0->height *
                    BYTES_PER_PIXEL32
                )
            );
        }
	else
	{
	    /* Attempt to best guess by unsupported depth. */
            memcpy(  
                endeavour_image.padlock_tmp->data,
                endeavour_image.padlock0->data,   
                MIN((int)endeavour_image.padlock_tmp->width *
                    (int)endeavour_image.padlock_tmp->height *
                    (int)(osw_gui[0].depth >> 3),
                    (int)endeavour_image.padlock0->width *
                    (int)endeavour_image.padlock0->height *
                    (int)(osw_gui[0].depth >> 3)
                )
            );
	}


	/* *************************************************************** */

	/* Calculate color levels. */
        if(timmer[T_LGLOW_CODE].max != 0)
	    gam = (double)timmer[T_LGLOW_CODE].val /
                  (double)timmer[T_LGLOW_CODE].max;
        else
            gam = 1.0;
        if(gam > 0.5)
            gam = ((1 - gam) * 0.5) + 0.6;
        else
            gam = (gam * 0.5) + 0.6;

        WidgetAdjustImageGamma(
            endeavour_image.padlock_tmp, gam * 1.25, gam, gam   
        );


	/* **************************************************************** */

	/* Redraw on endeavour windows. */
	for(i = 0; i < (int)total_endeavour_wins; i++)
	{
	    if(endeavour_win[i] == NULL) continue;
            if(endeavour_win[i]->map_state < 1) continue;

	    EDWDraw(endeavour_win[i], DRAW_AMOUNT_STATUS);
	}

        /* Redraw on recycle bins. */
        for(i = 0; i < (int)total_recbin_wins; i++)
        {
            if(recbin_win[i] == NULL) continue;
            if(recbin_win[i]->map_state < 1) continue;

            RecBinDraw(recbin_win[i], DRAW_AMOUNT_STATUS);
        }

#ifdef USE_IMLIB
        /* Redraw on image browsers. */
        for(i = 0; i < (int)total_img_browsers; i++)
        {
            if(img_browser[i] == NULL) continue;
            if(img_browser[i]->map_state < 1) continue;

            IMBRDraw(img_browser[i], DRAW_AMOUNT_STATUS);
        }
#endif /* USE_IMLIB */ 

	/* Schedual next padlock glow redraw. */
	next.rd_padlock_glow = cur_millitime + timmer[T_LGLOW_CODE].i;


	return;
}



int ManageEvents()
{
	int events_handled = 0;
	event_t event;
        static long last_motionnotify;
        static long last_configurenotify;
	long start_time;


        if(!IDC())
            return(-1);

	start_time = MilliTime();


        /* ******************************************************* */
        /* Manage events. */

        while(OSWEventsPending() > 0)
        {
	    /* Break if management brings sync out too much. */
/*
	    if(MilliTime() > (start_time + 1000))
	    {
		OSWPurgeAllEvents();
		break;
	    }
*/

	    /* Fetch event. */
	    events_handled = 0;
            OSWWaitNextEvent(&event);

            /* ******************************************************* */

            /* Too soon for MotionNotify? */
            if(event.type == MotionNotify)
            {
                /* Sanitize last_motionnotify. */
                if(last_motionnotify > MilliTime())
                    last_motionnotify = 0;

                /* Too soon? */
                if((last_motionnotify + DEF_MOTIONNOTIFY_INT) > MilliTime())
                    continue;

                /* Update last_motionnotify time. */
                last_motionnotify = MilliTime();
            }


            /* Too soon for ConfigureNotify? */
            if(event.type == ConfigureNotify)
            {
                /* Sanitize last_configurenotify. */  
                if(last_configurenotify > MilliTime())
                    last_configurenotify = 0;
                
                /* Too soon? */
                if((last_configurenotify + DEF_CONFIGURENOTIFY_INT) > MilliTime())
                    continue;
 
                /* Update last_motionnotify time. */
                last_configurenotify = MilliTime();
            }


	    /* ******************************************************** */
	    /* Manage widgets' system. */
            WidgetManage(&event);


            /* ******************************************************** */
            /* Manage client created widgets. */

	    /* Manage endeavour windows. */
	    if(events_handled == 0)
	        events_handled += EDWManageAll(&event);

	    /* Manage internal viewer. */
            if(events_handled == 0)
                events_handled += IntViewerManageAll(&event);

#ifdef USE_IMLIB
            /* Manage image browsers. */
            if(events_handled == 0)
                events_handled += IMBRManageAll(&event);
#endif /* USE_IMLIB */

#ifdef USE_IMLIB
	    /* Manage image viewer. */
            if(events_handled == 0)
                events_handled += ImgViewerManageAll(&event);
#endif /* USE_IMLIB */

	    /* Manage properties dialog. */
            if(events_handled == 0)
                events_handled += PropDialogManageAll(&event);

	    /* Operations dialog. */
            if((events_handled == 0) && (total_op_dialogs > 0))
                events_handled += OPDialogManage(op_dialog[0], &event);

            /* Manage comfermation dialog. */
            if(events_handled == 0)
                events_handled += ComfDialogManage(&event);

            /* Manage error dialog. */
            if(events_handled == 0)
                events_handled += DialogWinManage(&err_dw, &event);

	    /* Recycle bin. */
	    if((events_handled == 0) && (total_recbin_wins > 0))
                events_handled += RecBinManage(recbin_win[0], &event);

            /* Manage drag (after all other window managements). */
            DragManage(&event);


	    /* ********************************************************** */

            /* Clean events if it was a client message. */
/*
            if(event.type == ClientMessage)
                OSWPurgeAllEvents();
 */
	}


	return(events_handled);
}



int main(int argc, char *argv[])
{
	int x, y, z;
	long t;
        int status;
	int num_edws_start;
	char *strptr;
	char stringa[1024];
	char startup_dir[PATH_MAX + NAME_MAX];
	struct stat stat_buf;

	bool_t user_set_visual;

	char warn_no_endeavour_dir = 0;
	char warn_bad_rcfile = 0;
/*	char warn_recbin_big = 0; */
	char warn_created_default_files = 0;
	char warn_bad_startup_dir = 0;
	char force_black_and_white = 0;

	char start_recycle_bin = 0;
	char start_image_browser = 0;

        /* Reset variables. */
	x = 0; y = 0; z = 0;
	status = 0;
	num_edws_start = 1;
	user_set_visual = False;

	userinfo.pid = getpid();

        runlevel = 1;
        cur_millitime = MilliTime();

	/* Options. */
	option.write_protect = (DEF_WRITE_PROTECT) ? 1 : 0;
	option.comferm_pointer_motion_copy =
	    (DEF_COMF_POINTER_MOTION_COPY) ? 1 : 0;
        option.comferm_pointer_motion_move =
	    (DEF_COMF_POINTER_MOTION_MOVE) ? 1 : 0;
	option.comferm_overwrites = (DEF_COMF_OVERWRITE) ? 1 : 0;
        option.comferm_recycle = (DEF_COMF_RECYCLE) ? 1 : 0;
	option.comferm_purge = (DEF_COMF_PURGE) ? 1 : 0;
	option.comferm_purge_all = (DEF_COMF_PURGE_ALL) ? 1 : 0;
	option.comferm_unmount_onexit = (DEF_COMF_UNMOUNT_ONEXIT) ? 1 : 0;

	option.show_prog_output = 1;	/* As needed. */
	option.show_expandable_dirs = 1;
	option.animated_padlock = 1;

	option.file_exe_max_burst = DEF_EXECUTION_PURST_MAX;
	option.mount_timeout = DEF_MOUNT_TIMEOUT;

	option.recbin_items_warn = 50;
	option.recbin_size_warn = 2000000;

	/* User information. */
	userinfo.uid = getuid();
	userinfo.euid = geteuid();
        userinfo.gid = getgid();
        userinfo.egid = getegid();
	/* user's name and group are fetched below. */

	/* Directory names. */
        getcwd(dname.current_working, PATH_MAX);   
        dname.current_working[PATH_MAX - 1] = '\0';

        strptr = getenv("HOME");
	if(strptr == NULL)
	{
	    fprintf(stderr,
		"%s: Cannot fetch value for the HOME enviroment.\n",
		argv[0]
	    );
	    return(-1);
	}
        strncpy(dname.home, strptr, PATH_MAX);
        dname.home[PATH_MAX - 1] = '\0';

	strncpy(dname.prog_home, PrefixPaths(dname.home, LOCAL_ENDEAVOUR_DIR),
	    PATH_MAX);
	dname.prog_home[PATH_MAX - 1] = '\0';

	strptr = PrefixPaths(dname.home, RECYCLE_LOCAL_DIR);
        strncpy(
	    dname.recycle_bin,
	    (strptr == NULL) ? "/" : strptr,
	    PATH_MAX
	);
	dname.recycle_bin[PATH_MAX - 1] = '\0';

	/* File names. */
	strptr = PrefixPaths(dname.home, LOCAL_ENDEAVOUR_DIR);
	strncpy(fname.rcfile, (strptr == NULL) ? "/" : strptr,
	    PATH_MAX + NAME_MAX
	);
	strptr = PrefixPaths(fname.rcfile, DEF_RCFILE);
	strncpy(fname.rcfile, (strptr == NULL) ? "/" : strptr,
	    PATH_MAX + NAME_MAX
        );

	/* Recycle bin file name is set farther below. */

	/* Program output spool file. */
	strptr = PrefixPaths(dname.prog_home, DEF_PROG_OUTPUT);
	sprintf(fname.prog_output, "%s%i", strptr, userinfo.pid);
	fname.prog_output[PATH_MAX + NAME_MAX - 1] = '\0';
        strncpy(
            fname.prog_output_ori,
            (strptr == NULL) ? "/prog_output" : strptr,
            PATH_MAX + NAME_MAX
        );
        fname.prog_output_ori[PATH_MAX + NAME_MAX - 1] = '\0';

	/* List output spool file. */
	strptr = PrefixPaths(dname.prog_home, DEF_LIST_OUTPUT);
	sprintf(fname.list_output, "%s%i", strptr, userinfo.pid);
	fname.list_output[PATH_MAX + NAME_MAX - 1] = '\0';
        strncpy(
            fname.list_output_ori,
            (strptr == NULL) ? "/list_output" : strptr,
            PATH_MAX + NAME_MAX
        );
        fname.list_output_ori[PATH_MAX + NAME_MAX - 1] = '\0';

	/* Com pipe. */
        strptr = PrefixPaths(dname.prog_home, DEF_ENDEAVOUR_COMPIPE);
        strncpy(
	    fname.com_pipe,
	    (strptr == NULL) ? "/endeavour_pipe" : strptr,
	    PATH_MAX + NAME_MAX
	);
        fname.com_pipe[PATH_MAX + NAME_MAX - 1] = '\0';
        strncpy(
            fname.com_pipe_ori,
            (strptr == NULL) ? "/endeavour_pipe" : strptr,
            PATH_MAX + NAME_MAX 
        );
        fname.com_pipe_ori[PATH_MAX + NAME_MAX - 1] = '\0';


	/* Commands. */
        command.move_file = StringCopyAlloc(SYSTEM_CMD_MOVE_FILE);
        command.move_dir = StringCopyAlloc(SYSTEM_CMD_MOVE_DIR);

        command.copy_file = StringCopyAlloc(SYSTEM_CMD_COPY_FILE);
        command.copy_dir = StringCopyAlloc(SYSTEM_CMD_COPY_DIR);

        command.rename_file = StringCopyAlloc(SYSTEM_CMD_RENAME_FILE);
        command.rename_dir = StringCopyAlloc(SYSTEM_CMD_RENAME_DIR);

        command.chmod_file = StringCopyAlloc(SYSTEM_CMD_CHMOD_FILE);
        command.chmod_dir = StringCopyAlloc(SYSTEM_CMD_CHMOD_DIR);
        
        command.chown_file = StringCopyAlloc(SYSTEM_CMD_CHOWN_FILE);
        command.chown_dir = StringCopyAlloc(SYSTEM_CMD_CHOWN_DIR);

        command.touch_file = StringCopyAlloc(SYSTEM_CMD_TOUCH_FILE);
        command.touch_dir = StringCopyAlloc(SYSTEM_CMD_TOUCH_DIR);

	command.text_editor = StringCopyAlloc(SYSTEM_CMD_TEXT_EDITOR);

	command.df = StringCopyAlloc(SYSTEM_CMD_DEVICE_FIND);

	yes_to_all = False;


	next.rd_padlock_glow = 0;
	next.sanity = 0;

	TimmerInitAll();

        file_ass = NULL;
        total_file_asses = 0;

	endeavour_win = NULL;
	total_endeavour_wins = 0;

	pri_fviewer = NULL;
	total_pri_fviewers = 0;

	op_dialog = NULL;
	total_op_dialogs = 0;

	recbin_win = NULL;
	total_recbin_wins = 0;

	prop_dialog = NULL;
	total_prop_dialogs = 0;

#ifdef USE_IMLIB
	img_viewer = NULL;
	total_img_viewers = 0;
#endif

#ifdef USE_IMLIB
        img_browser = NULL;
        total_img_browsers = 0;
#endif  

	devrec = NULL;
	total_devrecs = 0;

	id_record = NULL;
	total_id_records = 0;

	proc_hist = NULL;
	total_proc_hists = 0;

	/* Initialize time zone. */
	tzset();

	/* Fetch owner and group id's. */
	IDLoadFromFile(NULL);

	/* Update user name and group. */
	strptr = IDGetNameByUID(userinfo.uid);
	strncpy(userinfo.name,
	    ((strptr == NULL) ? "(null)" : strptr),
	    NAME_MAX
	);
	userinfo.name[NAME_MAX - 1] = '\0';

        strptr = IDGetGroupByGID(userinfo.gid);
        strncpy(userinfo.group,
            ((strptr == NULL) ? "(null)" : strptr),
            NAME_MAX
        );
        userinfo.group[NAME_MAX - 1] = '\0';

	/* Recycle bin file name. */
        strptr = PrefixPaths(dname.recycle_bin, userinfo.name);
        strncpy(
            fname.recycle_bin,
            (strptr == NULL) ? "/someones_recyclebin" : strptr,
            PATH_MAX + NAME_MAX
        );
        fname.recycle_bin[PATH_MAX + NAME_MAX - 1] = '\0';
        strncat(fname.recycle_bin, RECYCLE_BIN_FILE_NAME,
            PATH_MAX + NAME_MAX - strlen(fname.recycle_bin) - 1);
        fname.recycle_bin[PATH_MAX + NAME_MAX - 1] = '\0';


	/* ********************************************************** */
	/* Parse arguments. */

	for(x = 1; x < argc; x++)
	{
	    /* Help. */
	    if(strpfx(argv[x], "--h") ||
               strpfx(argv[x], "-h") ||
               strpfx(argv[x], "-?") ||
               strpfx(argv[x], "/?")
	    )
	    {
		printf(PROG_HELP_MESG);

                FreeGlobalMemory();
                return(0);
	    }
	    /* Version. */
            else if(strpfx(argv[x], "--ver") ||
                    strpfx(argv[x], "-ver")
	    )
            {
                printf("%s Version %s\n%s\n",
		    PROG_NAME, PROG_VERSION,
                    PROG_COPYRIGHT
		);

                FreeGlobalMemory();
                return(0);
            }
	    /* Number of browsers to start up with. */
	    else if(!strcmp(argv[x], "--count") ||
                    !strcmp(argv[x], "-count") ||
                    !strcmp(argv[x], "-coun") ||
                    !strcmp(argv[x], "-cou") ||
                    !strcmp(argv[x], "-co") ||
                    !strcmp(argv[x], "-cnt") ||
                    !strcmp(argv[x], "-c")
	    )
	    {
		x++;
		if(x >= argc)
		{
		    fprintf(stderr,
			"Error: Missing argument for parameter `%s'\n",
		        argv[x - 1]
		    );
		    FreeGlobalMemory();
		    return(1);
		}

		num_edws_start = atoi(argv[x]);

		if(num_edws_start < 1)
		{
                    FreeGlobalMemory();
                    return(0);
		}
	    }
	    /* Startup in recycle bin. */
	    else if(!strcmp(argv[x], "--recyclebin") ||
                    !strcmp(argv[x], "-recyclebin") ||
                    !strcmp(argv[x], "--recbin") ||
                    !strcmp(argv[x], "-recbin")
            )
            {
                start_recycle_bin = 1;
            }
            /* Startup in image browser. */
            else if(!strcmp(argv[x], "--imagebrowser") ||
                    !strcmp(argv[x], "-imagebrowser") ||
                    !strcmp(argv[x], "--imgbrowser") ||
                    !strcmp(argv[x], "-imgbrowser") ||
                    !strcmp(argv[x], "--imgbrowse") ||
                    !strcmp(argv[x], "-imgbrowse")
            )
            {
                start_image_browser = 1;
            }
	    /* Free form argument is startup directory. */
	    else if(argv[x][0] == '/')
	    {
		if(chdir(argv[x]) == -1)
		{
		    strncpy(startup_dir, argv[x], PATH_MAX + NAME_MAX);
		    startup_dir[PATH_MAX + NAME_MAX - 1] = '\0';
		    warn_bad_startup_dir = 1;
		}
		else
		{
		    getcwd(dname.current_working, PATH_MAX);
                    dname.current_working[PATH_MAX - 1] = '\0';
		}
	    }
	}



        /* ********************************************************** */
        /* Select signals. */
        signal(SIGINT, HandleSignal);
        signal(SIGTERM, HandleSignal);
	signal(SIGUSR1, HandleSignal);
        signal(SIGSEGV, HandleSignal);


	/* ********************************************************** */
	/* Fetch parameters from restart configuration. */
	strptr = PrefixPaths(dname.home, LOCAL_ENDEAVOUR_DIR);
	if(!stat(strptr, &stat_buf))
	{
	    /* Local program directory exists. */

	    if(!stat(fname.rcfile, &stat_buf))
	    {
	        if(RCLoadFromFile(fname.rcfile) < 0)
	            warn_bad_rcfile = 1;
	    }
	    else
	    {
		if(InstallDefaultFiles())
		    warn_bad_rcfile = 1;
		else
		{
                    if(RCLoadFromFile(fname.rcfile) < 0)
                        warn_bad_rcfile = 1;
		    else
		        warn_created_default_files = 1;
		}
	    }
	}
	else
	{
	    /* Local program directory does NOT exist. */

	    if(InstallDefaultFiles())
	    {
                warn_no_endeavour_dir = 1;
	    }
	    else
	    {
                if(RCLoadFromFile(fname.rcfile) < 0)
                    warn_bad_rcfile = 1;
		else
		    warn_created_default_files = 1;
	    }
	}

        /* ********************************************************** */
        /* Set up GUI and widget globals. */
        status = OSWGUIConnect(argc, argv);
        if(status)
	{
	    FreeGlobalMemory();
            return(1);
	}

#ifdef USE_IMLIB
	if(IMLIBOSWInit())
	{
	    FreeGlobalMemory();
	    OSWGUIDisconnect();
            return(1);
	}
#endif

	/* Widget globals. */
        status = WidgetInitGlobals(argc, argv);
        if(status)
	{
	    FreeGlobalMemory();
            OSWGUIDisconnect();
            return(1);
	}

	/* Need force black and white? */
	if(((int)osw_gui[0].depth < 8) ||
            (force_black_and_white == 1)
        )
	    widget_global.force_mono = 1;

	/* Set scroll bar repeat notify handler. */
	SBarSetNotifyFunction(ScrollBarNotifyHandler);

	/* Fonts. */
        OSWLoadFont(&font.std, "7x14");
        OSWLoadFont(&font.bold, "7x14");
        OSWSetFont(font.std);


	/* Load all Endeavour cursors. */
	if(LoadAllCursors())
	{
	    FreeGlobalMemory();
            WidgetDestroyGlobals();
            OSWGUIDisconnect();
	    return(1);
	}

	/* ********************************************************** */
	if(LoadEndeavourImages())
	{
	    FreeGlobalMemory();
            WidgetDestroyGlobals();
            OSWGUIDisconnect();
	    return(1);
	}

        /* ********************************************************** */
        /* Set up window and widgets. */

	/* Error dialog. */
	if(
	    DialogWinInit(
                &err_dw,
                osw_gui[0].root_win,
                0,
                0,
                endeavour_image.error_icon
	    )
        )
	{
            FreeGlobalMemory();
            WidgetDestroyGlobals();
            OSWGUIDisconnect();
            return(1);
	}

	/* Operations dialog. */
	total_op_dialogs = 1;
	op_dialog = (op_dialog_struct **)realloc(op_dialog,
            total_op_dialogs * sizeof(op_dialog_struct *)
        );
        if(op_dialog == NULL)
        {
            FreeGlobalMemory();
            WidgetDestroyGlobals();
            OSWGUIDisconnect();
            return(1);
        }
        op_dialog[0] = (op_dialog_struct *)calloc(1,
            sizeof(op_dialog_struct));
        if(op_dialog[0] == NULL)
        { 
            FreeGlobalMemory();
            WidgetDestroyGlobals();
            OSWGUIDisconnect();
            return(1);
        }
        if(OPDialogInit(op_dialog[0]))
        {
            FreeGlobalMemory();
            WidgetDestroyGlobals();
            OSWGUIDisconnect();
            return(1);
        }

	/* Recycle bin. */
	total_recbin_wins = 1;
	recbin_win = (recbin_win_struct **)realloc(recbin_win,
	    total_recbin_wins * sizeof(recbin_win_struct *)
	);
	if(recbin_win == NULL)
	{
            FreeGlobalMemory();
            WidgetDestroyGlobals();
            OSWGUIDisconnect();
            return(1);
	}
	recbin_win[0] = (recbin_win_struct *)calloc(1,
	    sizeof(recbin_win_struct));
	if(recbin_win[0] == NULL)
	{   
            FreeGlobalMemory();
            WidgetDestroyGlobals();
            OSWGUIDisconnect();
            return(1);
        }
	if(
	    RecBinInit(
		recbin_win[0],
		0, 0,
		0, 0,
		fname.recycle_bin
	    )
	)
	{
            FreeGlobalMemory();
            WidgetDestroyGlobals();
            OSWGUIDisconnect();
            return(1);
        }

	/* Comfermation dialog. */
	if(ComfDialogInit())
	{
            FreeGlobalMemory();
            WidgetDestroyGlobals();
            OSWGUIDisconnect();
            return(1);
	}


	/* ********************************************************** */

        /* Spool files startup warnings. */
        SFDoStartupWarn();


	/* ********************************************************** */

	/* Refresh mounted device listing. */
        DevRecRefreshList();


	/* Create recycle bin on startup? */
	if(start_recycle_bin)
	{
	    if((int)total_recbin_wins > 0)
		RecBinMap(recbin_win[0]);
	}
	/* Create image browser on startup? */
	else if(start_image_browser)
	{
#ifdef USE_IMLIB

	    x = IMBRAllocate();
	    if(IMBRIsAllocated(x))
	    {
		if(IMBRInit(img_browser[x]))
		{
		    FreeGlobalMemory();
                    WidgetDestroyGlobals();
                    OSWGUIDisconnect();

		    return(1);
		}
		IMBRLoad(img_browser[x], dname.current_working);
		IMBRMap(img_browser[x]);
	    }

#else /* USE_IMLIB */
/* If system is not compiled with USE_IMLIB support, map endeavour and show error. */

            x = EDWAllocate();
            if(EDWIsAllocated(x))
            {
                status = EDWInit(endeavour_win[y], 0, 0, 0, 0,
                    dname.current_working);
                if(status)
                {
                    FreeGlobalMemory();
                    WidgetDestroyGlobals();
                    OSWGUIDisconnect();

                    return(1);
                }
                EDWMap(endeavour_win[x]);
            }

	    /* Show error. */
            sprintf(stringa,
"Cannot start up in image browser:\n\n\
    USE_IMLIB flag was not used at compile time.\n\n"
            );
            printdw(&err_dw, stringa);

#endif /* USE_IMLIB */
	}
	/* Default to create endeavour window(s) on startup. */
	else
	{
	    for(x = 0; x < num_edws_start; x++)
	    {
	        y = EDWAllocate();
	        if(EDWIsAllocated(y))
	        {
	            status = EDWInit(endeavour_win[y], 0, 0, 0, 0,
		        dname.current_working);
	            if(status)
		    {
		        FreeGlobalMemory();
                        WidgetDestroyGlobals();
                        OSWGUIDisconnect();

                        return(1);
		    }
		    EDWMap(endeavour_win[y]);
		}
	    }
	}

	/* Startup warning messages? */
	if(warn_no_endeavour_dir)
	{
	    sprintf(stringa,
"Unable to find local program directory:\n\n\
    %s\n\n\
This directory is used to store all configuration\n\
and data files that this program uses.\n\
\n\
You may need to reinstall these files and directories.\n",
		LOCAL_ENDEAVOUR_DIR
	    );

	    printdw(&err_dw, stringa);
	}
	else if(warn_bad_rcfile)
	{
            sprintf(stringa,
"Error in configuration file:\n\n\
    %s\n\n\
Review the contents of the configuration file to\n\
make sure all parameters and their values are stated\n\
correctly.\n",
                fname.rcfile
            );
                    
            printdw(&err_dw, stringa);
        }
	else if(warn_created_default_files)
	{
            sprintf(stringa,
"Local program directory and configuration files\n\
created.   You may want to edit the configuration files\n\
to customize them to your system.  These files are\n\
located in `%s'.\n",
                PrefixPaths(dname.home, LOCAL_ENDEAVOUR_DIR)
            );

            printdw(&err_dw, stringa);
	}
	else if(warn_bad_startup_dir)
	{
	    sprintf(stringa,
"Unable to find the directory named:\n\n\
    %s\n\n\
Check the directory name and try again.\n",
		startup_dir
            );

            printdw(&err_dw, stringa);
	}


        /* *********************************************************** */
        /* Purge all events and enter runlevel 2. */
/*        OSWPurgeAllEvents(); */
        runlevel = 2;

        /* Main loop. */
        while(runlevel >= 2)
        {
            /* Update timming. */
	    t = MilliTime();

	    if(t < cur_millitime)
	    {
		/* Need to reset timmings. */
		ResetGlobalTimming();
	    }
	    cur_millitime = t;

	    /* Manage timmers. */
	    TimmerManageAll();

#ifdef USE_IMLIB
	    /* Image viewers repeat callback. */
	    ImgViewerRepeatCB();
#endif

#ifdef USE_IMLIB
            /* Image browsers repeat callback. */
	    IMBRRepeatCB();
#endif

	    /* Call widget manager. */
	    WidgetManage(NULL);

            /* Manage events. */
            ManageEvents();


	    /* Redraws. */
	    ManagePadlockDraw();


	    /* Sanity check. */
	    SanityCheck();


	    /* Sleep. */
            usleep(16000);
        }



        /* *********************************************************** */
/*
printf("On normal shutdown: edw: %i   viewers: %i\n",
	total_endeavour_wins, total_pri_fviewers);
*/

	/*   If there any unmountable devices still mounted,
	 *   comferm with user if they want it unmounted.
	 */
	DevRecDoComfUnmountAll();

	/*
	 *   Check spool files, only on normal shutdown.
	 */
	SFDoShutdownWarn();


	/* Free all global memory. */
	FreeGlobalMemory();


        /* Destroy widget globals and connection to GUI. */
        WidgetDestroyGlobals();
        OSWGUIDisconnect();


	return(0);
}
