/*
                        Operations Dialog Management

	Functions:

	void OPDialogChangeCursors(op_dialog_struct *od, int code)

	int OPDialogAddObject(
	        op_dialog_struct *od,
	        int op_code,
	        char *current_dir,
	        char *object
	)  
	void OPDialogClearObjects(op_dialog_struct *od)

        int OPDialogApply(op_dialog_struct *od)
	void OPDialogCancel(op_dialog_struct *od)

	int OPDialogInit(op_dialog_struct *od)
	int OPDialogResize(op_dialog_struct *od)
	int OPDialogDraw(op_dialog_struct *od)
	int OPDialogManage(op_dialog_struct *od, event_t *event)
	void OPDialogMap(op_dialog_struct *od)
        void OPDialogUnmap(op_dialog_struct *od)
	void OPDialogDestroy(op_dialog_struct *od)






*/

#include "endeavour.h"



#define OD_WIDTH	512
#define OD_HEIGHT	200

#define STATUS_HEIGHT	30

#define OD_PBAR_X	360
#define OD_PBAR_Y	8
#define OD_PBAR_WIDTH	140
#define OD_PBAR_HEIGHT	14


/*
 *	Change all cursors on operations dialog.
 */
void OPDialogChangeCursors(op_dialog_struct *od, int code)
{
	cursor_t std_cursor = osw_gui[0].std_cursor;


	if(od == NULL)
	    return;


        switch(code)
        {
          case EDW_CURSORS_CODE_WORKING:
            OSWSetWindowCursor(od->toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(od->status_win, endeavour_cursor.working);
            OSWSetWindowCursor(od->move_btn.toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(od->copy_btn.toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(od->rename_btn.toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(od->ok_btn.toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(od->cancel_btn.toplevel, endeavour_cursor.working);
	    break;

          default: /* EDW_CURSORS_CODE_NORMAL */
            OSWSetWindowCursor(od->toplevel, std_cursor);
            OSWSetWindowCursor(od->status_win, std_cursor);
            OSWSetWindowCursor(od->move_btn.toplevel, std_cursor);
            OSWSetWindowCursor(od->copy_btn.toplevel, std_cursor);
            OSWSetWindowCursor(od->rename_btn.toplevel, std_cursor);
            OSWSetWindowCursor(od->ok_btn.toplevel, std_cursor);
            OSWSetWindowCursor(od->cancel_btn.toplevel, std_cursor);
            break;
	}

	OSWGUISync(False);


	return;
}



/*
 *	Frees all objects on od.
 */
void OPDialogClearObjects(op_dialog_struct *od)
{
	int i;

	if(od == NULL) return;

	for(i = 0; i < od->total_objects; i++)
	{
	    free(od->object[i]);
	}
	free(od->object);
	od->object = NULL;

	od->total_objects = 0;

	od->last_object = -1;

	return;
}



/*
 *	Add object to od list to be operated on.
 */
int OPDialogAddObject(
	op_dialog_struct *od,
	int op_code,
	char *current_dir,
	char *object
)
{
	int i;


	/* Error checks. */
        if(!IDC() ||
           (od == NULL) ||
	   (op_code == OP_CODE_NONE) ||
           (object == NULL)
	)
            return(-1);

	if(!ISPATHABSOLUTE(object))
	{
	    fprintf(stderr, "OPDialogAddObject(): Path must be absolute.\n");
	    return(-3);
	}


	/* Set up values. */
	od->op_code = op_code;

	/* Clear status message. */
	free(od->status_mesg);
	od->status_mesg = NULL;

	/* Set current dir. */
	free(od->current_dir);
	od->current_dir = StringCopyAlloc(current_dir);


	/* Set title message. */
	switch(od->op_code)
	{
	  case OP_CODE_COPY:
	    OSWSetWindowTitle(od->toplevel, "Copy");
	    od->src_prompt.is_in_focus = 0;
	    od->tar_prompt.is_in_focus = 1;
	    break;

          case OP_CODE_MOVE:
            OSWSetWindowTitle(od->toplevel, "Move");
            od->src_prompt.is_in_focus = 0;
            od->tar_prompt.is_in_focus = 1;
            break;

          case OP_CODE_RENAME:
            OSWSetWindowTitle(od->toplevel, "Rename");
            od->src_prompt.is_in_focus = 0;
            od->tar_prompt.is_in_focus = 1;
            break;

	  default:
            OSWSetWindowTitle(od->toplevel, "Operation Dialog");
            od->src_prompt.is_in_focus = 0;
            od->tar_prompt.is_in_focus = 1;
            break;
	}


	/* Select entire buffer of the target prompt. */
	PromptMarkBuffer(&od->tar_prompt, PROMPT_POS_END);


	/* ******************************************************** */
	/* Add object to list. */
	if(od->total_objects < 0)
	    od->total_objects = 0;

	i = od->total_objects;
	od->total_objects++;

	/* Reallocate object pointers. */
	od->object = (char **)realloc(od->object,
	    od->total_objects * sizeof(char *)
	);
	if(od->object == NULL)
	{
	    od->total_objects = 0;
	    return(-1);
	}

	/* Allocate new object. */
	od->object[i] = StringCopyAlloc(object);
	if(od->object[i] == NULL)
	{
	    od->total_objects--;
	    return(-1);
	}


	return(0);
}



/*
 *	Performs a cancel procedure on od, clearing its objects
 *	list and unmapping it.
 */
void OPDialogCancel(op_dialog_struct *od)
{
        /* Error checks. */
        if(!IDC() || (od == NULL))
            return;

        /* Clear objects list on od. */
        OPDialogClearObjects(od);
  
        /* Unmap the operations dialog. */
        OPDialogUnmap(od);
  
        
        return;
}



/*
 *	Applies changes to disk on od.
 */
int OPDialogApply(op_dialog_struct *od)             
{
	int i, n, x;
	int op_count, op_total;
	char *strptr;

	char tar_path[PATH_MAX + NAME_MAX];
	char tmp_path[PATH_MAX + NAME_MAX];
	char stringa[512];

	char src_was_dir = 0;


	/* Error and missing information check. */
	if((od == NULL) ||
           (od->total_objects <= 0) ||
           (od->object == NULL) ||
	   (od->op_code == OP_CODE_NONE) ||
	   (od->current_dir == NULL) ||
           (od->tar_prompt.buf == NULL)
	)
	    return(-1);

	if(strlen(od->tar_prompt.buf) < 1)
	{
	    OPDialogClearObjects(od);
            return(-1);
	}

	/* ************************************************************ */

	/* Turn off global yes_to_all. */
	yes_to_all = False;


	/* **************************************************************** */
	/* Check write protect. */
        if(option.write_protect)
        {
	    switch(od->op_code)
	    {
	      case OP_CODE_COPY:
                printdw(&err_dw,
"Cannot copy object(s):\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                );
		break;

	      case OP_CODE_MOVE:
                printdw(&err_dw,
"Cannot move object(s):\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                );
		break;

              case OP_CODE_RENAME:
                printdw(&err_dw,
"Cannot rename object(s):\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                );
                break;

              default:
                printdw(&err_dw,
"Cannot operate on object(s):\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                );
                break;
	    }

            OPDialogClearObjects(od);
            return(-3);
        }

        /* **************************************************************** */
        /* Set current directory from location of endeavour browser. */
        if(chdir(od->current_dir) == 0)
        {
            getcwd(dname.current_working, PATH_MAX);
            dname.current_working[PATH_MAX - 1] = '\0';
        }
        else
        {
            printdw(&err_dw,
"Cannot access current directory.\n\n\
The directory you are in may not permit read or write\n\
operations or it no longer exists.  Try canceling\n\
and redoing the procedure that led up to this error.\n\n"
            );

            OPDialogClearObjects(od);
            return(-3);
        }

	/* Get tar_path, this will be the destination for all object(s). */
	strncpy(tar_path, od->tar_prompt.buf, PATH_MAX + NAME_MAX);
	tar_path[PATH_MAX + NAME_MAX - 1] = '\0';
        strptr = PrefixPaths(dname.current_working, tar_path);  
        if(strptr == NULL) return(-1);

        strncpy(tar_path, strptr, PATH_MAX + NAME_MAX);
        tar_path[PATH_MAX + NAME_MAX - 1] = '\0';
        SimplifyPath(tar_path);

        /*   Check for multiple coppies or moves, then if
         *   target path is a dir.
         */
        op_count = 0;
        op_total = od->total_objects;
        if(op_total > 1)
        {
            if(!ISPATHDIR(tar_path))
            {
                sprintf(stringa,
"Invalid operation, cannot copy or move multiple objects\n\
to a non-directory object:\n\n\
    %s\n\n",
                    tmp_path
                );
                printdw(&err_dw, stringa);

	        OPDialogClearObjects(od);
                return(-4);
            }
        }

        /* Move comfermation dialog to pointer's center. */
        WidgetCenterWindow(comf_dialog.toplevel,
            WidgetCenterWindowToPointer);

	/* Change op dialog cursors to working. */
        OPDialogChangeCursors(od, EDW_CURSORS_CODE_WORKING);

        /* Process objects. */
	for(i = 0; i < od->total_objects; i++)
	{
	    if(od->object[i] == NULL) continue;

/* Check if source object and target destination are the same. */
if(!strcmp(tar_path, od->object[i]))
{
	printf("%s: Object identical with target, skipping.\n", od->object[i]);
	continue;
}
            /* Check if target exists and if we can overwrite. */
            if(DESTEXISTS(tar_path, od->object[i]))
            {
                if(option.comferm_overwrites &&
                   !yes_to_all
                )
                {
                    x = ComfDialogMapQueryOverWrite(
                        tar_path,
                        od->object[i],
                        od->op_code
                    );
                    if(x == NO_PBTN_NUM)
                    {
                        continue;
		    }
                    else if(x == ALL_PBTN_NUM)
                    {
                        /* Switch yes to all on. */
                        yes_to_all = True;
                    }
                }
                /* Remove existing target file first. */
                if(!ISPATHDIR(tar_path))   
                    unlink(tar_path);   
            }

	    switch(od->op_code)
	    {
	      case OP_CODE_COPY:
                /* Update status message. */
                strncpy(tmp_path, od->object[i], PATH_MAX + NAME_MAX);
                tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
	        StringShortenFL(tmp_path, 36);
                free(od->status_mesg);
                sprintf(stringa, "Copying: %s", tmp_path);
                od->status_mesg = StringCopyAlloc(stringa);
                OPDialogDraw(od);
	        OSWGUISync(False);

		if(ISPATHDIR(od->object[i]))
		{
		    src_was_dir = 1;
	            FileOpCopyDir(tar_path, od->object[i]);
		}
		else
		{
		    FileOpCopyFile(tar_path, od->object[i]);
		}
		break;

	      case OP_CODE_MOVE:
                /* Update status message. */
                strncpy(tmp_path, od->object[i], PATH_MAX + NAME_MAX);
                tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
                StringShortenFL(tmp_path, 36);
                free(od->status_mesg);
                sprintf(stringa, "Moving: %s", tmp_path);
                od->status_mesg = StringCopyAlloc(stringa);
                OPDialogDraw(od);
                OSWGUISync(False);

                if(ISPATHDIR(od->object[i]))
		{
		    src_was_dir = 1;
                    FileOpMoveDir(tar_path, od->object[i]);

                    /*   Delete any source entries that exist in listed
                     *   memory in Endeavour Browsers.
		     */
                    for(n = 0; n < total_endeavour_wins; n++)
                    {
                        if(endeavour_win[n] == NULL) continue;

                        x = EDWMatchExistingDirFromPath(endeavour_win[n],
                            od->object[i]);
                        EDWColapseDir(endeavour_win[n], x);
                        EDWDeleteDirItemPath(endeavour_win[n],
			    od->object[i]);
                    }
		}
                else
		{
                    FileOpMoveFile(tar_path, od->object[i]);
		}
                break;

	      default:
		sprintf(stringa,
"Internal non-critical error:\n\n\
    Unsupported operation code %i\n\n\
Recieved in operations dialog apply function.\n\n",
		    od->op_code
		);
		printdw(&err_dw, stringa);

		break;
	    }

	    /* Update progress bar. */
	    op_count++;
	    PBarUpdate(&od->pbar, (double)op_count /
                (double)op_total, NULL);
	    OSWGUISync(False);

            usleep(16000);
	}


	/* ************************************************************ */

	/* Turn off global yes to all. */
	yes_to_all = False;

	/* Update status message. */
        free(od->status_mesg);
        od->status_mesg = StringCopyAlloc("Operation completed.");
        OPDialogDraw(od);
	OSWGUISync(False);


	/* Need to refresh all listings. */
	for(i = 0; i < (int)total_endeavour_wins; i++)
	{
	    if(!EDWIsInit(i)) continue;

            EDWRefreshAllLists(endeavour_win[i]);
	    if(ISPATHDIR(tar_path) && src_was_dir)
	    {
                n = EDWMatchExistingDirFromPath(endeavour_win[i], tar_path);
                EDWColapseDir(endeavour_win[i], n);
                EDWExpandDir(endeavour_win[i], n);
	    }

            if(endeavour_win[i]->map_state == 1)
	        EDWDraw(endeavour_win[i], DRAW_AMOUNT_LISTS);
	}
        for(i = 0; i < (int)total_recbin_wins; i++)
        {
            if(recbin_win[i] == NULL) continue;
            if(recbin_win[i]->map_state == 0) continue;

            RECBINREFRESHLISTING(recbin_win[i]);
            RecBinDraw(recbin_win[i], DRAW_AMOUNT_LISTS);
        }
#ifdef USE_IMLIB
        for(i = 0; i < (int)total_img_browsers; i++)
        {
            if(!IMBRIsInit(i)) continue;
            if(img_browser[i]->map_state == 0) continue;
	    if(img_browser[i]->current_dir == NULL) continue;

	    strncpy(tmp_path, img_browser[i]->current_dir, PATH_MAX + NAME_MAX);
	    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

            IMBRLoad(img_browser[i], tmp_path);
        }
#endif /* USE_IMLIB */



	/* Clear all objects on operations dialog. */
	OPDialogClearObjects(od);

        /* Change back all cursors on op_dialog. */
        OPDialogChangeCursors(od, EDW_CURSORS_CODE_NORMAL);



	return(0);
}



/* ************************************************************************
 *
 */
int OPDialogInit(op_dialog_struct *od)
{
        int btn_x, btn_y;
        win_attr_t wattr;
        pixmap_t pixmap;  
	char hotkey_str[PBTN_MAX_HOTKEYS + 1];


        if(!IDC() ||
           (od == NULL)
        )
            return(-1);

        od->map_state = 0;
        od->visibility_state = VisibilityFullyObscured;
        od->x = 0;
        od->y = 0;
        od->is_in_focus = 0;
        od->width = OD_WIDTH;
        od->height = OD_HEIGHT;
        od->disabled = False;

        od->status_mesg = NULL;

        /* ************************************************************* */

        /* Toplevel. */
        if(
            OSWCreateWindow(
                &od->toplevel,
                osw_gui[0].root_win,
                od->x, od->y,
                od->width, od->height
            )
        )
            return(-1);
        OSWSetWindowInput(od->toplevel, KeyPressMask |
            KeyReleaseMask | ButtonPressMask | ButtonReleaseMask |
            ExposureMask | StructureNotifyMask | FocusChangeMask |
            VisibilityChangeMask
        );
        OSWSetWindowBkg(od->toplevel, 0, widget_global.std_bkg_pm);

        WidgetCenterWindow(od->toplevel, WidgetCenterWindowToRoot);
        OSWGetWindowAttributes(od->toplevel, &wattr);
        od->x = wattr.x;
        od->y = wattr.y;
        od->width = wattr.width; 
        od->height = wattr.height;

        if(
            OSWCreatePixmap(
                &od->toplevel_buf,
                wattr.width, wattr.height
            )
        )
            return(-1);


        pixmap = OSWCreatePixmapFromImage(endeavour_image.endeavour_icon);

	/* WM Propeties. */
	OSWSetWindowWMProperties(
	    od->toplevel,
            "Operation Dialog",
            "Operation Dialog",
            (pixmap == 0) ? widget_global.std_icon_pm : pixmap,
            False,
            od->x, od->y,
            od->width, od->height,
            od->width, od->height,
	    WindowFrameStyleFixed,
	    NULL, 0
	);
        pixmap = 0;     /* Do not destroy pixmap. */


        /* Status window. */
        if(
            OSWCreateWindow(
                &od->status_win,
                od->toplevel,
                0, (int)od->height - STATUS_HEIGHT,
                od->width, STATUS_HEIGHT
            )
        )
            return(-1);
        OSWSetWindowInput(od->status_win, ButtonPressMask |
            ButtonReleaseMask | PointerMotionMask |
            ExposureMask
        );
        OSWGetWindowAttributes(od->status_win, &wattr);
        if(
            OSWCreatePixmap(
                &od->status_win_buf,
                wattr.width, wattr.height
            )
        )
            return(-1);


        /* Source prompt. */
        if(
            PromptInit(
                &od->src_prompt,
                od->toplevel, 
                32, 32,
                360, 30,
                PROMPT_STYLE_FLUSHED,
                "From:", 
                PATH_MAX + NAME_MAX,
                0,
                NULL
            )
        )
            return(-1);

        /* Target prompt. */
        if(
            PromptInit(
                &od->tar_prompt,
                od->toplevel,
                32, 60,
                360, 30,
                PROMPT_STYLE_FLUSHED,
                "To:",
                PATH_MAX + NAME_MAX,
                0,
                NULL
            )
        )
            return(-1);


	/* Copy Button. */
	btn_x = 32;
	btn_y = 130;
        if(
            PBtnInit(
                &od->copy_btn,
                od->toplevel,
                btn_x, btn_y,
                70, 28,
                "Copy",
                PBTN_TALIGN_CENTER,
                NULL,
		(void *)&od->copy_btn,
                NULL
            )
        )
            return(-1);
	PBtnSetHotKeys(&od->copy_btn, "\n");


	/* Move button. */
        if(
            PBtnInit(
                &od->move_btn,
                od->toplevel,
                btn_x, btn_y,
                70, 28,
                "Move",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&od->move_btn,
                NULL
            )
        )
            return(-1);
        PBtnSetHotKeys(&od->move_btn, "\n");

        /* Rename button. */
        if(
            PBtnInit(
                &od->rename_btn,
                od->toplevel,
                btn_x, btn_y,
                70, 28,
                "Rename",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&od->rename_btn,
                NULL
            )
        )
            return(-1);
        PBtnSetHotKeys(&od->rename_btn, "\n");

        /* OK button. */
        if(
            PBtnInit(
                &od->ok_btn,
                od->toplevel,
                btn_x, btn_y,
                70, 28,
                "OK",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&od->ok_btn,
                NULL
            )
        )
            return(-1);
        PBtnSetHotKeys(&od->ok_btn, "\n");

        /* Cancel button. */
	btn_x = (int)od->width - 100;
        if(
            PBtnInit(
                &od->cancel_btn,
                od->toplevel,
                btn_x, btn_y,
                70, 28,
                "Cancel",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&od->cancel_btn,
                NULL
            )
        )
            return(-1);
	hotkey_str[0] = 0x1b;	/* Escape. */
	hotkey_str[1] = '\0';
        PBtnSetHotKeys(&od->cancel_btn, hotkey_str);


	/* Progress bar. */
	if(
	    PBarInit(
		&od->pbar,
		od->status_win,
		OD_PBAR_X,
		OD_PBAR_Y,
		OD_PBAR_WIDTH,
		OD_PBAR_HEIGHT,
		0,		/* Start value. */
		0, 1,		/* Min and max. */
		NULL,
		PBAR_COMPLETION_HOLD
	    )
	)
	    return(-1);


	return(0);
}



int OPDialogResize(op_dialog_struct *od)
{
        if(!IDC() ||
           (od == NULL)
        )
            return(-1);


	/* No resizing. */
	return(0);
}



int OPDialogDraw(op_dialog_struct *od)  
{
	int x;
	win_attr_t wattr;
	char stringa[512];


        if(od == NULL)
            return(-1);


	/* Map as needed. */
	if(!od->map_state)
	{
	    OSWMapRaised(od->toplevel);

	    OSWMapWindow(od->status_win);

	    PBarMap(&od->pbar);

	    /* Map/unmap prompt. */
	    PromptUnmap(&od->src_prompt);
	    PromptMap(&od->tar_prompt);		/* Target prompt always mapped. */

	    /* Map/unmap buttons. */
	    PBtnUnmap(&od->copy_btn);
            PBtnUnmap(&od->move_btn);
            PBtnUnmap(&od->rename_btn);
            PBtnUnmap(&od->ok_btn);
            PBtnMap(&od->cancel_btn);
	    switch(od->op_code)
	    {
	      case OP_CODE_COPY:
		PBtnMap(&od->copy_btn);
		break;

              case OP_CODE_MOVE:
                PBtnMap(&od->move_btn);
                break;

              case OP_CODE_RENAME:
                PBtnMap(&od->rename_btn);
                break;

              default:
                PBtnMap(&od->ok_btn);
                break;
	    }

	    od->map_state = 1;
	}


	/* Toplevel. */
	if(1)
	{
	    OSWGetWindowAttributes(od->toplevel, &wattr);

	    WidgetPutImageTile(od->toplevel_buf,
		widget_global.std_bkg_img,
		wattr.width, wattr.height
	    );

	    /* Draw operation label. */
	    if(od->object != NULL)
	    {
		x = od->total_objects;

	        switch(od->op_code)
                {
                  case OP_CODE_COPY:
		    if(x <= 0)
		    {
			sprintf(stringa, "Copy: *nothing*");
		    }
		    else if(x == 1)
		    {
                        sprintf(stringa, "Copy: %s", od->object[0]);
		    }
		    else
		    {
		        sprintf(stringa, "Copy: %i objects", x);
		    }
		    break;

                  case OP_CODE_MOVE:
                    if(x <= 0)
                    {
                        sprintf(stringa, "Move: *nothing*");
                    }
                    else if(x == 1)
                    {
                        sprintf(stringa, "Move: %s", od->object[0]);
                    }
		    else
		    {
                        sprintf(stringa, "Move: %i objects", x);
		    }
                    break;

                  case OP_CODE_RENAME:
		    if(x > 0)
		    {
                        sprintf(stringa, "Rename: %s", od->object[0]);
		    }
                    break;
		}

		OSWSetFgPix(widget_global.normal_text_pix);
		OSWDrawString(od->toplevel_buf,
		    32, 32,
		    stringa
		);
	    }



	    /* Put buffer to window. */
	    OSWPutBufferToWindow(od->toplevel, od->toplevel_buf);
	}


	/* Status win. */
	if(1)
	{
            OSWGetWindowAttributes(od->status_win, &wattr);
 
            WidgetPutImageTile(od->status_win_buf,
                widget_global.std_bkg_img,
                wattr.width, wattr.height
            );


	    /* Draw status message. */
	    if(od->status_mesg != NULL)
	    {
		OSWSetFgPix(widget_global.normal_text_pix);
		OSWDrawString(
		    od->status_win_buf,
		    16, ((int)wattr.height / 2) + 5,
		    od->status_mesg
		);
	    }

	    /* Draw frame. */
	    WidgetFrameButtonPixmap(
		od->status_win_buf,
		False,
		wattr.width, wattr.height,
		widget_global.surface_highlight_pix,
		widget_global.surface_shadow_pix
	    );

            /* Put buffer to window. */
            OSWPutBufferToWindow(od->status_win, od->status_win_buf);
	}

	return(0);
}



int OPDialogManage(op_dialog_struct *od, event_t *event)
{
	int events_handled = 0;


        if((od == NULL) ||
           (event == NULL)
        )
            return(events_handled);

	if(!od->map_state &&
           (event->type != MapNotify)
	)
	    return(events_handled);


	switch(event->type)
	{
          /* ********************************************************* */
          case KeyPress:
            /* Skip KeyPress if not in focus. */  
            if(!od->is_in_focus)
                return(events_handled);

	    break;

          /* ********************************************************* */
          case KeyRelease:
            /* Skip KeyPress if not in focus. */   
            if(!od->is_in_focus)
                return(events_handled);

	    /* Escape key. */
/*
	    if(event->xkey.keycode == osw_keycode.esc)
	    {
                OPDialogCancel(od);
                return(events_handled);
	    }
*/
	    break;


          /* ********************************************************* */
          case ButtonPress:
            /* Change focus as needed. */
            if((event->xany.window == od->toplevel) ||
               (event->xany.window == od->status_win)
            )
                od->is_in_focus = 1;


            break;


          /* ********************************************************* */
          case ButtonRelease:
          
            break;

          /* ********************************************************* */
          case Expose:
            if((event->xany.window == od->toplevel) ||
               (event->xany.window == od->status_win)
            )
                events_handled++;

            break;

          /* ********************************************************* */
          case ConfigureNotify:
            if(event->xany.window == od->toplevel)  
            {
                OPDialogResize(od);

                events_handled++;
                return(events_handled);
            }
            break;

          /* ********************************************************* */
          case FocusIn:
            if(event->xany.window == od->toplevel)
            {
                od->is_in_focus = 1;
                events_handled++;
            }
            else
            {
                od->is_in_focus = 0;
                /* Do not report any event for focus out. */
            }
            return(events_handled);
            break;

          /* ********************************************************* */
          case FocusOut:
            if(event->xany.window == od->toplevel)
            {
                od->is_in_focus = 0;
                events_handled++;
            }
            return(events_handled);
            break;

          /* ********************************************************* */
          case VisibilityNotify:
            if(event->xany.window == od->toplevel)
            {
                od->visibility_state = event->xvisibility.state;

                events_handled++;
                return(events_handled);
            }
            break;
                
          /* ********************************************************* */
          case ClientMessage:
            if(OSWIsEventDestroyWindow(od->toplevel, event))
            {
		/* This will unmap it and free its objects list. */
                OPDialogCancel(od);

                events_handled++;
                return(events_handled);
            }
            break;

          /* ********************************************************* */
          case LeaveNotify:
            break;
	}

        /* Redraw as needed. */
        if(events_handled > 0)   
        {
            OPDialogDraw(od);
            return(events_handled);
        }


        /* ************************************************************ */
        /* Manage widgets. */

	/* Copy button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &od->copy_btn,
                event
            );
            if(((event->type == ButtonRelease) ||
               (event->type == KeyRelease)) &&
               (events_handled > 0)
            )
            {
                OPDialogApply(od);
                OPDialogUnmap(od);

                return(events_handled);
	    }
	}

        /* Move button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &od->move_btn,
                event
            );
            if(((event->type == ButtonRelease) ||
               (event->type == KeyRelease)) &&
               (events_handled > 0)
            )
            {
                OPDialogApply(od);
                OPDialogUnmap(od);

                return(events_handled);
            }
        }

        /* Rename button. */
        if(events_handled == 0)  
        {
            events_handled += PBtnManage(  
                &od->rename_btn,  
                event
            );
            if(((event->type == ButtonRelease) ||
               (event->type == KeyRelease)) &&
               (events_handled > 0)
            )
            {
		OPDialogApply(od);
                OPDialogUnmap(od);
                
                return(events_handled);
            } 
 	}

        /* OK button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &od->ok_btn,
                event
            );
            if(((event->type == ButtonRelease) ||
               (event->type == KeyRelease)) &&
               (events_handled > 0)
            )
            {
                OPDialogApply(od);
                OPDialogUnmap(od);

                return(events_handled);
            }
        }

	/* Cancel button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &od->cancel_btn,
                event
            );
            if(((event->type == ButtonRelease) ||
               (event->type == KeyRelease)) &&
               (events_handled > 0)
            )
            {
		/* This will clear its objects list and unmap it. */
                OPDialogCancel(od);
		return(events_handled);
            }
        }


        /* Prompt widgets (must be handled after buttons). */
        if(events_handled == 0)
        {
            events_handled += PromptManage(
                &od->tar_prompt,
                event
            );
        }
        if(events_handled == 0)
        {
            events_handled += PromptManage(
                &od->src_prompt,
                event
            );
        }

	/* Progress bar. */
	if(events_handled == 0)
        {
            events_handled += PBarManage(
		&od->pbar,
		event
	    );
	}



	return(events_handled);
}



void OPDialogMap(op_dialog_struct *od)
{
	if(!IDC() ||
           (od == NULL)
	)
	    return;


        UnfocusAllTopLvlWins();

	/* Change cursors. */
	OPDialogChangeCursors(od, EDW_CURSORS_CODE_NORMAL);

	/* Need to reset key states. */
	od->copy_btn.state = PBTN_UNARMED;
        od->move_btn.state = PBTN_UNARMED;
        od->rename_btn.state = PBTN_UNARMED;
        od->ok_btn.state = PBTN_UNARMED;
        od->cancel_btn.state = PBTN_UNARMED;

        /* Update progress bar to 0. */
        PBarUpdate(&od->pbar, 0, NULL);

	/* Put source prompt out of focus and target prompt in focus. */
	od->src_prompt.is_in_focus = 0;
	od->tar_prompt.is_in_focus = 1;

	/* Map by drawing. */
        od->map_state = 0;
	OPDialogDraw(od);

	od->map_state = 1;
	od->is_in_focus = 1;
	od->visibility_state = VisibilityUnobscured;

	return;
}



void OPDialogUnmap(op_dialog_struct *od)
{
        if(!IDC() ||   
           (od == NULL)
        )
            return;


	OSWUnmapWindow(od->toplevel);
	od->map_state = 0;
	od->is_in_focus = 0;


	/* Unmapping should reset all values. */
	od->op_code = OP_CODE_NONE;


	return;
}



void OPDialogDestroy(op_dialog_struct *od)
{
        if(od == NULL)
            return;


	/* Reset values. */
	od->op_code = OP_CODE_NONE;

	free(od->status_mesg);
	od->status_mesg = NULL;

	free(od->current_dir);
	od->current_dir = NULL;

	OPDialogClearObjects(od);

	if(IDC())
	{
	    PBarDestroy(&od->pbar);

	    PBtnDestroy(&od->cancel_btn);
            PBtnDestroy(&od->ok_btn);
            PBtnDestroy(&od->rename_btn);
            PBtnDestroy(&od->move_btn);
            PBtnDestroy(&od->copy_btn);

	    PromptDestroy(&od->src_prompt);
	    PromptDestroy(&od->tar_prompt);

	    OSWDestroyWindow(&od->status_win);
	    OSWDestroyPixmap(&od->status_win_buf);

            OSWDestroyWindow(&od->toplevel);
            OSWDestroyPixmap(&od->toplevel_buf);
	}


	return;
}
