/*
                    Process history recording.

	pid_t ProcHistLastPid()
	int ProcHistIsAllocated(int n)
	int ProcHistAdd(
		pid_t pid,
		time_t s_time
	)
	void ProcHistDelete(int n)
	void ProcHistDeleteAll()

	int ProcHistKill(int n)

	---

	Currently we only support one proc_history in the list
	since the rest is not really important.

 */

#include "endeavour.h"



/*
 *	Macro to get the pid of the latest recorded pid.
 */
pid_t ProcHistLastPid()
{

/* Use 0 for now. */
	if(!ProcHistIsAllocated(0))
	{
	    /* 2 is a safe number if we can't find anything else. */
	    return(2);
	}


	return(proc_hist[0]->pid);
}



int ProcHistIsAllocated(int n)
{
	if((n < 0) || (n >= total_proc_hists) ||
           (proc_hist == NULL)
        )
	{
	    return(0);
	}
	else if(proc_hist[n] == NULL)
	{
            return(0);
	}
	else
        {
            return(1);
        }
}



/*
 *	Add an item to the process history.
 *	The item addeded will be marked as running.
 */
int ProcHistAdd(
	pid_t pid,
	time_t s_time
)
{
	int n;


	if(total_proc_hists < 0)
	    total_proc_hists = 0;

/* Only allocating one for now. */
	if(total_proc_hists < 1)
	{
	    n = 0;
	    total_proc_hists = 1;
	    proc_hist = (proc_hist_struct **)realloc(proc_hist,
		total_proc_hists * sizeof(proc_hist_struct *)
	    );
	    if(proc_hist == NULL)
	    {
		total_proc_hists = 0;
		return(-1);
	    }

	    proc_hist[n] = (proc_hist_struct *)calloc(1,
		sizeof(proc_hist_struct));
	    if(proc_hist[n] == NULL)
            { 
                total_proc_hists--;
                return(-1);
            }
	}
	else
	{
	    n = 0;
	}


	/* Set values. */
	proc_hist[n]->is_running = 1;
	proc_hist[n]->pid = pid;
	proc_hist[n]->s_time = s_time;


	return(n);
}



void ProcHistDelete(int n)
{
	if(!ProcHistIsAllocated(n)) return;

	free(proc_hist[n]);
	proc_hist[n] = NULL;


	return;
}



void ProcHistDeleteAll()
{
	int i;

	for(i = 0; i < total_proc_hists; i++)
	{
	    ProcHistDelete(i);
	}
	free(proc_hist);
	proc_hist = NULL;

	total_proc_hists = 0;


	return;
}



/*
 *	Sends a SIGKILL signal to the pid referanced in the
 *	process history item n.
 */
int ProcHistKill(int n)
{
	int status;


	/* Error checks. */
	if(!ProcHistIsAllocated(n)) return(-2);
	if(!proc_hist[n]->is_running) return(-2);

	/* Leave process 1 alone, otherwise major crash. */
	if(proc_hist[n]->pid <= 1) return(-1);


	/*   Do not kill process numbers younger than parent or
         *   parent itself.
	 */
	if(proc_hist[n]->pid <= userinfo.pid)
	    return(-3);

	/* Send kill signal to pid. */
	status = kill(proc_hist[n]->pid, SIGKILL);
	if(status == 0)
	    proc_hist[n]->is_running = 0;


	return(status);
}
