/*
                        Edit properties dialog

	Functions:

	void PD_UNFOCUS_ALL_PROMPTS(prop_dialog_struct *pd)
	int PD_LINK_DEST_EXISTS(prop_dialog_struct *pd, char *path)

	int PropDialogIsAllocated(int dialog_num)
	int PropDialogIsInit(int dialog_num)

	int PropDialogLoad(prop_dialog_struct *pd, char *path)
	int PropDialogApply(prop_dialog_struct *pd)

	int PropDialogReclaimMemory()
	int PropDialogAllocate()
	void PropDialogDelete(int pd_num)
	void PropDialogDeleteAll()
	int PropDialogManageAll(event_t *event)

	int PropDialogInit(prop_dialog_struct *pd)
	int PropDialogResize(prop_dialog_struct *pd)
	int PropDialogDraw(prop_dialog_struct *pd)
	int PropDialogManage(prop_dialog_struct *pd, event_t *event)
	void PropDialogMap(prop_dialog_struct *pd)
	void PropDialogUnmap(prop_dialog_struct *pd)
	void PropDialogDestroy(prop_dialog_struct *pd)

*/


#include "endeavour.h"



#define PD_WIDTH	480
#define PD_HEIGHT	480

#define OUTLINE_NAME_POS        0.10    /* Coefficient. */
#define CHAR_WIDTH		7
#define CHAR_HEIGHT		14


/*
 *	Sets all prompts out of focus on pd.
 */
void PD_UNFOCUS_ALL_PROMPTS(prop_dialog_struct *pd)
{
	if(pd == NULL) return;

        PromptUnmarkBuffer(&pd->name_prompt, PROMPT_POS_CUR);
	pd->name_prompt.is_in_focus = 0;

        PromptUnmarkBuffer(&pd->link_dest_prompt, PROMPT_POS_CUR);
        pd->link_dest_prompt.is_in_focus = 0;

        PromptUnmarkBuffer(&pd->owner_prompt, PROMPT_POS_CUR);
        pd->owner_prompt.is_in_focus = 0;

        PromptUnmarkBuffer(&pd->group_prompt, PROMPT_POS_CUR);
        pd->group_prompt.is_in_focus = 0;


	return;
}



/*
 *	Returns 1 if the link specified in path has a valid
 *	existing final destination.
 */
int PD_LINK_DEST_EXISTS(prop_dialog_struct *pd, char *path)
{
	struct stat stat_buf;
	char *strptr;
/*
	char lpath[PATH_MAX + NAME_MAX];
*/
	char tmp_path[PATH_MAX + NAME_MAX];


	if(path == NULL)
	    return(0);

	if(!ISPATHABSOLUTE(path))
	{
	    if(pd == NULL) return(0);
	    strptr = GetParentDir(pd->ori_info.name);
	    if(strptr == NULL) return(0);

	    strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
	    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

	    strptr = PrefixPaths(tmp_path, path);
	    if(strptr == NULL) return(0);

	    strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
	    SimplifyPath(tmp_path);
	}
	else
	{
	    strncpy(tmp_path, path, PATH_MAX + NAME_MAX);
	}
	tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';


	if(stat(tmp_path, &stat_buf))
	    return(0);
	else
	    return(1);
}


/*
 *	Macro to draw outlines around items.
 */
void DRAW_OUTLINE(
        drawable_t d,
        int x, int y,
        unsigned int width,
        unsigned int height,
        char *name,
        unsigned long fg_pix,
        unsigned long bg_pix,
        unsigned long text_pix
)
{
        int text_start_pos;     /* In pixels. */
        int text_len;           /* In pixels. (inc margin) */


        /* Error checks. */
        if( !IDC() ||
            (d == 0) ||
            (name == NULL)
        )
            return;

        text_start_pos = (int)((double)width * OUTLINE_NAME_POS);
        text_len = (strlen(name) + 2) * CHAR_WIDTH;


        /* Draw shadows. */
        OSWSetFgPix(bg_pix);
        OSWDrawLine(d,                  /* left */
             x,
             y + (int)height,
             x,
             y
        );
        OSWDrawLine(d,                  /* top1 */
             x,
             y,
             x + (int)text_start_pos,
             y
        );
        OSWDrawLine(d,                  /* Top2 */
             x + (int)text_len + (int)text_start_pos,
             y,
             x + (int)width,
             y
        );
        OSWDrawLine(d,                  /* right */
            x + (int)width - 1,
            y + 1,
            x + (int)width - 1,
            y + (int)height - 1
        );
        OSWDrawLine(d,                  /* bottom */
            x + 1,
            y + (int)height - 1,
            x + (int)width - 1,
            y + (int)height - 1
        );

        /* Draw highlights. */
        OSWSetFgPix(fg_pix);
        OSWDrawLine(d,                  /* left */
             x + 1,
             y + (int)height - 2,
             x + 1,
             y + 1
        );
        OSWDrawLine(d,                  /* top1 */
             x + 1,
             y + 1,
             x + (int)text_start_pos,
             y + 1
        );
        OSWDrawLine(d,                  /* Top2 */
             x + (int)text_len + (int)text_start_pos,
             y + 1,
             x + (int)width - 2,
             y + 1
        );
        OSWDrawLine(d,                  /* right */
            x + (int)width,
            y + 1,
            x + (int)width,
            y + (int)height
        );
        OSWDrawLine(d,                  /* bottom */
            x + 1,
            y + (int)height,
            x + (int)width,
            y + (int)height
        );

        /* Draw text. */
        OSWSetFgPix(text_pix);
        OSWDrawString(
            d,
            x + (int)text_start_pos + CHAR_WIDTH,
            y + 6,
            name
        );


        return;
}



/* ******************************************************************
 *
 */

int PropDialogIsAllocated(int dialog_num)
{
	if((prop_dialog == NULL) ||
           (dialog_num < 0) ||
           (dialog_num >= (int)total_prop_dialogs)
	)
	{
	    return(0);
	}
	else if(prop_dialog[dialog_num] == NULL)
	{
	    return(0);
	}
	else
	{
	    return(1);
	}
}



int PropDialogIsInit(int dialog_num)
{
        if(!PropDialogIsAllocated(dialog_num))
        {
            return(0);
        }
	else if(prop_dialog[dialog_num]->toplevel == 0)
	{
	    return(0);
	}
        else
        {
            return(1);
        }
}


/* ************************************************************************
 *
 */

int PropDialogLoad(prop_dialog_struct *pd, char *path)  
{
	int len;
	int status;
	char tmp_path[PATH_MAX + NAME_MAX];
	struct stat stat_buf;
	char *strptr;


	if((pd == NULL) ||
           (path == NULL)
	)
	    return(-1);

	/* Get local stats of object. */
	if(lstat(path, &stat_buf))
	    return(-1);


	/* Set original information. */
	free(pd->ori_info.name);
	pd->ori_info.name = StringCopyAlloc(path);
	SimplifyPath(pd->ori_info.name);

	free(pd->ori_info.link_dest);
        pd->ori_info.link_dest = NULL;

        pd->ori_info.inode = stat_buf.st_ino;
        pd->ori_info.mode = stat_buf.st_mode;
        pd->ori_info.nlink = stat_buf.st_nlink;
        pd->ori_info.uid = stat_buf.st_uid;
        pd->ori_info.gid = stat_buf.st_gid;
        pd->ori_info.rdevice = stat_buf.st_rdev;
        pd->ori_info.size = stat_buf.st_size;
        pd->ori_info.last_accessed = stat_buf.st_atime;
        pd->ori_info.last_modified = stat_buf.st_mtime;
        pd->ori_info.create_time = stat_buf.st_ctime;

	/* If object is a link, get destination. */
	if(S_ISLNK(stat_buf.st_mode))
	{
	    status = readlink(path, tmp_path, PATH_MAX + NAME_MAX);
	    if(status >= (PATH_MAX + NAME_MAX))
		status = PATH_MAX + NAME_MAX - 1;
	    if(status < 0)
		status = 0;

	    /* Need to NULL terminate! */
	    tmp_path[status] = '\0';

	    pd->ori_info.link_dest = StringCopyAlloc(tmp_path);
	}


	/* ************************************************************ */
	/* Set up values in widgets. */

	/* Name. */
	len = pd->name_prompt.buf_len;
	strncpy(pd->name_prompt.buf, pd->ori_info.name, len);
	pd->name_prompt.buf[len - 1] = '\0';
	StripAbsolutePath(pd->name_prompt.buf);
	/* Set name prompt into focus. */
        PromptMarkBuffer(&pd->name_prompt, PROMPT_POS_END);
        pd->name_prompt.is_in_focus = 1;

	/* Link destination. */
	if(S_ISLNK(stat_buf.st_mode) &&
           (pd->ori_info.link_dest != NULL)
	)
        {
	    len = pd->link_dest_prompt.buf_len;
            strncpy(pd->link_dest_prompt.buf, pd->ori_info.link_dest, len);
            pd->link_dest_prompt.buf[len - 1] = '\0';
	}
	else
	{
	    len = pd->link_dest_prompt.buf_len;
	    memset(pd->link_dest_prompt.buf, 0x00, len);
	}

	/* Permissions. */
	if(stat_buf.st_mode & S_ISUID)
	    pd->perm_uid_tb.state = True;
	else
	    pd->perm_uid_tb.state = False;

        if(stat_buf.st_mode & S_ISGID)
            pd->perm_gid_tb.state = True;
        else
            pd->perm_gid_tb.state = False;

        if(stat_buf.st_mode & S_ISVTX)
            pd->perm_sticky_tb.state = True;
        else
            pd->perm_sticky_tb.state = False;



        if(stat_buf.st_mode & S_IRUSR)
            pd->perm_usrr_tb.state = True;
        else
            pd->perm_usrr_tb.state = False;

        if(stat_buf.st_mode & S_IWUSR)
            pd->perm_usrw_tb.state = True;
        else
            pd->perm_usrw_tb.state = False;

        if(stat_buf.st_mode & S_IXUSR)
            pd->perm_usrx_tb.state = True;
        else
            pd->perm_usrx_tb.state = False;


        if(stat_buf.st_mode & S_IRGRP)
            pd->perm_grpr_tb.state = True;
        else
            pd->perm_grpr_tb.state = False;

        if(stat_buf.st_mode & S_IWGRP)
            pd->perm_grpw_tb.state = True; 
        else
            pd->perm_grpw_tb.state = False;

        if(stat_buf.st_mode & S_IXGRP)
            pd->perm_grpx_tb.state = True; 
        else
            pd->perm_grpx_tb.state = False;


        if(stat_buf.st_mode & S_IROTH)
            pd->perm_othr_tb.state = True;
        else
            pd->perm_othr_tb.state = False;

        if(stat_buf.st_mode & S_IWOTH)
            pd->perm_othw_tb.state = True; 
        else
            pd->perm_othw_tb.state = False;

        if(stat_buf.st_mode & S_IXOTH)
            pd->perm_othx_tb.state = True; 
        else
            pd->perm_othx_tb.state = False;


	/* Owner and group. */
	strptr = IDGetNameByUID(stat_buf.st_uid);
	if(strptr != NULL)
	{
            len = pd->owner_prompt.buf_len;
            strncpy(pd->owner_prompt.buf, strptr, len);
            pd->owner_prompt.buf[len - 1] = '\0';
        }
	else
	{
            len = pd->owner_prompt.buf_len;
	    memset(pd->owner_prompt.buf, 0x00, len);
	}

        strptr = IDGetGroupByGID(stat_buf.st_gid);
        if(strptr != NULL)
        {
            len = pd->group_prompt.buf_len;
            strncpy(pd->group_prompt.buf, strptr, len);
            pd->group_prompt.buf[len - 1] = '\0';
        }
        else
        {
            len = pd->group_prompt.buf_len;
            memset(pd->group_prompt.buf, 0x00, len);
        }




	return(0);
}



/*
 *	Applies changes on the pd.
 */
int PropDialogApply(prop_dialog_struct *pd)
{
	int i, len;
	int status;
	umode_t mode;
	char stringa[1024];
	char old_name[PATH_MAX + NAME_MAX];
	char tmp_path[PATH_MAX + NAME_MAX];
	struct stat stat_buf;
	char *strptr;


	if(pd == NULL)
	    return(-1);

	/* Must have a valid original file name */
	if(pd->ori_info.name == NULL)
	    return(-2);

	/* Make sure object still exists. */
	if(lstat(pd->ori_info.name, &stat_buf) == -1)
	{
	    sprintf(stringa,
"Non-existant object:\n\n\
    %s\n\n\
Another thread of execution or another process may have\n\
changed the name or removed the object.\n",
                pd->ori_info.name
            );
            printdw(&err_dw, stringa);

            return(-2);
	}


	/* Master write protect check. */
        if(option.write_protect)
        {
            printdw(&err_dw,
"Cannot apply object property changes:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
            );
            return(-3);
	}


	/* ******************************************************** */

	/* Change name. */
	strncpy(old_name, pd->ori_info.name, PATH_MAX + NAME_MAX);
	old_name[PATH_MAX + NAME_MAX - 1] = '\0';

	strptr = GetParentDir(pd->ori_info.name);
	if(strptr == NULL) return(-1);
	strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
	tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
	strptr = PrefixPaths(tmp_path, pd->name_prompt.buf);
	if(strptr == NULL) return(-1);

	/* Make sure new name does not exist. */
	if(lstat(strptr, &stat_buf) == 0)
	{
	    /* Dose target name differ from original name? */
	    if(strcmp(strptr, pd->ori_info.name))
	    {
		/* Names differ, implying target exists.  Cannot overwrite. */
                printdw(&err_dw,
"Cannot change object name:\n\n\
    Target name exists and differs from original name.\n\n\
You must choose a name that matches the original or\n\
that of an object that does not exist.\n"
                );

		/* Set original name back. */
		len = pd->name_prompt.buf_len;
		strncpy(pd->name_prompt.buf,
		    pd->ori_info.name,
		    len
		);
		pd->name_prompt.buf[len - 1] = '\0';
		StripAbsolutePath(pd->name_prompt.buf);

	        return(-2);
	    }
	}
	status = 0;
	if(S_ISDIR(pd->ori_info.mode))
	{
	    if(strcmp(strptr, pd->ori_info.name))
	        status = FileOpMoveDir(strptr, pd->ori_info.name);
	}
	else
	{
	    if(strcmp(strptr, pd->ori_info.name))
		status = FileOpMoveFile(strptr, pd->ori_info.name);
	}
	if(status < 0)
	    return(-1);

	/* Update original name. */
	strptr = PrefixPaths(tmp_path, pd->name_prompt.buf);
	free(pd->ori_info.name);
	pd->ori_info.name = StringCopyAlloc(strptr);
	if(pd->ori_info.name == NULL)
	    return(-1);


	/* Set link destination. */
	if(S_ISLNK(pd->ori_info.mode) &&
           (pd->link_dest_prompt.buf != NULL)
	)
	{
	    /* Need to change current working directory. */
	    strptr = GetParentDir(pd->ori_info.name);
	    if(strptr == NULL) return(-1);
	    if(chdir(strptr) == 0)
	    {
		getcwd(dname.current_working, PATH_MAX);
		dname.current_working[PATH_MAX - 1] = '\0';
	    }
	    unlink(pd->ori_info.name);
	    status = symlink(pd->link_dest_prompt.buf, pd->ori_info.name);
	    if(status == -1)
	    {
		sprintf(stringa,
"Error defining link:\n\n\
    %s\n\n\
 To destination:\n\n\
    %s\n\n\
Check to make sure that the destination is valid.\n",
                    pd->ori_info.name,
		    pd->link_dest_prompt.buf
		);
		printdw(&err_dw, stringa);

		return(-1);
	    }
	}


        /* Set owner and group. */
        if((pd->owner_prompt.buf != NULL) &&
           (pd->group_prompt.buf != NULL)
        )
        {
            if((strlen(pd->owner_prompt.buf) > 0) &&
               (strlen(pd->group_prompt.buf) > 0)
            )
            {
                status = chown(pd->ori_info.name,
                    IDGetUIDbyName(pd->owner_prompt.buf),
                    IDGetGIDbyGroup(pd->group_prompt.buf)
                );
            }
        }   


	/*    Set permissions.
	 *
	 *    Note: Setting permissions must be done *after* chowning
	 *    since chowning will change permissions!
	 */
	if(!S_ISLNK(pd->ori_info.mode))
	{
	    mode = 0;

            mode |= (pd->perm_usrr_tb.state) ? S_IRUSR : 0;
            mode |= (pd->perm_usrw_tb.state) ? S_IWUSR : 0;
            mode |= (pd->perm_usrx_tb.state) ? S_IXUSR : 0;
	    if(pd->perm_uid_tb.state)
		mode |= S_ISUID;

            mode |= (pd->perm_grpr_tb.state) ? S_IRGRP : 0;
            mode |= (pd->perm_grpw_tb.state) ? S_IWGRP : 0;
            mode |= (pd->perm_grpx_tb.state) ? S_IXGRP : 0;
            if(pd->perm_gid_tb.state)
                mode |= S_ISGID;

            mode |= (pd->perm_othr_tb.state) ? S_IROTH : 0;
            mode |= (pd->perm_othw_tb.state) ? S_IWOTH : 0;
            mode |= (pd->perm_othx_tb.state) ? S_IXOTH : 0;
            if(pd->perm_sticky_tb.state)
		mode |= S_ISVTX;

	    status = chmod(pd->ori_info.name, mode);

	    /* Update mode. */
	    if(lstat(pd->ori_info.name, &stat_buf) == 0)
	    {
	        pd->ori_info.mode = stat_buf.st_mode;
	    }
	}
	else
	{
	    /* It is a link, so set appropriate permission values. */
            pd->perm_uid_tb.state = False;
            pd->perm_gid_tb.state = False;
            pd->perm_sticky_tb.state = False;

            pd->perm_usrr_tb.state = True;
            pd->perm_usrw_tb.state = True;
            pd->perm_usrx_tb.state = True;

            pd->perm_grpr_tb.state = True;
            pd->perm_grpw_tb.state = True;
            pd->perm_grpx_tb.state = True;

            pd->perm_grpr_tb.state = True;
            pd->perm_grpw_tb.state = True;
            pd->perm_grpx_tb.state = True;	    
	}


	/* *********************************************************** */

	/* Update listings on all endeavour windows. */
	for(i = 0; i < (int)total_endeavour_wins; i++)
	{
	    if(endeavour_win[i] == NULL) continue;

            if(ISPATHDIR(pd->ori_info.name))
            {
                /* Update dir item name on the list. */
                EDWUpdateDirItemName(
                    endeavour_win[i], 
                    old_name,                /* Old name. */
                    pd->ori_info.name        /* New name. */
                );
            }
            else
            {
                /* Update file listing. */
                EDWRefreshFileList(
                    endeavour_win[i],
                    endeavour_win[i]->loc_prompt.buf
                );
            }
 
	    if(endeavour_win[i]->map_state == 1)
	        EDWDraw(endeavour_win[i], DRAW_AMOUNT_COMPLETE);
	}


	return(0);
}



int PropDialogDoTouch(prop_dialog_struct *pd)
{
	struct stat stat_buf;
	char *strptr;
	int i, status;


	if(pd == NULL)
	    return(-1);

	if((pd->ori_info.name == NULL) ||
	   (pd->toplevel == 0)
	)
	    return(-1);

	if(lstat(pd->ori_info.name, &stat_buf) == -1)
	    return(-1);


	/* Check master write-protect. */
        if(option.write_protect)
        {
            printdw(&err_dw,
"Cannot update date and time stamps on object:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
            );
	    return(-3);
	}


        /* Need to change current working directory. */
        strptr = GetParentDir(pd->ori_info.name);
        if(strptr == NULL) return(-1);
        if(chdir(strptr) == 0)
        {
            getcwd(dname.current_working, PATH_MAX);   
            dname.current_working[PATH_MAX - 1] = '\0';
        }

	/* Touch object. */
	if(S_ISDIR(pd->ori_info.mode))
	{
	    status = FileOpTouchDir(pd->ori_info.name);
	}
	else
	{
	    status = FileOpTouchFile(pd->ori_info.name);
	}

	/* Update date stats. */
        if(lstat(pd->ori_info.name, &stat_buf) == -1)
            return(-1);

        pd->ori_info.last_accessed = stat_buf.st_atime;
        pd->ori_info.last_modified = stat_buf.st_mtime;
        pd->ori_info.create_time = stat_buf.st_ctime;


        /* Must refresh all endeavour windows. */
        for(i = 0; i < (int)total_endeavour_wins; i++)
        {
            if(endeavour_win[i] == NULL) continue;
            if(endeavour_win[i]->map_state == 0) continue;
        
            EDWRefreshAllLists(endeavour_win[i]);
            EDWDraw(endeavour_win[i], DRAW_AMOUNT_COMPLETE);
        }


	return(0);
}


/* ************************************************************************
 *
 */

int PropDialogReclaimMemory()
{
        int i, highest_pd_num;
         
            
        /* Skip if nothing is allocated. */
        if( (prop_dialog == NULL) ||
            ((int)total_prop_dialogs <= 0)
        )
            return(-1);

        /* ******************************************************** */
        
        /* Reclaim memory. */
        highest_pd_num = -1;
        for(i = 0; i < (int)total_prop_dialogs; i++)
        {
            if(prop_dialog[i] == NULL)
                continue;
            highest_pd_num = i;
        }

        if(highest_pd_num < ((int)total_prop_dialogs - 1))
        {
            /* Adjust global variable total_prop_dialogs. */
            total_prop_dialogs = highest_pd_num + 1;
                 
            /* Reallocate or free pointers. */
            if((int)total_prop_dialogs <= 0)
            {
                /* total_prop_dialogs has dropped to 0, free all pointers. */
                free(prop_dialog);
                prop_dialog = NULL;
                total_prop_dialogs = 0;
            }
            else
            {
                prop_dialog = (prop_dialog_struct **)realloc(
                    prop_dialog,
                    total_prop_dialogs * sizeof(prop_dialog_struct *)
                );
                if(prop_dialog == NULL)
                {
                    total_prop_dialogs = 0;
                }
            }
        }

        return(0);
}



int PropDialogAllocate()
{
        int i;
        int new_pd_num;

 
        /* Sanitize global total_prop_dialogs. */
        if((int)total_prop_dialogs < 0)
            total_prop_dialogs = 0;


        /* Check for available entry. */
        for(i = 0; i < (int)total_prop_dialogs; i++)
        {
            if(prop_dialog[i] == NULL)
                break;
        }
        if(i < (int)total_prop_dialogs)
        {
            /* Found an unallocated pointer. */
            new_pd_num = i;
	}
	else
	{
	    /* Need to allocate more pointers. */

	    new_pd_num = total_prop_dialogs;
            total_prop_dialogs++;	/* Increment global total. */

            /* Allocate more prop dialog window pointers. */
            prop_dialog = (prop_dialog_struct **)realloc(
                prop_dialog, total_prop_dialogs * sizeof(prop_dialog_struct *)
            );
            if(prop_dialog == NULL)
            {
                total_prop_dialogs = 0;
                return(-1);
            }
	}


        /* Allocate a new pd structure for that pointer. */
        prop_dialog[new_pd_num] = (prop_dialog_struct *)calloc(1,
            sizeof(prop_dialog_struct));
        if(prop_dialog[new_pd_num] == NULL)
        {
            return(-1);
        }


        return(new_pd_num);
}



void PropDialogDelete(int pd_num)
{
	if(!PropDialogIsAllocated(pd_num))
	    return;

	/* Destroy allocated GUI resources and substructures. */
	PropDialogDestroy(prop_dialog[pd_num]);

	free(prop_dialog[pd_num]);
	prop_dialog[pd_num] = NULL;


	return;
}



void PropDialogDeleteAll()
{ 
        int i;

	for(i = 0; i < (int)total_prop_dialogs; i++)
	{
	    /* This will deallocate all substructures and itself. */
	    PropDialogDelete(i);
	}

	/* Free pointers and global total. */
	free(prop_dialog);
	prop_dialog = NULL; 
 
        total_prop_dialogs = 0;


	return;
}



int PropDialogManageAll(event_t *event)
{
        static int i, events_handled;

            
        /* Error checks. */
        if(!IDC() ||
           (event == NULL)
        )
            return(0);


        /* ********************************************************** */
        events_handled = 0;

        for(i = 0; i < (int)total_prop_dialogs; i++)
        {
            /* Skip unallocated ones. */
            if(prop_dialog[i] == NULL)
                continue;

            /* ClientMessage must be handled here. */
            if((event->type == ClientMessage) &&
               OSWIsEventDestroyWindow(prop_dialog[i]->toplevel, event)
            )
            {
		/* Delete and reclaim memory. */
                PropDialogDelete(i);
                PropDialogReclaimMemory();

                OSWPurgeAllEvents();

                events_handled++;
                break;		/* Break because events are messed up now. */
            }

	    /* Manage events normally. */
            events_handled += PropDialogManage(prop_dialog[i], event);

	    /* If the dialog unmapped, then delete it. */
	    if(prop_dialog[i]->map_state == 0)
	    {
                /* Delete and reclaim memory. */
	        PropDialogDelete(i);
		PropDialogReclaimMemory();

                OSWPurgeAllEvents();

                break;          /* Break because events are messed up now. */
	    }
	}




	return(events_handled);
}



/* **********************************************************************
 *
 *
 */
int PropDialogInit(prop_dialog_struct *pd)
{
	int btn_x, btn_y;
        win_attr_t wattr;
        pixmap_t pixmap;  


	if(!IDC() || (pd == NULL))
	    return(-1);

	/* Reset values. */
	pd->map_state = 0;
	pd->visibility_state = VisibilityFullyObscured;
        pd->x = 0;
        pd->y = 0;
        pd->is_in_focus = 0;
        pd->width = PD_WIDTH;
        pd->height = PD_HEIGHT;
        pd->disabled = False;

	memset(&pd->ori_info, 0x00, sizeof(object_struct));


	/* Toplevel. */
	if(
	    OSWCreateWindow(
		&pd->toplevel,
		osw_gui[0].root_win,
		pd->x, pd->y,
		pd->width, pd->height
	    )
	)
	    return(-1);
	OSWSetWindowInput(pd->toplevel, KeyPressMask |
            KeyReleaseMask | ButtonPressMask | ButtonReleaseMask |
            ExposureMask | StructureNotifyMask | FocusChangeMask |
            VisibilityChangeMask
        );
        OSWSetWindowBkg(pd->toplevel, 0, widget_global.std_bkg_pm);

        WidgetCenterWindow(pd->toplevel, WidgetCenterWindowToRoot);
        OSWGetWindowAttributes(pd->toplevel, &wattr);
        pd->x = wattr.x;
        pd->y = wattr.y;
        pd->width = wattr.width; 
        pd->height = wattr.height;

        if(
            OSWCreatePixmap(
                &pd->toplevel_buf,
                wattr.width, wattr.height
            )
        )
            return(-1);


        /* Standard properties for toplevel. */
        pixmap = OSWCreatePixmapFromImage(endeavour_image.endeavour_icon);


	/* WM Properties. */
	OSWSetWindowWMProperties(
	    pd->toplevel,
            "Properties",
            "Properties",
            (pixmap == 0) ? widget_global.std_icon_pm : pixmap,
	    False,
            pd->x, pd->y,
            pd->width, pd->height,
            pd->width, pd->height,
            WindowFrameStyleFixed,
            NULL, 0
	);
        pixmap = 0;     /* Do not destroy pixmap. */


	/* Name prompt. */
	if(
	    PromptInit(
		&pd->name_prompt,
		pd->toplevel,
		32, 32,
		335, 30,
		PROMPT_STYLE_FLUSHED,
		"Name:",
		PATH_MAX + NAME_MAX,
		0,
		NULL
	    )
	)
	    return(-1);

        /* Link destination prompt. */
        if(
            PromptInit(
                &pd->link_dest_prompt, 
                pd->toplevel,
                32, 72,
                285, 30,
                PROMPT_STYLE_FLUSHED,
                "Link To:",
                PATH_MAX + NAME_MAX,
                0,
                NULL
            )
        )
            return(-1);


        /* Owner prompt. */
        if(
            PromptInit(
                &pd->owner_prompt,
                pd->toplevel,
                32, 290,
                180, 30,
                PROMPT_STYLE_FLUSHED,
                "Owner:",
                NAME_MAX,
                0,
                NULL
            )
        )
            return(-1);

        /* Group prompt. */
        if(
            PromptInit(
                &pd->group_prompt,    
                pd->toplevel,
                215, 290,
                180, 30,
                PROMPT_STYLE_FLUSHED,
                "Group:", 
                NAME_MAX, 
                0,
                NULL
            )
        )
            return(-1);


	/* Permission toggle buttons. */
	/* SUID, SGID, Sticky. */
	btn_x = 240;
	btn_y = 140;
	if(
	    TgBtnInit(
		&pd->perm_uid_tb,
		pd->toplevel,
		btn_x, btn_y,
		False,
		"SUID"
	    )
	)
	    return(-1);

	btn_x += 70;
        if(
            TgBtnInit(    
                &pd->perm_gid_tb,
                pd->toplevel,
                btn_x, btn_y,
                False,
                "SGID"
            )
        )
            return(-1);

        btn_x += 70;
        if(
            TgBtnInit( 
                &pd->perm_sticky_tb,
                pd->toplevel,
                btn_x, btn_y,
                False,
                "Sticky"
            )
        )   
            return(-1);

        /* Standard permissions. */
        btn_x = 32;
        btn_y = 220;
        if(
            TgBtnInit(
                &pd->perm_usrr_tb,
                pd->toplevel,
                btn_x, btn_y,
                False,
                NULL
            )
        )
            return(-1);
        btn_x += 40;
        if(
            TgBtnInit(
                &pd->perm_usrw_tb,
                pd->toplevel,
                btn_x, btn_y,
                False,
		NULL
            )
        )
            return(-1);
        btn_x += 40;
        if(  
            TgBtnInit(
                &pd->perm_usrx_tb,
                pd->toplevel,
                btn_x, btn_y,
                False,
                NULL
            )
        )
            return(-1);

        btn_x += 60;
        if(     
            TgBtnInit(
                &pd->perm_grpr_tb,
                pd->toplevel,
                btn_x, btn_y,
                False,
                NULL  
            )
        )
            return(-1);
        btn_x += 40;
        if(
            TgBtnInit(
                &pd->perm_grpw_tb,
                pd->toplevel,
                btn_x, btn_y,
                False,
                NULL  
            )
        )
            return(-1);
        btn_x += 40;   
        if(
            TgBtnInit(
                &pd->perm_grpx_tb,
                pd->toplevel,
                btn_x, btn_y,
                False,
                NULL  
            )
        )
            return(-1);

	btn_x += 60;
        if(
            TgBtnInit(
                &pd->perm_othr_tb,
                pd->toplevel,
                btn_x, btn_y,
                False,
                NULL  
            )
        )
            return(-1);
        btn_x += 40;
        if(
            TgBtnInit(
                &pd->perm_othw_tb,
                pd->toplevel,
                btn_x, btn_y,
                False,
                NULL  
            )
        )
            return(-1);
        btn_x += 40;   
        if(
            TgBtnInit(
                &pd->perm_othx_tb,
                pd->toplevel,
                btn_x, btn_y,
                False,
                NULL  
            )
        )
            return(-1);





        /* Touch button. */
        if(
            PBtnInit(
                &pd->touch_btn,
                pd->toplevel,
                360, 360,
                70, 28,
                "Touch",
                PBTN_TALIGN_CENTER,
                NULL,
		(void *)&pd->touch_btn,
                NULL
            )
        )
            return(-1);
	PBtnSetHintMessage(
	    &pd->touch_btn,
	    "Set date and time stamps to current time."
	);


	/* OK button. */
	if(
	    PBtnInit(
		&pd->ok_btn,
		pd->toplevel,
		32, 440,
		70, 28,
		"OK",
		PBTN_TALIGN_CENTER,
		NULL,
                (void *)&pd->ok_btn,
		NULL
	    )
	)
            return(-1);

	/* Apply button. */
        if(
            PBtnInit(
                &pd->apply_btn,
                pd->toplevel,
                205, 440,
                70, 28,
                "Apply",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&pd->apply_btn,
                NULL
            )
        )  
            return(-1);

        /* Cancel button. */
        if(
            PBtnInit(
                &pd->cancel_btn,
                pd->toplevel,
                380, 440,
                70, 28,
                "Cancel",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&pd->cancel_btn,
                NULL
            )
        )
            return(-1);



	return(0);
}



int PropDialogResize(prop_dialog_struct *pd)
{

	if(!IDC() ||
           (pd == NULL)
	)
	    return(-1);

	/* Prop dialogs are fixed sized. */



	return(0);
}



int PropDialogDraw(prop_dialog_struct *pd)
{
	win_attr_t wattr;
	char stringa[1024];
	image_t *image, *image2;
	umode_t mode;


        if(!IDC() || (pd == NULL))
            return(-1);

	if(pd->toplevel == 0)
	    return(-1);


	/* Map as needed. */
	if(pd->map_state == 0)
	{
	    OSWMapRaised(pd->toplevel);

	    PromptMap(&pd->name_prompt);
            PromptMap(&pd->link_dest_prompt);
            PromptMap(&pd->owner_prompt);
            PromptMap(&pd->group_prompt);

	    TgBtnMap(&pd->perm_uid_tb);
            TgBtnMap(&pd->perm_gid_tb);
            TgBtnMap(&pd->perm_sticky_tb);

            TgBtnMap(&pd->perm_usrr_tb);
            TgBtnMap(&pd->perm_usrw_tb);
            TgBtnMap(&pd->perm_usrx_tb);
            TgBtnMap(&pd->perm_grpr_tb);
            TgBtnMap(&pd->perm_grpw_tb);
            TgBtnMap(&pd->perm_grpx_tb);
            TgBtnMap(&pd->perm_othr_tb);
            TgBtnMap(&pd->perm_othw_tb);
            TgBtnMap(&pd->perm_othx_tb);


	    PBtnMap(&pd->ok_btn);
            PBtnMap(&pd->cancel_btn);
            PBtnMap(&pd->apply_btn);
            PBtnMap(&pd->touch_btn);

	    pd->map_state = 1;
	}


	/* Map or unmap the link destination prompt. */
	if(S_ISLNK(pd->ori_info.mode))
	    PromptMap(&pd->link_dest_prompt);
	else
	    PromptUnmap(&pd->link_dest_prompt);


	/* Toplevel. */
	if(1)
	{
	    OSWGetWindowAttributes(pd->toplevel, &wattr);

	    WidgetPutImageTile(pd->toplevel_buf,
		widget_global.std_bkg_img,
		wattr.width, wattr.height
	    );

	    /* Object icon. */
	    mode = pd->ori_info.mode;
	    if(S_ISLNK(mode))
		image = widget_global.linkicon_selected_img;
	    else if(S_ISDIR(mode))
                image = widget_global.diricon_selected_img;
	    else if((S_IXUSR & mode) ||
                    (S_IXGRP & mode) ||
                    (S_IXOTH & mode)
	    )
		image = widget_global.execicon_selected_img;
            else if(S_ISREG(mode))
                image = widget_global.fileicon_selected_img;
	    else if(S_ISCHR(mode) ||
                    S_ISBLK(mode) ||  
                    S_ISFIFO(mode) ||
                    S_ISSOCK(mode)
            )
                image = widget_global.pipeicon_selected_img;
            else
                image = widget_global.fileicon_selected_img;

	    if(!OSWCreateImage(&image2, 32, 32) && (image != NULL))
	    {
	        WidgetResizeImageBuffer(
		    osw_gui[0].depth,
		    (u_int8_t *)image2->data,
                    (u_int8_t *)image->data,
		    image2->width, image2->height,
		    image->width, image->height
	        );
	        WidgetPutImageRaised(
		    pd->toplevel_buf,
		    image2,
		    390, 50,
		    16
	        );
	    }
	    OSWDestroyImage(&image2);

	    /* If link, check if it exists. */
	    if(S_ISLNK(mode))
	    {
		if(PD_LINK_DEST_EXISTS(pd, pd->link_dest_prompt.buf))
		{
		    OSWSetFgPix(widget_global.normal_text_pix);
		    OSWDrawString(pd->toplevel_buf,
		        324, 92,
		        "Valid"
		    );
		}
		else
		{
                    OSWSetFgPix(widget_global.normal_text_pix);
                    OSWDrawString(pd->toplevel_buf,
                        324, 92,
                        "Missing"
                    ); 
                }
	    }


	    /* Permission labels. */
	    OSWSetFgPix(widget_global.normal_text_pix);
            OSWDrawString(pd->toplevel_buf,
                32, 197,
 "      Owner              Group               Other"
            );

	    OSWDrawString(pd->toplevel_buf,
		32, 218,
 " R     W     X        R    W     X        R    W    X"
	    );

	    /* Dates. */
	    sprintf(stringa, "      Created: %s",
		EDW_GETTIMESTR(pd->ori_info.create_time));
	    OSWDrawString(pd->toplevel_buf,
		32, 369,
		stringa
	    );
            sprintf(stringa, "     Modified: %s",
                EDW_GETTIMESTR(pd->ori_info.last_modified));
            OSWDrawString(pd->toplevel_buf,
                32, 389,
                stringa
            );
            sprintf(stringa, "Last Accessed: %s",
                EDW_GETTIMESTR(pd->ori_info.last_accessed));
            OSWDrawString(pd->toplevel_buf,
                32, 409,
                stringa
            );



	    /* Name outline. */
	    DRAW_OUTLINE(
                pd->toplevel_buf,
		16, 16,
		445, 97,
                "Identification",
                widget_global.surface_highlight_pix,
		widget_global.surface_shadow_pix,
		widget_global.normal_text_pix
	    );

            /* Permissions outline. */
            DRAW_OUTLINE(
                pd->toplevel_buf,
                16, 130,
                445, 128,
                "Permissions",
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix,
                widget_global.normal_text_pix
            );

            /* Ownership outline. */
            DRAW_OUTLINE(
                pd->toplevel_buf,
                16, 275,
                445, 55,
                "Ownership",
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix,
                widget_global.normal_text_pix
            );

            /* Dates outline. */
            DRAW_OUTLINE(
                pd->toplevel_buf,
                16, 345,
                445, 75,
                "Time Stamps",
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix,
                widget_global.normal_text_pix
            );


	    /* Put buffer. */
	    OSWPutBufferToWindow(pd->toplevel, pd->toplevel_buf);
	}




	return(0);
}


int PropDialogManage(prop_dialog_struct *pd, event_t *event)
{
	static int events_handled;


	/* Error checks. */
	if(!IDC() ||
           (pd == NULL) ||
           (event == NULL)
	)
	    return(0);

	if((pd->toplevel == 0) ||
           (pd->map_state == 0)
	)
	    return(0);


	/* ************************************************************* */

	events_handled = 0;

	switch(event->type)
	{
	  /* *********************************************************** */
	  case KeyPress:
	    if(pd->is_in_focus == 0)
		return(events_handled);

	    /* Tab. */
	    if(event->xkey.keycode == osw_keycode.tab)
	    {
		if(pd->name_prompt.is_in_focus == 1)
		{
                    PD_UNFOCUS_ALL_PROMPTS(pd);

		    if(S_ISLNK(pd->ori_info.mode))
		    {
			pd->link_dest_prompt.is_in_focus = 1;
		        PromptMarkBuffer(
                            &pd->link_dest_prompt,
                            PROMPT_POS_END
                        );
		    }
		    else
		    {
			pd->owner_prompt.is_in_focus = 1;
                        PromptMarkBuffer(
                            &pd->owner_prompt,
                            PROMPT_POS_END
                        );
		    }
		}
		else if((pd->link_dest_prompt.is_in_focus == 1) &&
                        S_ISLNK(pd->ori_info.mode)
		)
		{
                    PD_UNFOCUS_ALL_PROMPTS(pd);

                    pd->owner_prompt.is_in_focus = 1;
                    PromptMarkBuffer(
                        &pd->owner_prompt,
                        PROMPT_POS_END
                    );
		}
		else if(pd->owner_prompt.is_in_focus == 1)
		{
                    PD_UNFOCUS_ALL_PROMPTS(pd);

                    pd->group_prompt.is_in_focus = 1;
                    PromptMarkBuffer(
                        &pd->group_prompt,
                        PROMPT_POS_END
                    );
		}
                else if(pd->group_prompt.is_in_focus == 1)
                {
                    PD_UNFOCUS_ALL_PROMPTS(pd);

                    pd->name_prompt.is_in_focus = 1;
                    PromptMarkBuffer(
                        &pd->name_prompt,
                        PROMPT_POS_END
                    );
                }
		else
		{
		    /* Default to having name prompt into focus. */
		    PD_UNFOCUS_ALL_PROMPTS(pd);

		    pd->name_prompt.is_in_focus = 1;
                    PromptMarkBuffer( 
                        &pd->name_prompt,
                        PROMPT_POS_END
                    );
		}

		/* Redraw all prompts. */
		PromptDraw(&pd->name_prompt, PROMPT_DRAW_AMOUNT_TEXTONLY);
		if(S_ISLNK(pd->ori_info.mode))
		{
                    PromptDraw(
			&pd->link_dest_prompt,
			PROMPT_DRAW_AMOUNT_TEXTONLY
		    );
		}
                PromptDraw(&pd->owner_prompt, PROMPT_DRAW_AMOUNT_TEXTONLY);
                PromptDraw(&pd->group_prompt, PROMPT_DRAW_AMOUNT_TEXTONLY);

                events_handled++;
                return(events_handled);
	    }
	    /* Escape key. */
	    else if(event->xkey.keycode == osw_keycode.esc)
	    {
		pd->cancel_btn.state = PBTN_ARMED;
		PBtnDraw(&pd->cancel_btn);

                OSWKBAutoRepeatOff();

		events_handled++;
		return(events_handled);
	    }
	    /* Enter key. */
            else if((event->xkey.keycode == osw_keycode.enter) ||
                    (event->xkey.keycode == osw_keycode.np_enter)
	    )
            {
                PropDialogApply(pd); 
                PropDialogUnmap(pd);
                
                events_handled++;
                return(events_handled);
            }
	    break;

          /* *********************************************************** */
          case KeyRelease:
            if(pd->is_in_focus == 0)
                return(events_handled);

            /* Escape key. */
            if(event->xkey.keycode == osw_keycode.esc)
            {
                pd->cancel_btn.state = PBTN_UNARMED;
                PBtnDraw(&pd->cancel_btn);

                PropDialogUnmap(pd);
                OSWKBAutoRepeatOn();

                events_handled++;
                return(events_handled);
            }

            break;  

          /* *********************************************************** */
          case ButtonPress:
	    if(event->xany.window == pd->toplevel)
		pd->is_in_focus = 1;

            break;

          /* *********************************************************** */
          case ButtonRelease:

            break;

          /* *********************************************************** */
          case MotionNotify:
            if(event->xany.window == pd->toplevel)
	    {

		events_handled++;
		return(events_handled);
	    }
            break;

          /* *********************************************************** */
          case Expose:
            if(event->xany.window == pd->toplevel)
	    {
                events_handled++;
	    }
            break;

          /* ********************************************************* */
          case ConfigureNotify:
            if(event->xany.window == pd->toplevel)
            {
                PropDialogResize(pd);

                events_handled++;
            }
            break;

          /* ********************************************************* */
          case FocusIn:
            if(event->xany.window == pd->toplevel)
            {
                pd->is_in_focus = 1;
                events_handled++;
            }
            else
            {
                pd->is_in_focus = 0;
                /* Do not report any event for focus out. */
            }
            return(events_handled);
            break;

          /* ********************************************************* */
          case FocusOut:
            if(event->xany.window == pd->toplevel)
            {
                pd->is_in_focus = 0;
                events_handled++;
            }
            return(events_handled);
            break;

          /* ********************************************************* */
          case VisibilityNotify:
            if(event->xany.window == pd->toplevel)
            {
                pd->visibility_state = event->xvisibility.state;

                events_handled++;
                return(events_handled);
            }
            break;
	}

	if(events_handled > 0)
	{
	    PropDialogDraw(pd);
	    return(events_handled);
	}

        /* ************************************************************* */
	/* Manage widgets. */

	/* Take all prompts out of focus on ButtonPress. */
	if(event->type == ButtonPress)
	    PD_UNFOCUS_ALL_PROMPTS(pd);

	/* Name prompt. */
	if(events_handled == 0)
	{
	    events_handled += PromptManage(&pd->name_prompt, event);
	}

	/* Link destination. */
        if(events_handled == 0)
        {
            events_handled += PromptManage(&pd->link_dest_prompt, event);
	    if((events_handled > 0) &&
               ((event->type == KeyPress) ||
                (event->type == ButtonPress)
	       )
	    )
	    {
		PropDialogDraw(pd);
	    }
        }

        /* Owner prompt. */
        if(events_handled == 0)
        {
            events_handled += PromptManage(&pd->owner_prompt, event);
        }

        /* Group prompt. */
        if(events_handled == 0)
        {
            events_handled += PromptManage(&pd->group_prompt, event);
        }


	/* OK button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(&pd->ok_btn, event);

	    if(((event->type == ButtonRelease) ||
               (event->type == KeyRelease)) &&
               (events_handled > 0)
	    )
	    {
		PropDialogApply(pd);
		PropDialogUnmap(pd);
		return(events_handled);
	    }
        }
        /* Apply button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(&pd->apply_btn, event);

            if(((event->type == ButtonRelease) ||
               (event->type == KeyRelease)) &&
               (events_handled > 0)
            )
            {
                PropDialogApply(pd);
		PropDialogDraw(pd);
                return(events_handled);
            }
        }
        /* Cancel button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(&pd->cancel_btn, event);

            if(((event->type == ButtonRelease) ||
               (event->type == KeyRelease)) &&
               (events_handled > 0)
            )
            {
                PropDialogUnmap(pd);
                return(events_handled);
            }
        }
        /* Touch button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(&pd->touch_btn, event);

            if(((event->type == ButtonRelease) ||
               (event->type == KeyRelease)) &&
               (events_handled > 0)
            )
            {
	        PropDialogDoTouch(pd);
	        PropDialogDraw(pd);

		return(events_handled);
	    }
        }
 
	/* Permission toggle buttons. */
	if(events_handled == 0)
        {
            events_handled += TgBtnManage(&pd->perm_uid_tb, event);
	}
        if(events_handled == 0)
        {
            events_handled += TgBtnManage(&pd->perm_gid_tb, event);
        }
        if(events_handled == 0)
        {
            events_handled += TgBtnManage(&pd->perm_sticky_tb, event);
        }
        if(events_handled == 0)
        {
            events_handled += TgBtnManage(&pd->perm_usrr_tb, event);
        }
        if(events_handled == 0)
        {
            events_handled += TgBtnManage(&pd->perm_usrw_tb, event);
        }
        if(events_handled == 0)
        {
            events_handled += TgBtnManage(&pd->perm_usrx_tb, event);
        }
        if(events_handled == 0)
        {
            events_handled += TgBtnManage(&pd->perm_grpr_tb, event);
        }
        if(events_handled == 0)
        {
            events_handled += TgBtnManage(&pd->perm_grpw_tb, event);
        }
        if(events_handled == 0)
        {
            events_handled += TgBtnManage(&pd->perm_grpx_tb, event);
        }
        if(events_handled == 0)
        {
            events_handled += TgBtnManage(&pd->perm_othr_tb, event);
        }
        if(events_handled == 0)
        {
            events_handled += TgBtnManage(&pd->perm_othw_tb, event);
        }
        if(events_handled == 0)
        {
            events_handled += TgBtnManage(&pd->perm_othx_tb, event);
        }



	return(events_handled);
}



void PropDialogMap(prop_dialog_struct *pd)
{
        if(!IDC() ||
           (pd == NULL)
	)
            return;

        UnfocusAllTopLvlWins();

        pd->ok_btn.state = PBTN_UNARMED;
        pd->apply_btn.state = PBTN_UNARMED;
        pd->cancel_btn.state = PBTN_UNARMED;
        pd->touch_btn.state = PBTN_UNARMED;

	pd->map_state = 0;
	PropDialogDraw(pd);

	pd->is_in_focus = 1;
	pd->visibility_state = VisibilityUnobscured;

	return;
}



void PropDialogUnmap(prop_dialog_struct *pd)
{
        if(!IDC() || 
           (pd == NULL)
        )   
            return;


	OSWUnmapWindow(pd->toplevel);
	pd->map_state = 0;
	pd->is_in_focus = 0;


	return;
}



void PropDialogDestroy(prop_dialog_struct *pd)
{
	if(pd == NULL)
	    return;


	if(IDC())
	{
	    TgBtnDestroy(&pd->perm_uid_tb);
            TgBtnDestroy(&pd->perm_gid_tb);
            TgBtnDestroy(&pd->perm_sticky_tb);

            TgBtnDestroy(&pd->perm_usrr_tb);
            TgBtnDestroy(&pd->perm_usrw_tb);
            TgBtnDestroy(&pd->perm_usrx_tb);

            TgBtnDestroy(&pd->perm_grpr_tb);
            TgBtnDestroy(&pd->perm_grpw_tb);
            TgBtnDestroy(&pd->perm_grpx_tb);

            TgBtnDestroy(&pd->perm_othr_tb);
            TgBtnDestroy(&pd->perm_othw_tb);
            TgBtnDestroy(&pd->perm_othx_tb);

	    PromptDestroy(&pd->group_prompt);
            PromptDestroy(&pd->owner_prompt);
            PromptDestroy(&pd->link_dest_prompt);
            PromptDestroy(&pd->name_prompt);

	    PBtnDestroy(&pd->touch_btn);
            PBtnDestroy(&pd->apply_btn);
            PBtnDestroy(&pd->cancel_btn);
            PBtnDestroy(&pd->ok_btn);

	    OSWDestroyWindow(&pd->toplevel);
	    OSWDestroyPixmap(&pd->toplevel_buf);
	}


	/* Free original object information. */
	free(pd->ori_info.name);
	pd->ori_info.name = NULL;

	free(pd->ori_info.link_dest);
	pd->ori_info.link_dest = NULL;

	memset(&pd->ori_info, 0x00, sizeof(object_struct));

	return;
}
