/*
                      Restart Configuration File

	Functions:

	int RCLoadFromFile(char *filename)

	---


*/

#include "endeavour.h"


int RCLoadFromFile(char *filename)
{
	int lines_read;
	int fass_num, rec_num;
	FILE *fp;
	char *strptr;
	char *strptr2;

	char parm[CFG_PARAMETER_MAX];
	char val[CFG_VALUE_MAX];
	struct stat stat_buf;


	if(filename == NULL)
	    return(-1);

	if(stat(filename, &stat_buf))
	{
            fprintf(stderr,
                "%s: No such file.\n",
                filename
            );
	    return(-1);
	}

	fp = fopen(filename, "rb");
	if(fp == NULL)
	{
	    fprintf(stderr,
		"%s: Cannot read.\n",
		filename
	    );
	    return(-1);
	}

	/* Need to delete all file associations. */
	FAssDeleteAll();


	/* ************************************************************* */

	lines_read = 0;
	strptr = NULL;

	while(1)
	{
            free(strptr); strptr = NULL;
	    strptr = FReadNextLineAllocCount(
		fp, UNIXCFG_COMMENT_CHAR, &lines_read
	    );
	    if(strptr == NULL) break;

	    strptr2 = StringCfgParseParm(strptr);
	    if(strptr2 == NULL) continue;
	    strncpy(parm, strptr2, CFG_PARAMETER_MAX);
	    parm[CFG_PARAMETER_MAX - 1] = '\0';

            strptr2 = StringCfgParseValue(strptr);
            if(strptr2 == NULL) strptr2 = "0";	/* Set it to "0" if NULL. */
            strncpy(val, strptr2, CFG_VALUE_MAX);
            val[CFG_VALUE_MAX - 1] = '\0';


	    /* VersionMajor */
	    if(!strcmp(parm, "VersionMajor"))
	    {
                option.rc_version_major = atoi(val);
	    }
	    /* VersionMinor */
	    else if(!strcmp(parm, "VersionMinor"))
            {
		option.rc_version_minor = atoi(val);
            }

	    /* MasterWriteProtect (at startup) */
	    else if(!strcmp(parm, "MasterWriteProtect"))
            {
                option.write_protect = StringIsYes(val);
            }
            /* ComfermPointerMotionMove */
            else if(!strcmp(parm, "ComfermPointerMotionMove"))
            {
                option.comferm_pointer_motion_move = StringIsYes(val);
            }
            /* ComfermPointerMotionCopy */
            else if(!strcmp(parm, "ComfermPointerMotionCopy"))
            {
                option.comferm_pointer_motion_copy = StringIsYes(val);
            }
            /* ComfermOverwrite */
            else if(!strcmp(parm, "ComfermOverwrite"))
            {
                option.comferm_overwrites = StringIsYes(val);
            }
            /* ComfermRecycles or ComfermDeletes */
            else if(!strcmp(parm, "ComfermRecycles") ||
                    !strcmp(parm, "ComfermDeletes") ||
                    !strcmp(parm, "ComfermRecycle") ||
                    !strcmp(parm, "ComfermDelete")
	    )
            {
                option.comferm_recycle = StringIsYes(val);
            }
            /* ComfermPurge */
            else if(!strcmp(parm, "ComfermPurge"))
            {
                option.comferm_purge = StringIsYes(val);
            }
            /* ComfermPurgeAll */
            else if(!strcmp(parm, "ComfermPurgeAll"))
            {
                option.comferm_purge_all = StringIsYes(val);
            }
            /* ComfermUnmountOnExit */
            else if(!strcmp(parm, "ComfermUnmountOnExit") ||
                    !strcmp(parm, "ComfermUnmountOnexit")
	    )
            {
                option.comferm_unmount_onexit = StringIsYes(val);
            }
	    /* ExecutionMaxBurst */
            else if(!strcmp(parm, "ExecutionMaxBurst"))
            {
                option.file_exe_max_burst = atoi(val);
	    }
            /* MountTimeout */
            else if(!strcmp(parm, "MountTimeout") ||
                    !strcmp(parm, "MountTimeOut")
	    )
            {
                option.mount_timeout = atol(val);

		/* Must be atleast 1 second long. */
		if((long)option.mount_timeout < 1)
		    option.mount_timeout = 1;
            }
	    /* ShowProgramOutput */
            else if(!strcmp(parm, "ShowProgramOutput") ||
                    !strcmp(parm, "ShowProgramOutputs")
	    )
	    {
		option.show_prog_output = atoi(val);
		if((option.show_prog_output != 0) &&
                   (option.show_prog_output != 1) &&
                   (option.show_prog_output != 2)
		)
		{
                    fprintf(stderr,
            "%s: Line %i: ShowProgramOutput: Unsupported value.\n",
                        filename, lines_read
                    );
		}
	    }
	    /* ShowExpandableDirectories */
            else if(!strcmp(parm, "ShowExpandableDirectories") ||
                    !strcmp(parm, "ShowExpandableDirs")
            )
            {
                option.show_expandable_dirs = StringIsYes(val);
            }

            /* AnimatePadlock */
            else if(!strcmp(parm, "AnimatePadlock") ||
                    !strcmp(parm, "AnimatePadLock")
            )
            {
                option.animated_padlock = StringIsYes(val);
            }

	    /* UserName */
            else if(!strcmp(parm, "UserName") ||
                    !strcmp(parm, "User")
	    )
            {
                strncpy(userinfo.name, val, NAME_MAX);
		userinfo.name[NAME_MAX - 1] = '\0';
            }

	    /* RecycleBinDir (local) */
	    else if(!strcmp(parm, "RecycleBinDir"))
	    {
		if(ISPATHABSOLUTE(val))
		{
		    if(stat(val, &stat_buf))
		    {
		        strncpy(dname.recycle_bin, val, PATH_MAX);
		        dname.recycle_bin[PATH_MAX - 1] = '\0';
		    }
		    else
		    {
		        fprintf(stderr,
			    "%s: Line %i: %s: No such directory.\n",
			    filename, lines_read, val
		        );
		    }
		}
		else
		{
		    strptr2 = PrefixPaths(dname.home, val);
		    strncpy(dname.recycle_bin,
			(strptr2 == NULL) ? "/" : strptr2,
			PATH_MAX
		    );
		    dname.recycle_bin[PATH_MAX - 1] = '\0';

                    if(stat(dname.recycle_bin, &stat_buf))
                    {
                        fprintf(stderr,
                            "%s: Line %i: %s: No such directory.\n",
                            filename, lines_read, dname.recycle_bin
                        );
		    }
		}
	    }
            /* RecycleBin (local) */
            else if(!strcmp(parm, "RecycleBin"))
            {
		if(ISPATHABSOLUTE(val))
		{
                    strncpy(fname.recycle_bin, val, PATH_MAX + NAME_MAX);
                    fname.recycle_bin[PATH_MAX + NAME_MAX - 1] = '\0';
		}
		else
		{
		    strptr2 = PrefixPaths(dname.recycle_bin, val);
		    strncpy(fname.recycle_bin,
			(strptr2 == NULL) ? "/" : strptr2,
			PATH_MAX + NAME_MAX
		    );
                    fname.recycle_bin[PATH_MAX + NAME_MAX - 1] = '\0';
		}

                /* Its okay if this file does not exist. */
            }
	    /* ProgramOutputFile */
            else if(!strcmp(parm, "ProgramOutputFile"))
	    {
		strncpy(fname.prog_output_ori, val, PATH_MAX + NAME_MAX);
		fname.prog_output_ori[PATH_MAX + NAME_MAX - 1] = '\0';

		sprintf(
		    fname.prog_output,
		    "%s%i",
		    val,
		    userinfo.pid
		);
		if(!stat(fname.prog_output, &stat_buf))
		{
		    fprintf(stderr, "%s: Line %i: Warning, `%s' file exists.\n",
			filename, lines_read, fname.prog_output
		    );
		}
	    }
            /* ListOutputFile */
            else if(!strcmp(parm, "ListOutputFile"))
            {
                strncpy(fname.list_output_ori, val, PATH_MAX + NAME_MAX);
                fname.list_output_ori[PATH_MAX + NAME_MAX - 1] = '\0';

                sprintf(
                    fname.list_output,
                    "%s%i",
                    val,
                    userinfo.pid
                );
                if(!stat(fname.list_output, &stat_buf))
                {
                    fprintf(stderr, "%s: Line %i: Warning, `%s' file exists.\n",
                        filename, lines_read, fname.list_output
                    );
                }
            }
            /* ComPipe */
            else if(!strcmp(parm, "ComPipe"))
            {
                strncpy(fname.com_pipe_ori, val, PATH_MAX + NAME_MAX);
                fname.com_pipe_ori[PATH_MAX + NAME_MAX - 1] = '\0';

                sprintf(
                    fname.com_pipe,
                    "%s%i",
                    val,
                    userinfo.pid
                );
            }

            /* MoveFileCommand */
            else if(!strcmp(parm, "MoveFileCommand"))
            {
                free(command.move_file);
                command.move_file = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.move_file)) 
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
            }
            /* MoveDirectoryCommand */
            else if(!strcmp(parm, "MoveDirectoryCommand") ||
                    !strcmp(parm, "MoveDirCommand")
            )
            {
		free(command.move_dir);
		command.move_dir = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.move_dir))
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
	    }
            /* CopyFileCommand */
            else if(!strcmp(parm, "CopyFileCommand"))
            {
                free(command.copy_file);
                command.copy_file = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.copy_file))
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
            }
	    /* CopyDirectoryCommand */
            else if(!strcmp(parm, "CopyDirectoryCommand") ||
                    !strcmp(parm, "CopyDirCommand")
	    )
            {
                free(command.copy_dir);
                command.copy_dir = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.copy_dir))
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
	    }
            /* RenameFileCommand */
            else if(!strcmp(parm, "RenameFileCommand"))
            {
                free(command.rename_file);
                command.rename_file = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.rename_file))
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
            }
            /* RenameDirectoryCommand */
            else if(!strcmp(parm, "RenameDirectoryCommand") ||
                    !strcmp(parm, "RenameDirCommand")
            )
            {
                free(command.rename_dir);
                command.rename_dir = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.rename_dir))
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
            }
            /* ChmodFileCommand */
            else if(!strcmp(parm, "ChmodFileCommand"))
            {
                free(command.chmod_file);
                command.chmod_file = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.chmod_file)) 
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
            }
            /* ChmodDirectoryCommand */
            else if(!strcmp(parm, "ChmodDirectoryCommand") ||
                    !strcmp(parm, "ChmodDirCommand")
            )
            {
                free(command.chmod_dir);
                command.chmod_dir = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.chmod_dir))
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
            }
            /* ChownFileCommand */
            else if(!strcmp(parm, "ChownFileCommand"))
            {
                free(command.chown_file);
                command.chown_file = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.chown_file))
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
            }
            /* ChownDirectoryCommand */
            else if(!strcmp(parm, "ChownDirectoryCommand") ||
                    !strcmp(parm, "ChownDirCommand")
            )
            {
                free(command.chown_dir);
                command.chown_dir = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.chown_dir))
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
            }
            /* TouchFileCommand */
            else if(!strcmp(parm, "TouchFileCommand"))
            {
                free(command.touch_file);
                command.touch_file = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.touch_file))
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
            }
            /* TouchDirectoryCommand */
            else if(!strcmp(parm, "TouchDirectoryCommand") ||
                    !strcmp(parm, "TouchDirCommand")
            )
            {
                free(command.touch_dir);
                command.touch_dir = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.touch_dir))
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
            }
	    /* TextEditorCommand */
            else if(!strcmp(parm, "TextEditorCommand"))
	    {
	        free(command.text_editor);
		command.text_editor = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.text_editor))
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
	    }
            /* DeviceFindCommand */
            else if(!strcmp(parm, "DeviceFindCommand"))
            {
                free(command.df);
                command.df = StringCopyAlloc(val);
                if(ISCMDREMOVE(command.df))    
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                        filename, lines_read
                    );
                }
            }


            /* ********************************************************** */
            /* Device mount information. */

	    /* BeginDeviceRecord */
            else if(!strcmp(parm, "BeginDeviceRecord"))
            {
                /* Allocate a new device record. */
                rec_num = DevRecAllocate();
                if(rec_num < 0)
                    continue;

		devrec[rec_num]->flags = 0;


                while(1)
                {
                    free(strptr); strptr = NULL;
                    strptr = FReadNextLineAllocCount(   
                        fp, UNIXCFG_COMMENT_CHAR, &lines_read  
                    );
                    if(strptr == NULL) break;

                    strptr2 = StringCfgParseParm(strptr);
                    if(strptr2 == NULL) continue;
                    strncpy(parm, strptr2, CFG_PARAMETER_MAX);
                    parm[CFG_PARAMETER_MAX - 1] = '\0';

                    strptr2 = StringCfgParseValue(strptr);
                    if(strptr2 == NULL) strptr2 = "0";
                    strncpy(val, strptr2, CFG_VALUE_MAX);
                    val[CFG_VALUE_MAX - 1] = '\0';


                    /* Label */
                    if(!strcmp(parm, "Label") ||
                       !strcmp(parm, "FriendlyName")
                    )
                    {
                        free(devrec[rec_num]->label);
                        devrec[rec_num]->label = StringCopyAlloc(val);
                    }
                    /* Device */
                    else if(!strcmp(parm, "Device") ||
                            !strcmp(parm, "DeviceName") ||
                            !strcmp(parm, "DevName")
		    )
                    {
                        free(devrec[rec_num]->devname);
                        devrec[rec_num]->devname = StringCopyAlloc(val);
                    }
                    /* Path */
                    else if(!strcmp(parm, "Path") ||
                            !strcmp(parm, "MountedPath") ||
                            !strcmp(parm, "MountPath") ||
                            !strcmp(parm, "Destination") ||
                            !strcmp(parm, "TargetPath") ||
                            !strcmp(parm, "TarPath") ||
                            !strcmp(parm, "Target")
		    )
                    {
                        free(devrec[rec_num]->path);
                        devrec[rec_num]->path = StringCopyAlloc(val);
                    }
		    /* ReadOnly */
                    else if(!strcmp(parm, "ReadOnly") ||
                            !strcmp(parm, "Readonly") ||
                            !strcmp(parm, "RO")
                    )
                    {
			if(StringIsYes(val))
			{
                            devrec[rec_num]->flags =
				(devrec[rec_num]->flags | DEV_TF_READONLY);
			}
                    }   
		    /* Unmountable */
                    else if(!strcmp(parm, "Unmountable"))
                    {
                        if(!StringIsYes(val))
                        {
                            devrec[rec_num]->flags =
                                (devrec[rec_num]->flags | DEV_TF_NOUNMOUNT);
                        }
                    }
                    /* MountCommand */
                    else if(!strcmp(parm, "MountCommand") ||
                            !strcmp(parm, "MountCmd")
                    )
                    {
                        free(devrec[rec_num]->cmd_mount);
                        devrec[rec_num]->cmd_mount = StringCopyAlloc(val);
			if(ISCMDREMOVE(devrec[rec_num]->cmd_mount))
			{
                            fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
				filename, lines_read
                            );
			}
                    }
                    /* UnmountCommand */
                    else if(!strcmp(parm, "UnmountCommand") ||
                            !strcmp(parm, "UnmountCmd")   
                    )
                    {
                        free(devrec[rec_num]->cmd_unmount);
                        devrec[rec_num]->cmd_unmount = StringCopyAlloc(val);
                        if(ISCMDREMOVE(devrec[rec_num]->cmd_unmount))
                        {
                            fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                                filename, lines_read                    
                            ); 
                        }
                    }
                    /* StatusCommand */
                    else if(!strcmp(parm, "StatusCommand") ||
                            !strcmp(parm, "StatusCmd")
                    )
                    {
                        free(devrec[rec_num]->cmd_status);
                        devrec[rec_num]->cmd_status = StringCopyAlloc(val);
                        if(ISCMDREMOVE(devrec[rec_num]->cmd_status))
                        {
                            fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                                filename, lines_read                    
                            ); 
                        }
                    }
                    /* ImageCode */
                    else if(!strcmp(parm, "ImageCode") ||
                            !strcmp(parm, "Image") ||
                            !strcmp(parm, "IconCode") ||
                            !strcmp(parm, "Icon")
                    )
                    {
                        devrec[rec_num]->image_code = atoi(val);
                    }



                    /* EndDeviceRecord */
                    else if(!strcmp(parm, "EndDeviceRecord"))
                    {
                        /* Make sure the minimum number of values are set. */
             
                 
             
                        break;
                    }
		}
	    }

	    /* ********************************************************** */
	    /* File associations. */

	    /* BeginFileAssociation */
	    else if(!strcmp(parm, "BeginFileAssociation"))
	    {
		/* Allocate a new file association. */
		fass_num = FAssAdd("yiff", "yiff");
		if(fass_num < 0)
		    continue;

                while(1)
                {
                    free(strptr); strptr = NULL;
                    strptr = FReadNextLineAllocCount(   
                        fp, UNIXCFG_COMMENT_CHAR, &lines_read  
                    );
                    if(strptr == NULL) break;

                    strptr2 = StringCfgParseParm(strptr);
                    if(strptr2 == NULL) continue;
                    strncpy(parm, strptr2, CFG_PARAMETER_MAX);
                    parm[CFG_PARAMETER_MAX - 1] = '\0';

                    strptr2 = StringCfgParseValue(strptr);
                    if(strptr2 == NULL) strptr2 = "0";
                    strncpy(val, strptr2, CFG_VALUE_MAX);
                    val[CFG_VALUE_MAX - 1] = '\0';


		    /* Extensions */
		    if(!strcmp(parm, "Extensions") ||
                       !strcmp(parm, "Extension") ||
                       !strcmp(parm, "Ext")
		    )
		    {
			free(file_ass[fass_num]->ext);
			file_ass[fass_num]->ext = StringCopyAlloc(val);
		    }
                    /* Command */  
                    else if(!strcmp(parm, "Command") ||
                            !strcmp(parm, "Cmd")
                    )
                    {
                        free(file_ass[fass_num]->cmd);
                        file_ass[fass_num]->cmd = StringCopyAlloc(val);
                        if(ISCMDREMOVE(file_ass[fass_num]->cmd))
                        {
                            fprintf(stderr,
 "%s: Line %i: Warning: Command may cause object removal.\n",
                                filename, lines_read
                            );
                        }
                    }
		    /* EndFileAssociation */
		    else if(!strcmp(parm, "EndFileAssociation"))
		    {
			/* Make sure the minimum number of values are set. */



			break;
		    }
		}
	    }
	    else
	    {
		/* Unsupported parameter. */


	    }

	}


	/* Close file. */
	fclose(fp);
	fp = NULL;



	return(lines_read);
}
