/*
                                   Recycle Bin



	Note: The term 'real object' reffers to an unrecycled object.

 */

#ifndef RECBIN_H
#define RECBIN_H

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <time.h>
extern char *tzname[2];
#include <signal.h>
 
   

#include "../include/graphics.h"
#include "../include/disk.h"
#include "../include/osw-x.h"
#include "../include/widget.h"
#include "../include/tga.h"
  
#include "object.h"
 

/*
 *	Recycled contents directories:
 *
 *	These directories contain the recycled files or directories.
 *
 *	The RECYCLE_LOCAL_DIR will have $HOME prefixed to it.
 */
#define RECYCLE_GLOBAL_DIR	"/var/preserve"
#define RECYCLE_LOCAL_DIR	".endeavour"


/*
 *	Recycled binary file name:
 *
 *	User's name (fetch from the USER enviroment) will be prepended
 *	to it.
 */
#define RECYCLE_BIN_FILE_NAME	".recycled.bin"


/*      
 *      Recycle bin window:
 */
typedef struct {
   
        int map_state;   
        int visibility_state;
        int x, y;
        unsigned width, height;
        char is_in_focus;   
        bool_t disabled;

	pixmap_t icon;

	win_t toplevel;

	win_t list;
	pixmap_t list_buf;
	scroll_bar_struct list_sb;        

	prompt_window_struct recbin_path_prompt;

        push_button_struct recover_btn;
        push_button_struct refresh_btn;
        push_button_struct purge_btn;
        push_button_struct purge_all_btn;
        push_button_struct close_btn;

	win_t status_win;
	pixmap_t status_win_buf;

	char *status_mesg;

	int *sel_object;
	int total_sel_objects;

	object_struct **object;		/* List of recycled objects. */
	int total_objects;

} recbin_win_struct;  

extern recbin_win_struct **recbin_win;
extern int total_recbin_wins;	/* Should always be 1. */


/* *********************************************************************
 *
 *                            Functions
 */

/* In recbinevent.c */
extern void RECBINSETALLCURSORS(recbin_win_struct *rb, int code);
extern int RECBINREFRESHLISTING(recbin_win_struct *rb);

extern int RecBinProcessMovement(
        void *tar_ptr,
        int tar_ptr_type,
        void *src_ptr,
        int src_ptr_type,
        int tar_drag_type,
        int src_drag_type,
        int tar_item_num,
        int op_code,
        bool_t comfermed
);

extern int RecBinInit(
	recbin_win_struct *rb,
	int x, int y,
	unsigned int width, unsigned int height,
	char *recbin_path
);
extern int RecBinResize(recbin_win_struct *rb);
extern int RecBinDraw(recbin_win_struct *rb, int amount);
extern int RecBinManage(recbin_win_struct *rb, event_t *event);
extern void RecBinMap(recbin_win_struct *rb);
extern void RecBinUnmap(recbin_win_struct *rb);
extern void RecBinDestroy(recbin_win_struct *rb);


/* In recbinio.c */
extern int RecycleAddObject(const char *path, const char *recbin);
extern int RecycleRemoveObject(int entry_num, char *recbin);
extern int RecycleRecoverObject(int entry_num, char *recbin, char *path);

extern void RecycleDoPurgeObject(char *string, char *recbin);
extern int RecycleDoRecycleObject(const char *path, const char *recbin);
extern int RecycleDoRestoreObject(char *string, char *recbin, char *path);
extern object_struct **RecycleGetObjectListing(char *recbin, int *nobjs);










#endif /* RECYCLE_H */
