/*
                       Recycle Bin Window Management

	Functions:

	void RECBINSETALLCURSORS(recbin_win_struct *rb, int code)
	void RECBINCHANGEICON(recbin_win_struct *rb)

	int RECBIN_SELECT(recbin_win_struct *rb, int item_num)
        int RECBIN_ISSELECTED(recbin_win_struct *rb, int item_num)
	void RECBIN_UNSELECT(recbin_win_struct *rb, int item_num)
        void RECBIN_UNSELECTALL(recbin_win_struct *rb)

	void RECBINFREEALLOBJECTS(recbin_win_struct *rb)
	int RECBINREFRESHLISTING(recbin_win_struct *rb)

        int RecBinProcessMovement(
                void *tar_ptr,
                int tar_ptr_type,
                void *src_ptr,
                int src_ptr_type,
                int tar_drag_type,
                int src_drag_type,
                int tar_item_num,
                int op_code,
                bool_t comfermed
        )

	int RecBinInit(
	        recbin_win_struct *rb,
	        int x, int y,
	        unsigned int width, unsigned int height,
	        char *recbin_path
	)
	int RecBinResize(recbin_win_struct *rb)
	int RecBinDraw(recbin_win_struct *rb, int amount)
	int RecBinManage(recbin_win_struct *rb, event_t *event)
	void RecBinMap(recbin_win_struct *rb)
	void RecBinUnmap(recbin_win_struct *rb)
	void RecBinDestroy(recbin_win_struct *rb)

	---


 */

#include "endeavour.h"


#define RECBIN_MIN_WIDTH	380
#define RECBIN_MIN_HEIGHT	250

#define ROW_SPACING             28

#define CHAR_WIDTH              8   
#define CHAR_HEIGHT             16  


#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))


#define STATUS_HEIGHT           30

#define HINT_MESG_RECOVER	"Recover the selected object(s) (ENTER)"
#define HINT_MESG_REFRESH	"Refresh list (F5)"
#define HINT_MESG_PURGE		"Permanently discard the selected object(s) (DEL)"
#define HINT_MESG_PURGEALL	"Permanently discard all the objects"
#define HINT_MESG_PADLOCK	"Toggles master write-protect on/off"



void RECBINSETALLCURSORS(recbin_win_struct *rb, int code)
{
	cursor_t std_cursor = osw_gui[0].std_cursor;


        if(rb == NULL)   
            return;


        switch(code)
        {
          case EDW_CURSORS_CODE_WORKING:
            OSWSetWindowCursor(rb->toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(rb->list, endeavour_cursor.working);
            OSWSetWindowCursor(rb->recover_btn.toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(rb->refresh_btn.toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(rb->purge_btn.toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(rb->purge_all_btn.toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(rb->close_btn.toplevel, endeavour_cursor.working);
            OSWSetWindowCursor(rb->status_win, endeavour_cursor.working);
	    break;

          default: /* EDW_CURSORS_CODE_NORMAL */
            OSWSetWindowCursor(rb->toplevel, std_cursor);
            OSWSetWindowCursor(rb->list, std_cursor);
            OSWSetWindowCursor(rb->recover_btn.toplevel, std_cursor);
            OSWSetWindowCursor(rb->refresh_btn.toplevel, std_cursor);
            OSWSetWindowCursor(rb->purge_btn.toplevel, std_cursor);
            OSWSetWindowCursor(rb->purge_all_btn.toplevel, std_cursor);
            OSWSetWindowCursor(rb->close_btn.toplevel, std_cursor);
            OSWSetWindowCursor(rb->status_win, std_cursor);
            break;
	}

	OSWGUISync(False);


	return;
}



void RECBINCHANGEICON(recbin_win_struct *rb)
{
	win_attr_t wattr;


	if(rb == NULL)
	    return;


	OSWGetWindowAttributes(rb->toplevel, &wattr);

	if(rb->icon != 0)
	    OSWDestroyPixmap(&rb->icon);
	if((int)rb->total_objects > 0)
            rb->icon = OSWCreatePixmapFromImage(
		endeavour_image.recycle_bin_large1);
	else
	    rb->icon = OSWCreatePixmapFromImage(
		endeavour_image.recycle_bin_large0);


	rb->x = wattr.x;
	rb->y = wattr.y;


	/* WM Properties. */
	OSWSetWindowWMProperties(
            rb->toplevel,
            "Recycle Bin",
            "Recycle Bin",
            (rb->icon == 0) ? widget_global.std_icon_pm : rb->icon,
            False,
            rb->x, rb->y,
            RECBIN_MIN_WIDTH, RECBIN_MIN_HEIGHT,
            osw_gui[0].display_width, osw_gui[0].display_height,
            WindowFrameStyleStandard,
            NULL, 0
	);


	return;
}



void RECBIN_UNSELECTALL(recbin_win_struct *rb)
{
        if(rb == NULL)
            return;

        free(rb->sel_object);
        rb->sel_object = NULL;

        rb->total_sel_objects = 0;


	return;
}



int RECBIN_SELECT(recbin_win_struct *rb, int item_num)
{
	int i;

        if(rb == NULL)
            return(-1);

	if((int)rb->total_sel_objects < 0)
	    rb->total_sel_objects = 0;


	/* Check if already selected. */
	for(i = 0; i < (int)rb->total_sel_objects; i++)
	{
	    if(rb->sel_object[i] == item_num)
		break;
	}

	rb->total_sel_objects++;
	rb->sel_object = (int *)realloc(rb->sel_object,
	    (int)rb->total_sel_objects * sizeof(int));
	if(rb->sel_object == NULL)
	{
	    rb->total_sel_objects = 0;
	    return(-1);
	}


	rb->sel_object[(int)rb->total_sel_objects - 1] = item_num;


	return(0);
}



int RECBIN_ISSELECTED(recbin_win_struct *rb, int item_num)
{
	int i;

	if(rb == NULL)
	    return(0);
	if((item_num < 0) || (item_num >= (int)rb->total_objects))
	    return(0);


	for(i = 0; i < (int)rb->total_sel_objects; i++)
	{
	    if(rb->sel_object[i] == item_num)
		return(1);
	}

	return(0);
}



/*
 *	Unselects item_num on the recycle bin rb.
 */
void RECBIN_UNSELECT(recbin_win_struct *rb, int item_num)
{
	int i;

        if(rb == NULL)
            return;

	if((item_num < 0) || (item_num >= (int)rb->total_objects))
	    return;

	for(i = 0; i < (int)rb->total_sel_objects; i++)
	{
	    if(rb->sel_object[i] == item_num)
		rb->sel_object[i] = -1;
	}

	return;
}



int GETITEMBYPOSITION(recbin_win_struct *rb, int x, int y)
{
	int item_num;

	if(rb == NULL)
	    return(-1);

	item_num = ((int)rb->list_sb.y_win_pos / ROW_SPACING) +
            ((y + ((int)rb->list_sb.y_win_pos % ROW_SPACING)) /
            ROW_SPACING);

	/* Sanitize. */
	if(item_num >= (int)rb->total_objects)
	    item_num = -1;
	else if(item_num < 0)
	    item_num = -1;

	return(item_num);
}



void RECBINFREEALLOBJECTS(recbin_win_struct *rb)
{
	int i;

	if(rb == NULL)
	    return;


	for(i = 0; i < (int)rb->total_objects; i++)
	{
	    if(rb->object[i] == NULL) continue;

	    free(rb->object[i]->name);
	    free(rb->object[i]->link_dest);

	    free(rb->object[i]);
	    rb->object[i] = NULL;
	}
	free(rb->object);
	rb->object = NULL;

	rb->total_objects = 0;


	return;
}



int RECBINREFRESHLISTING(recbin_win_struct *rb)
{
	int i;
	off_t tsize;
	char stringa[512];


	if(rb == NULL)
            return(-1);

	/* Free previous object listings and selections. */
	RECBIN_UNSELECTALL(rb);
	RECBINFREEALLOBJECTS(rb);

	/* Get object listing from recbin. */
	rb->object = RecycleGetObjectListing(
	    rb->recbin_path_prompt.buf,
	    &rb->total_objects
	);

	rb->list_sb.y_win_pos = 0;


	/* Need to change the recycle bin icon. */
	RECBINCHANGEICON(rb);


	/* Check if any objects were allocated. */
	if(rb->object == NULL)
	    return(-1);


	/* ********************************************************* */
	/* Update status mesg. */

	for(i = 0, tsize = 0; i < (int)rb->total_objects; i++)
	{
	    if(rb->object[i] == NULL) continue;

	    tsize += rb->object[i]->size;
	}

	sprintf(
	    stringa,
	    "%i objects total, consuming %ld bytes.",
	    rb->total_objects, tsize
	);

	free(rb->status_mesg);
	rb->status_mesg = StringCopyAlloc(stringa);


	return(0);
}



/*
 *	Processes a drag release on the recycle bin.
 */
int RecBinProcessMovement(
	void *tar_ptr,
        int tar_ptr_type,
        void *src_ptr,
        int src_ptr_type,
        int tar_drag_type,
        int src_drag_type,
        int tar_item_num,
        int op_code,
        bool_t comfermed
)
{
        int i, n, x;
        char tmp_path[PATH_MAX + NAME_MAX];

        endeavour_win_struct *src_edw;
        recbin_win_struct *src_rb, *tar_rb;
#ifdef USE_IMLIB
        img_viewer_struct *src_iv;
        img_browser_struct *src_ib;
#endif /* USE_IMLIB */


        /* Error checks. */
        if((tar_ptr == NULL) ||         /* Target must exist. */
           (tar_ptr_type == POINTER_TYPE_UNDEFINED)
        )
            return(-1);   

        /* Target must be endeavour window. */
        if(tar_ptr_type != POINTER_TYPE_RB)
        {
            fprintf(stderr,
    "RecBinProcessMovement(): tar_ptr_type must be POINTER_TYPE_RB.\n"
            );
            return(-1);
        }

        /* ************************************************************ */
        /* Get target and source pointers. */
        src_edw = NULL;
        src_rb = NULL;
	tar_rb = (recbin_win_struct *)tar_ptr;	/* Always not NULL. */

#ifdef USE_IMLIB
        src_iv = NULL;
        src_ib = NULL;
#endif /* USE_IMLIB */

	/* Check source pointer type. */
        switch(src_ptr_type)
        {
          case POINTER_TYPE_EDW:
            src_edw = (endeavour_win_struct *)src_ptr;
            break;

          case POINTER_TYPE_RB:
            src_rb = (recbin_win_struct *)src_ptr; 
            break;
            
#ifdef USE_IMLIB
          case POINTER_TYPE_IV:
            src_iv = (img_viewer_struct *)src_ptr;
            break;
         
          case POINTER_TYPE_IB:
            src_ib = (img_browser_struct *)src_ptr;
            break;
#endif /* USE_IMLIB */ 
         
          default:
            fprintf(stderr,
        "EDWProcessMovement(): Unknown source pointer type %i.\n",
                src_ptr_type
            );
            return(-1);
            break;
        }


        /* ************************************************************ */
        /* Endeavour Browser to Recycle Bin. */
        if((src_edw != NULL) &&
           (tar_rb != NULL)
        )
        {
	    /* src_edw must have location prompt buffer. */
	    if(src_edw->loc_prompt.buf == NULL)
		return(-1);

            /* Need to clean up drag here or else comf dialog does not   
             * map properly with normal cursors.  
             */
            DragCleanup();

            /* Check if internal write-protect is on. */
            if(option.write_protect)
            {
                printdw(&err_dw,
"Cannot recycle object:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                );
                return(-3);
            }

            /* Need to turn off yes_to_all just in case. */
            yes_to_all = False;

	    /* Recycle dirs on src_edw. */
            for(i = 0; i < (int)src_edw->total_sel_dirs; i++)
            {
                n = src_edw->sel_dir[i];
                if((n < 0) || (n >= (int)src_edw->total_dirs))
                    continue;
                if(src_edw->dir[n] == NULL) continue;
                if(src_edw->dir[n]->fullname == NULL) continue;

                /* Get tmp_path. */
                strncpy(
                    tmp_path,
		    src_edw->dir[n]->fullname,
                    PATH_MAX + NAME_MAX
                );
                tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

		/* Comfermation needed for each item? */
                if(!comfermed)
                {
                    /* Prompt for recycle comfermation. */
                    if(option.comferm_recycle &&
                       !yes_to_all
                    )
                    {
                        x = ComfDialogMapQueryOverWrite(
                            NULL,
                            tmp_path,
                            op_code
                        );
                        if(x == NO_PBTN_NUM)
                            continue;
                        else if(x == ALL_PBTN_NUM)
                        {
                            /* Switch yes to all on. */
                            yes_to_all = True;
                        }
                    }
		}

		/* Recycle tmp_path. */
		RecycleDoRecycleObject(tmp_path, fname.recycle_bin);
	    }
            /* Recycle files on src_edw. */
            for(i = 0; i < (int)src_edw->total_sel_files; i++)
            {
                n = src_edw->sel_file[i];
                if((n < 0) || (n >= (int)src_edw->total_files))
                    continue;
                if(src_edw->file[n] == NULL) continue;
                if(src_edw->file[n]->name == NULL) continue;
                
                /* Get tmp_path. */  
                strncpy(
                    tmp_path,
                    PrefixPaths(
			src_edw->loc_prompt.buf,
			src_edw->file[n]->name
		    ),
                    PATH_MAX + NAME_MAX
                );
                tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

                /* Comfermation needed for each item? */
                if(!comfermed)
                {
                    /* Prompt for recycle comfermation. */
                    if(option.comferm_recycle &&
                       !yes_to_all
                    )
                    {
                        x = ComfDialogMapQueryOverWrite(
                            NULL,
                            tmp_path,
                            op_code
                        );
                        if(x == NO_PBTN_NUM)
                            continue;
                        else if(x == ALL_PBTN_NUM)
                        {
                            /* Switch yes to all on. */
                            yes_to_all = True;
                        }
                    }
                }
                
                /* Recycle tmp_path. */
                RecycleDoRecycleObject(tmp_path, fname.recycle_bin);
            }


            /* Refresh recycle bin list and redraw it. */
            RECBINREFRESHLISTING(tar_rb);
            RecBinDraw(tar_rb, DRAW_AMOUNT_COMPLETE);

            /* Need to refresh edw list (after refreshing recbin). */
            EDWRefreshAllLists(src_edw);
	    EDWDraw(src_edw, DRAW_AMOUNT_COMPLETE);
	}
#ifdef USE_IMLIB
        /* ************************************************************* */
        /* Image Browser to Recycle Bin. */
	else if((src_ib != NULL) &&
                (tar_rb != NULL)
        )
        {
            /* src_ib must have current_dir buffer. */
            if(src_ib->current_dir == NULL)
                return(-1);

            /* Need to clean up drag here or else comf dialog does not
             * map properly with normal cursors.
             */
            DragCleanup();
                        
            /* Check if internal write-protect is on. */
            if(option.write_protect)     
            {
                printdw(&err_dw,
"Cannot recycle object:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                );
                return(-3);
            }

            /* Need to turn off yes_to_all just in case. */
            yes_to_all = False;
                            
            /* Recycle objects on src_ib. */
            for(i = 0; i < (int)src_ib->total_sel_thumbs; i++)
            {
/* Warn about directories since they cannot be
 * recycled, only deleted?
 */
                n = src_ib->sel_thumb[i];
                if((n < 0) || (n >= (int)src_ib->total_thumbs))
                    continue;
                if(src_ib->thumb[n] == NULL) continue;
                if(src_ib->thumb[n]->filename == NULL) continue;

                /* Get tmp_path. */
                strncpy(
                    tmp_path,
                    PrefixPaths(
                        src_ib->current_dir,
                        src_ib->thumb[n]->filename
                    ),
                    PATH_MAX + NAME_MAX
                );
                tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';


                /* Comfermation needed for each item? */
                if(!comfermed)
                {
                    /* Prompt for recycle comfermation. */
                    if(option.comferm_recycle &&
                       !yes_to_all
                    )
                    {
                        x = ComfDialogMapQueryOverWrite(
                            NULL,
                            tmp_path,
                            op_code
                        );
                        if(x == NO_PBTN_NUM)
                            continue;
                        else if(x == ALL_PBTN_NUM)
                        {
                            /* Switch yes to all on. */
                            yes_to_all = True;
                        }
                    }
                }

                /* Recycle tmp_path. */
                RecycleDoRecycleObject(tmp_path, fname.recycle_bin);
	    }


            /* Refresh recycle bin list and redraw it. */
            RECBINREFRESHLISTING(tar_rb);
            RecBinDraw(tar_rb, DRAW_AMOUNT_COMPLETE);

            /* Need to refresh imgbrowser list (after refreshing recbin). */
	    strncpy(tmp_path, src_ib->current_dir, PATH_MAX + NAME_MAX);
	    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
            IMBRLoad(src_ib, tmp_path);
            IMBRDraw(src_ib, DRAW_AMOUNT_COMPLETE);
        }
#endif /* USE_IMLIB */


        /* Turn yes to all off. */
        yes_to_all = False;


        return(0);
}



/* *********************************************************************
 *
 */
int RecBinInit(
        recbin_win_struct *rb,
        int x, int y,
        unsigned int width, unsigned int height,
        char *recbin_path
)
{
        int btn_x, btn_y, status;
        win_attr_t wattr;
        pixmap_t pixmap;
	char hotkeystr[PBTN_MAX_HOTKEYS];


        if(!IDC() || (rb == NULL))
            return(-1);


        /* Clear structure. */
	memset(rb, 0x00, sizeof(recbin_win_struct));


        /* Set default values. */
        rb->map_state = 0;   
        rb->visibility_state = VisibilityFullyObscured;
        rb->x = 0;
        rb->y = 0;
        rb->width = 580;
        rb->height = 480;
        rb->is_in_focus = 0;
        rb->disabled = False;

	rb->icon = 0;

	if((int)width > RECBIN_MIN_WIDTH)
	    rb->width = width;
        if((int)height > RECBIN_MIN_HEIGHT)
            rb->height = height;


        /* ************************************************************** */

        /* Create toplevel. */
        status = OSWCreateWindow(
            &rb->toplevel,
            osw_gui[0].root_win,
            rb->x, rb->y,
            rb->width, rb->height
        );
        if(status)
            return(-1);
        OSWSetWindowInput(rb->toplevel, KeyPressMask |
            KeyReleaseMask | ButtonPressMask | ButtonReleaseMask |
            PointerMotionMask | ExposureMask | StructureNotifyMask |
            FocusChangeMask | VisibilityChangeMask
        );
        OSWSetWindowBkg(rb->toplevel, 0, widget_global.std_bkg_pm);

        WidgetCenterWindow(rb->toplevel, WidgetCenterWindowToRoot);
        OSWGetWindowAttributes(rb->toplevel, &wattr);
        rb->x = wattr.x;
        rb->y = wattr.y;
        rb->width = wattr.width;
        rb->height = wattr.height;


        pixmap = OSWCreatePixmapFromImage(
	    endeavour_image.recycle_bin_large0);
	rb->icon = pixmap;


	/* WM Properties. */
        OSWSetWindowWMProperties(
            rb->toplevel,
            "Recycle Bin",
            "Recycle Bin",
            (pixmap == 0) ? widget_global.std_icon_pm : pixmap,
            False,
            rb->x, rb->y,
            RECBIN_MIN_WIDTH, RECBIN_MIN_HEIGHT,
            osw_gui[0].display_width, osw_gui[0].display_height,
            WindowFrameStyleStandard,
            NULL, 0
        );
        pixmap = 0;     /* Do not destroy pixmap. */


	/* Recycled objects list window. */
	if(
	    OSWCreateWindow(
		&rb->list,
		rb->toplevel,
		0, 0,
		rb->width,
		MAX(
		    (int)rb->height - 165,
		    1
		)
	    )
	)
	    return(-1);
	OSWSetWindowBkg(rb->list, widget_global.surface_editable_pix, 0);
	OSWSetWindowInput(rb->list, ButtonPressMask | ButtonReleaseMask |
            PointerMotionMask | ExposureMask);
	OSWGetWindowAttributes(rb->list, &wattr);
	if(
	    OSWCreatePixmap(
		&rb->list_buf,
		wattr.width, wattr.height
	    )
	)
	    return(-1);
	if(
	    SBarInit(
		&rb->list_sb,
		rb->list,
		wattr.width, wattr.height
	    )
	)
	    return(-1);

	/* Status window. */
	if(
	    OSWCreateWindow(
		&rb->status_win,
		rb->toplevel,
		0, (int)rb->height - STATUS_HEIGHT,
		rb->width, STATUS_HEIGHT
	    )
	)
	    return(-1);
	OSWGetWindowAttributes(rb->status_win, &wattr);
	if(
            OSWCreatePixmap( 
                &rb->status_win_buf,
                wattr.width, wattr.height
            )
        )
            return(-1);
	OSWSetWindowInput(rb->status_win, ButtonPressMask | ButtonReleaseMask |
            PointerMotionMask | ExposureMask | EnterWindowMask |
	    LeaveWindowMask
	);

	/* Status window hint message. */
	OSWGetWindowAttributes(rb->toplevel, &wattr);
        status = HintWinAddMessage(   
            rb->status_win,
            rb->toplevel,
            (int)wattr.width - 275,
            (int)wattr.height - 50,
            HINT_MESG_PADLOCK
        );
        if(status < 0)
            return(-1);


	/* Recycle bin path prompt. */
	if(
	    PromptInit(
		&rb->recbin_path_prompt,
		rb->toplevel,
		10, (int)rb->height - 160,
                MAX((int)rb->width - 20, 50), 30,
                PROMPT_STYLE_FLUSHED,
                "Recycle Bin:",
                PATH_MAX + NAME_MAX,
                25,  
                NULL
	    )
        )
            return(-1);
	if(recbin_path != NULL)
	{
	    strncpy(rb->recbin_path_prompt.buf, recbin_path,
		rb->recbin_path_prompt.buf_len);

	    rb->recbin_path_prompt.buf_pos =
		strlen(rb->recbin_path_prompt.buf);
	}


	/* Refresh button. */
        btn_x = 20;
        btn_y = (int)rb->height - STATUS_HEIGHT - (3 * 10) - (2 * 28);

        if(
            PBtnInit(  
                &rb->refresh_btn,
                rb->toplevel,
                btn_x, btn_y,
                70, 28,   
                "Refresh",
                PBTN_TALIGN_CENTER,
                NULL,
		(void *)&rb->refresh_btn,
                NULL
            )
        )
            return(-1);
        PBtnSetHintMessage(
            &rb->refresh_btn,
            HINT_MESG_REFRESH
        );

	/* Purge button. */
	btn_x += 85;
        if(
            PBtnInit(
                &rb->purge_btn,
                rb->toplevel,
                btn_x, btn_y,
                70, 28,
                "Purge",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&rb->purge_btn,
                NULL
            )
        )
            return(-1);
        PBtnSetHintMessage(
            &rb->purge_btn,
            HINT_MESG_PURGE
        );
        hotkeystr[0] = 0x04;
	hotkeystr[1] = '\0';
        PBtnSetHotKeys(&rb->purge_btn, hotkeystr);

	/* Purge all. */
        btn_x += 85;
        if(
            PBtnInit(
                &rb->purge_all_btn,
                rb->toplevel,
                btn_x, btn_y,
                70, 28,
                "Purge All",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&rb->purge_all_btn,
                NULL
            )
        )
            return(-1);
        PBtnSetHintMessage(
            &rb->purge_all_btn,
            HINT_MESG_PURGEALL
        );


	/* Recover button. */
	btn_x = 20;
	btn_y = (int)rb->height - STATUS_HEIGHT - 10 - 28;
	if(
	    PBtnInit(
                &rb->recover_btn,
		rb->toplevel,
                btn_x, btn_y,
                70, 28,
                "Recover",
                PBTN_TALIGN_CENTER,
                NULL,
                (void *)&rb->recover_btn,
                NULL
	    )
        )
	    return(-1);
        PBtnSetHintMessage(
            &rb->recover_btn,
            HINT_MESG_RECOVER
        );
	hotkeystr[0] = '\n';
	hotkeystr[1] = ' ';
	hotkeystr[2] = '\0';
        PBtnSetHotKeys(&rb->recover_btn, hotkeystr);

	/* Close button. */
	btn_x = (int)rb->width - 90;
        if(
            PBtnInit(
                &rb->close_btn,
                rb->toplevel,
                btn_x, btn_y,
                70, 28,
                "Close",
                PBTN_TALIGN_CENTER, 
                NULL,
                (void *)&rb->close_btn,
                NULL
            )
        )
            return(-1);
	hotkeystr[0] = 0x1b;
        hotkeystr[1] = 'c';
        hotkeystr[2] = '\0';
        PBtnSetHotKeys(&rb->close_btn, hotkeystr);


	/* Update listing. */
	RECBINREFRESHLISTING(rb);


	return(0);
}



int RecBinResize(recbin_win_struct *rb)
{
	int btn_x, btn_y;
        win_attr_t wattr;   
         

        if(rb == NULL)
            return(-1);

        /* Get toplevel attributes. */
        OSWGetWindowAttributes(rb->toplevel, &wattr);

        /* Skip if no change. */
        if(((int)rb->width == wattr.width) &&
           ((int)rb->height == wattr.height)
        )
            return(-2);


        /* Set new size values of toplevel. */
        rb->x = wattr.x;
        rb->y = wattr.y;
        rb->width = wattr.width;
        rb->height = wattr.height;


	/* List window. */
	OSWMoveResizeWindow(rb->list,
	    0,
	    0,
	    rb->width,
            MAX(
                (int)rb->height - 165,
                1
            )
 	);
	OSWGetWindowAttributes(rb->list, &wattr);
	OSWDestroyPixmap(&rb->list_buf);
	if(
	    OSWCreatePixmap(
                &rb->list_buf,
                wattr.width, wattr.height
            )
        )
            return(-1);
	SBarResize(&rb->list_sb, wattr.width, wattr.height);
	/* Reset positions. */
	rb->list_sb.y_win_pos = 0;


	/* Status window. */
	OSWMoveResizeWindow(rb->status_win,
            0, (int)rb->height - STATUS_HEIGHT,
            rb->width, STATUS_HEIGHT
	);
	OSWGetWindowAttributes(rb->status_win, &wattr);
	OSWDestroyPixmap(&rb->status_win_buf);
        if(
            OSWCreatePixmap(
                &rb->status_win_buf,
                wattr.width, wattr.height
            )
        ) 
            return(-1);

        OSWGetWindowAttributes(rb->toplevel, &wattr);



	/* Recycle bin path prompt. */
	OSWMoveResizeWindow(rb->recbin_path_prompt.toplevel,
            10, (int)rb->height - 160,
            MAX((int)rb->width - 20, 50), 30
        );
	PromptMap(&rb->recbin_path_prompt);

	/* Buttons. */
        btn_x = 20;
	btn_y = (int)rb->height - STATUS_HEIGHT - (3 * 10) - (2 * 28);
        OSWMoveWindow(rb->refresh_btn.toplevel,
            btn_x, btn_y
        );

        btn_x += 85;
        OSWMoveWindow(rb->purge_btn.toplevel,
            btn_x, btn_y
        );

        btn_x += 85;
        OSWMoveWindow(rb->purge_all_btn.toplevel,
            btn_x, btn_y
        );


        btn_x = 20;
        btn_y = (int)rb->height - STATUS_HEIGHT - 10 - 28;
        OSWMoveWindow(rb->recover_btn.toplevel,
            btn_x, btn_y
        );

	btn_x = (int)rb->width - 90;
        OSWMoveWindow(rb->close_btn.toplevel,
            btn_x, btn_y 
        );


        /* Sync with GUI. */
        OSWGUISync(False);


	return(0);
}



int RecBinDraw(recbin_win_struct *rb, int amount)
{
	int y;
	win_attr_t wattr_toplevel, wattr;

        int list_y_pos;            
        int item_num;
        int items_visiable, items_drawn;

        image_t *ximage;
        char tmp_path[2 * (PATH_MAX + NAME_MAX)];
        char stringa[NAME_MAX + 256];


        /* Error checks. */
        if( !IDC() ||
            (rb == NULL)
        )   
            return(-1);

        if(rb->toplevel == 0)
            return(-1);


        /* ************************************************************ */
         
        /* Get toplevel attributes. */
        OSWGetWindowAttributes(rb->toplevel, &wattr_toplevel);


        /* Map as needed. */
        if(!rb->map_state)
        {
	    /* Refresh list. */
	    RECBINREFRESHLISTING(rb);

	    /* Map windows and widgets. */
	    OSWMapRaised(rb->toplevel);

	    OSWMapWindow(rb->list);
	    OSWMapWindow(rb->status_win);

	    PromptMap(&rb->recbin_path_prompt);
	    PBtnMap(&rb->recover_btn);
	    PBtnMap(&rb->refresh_btn);
            PBtnMap(&rb->purge_btn);
            PBtnMap(&rb->purge_all_btn);
	    PBtnMap(&rb->close_btn);

	    OSWGetWindowAttributes(rb->list, &wattr);
	    SBarDraw(
		&rb->list_sb,
		wattr.width, wattr.height,
		wattr.width,
		((int)rb->total_objects + 1) * ROW_SPACING
	    );

	    rb->map_state = 1;
	    rb->visibility_state = VisibilityUnobscured;

	    /* Must change draw amount to complete. */
	    amount = DRAW_AMOUNT_COMPLETE;
	}


        /* ********************************************************** */
	/* Draw list window. */
	if((amount == DRAW_AMOUNT_COMPLETE) ||
	   (amount == DRAW_AMOUNT_LISTS) ||
           (amount == DRAW_AMOUNT_DIR_LIST) ||
           (amount == DRAW_AMOUNT_FILE_LIST) ||
           (amount == DRAW_AMOUNT_FILE_HEADING)
        )
	{
	    /* Draw list window. */
            OSWGetWindowAttributes(rb->list, &wattr);

            /* Clear buffer. */
            OSWClearPixmap(rb->list_buf,  
                wattr.width, wattr.height,
                widget_global.surface_editable_pix
            );

	    /* Draw listing. */
	    if((rb->object != NULL) && ((int)rb->total_objects > 0))
	    {
	        /* Calculate starting position for gradiant scrolling. */
                list_y_pos = 5 - ((int)rb->list_sb.y_win_pos % ROW_SPACING);

                /* Calculate starting dir_num. */
                item_num = rb->list_sb.y_win_pos / ROW_SPACING;
                if(item_num < 0)
                    item_num = 0;

                /* Calculate items_visiable. */
                items_visiable = ((int)wattr.height / ROW_SPACING) + 1;

                items_drawn = 0;

	        while(item_num < (int)rb->total_objects)
	        {
                    /* Don't draw too many then visable. */
                    if(items_drawn > items_visiable)
                        break;

                    /* Skip unallocated entries. */ 
                    if(rb->object[item_num] == NULL)  
                    {
                        /* Skip unallocated dir entries. */
                        item_num++; continue;
                    }

                    /* Allocate name for items with no name. */
                    if(rb->object[item_num]->name == NULL)
                    {
                        rb->object[item_num]->name =
			    StringCopyAlloc(NO_NAME_STR); 
                        if(rb->object[item_num]->name == NULL)
                        {
                            item_num++; continue;
                        }
                    }

                    /* Check if item is selected. */
                    if(RECBIN_ISSELECTED(rb, item_num))
                    {
                        /* Selected. */
                        OSWSetFgPix(widget_global.surface_selected_pix);
		        if(S_ISLNK(rb->object[item_num]->mode))
		        {
                            OSWDrawSolidRectangle(rb->list_buf,
                             24 + 4,
                             list_y_pos,
                             ( strlen(rb->object[item_num]->name) +
                               ((rb->object[item_num]->link_dest == NULL) ?
                               0 : strlen(rb->object[item_num]->link_dest) + 4
                               )
                             ) * CHAR_WIDTH,
                             ROW_SPACING
                            );
		        }
		        else
                        {
                            OSWDrawSolidRectangle(rb->list_buf,
                             24 + 4,
                             list_y_pos,
                             strlen(rb->object[item_num]->name) *
                                 CHAR_WIDTH,
                             ROW_SPACING
                            );
                        }


                        /* See which ximage to use. */
                        if(S_ISLNK(rb->object[item_num]->mode))
                            ximage = widget_global.linkicon_selected_img;
                        else if(S_ISDIR(rb->object[item_num]->mode))
                            ximage = widget_global.diricon_selected_img;
                        else if( (S_IXUSR & rb->object[item_num]->mode) ||
                                 (S_IXGRP & rb->object[item_num]->mode) ||
                                 (S_IXOTH & rb->object[item_num]->mode)
                        )
                            ximage = widget_global.execicon_selected_img;
                        else if(S_ISREG(rb->object[item_num]->mode))
                            ximage = widget_global.fileicon_selected_img;
                        else if(S_ISCHR(rb->object[item_num]->mode) ||
                                S_ISBLK(rb->object[item_num]->mode) ||
                                S_ISFIFO(rb->object[item_num]->mode) ||
                                S_ISSOCK(rb->object[item_num]->mode)
                        )
                            ximage = widget_global.pipeicon_selected_img;
                        else
                            ximage = widget_global.fileicon_selected_img;

                        OSWSetFgPix(widget_global.selected_text_pix);
                    } 
		    else
		    {
                        /* See which ximage to use. */
                        if(S_ISLNK(rb->object[item_num]->mode))
                            ximage = widget_global.linkicon_normal_img;
                        else if(S_ISDIR(rb->object[item_num]->mode))
                            ximage = widget_global.diricon_normal_img;
                        else if( (S_IXUSR & rb->object[item_num]->mode) ||
                                 (S_IXGRP & rb->object[item_num]->mode) ||
                                 (S_IXOTH & rb->object[item_num]->mode)
                        )
                            ximage = widget_global.execicon_normal_img;
                        else if(S_ISREG(rb->object[item_num]->mode))
                            ximage = widget_global.fileicon_normal_img;
                        else if(S_ISCHR(rb->object[item_num]->mode) ||
                                S_ISBLK(rb->object[item_num]->mode) ||
                                S_ISFIFO(rb->object[item_num]->mode) ||
                                S_ISSOCK(rb->object[item_num]->mode)
                        )
                            ximage = widget_global.pipeicon_normal_img;
                        else
                            ximage = widget_global.fileicon_normal_img;

                        /* Normal directory entry. */
                        OSWSetFgPix(widget_global.editable_text_pix);
		    }

                    /* Put icon image. */
                    OSWPutImageToDrawablePos(
                        ximage,
                        (drawable_t)rb->list_buf,
                        0,
                        list_y_pos + 2
                    );

                    /* Name (print link name if needed). */
                    if(S_ISLNK(rb->object[item_num]->mode))
                    {
			if(rb->object[item_num]->link_dest != NULL)
		        {
                            sprintf(tmp_path, "%s -> %s",
                                rb->object[item_num]->name,
                                rb->object[item_num]->link_dest
                            );
                        }
		        else
			{
		            strncpy(tmp_path, rb->object[item_num]->name,
                                PATH_MAX + NAME_MAX);
			}
		    }
                    else
                    {
                        strncpy(tmp_path, rb->object[item_num]->name,
			    PATH_MAX + NAME_MAX);
                    }
                    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';

                    OSWDrawString(rb->list_buf,
                        24 + 8,
                        list_y_pos + 18,
                        tmp_path
                    );

		    /* Object size (files only). */
		    if(S_ISREG(rb->object[item_num]->mode))
		    {
                        OSWSetFgPix(widget_global.editable_text_pix);

			sprintf(
			    stringa,
			    "(%ld bytes)",
			    rb->object[item_num]->size
			);

                        OSWDrawString(rb->list_buf,
                            24 +
                                (strlen(rb->object[item_num]->name) * CHAR_WIDTH)
			    + 8 + 10,
                            list_y_pos + 18, 
                            stringa
                        );
		    }


                    /* Increment positions. */
                    list_y_pos += ROW_SPACING;
                    items_drawn++;
                    item_num++;
	        }
	    }

            /* Draw frame around list window. */
            WidgetFrameButtonPixmap(rb->list_buf, True,
                wattr.width, wattr.height,
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix
            );

            /* Put buffer. */
            OSWPutBufferToWindow(rb->list, rb->list_buf);
	}

	/* ********************************************************** */
	/* Status window. */
	if((amount == DRAW_AMOUNT_COMPLETE) ||
           (amount == DRAW_AMOUNT_STATUS)
	)
	{
	    OSWGetWindowAttributes(rb->status_win, &wattr);
            WidgetPutImageTile(
                rb->status_win_buf, widget_global.std_bkg_img,
                wattr.width, wattr.height
            );

	    /* Padlock icon. */
            if(option.write_protect)
            {
                ximage = endeavour_image.padlock1;
            }               
            else
            {
                ximage = endeavour_image.padlock_tmp;
            }
            if(ximage != NULL)
            {
                OSWPutImageToDrawablePos(
                    ximage,
                    (drawable_t)rb->status_win_buf,
                    (int)wattr.width - (int)ximage->width - 15,
                    ((int)wattr.height / 2) -
                        ((int)ximage->height / 2)
                );
            }


            /* Status message. */
            if((rb->status_mesg != NULL) && (ximage != NULL))
            {
                OSWSetFgPix(widget_global.normal_text_pix);
                OSWDrawStringLimited(
                    rb->status_win_buf,
                    10,
                    ((int)wattr.height / 2) + 5,
                    rb->status_mesg,
                    MIN((int)strlen(rb->status_mesg),
                        ((int)wattr.width - (int)ximage->width - 25) /
                        CHAR_WIDTH
                    )
                );
            }
            
            /* Raised frame. */
            WidgetFrameButtonPixmap(rb->status_win_buf, False,
                wattr.width, wattr.height,
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix
            );

            /* Put buffer. */
            OSWPutBufferToWindow(rb->status_win, rb->status_win_buf);
        }


        /* ********************************************************** */
        /* Horizontal rule. */
	if(amount == DRAW_AMOUNT_COMPLETE)
	{
	    OSWGetWindowAttributes(rb->toplevel, &wattr);

	    OSWClearWindow(rb->toplevel);

	    y = (int)wattr.height - STATUS_HEIGHT - 20 - 28;

	    OSWSetFgPix(widget_global.surface_shadow_pix);
	    OSWDrawLine((drawable_t)rb->toplevel,
		0, y,
		(int)wattr.width, y
	    );

            OSWSetFgPix(widget_global.surface_highlight_pix);
            OSWDrawLine((drawable_t)rb->toplevel,
                0, y + 1,
                (int)wattr.width, y + 1
            );
	}


	return(0);
}



int RecBinManage(recbin_win_struct *rb, event_t *event)
{
	int x, y, z;
        int events_handled = 0;
	win_attr_t wattr;
        image_t *image;

        char tmp_path[PATH_MAX + NAME_MAX];
	char stringa[1024];

	char **strv;
	int strc;


        /* Error checks. */
        if((rb == NULL) ||
           (event == NULL) 
        )
            return(events_handled);
             
        switch(event->type)
        {
          /* ********************************************************* */
          case KeyPress:
            /* Skip KeyPress if rb not in focus. */
            if(rb->is_in_focus == 0)
                return(events_handled);


	    /* Recycle bin path prompt. */
	    if((rb->recbin_path_prompt.is_in_focus == 1) &&
               (rb->recbin_path_prompt.buf != NULL)
            )
	    {
		/* Enter key. */
	        if((event->xkey.keycode == osw_keycode.enter) ||
                   (event->xkey.keycode == osw_keycode.np_enter)
	        )
	        {
		    /* Change global fname.recycle_bin if absolute path. */
		    if(ISPATHABSOLUTE(rb->recbin_path_prompt.buf))
		    {
			strncpy(fname.recycle_bin,
			    rb->recbin_path_prompt.buf,
			    PATH_MAX + NAME_MAX
			);
			fname.recycle_bin[PATH_MAX + NAME_MAX - 1] = '\0';
		    }
		    /* Not absolute path, change back to previous global. */
		    else
		    {
			strncpy(
			    rb->recbin_path_prompt.buf,
			    fname.recycle_bin,
			    rb->recbin_path_prompt.buf_len
			);
			rb->recbin_path_prompt.buf_pos = strlen(
			    rb->recbin_path_prompt.buf
			);
		    }


		    /* Refresh listing on recycle bin. */
                    RECBINREFRESHLISTING(rb);

                    /* Redraw recycle bin. */
                    RecBinDraw(rb, DRAW_AMOUNT_COMPLETE);

		    events_handled++;
                    return(events_handled);
	        }
	    }
	    /* F5 refresh listing. */
	    else if(event->xkey.keycode == osw_keycode.f5)
	    {
                rb->refresh_btn.state = PBTN_ARMED;
                PBtnDraw(&rb->refresh_btn);

		OSWKBAutoRepeatOff();

                events_handled++;
                return(events_handled);
	    }

            break;

          /* ********************************************************* */
          case KeyRelease:
            /* Skip KeyPress if rb not in focus. */  
            if(rb->is_in_focus == 0)
                return(events_handled);


            /* F5 refresh listing. */
            if(event->xkey.keycode == osw_keycode.f5)
            {
                /* Refresh listing on recycle bin. */
                RECBINREFRESHLISTING(rb);
             
                /* Redraw recycle bin. */
                RecBinDraw(rb, DRAW_AMOUNT_COMPLETE);

		rb->refresh_btn.state = PBTN_UNARMED;
		PBtnDraw(&rb->refresh_btn);

                OSWKBAutoRepeatOn();


                events_handled++;
                return(events_handled);
            }
	    break;

          /* ********************************************************* */
	  case ButtonPress:
            /* Change focus as needed. */
            if((event->xany.window == rb->toplevel) ||
               (event->xany.window == rb->list) ||
               (event->xany.window == rb->status_win)
            )
                rb->is_in_focus = 1;


	    /* Status window - padlock. */
	    if((event->xany.window == rb->status_win) &&
               (endeavour_image.padlock0 != NULL)
	    )
	    {
                image = endeavour_image.padlock0;

                OSWGetWindowAttributes(rb->status_win, &wattr);
                if( (event->xbutton.x >= ((int)wattr.width - image->width - 15)) &&
                    (event->xbutton.x < ((int)wattr.width - 15)) &&
                    (event->xbutton.y >= 2) &&
                    (event->xbutton.y < ((int)wattr.height - 2))
                )
                {
                    if(option.write_protect)
                        option.write_protect = 0;
                    else
                        option.write_protect = 1;

                    /* Need to redraw padlocks on all windows that have. */
                    for(x = 0; x < (int)total_endeavour_wins; x++)
		    {
			if(endeavour_win[x] == NULL) continue;
			if(endeavour_win[x]->map_state < 1) continue;
                        EDWDraw(endeavour_win[x], DRAW_AMOUNT_STATUS);
		    }
		    for(x = 0; x < (int)total_recbin_wins; x++)
                    {
			if(recbin_win[x] == NULL) continue;
                        if(recbin_win[x]->map_state < 1) continue;

                        RecBinDraw(recbin_win[x], DRAW_AMOUNT_STATUS);
                    }
#ifdef USE_IMLIB
                    for(x = 0; x < (int)total_img_browsers; x++)
                    {
                        if(img_browser[x] == NULL) continue;
                        if(img_browser[x]->map_state < 1) continue;

                        IMBRDraw(img_browser[x], DRAW_AMOUNT_STATUS);
                    }
#endif /* USE_IMLIB */

                    events_handled++;
                    return(events_handled);
                }
	    } 
	    /* List window. */
	    else if(event->xany.window == rb->list)
	    {
		/* Set scroll bar into focus. */
		rb->list_sb.is_in_focus = 1;

                events_handled++;
                return(events_handled);
	    }
	    break;

          /* ********************************************************* */
          case ButtonRelease:
            /* Change focus as needed. */
/*
            if((event->xany.window == rb->toplevel) ||
               (event->xany.window == rb->list) ||
               (event->xany.window == rb->status_win)
            )
                rb->is_in_focus = 1;
*/

	    /* List window: Button1. */
	    if((event->xany.window == rb->list) &&
               (event->xbutton.button == Button1) &&
               !drag_record[0].state
            )
            {
                /* Get attributes (will be needed below) */
                OSWGetWindowAttributes(rb->list, &wattr);

                /* Get item button was released on. */
                x = GETITEMBYPOSITION(
                    rb,
                    event->xbutton.x,
                    event->xbutton.y
		);

                /* Did we get a valid position? */
                if((x > -1) && (x < (int)rb->total_objects))
                {
                    /*   Shift key held down will select multiple files
	             *   in a row.
		     */
                    if(drag_record[0].shift_key_state)
                    {
                        /* Only works if more than 1 object selected. */
                        if((int)rb->total_sel_objects > 0)
                        {
			    y = 0;
			    z = 0;
                            /* Top to bottom select. */
                            if(x > (int)rb->sel_object[0])
                            {
                                y = rb->sel_object[0];
                                z = x + 1;
                            }
                            else if(x < (int)rb->sel_object[0])
                            {
                                y = x;
                                z = rb->sel_object[0] + 1;
                            }
                            RECBIN_UNSELECTALL(rb);
                            while(y < z)
                            {
                                RECBIN_SELECT(rb, y);
                                y++;
                            }
                        }
                    }
                    /* Ctrl key held down will select multiple objects. */
                    else if(drag_record[0].ctrl_key_state)
                    {
		        if(RECBIN_ISSELECTED(rb, x))
			    RECBIN_UNSELECT(rb, x);
			else
			    RECBIN_SELECT(rb, x);
                    }
                    /* Single select. */
                    else
                    {
                        RECBIN_UNSELECTALL(rb);
                        RECBIN_SELECT(rb, x);
                    }

                    /* Update status message. */
                    for(x = 0, z = 0; x < (int)rb->total_sel_objects; x++)
                    {
                            y = rb->sel_object[x];
                            if((y < 0) || (y >= (int)rb->total_objects)) continue;
                            if(rb->object[y] == NULL) continue;

                            z += rb->object[y]->size;
                    }
                    sprintf(tmp_path,
   "%i object%s selected, %i bytes.",
                        rb->total_sel_objects,
                        ((int)rb->total_sel_objects > 1) ? "s" : "",
                        z
                    );
		    free(rb->status_mesg);
                    rb->status_mesg = StringCopyAlloc(tmp_path);

 
                    /* Scroll if near edge. */
                    if(event->xbutton.y >
                       ((int)wattr.height - ROW_SPACING)
                    )
                    {
                            RecBinDraw(rb, DRAW_AMOUNT_LISTS);
			    OSWGUISync(False);

			    OSWGetWindowAttributes(rb->list, &wattr);
                            rb->list_sb.y_win_pos += (int)wattr.height / 2;
                            usleep(SELECT_SHIFT_DELAY * 1000);

                            y = (((int)rb->total_objects + 1) * ROW_SPACING) -
                                (int)wattr.height;
                            if(rb->list_sb.y_win_pos > y) 
                                rb->list_sb.y_win_pos = y;

                            SBarDraw(
				&rb->list_sb,
				wattr.width, wattr.height,
                                wattr.width,
                                ((int)rb->total_objects + 1) * ROW_SPACING
                            );
                    }
                    else if(event->xbutton.y < ROW_SPACING)
                    {
			    RecBinDraw(rb, DRAW_AMOUNT_LISTS);
                            OSWGUISync(False);

			    OSWGetWindowAttributes(rb->list, &wattr);
                            rb->list_sb.y_win_pos -= (int)wattr.height / 2;
                            usleep(SELECT_SHIFT_DELAY * 1000);

                            if(rb->list_sb.y_win_pos < 0) 
                                rb->list_sb.y_win_pos = 0;

			    SBarDraw(
				&rb->list_sb,
				wattr.width, wattr.height,
				wattr.width,
				((int)rb->total_objects + 1) * ROW_SPACING
			    );
                    }
                }	/* Did we get a valid position? */

		/* Redraw and return. */
                RecBinDraw(rb, DRAW_AMOUNT_LISTS);
                RecBinDraw(rb, DRAW_AMOUNT_STATUS);

                events_handled++;
                return(events_handled);
	    }

            break;

          /* ********************************************************* */
          case MotionNotify:

            /* Status window (padlock hints) */
            if((event->xany.window == rb->status_win) &&
               (endeavour_image.padlock0 != NULL)
            )
            {
                image = endeavour_image.padlock0;
                x = event->xmotion.x;
                y = event->xmotion.y;

                OSWGetWindowAttributes(rb->status_win, &wattr);
                if( (x >= ((int)wattr.width - (int)image->width - 15)) &&
                    (x < ((int)wattr.width - 15)) &&
                    (y >= 2) &&
                    (y < ((int)wattr.height - 2))
                )
                {
                    /* Schedual hints window map. */
                    HintWinSetSchedual(  
                        widget_global.hintwin_map_delay,
                        rb->status_win
                    );
                    events_handled++;
                }
                else
                {
                    /* Unmap/unschedual hints window. */
                    HintWinUnmap();
                }
		return(events_handled);
            }
            break;   

          /* ********************************************************* */
          case Expose:
	    if(event->xany.window == rb->toplevel)
	    {
		RecBinDraw(rb, DRAW_AMOUNT_COMPLETE);
		OSWPurgeWindowTypedEvent(event->xany.window, Expose);

		events_handled++;
                return(events_handled);
	    }
	    else if(event->xany.window == rb->list)
	    {
                RecBinDraw(rb, DRAW_AMOUNT_LISTS);
                OSWPurgeWindowTypedEvent(event->xany.window, Expose);

                events_handled++;
                return(events_handled);
	    }
            else if(event->xany.window == rb->status_win)
            {
                RecBinDraw(rb, DRAW_AMOUNT_STATUS);
                OSWPurgeWindowTypedEvent(event->xany.window, Expose);

                events_handled++;
                return(events_handled);
            }                            
            break;

          /* ********************************************************* */
          case ConfigureNotify:
            if(event->xany.window == rb->toplevel)
            {
                if(RecBinResize(rb) == 0)
		{
		    rb->map_state = 0;
                    RecBinDraw(rb, DRAW_AMOUNT_COMPLETE);
		}

                events_handled++;
		return(events_handled);
            }
            break;   

          /* ********************************************************* */
          case FocusIn:
            if(event->xany.window == rb->toplevel)
            {
                rb->is_in_focus = 1;
                events_handled++;
            }
            else
            {
                rb->is_in_focus = 0;
                /* Do not report any event for focus out. */
            }
            return(events_handled);
            break;  

          /* ********************************************************* */
          case FocusOut:
            if(event->xany.window == rb->toplevel)
            {
                rb->is_in_focus = 0;
		events_handled++;
            }
            return(events_handled);
            break;
                
          /* ********************************************************* */
          case VisibilityNotify:
            if(event->xany.window == rb->toplevel)
            {
                rb->visibility_state = event->xvisibility.state;

                events_handled++;
                return(events_handled);
            }
            break;

	  /* ********************************************************* */
	  case ClientMessage:
	    if(OSWIsEventDestroyWindow(rb->toplevel, event))
            {
		RecBinUnmap(rb);

		events_handled++;
                return(events_handled);
	    }
	    break;

          /* ********************************************************* */
          case LeaveNotify:
	    /* Status window (padlock). */
            if(event->xany.window == rb->status_win)
            {
                /* Unmap/unschedual hints window. */
                HintWinUnmap();
            }
	    break;
        }

        /* Redraw as needed. */
        if(events_handled > 0)
        {
            RecBinDraw(rb, DRAW_AMOUNT_COMPLETE);
            return(events_handled);
        }       

          
        /* ************************************************************ */
        /* Manage widgets. */

	/* List scroll bars. */
	if(events_handled == 0)
        {
	    OSWGetWindowAttributes(rb->list, &wattr);

	    if(rb->recbin_path_prompt.is_in_focus == 0)
	    {
                events_handled += SBarManage(
		    &rb->list_sb,
		    wattr.width,
		    wattr.height,
		    wattr.width,
		    ((int)rb->total_objects + 1) * ROW_SPACING,
		    event
	        );
	    }
	    else if((event->type != KeyPress) &&
                    (event->type != KeyRelease)
            )
            {
                events_handled += SBarManage(
                    &rb->list_sb,
                    wattr.width, 
                    wattr.height,
                    wattr.width, 
                    ((int)rb->total_objects + 1) * ROW_SPACING,
                    event
                );
            }


	    if(events_handled > 0)
	    {
		RecBinDraw(rb, DRAW_AMOUNT_LISTS);
                return(events_handled);
	    }
	}


	/* Prompt widget. */
	if(events_handled == 0)
	{
	    events_handled += PromptManage(
		&rb->recbin_path_prompt,
		event
	    );

	}


	/* Recover button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &rb->recover_btn,
                event  
            );
            if(((event->type == ButtonRelease) ||  
               (event->type == KeyRelease)) &&
               (events_handled > 0)
            )
            {
		if(option.write_protect)
		{
		    printdw(&err_dw,
"Cannot recover object:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
		    );
		}
		else
		{
                    /* Set all recbin window cursors to `working'. */
                    RECBINSETALLCURSORS(rb, EDW_CURSORS_CODE_WORKING);


		    /* Recover all selected objects. */
		    for(x = 0; x < (int)rb->total_sel_objects; x++)
		    {
			y = rb->sel_object[x];
			if((y < 0) || (y >= (int)rb->total_objects))
			    continue;
			if(rb->object[y] == NULL) continue;
                        if(rb->object[y]->name == NULL) continue;

			z = RecycleDoRestoreObject(
                            rb->object[y]->name,
                            fname.recycle_bin,
                            NULL	/* Use recorded target path. */
                        );
			if(z < 0)
			{
			    sprintf(stringa,
"Unable to recover object:\n
\n\
    %s\n\
\n\
For one the following reasons:\n\n\
    * An existing `real' copy of the object exists.\n\
    * Original recovery path no longer exists.\n\
    * Data in the recycle bin is damaged beyond recovery.\n\
\n\
Try dragging the object from the recycle bin to a\n\
browser window to recover to an explicit destination.\n",
			        rb->object[y]->name
			    );
			    printdw(&err_dw, stringa);

			    /* Break, do not attempt to recover more. */
			    break;
			}
		    }
		    RECBINREFRESHLISTING(rb);

		    /* Need to refresh listing and redraw all endeavour windows. */
		    for(x = 0; x < (int)total_endeavour_wins; x++)
		    {
			if(endeavour_win[x] == NULL) continue;
			EDWRefreshAllLists(endeavour_win[x]);
			EDWDraw(endeavour_win[x], DRAW_AMOUNT_COMPLETE);
		    }

                    /* Complete redraw. */
                    rb->map_state = 0;
                    RecBinDraw(rb, DRAW_AMOUNT_COMPLETE);

                    /* Set back all recbin window cursors. */
                    RECBINSETALLCURSORS(rb, EDW_CURSORS_CODE_NORMAL);
		}
	    }
        }

        /* Refresh button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &rb->refresh_btn,
                event
            );
            if(((event->type == ButtonRelease) ||  
               (event->type == KeyRelease)) &&
               (events_handled > 0)
            )
            {
                /* Set all recbin window cursors to `working'. */
                RECBINSETALLCURSORS(rb, EDW_CURSORS_CODE_WORKING);

		/* Refresh listing. */
		RECBINREFRESHLISTING(rb);

                /* Complete redraw. */
                rb->map_state = 0;
		RecBinDraw(rb, DRAW_AMOUNT_COMPLETE);

                /* Set back all recbin window cursors. */
                RECBINSETALLCURSORS(rb, EDW_CURSORS_CODE_NORMAL);

		return(events_handled);
	    }
        }

        /* Purge button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &rb->purge_btn,
                event
            );
            if(((event->type == ButtonRelease) ||  
               (event->type == KeyRelease)) &&
               (events_handled > 0)
            )
            {
                /* Check if internal write-protect is on. */
                if(option.write_protect)
                {
                    printdw(&err_dw,
"Cannot purge object(s):\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                    );
                    return(events_handled);
                }

		/* Set all recbin window cursors to `working'. */
		RECBINSETALLCURSORS(rb, EDW_CURSORS_CODE_WORKING);

		/* Create purge items list. */
		strv = NULL;
		strc = 0;

		for(x = 0; x < (int)rb->total_sel_objects; x++)
		{
		    y = rb->sel_object[x];
		    if((y < 0) || (y >= (int)rb->total_objects))
			continue;
		    if(rb->object[y] == NULL) continue;
		    if(rb->object[y]->name == NULL) continue;

		    strc++;
		    strv = (char **)realloc(strv, strc * sizeof(char *));
		    if(strv == NULL) return(events_handled);

		    strv[strc - 1] = StringCopyAlloc(rb->object[y]->name);
		}


		/* Comferm and purge each item. */

                /* Move comfermation dialog to pointer's center. */
                WidgetCenterWindow(comf_dialog.toplevel,
                    WidgetCenterWindowToPointer);

                yes_to_all = False;
		if(option.comferm_purge)
		{
		    for(x = 0; x < strc; x++)
		    {
		        if(strv[x] == NULL) continue;

		        /* Do comfermation. */
		        if(!yes_to_all)
		        {
			    z = ComfDialogMapQueryPurge(rb, strv[x]);
			    if(z == NO_PBTN_NUM)
			        continue;
			    if(z == ALL_PBTN_NUM)
			        yes_to_all = True;
		        }

		        /* Purge this object. */
		        RecycleDoPurgeObject(strv[x],
			    rb->recbin_path_prompt.buf);
		    }
		}

                /* Turn off yes_to_all. */
                yes_to_all = False;

		/* Free names. */
		for(x = 0; x < strc; x++)
		    free(strv[x]);
		free(strv);
		strv = NULL;
		strc = 0;

		/* Refresh listing. */
                RECBINREFRESHLISTING(rb);

                /* Complete redraw. */
                rb->map_state = 0;
                RecBinDraw(rb, DRAW_AMOUNT_COMPLETE); 

                /* Set back all recbin window cursors. */
                RECBINSETALLCURSORS(rb, EDW_CURSORS_CODE_NORMAL);


		/* Refresh endeavour windows. */
		for(x = 0; x < (int)total_endeavour_wins; x++)
		{
		    if(endeavour_win[x] == NULL) continue;
		    EDWRefreshAllLists(endeavour_win[x]);
		    EDWDraw(endeavour_win[x], DRAW_AMOUNT_COMPLETE);
		}


                return(events_handled);
	    }
	}

        /* Purge all button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &rb->purge_all_btn,
                event
            );
            if(((event->type == ButtonRelease) ||
               (event->type == KeyRelease)) &&
               (events_handled > 0)
            )
            {
                /* Check if internal write-protect is on. */
                if(option.write_protect)
                {
                    printdw(&err_dw,
"Cannot purge all contents of recycle bin:\n\n\
    Master write-protect is on.\n\n\
(To turn off the master write-protect,\n\
click on the lower right padlock icon.)\n"
                    );
                    return(events_handled);
                }

		/* Set all recbin window cursors to `working'. */
                RECBINSETALLCURSORS(rb, EDW_CURSORS_CODE_WORKING);

		/* Comferm. */

                /* Move comfermation dialog to pointer's center. */
                WidgetCenterWindow(comf_dialog.toplevel,
                    WidgetCenterWindowToPointer);

		if(option.comferm_purge_all)
		{
		    z = ComfDialogMapQueryPurgeAll(rb);
		    if(z == YES_PBTN_NUM)
		    {
		        /* Purge selected items by removing recycle bin. */
		        unlink(fname.recycle_bin);
		    }
		}

		/* Refresh listing. */
                RECBINREFRESHLISTING(rb);

                /* Complete redraw. */
                rb->map_state = 0;
		RecBinDraw(rb, DRAW_AMOUNT_COMPLETE);

                /* Set back all recbin window cursors. */
                RECBINSETALLCURSORS(rb, EDW_CURSORS_CODE_NORMAL);


                /* Refresh endeavour windows. */
                for(x = 0; x < (int)total_endeavour_wins; x++)
                {
                    if(endeavour_win[x] == NULL) continue;
                    EDWRefreshAllLists(endeavour_win[x]);
                    EDWDraw(endeavour_win[x], DRAW_AMOUNT_COMPLETE);
                }
                 

                return(events_handled);
            }
        }

        /* Close button. */
        if(events_handled == 0)
        {
            events_handled += PBtnManage(
                &rb->close_btn,
                event
            );
	    if(((event->type == ButtonRelease) ||
               (event->type == KeyRelease)) &&
               (events_handled > 0)
	    )
	    {
		RecBinUnmap(rb);
		return(events_handled);
	    }
        }






        return(events_handled);
}



void RecBinMap(recbin_win_struct *rb)
{
        if(rb == NULL)
            return;


        UnfocusAllTopLvlWins();

	/* Reset cursor codes. */
	RECBINSETALLCURSORS(rb, EDW_CURSORS_CODE_NORMAL);

	/* Reset button states. */
	rb->recover_btn.state = PBTN_UNARMED;
        rb->refresh_btn.state = PBTN_UNARMED;
        rb->purge_btn.state = PBTN_UNARMED;
        rb->purge_all_btn.state = PBTN_UNARMED;
        rb->close_btn.state = PBTN_UNARMED;


	/* Map recycle bin. */
	rb->map_state = 0;
        RecBinDraw(rb, DRAW_AMOUNT_COMPLETE);

        rb->visibility_state = VisibilityUnobscured;

	/* Must be in focus. */
        rb->is_in_focus = 1;

	return;
}



void RecBinUnmap(recbin_win_struct *rb)
{
	if(rb == NULL)
	    return;


	OSWUnmapWindow(rb->toplevel);
	rb->visibility_state = VisibilityFullyObscured;
	rb->is_in_focus = 0;
	rb->map_state = 0;

	return;
}



void RecBinDestroy(recbin_win_struct *rb)
{
        if(rb == NULL)
            return; 


	/* Free all selections and objects. */
	RECBIN_UNSELECTALL(rb);
	RECBINFREEALLOBJECTS(rb);

	free(rb->status_mesg);
	rb->status_mesg = NULL;


        if(IDC())
        {

            HintWinDeleteMessage(rb->status_win);          /* Padlock hint. */
	    OSWDestroyPixmap(&rb->status_win_buf);
	    OSWDestroyWindow(&rb->status_win);

            PromptDestroy(&rb->recbin_path_prompt);

	    PBtnDestroy(&rb->recover_btn);
            PBtnDestroy(&rb->refresh_btn);
            PBtnDestroy(&rb->purge_btn);
            PBtnDestroy(&rb->purge_all_btn);
            PBtnDestroy(&rb->close_btn);

	    SBarDestroy(&rb->list_sb);
	    OSWDestroyPixmap(&rb->list_buf);
	    OSWDestroyWindow(&rb->list);

            OSWDestroyWindow(&rb->toplevel);

	    OSWDestroyPixmap(&rb->icon);
	}


        rb->map_state = 0;
        rb->visibility_state = VisibilityFullyObscured;
        rb->x = 0;
	rb->y = 0;
	rb->width = 0;
	rb->height = 0;
        rb->is_in_focus = 0;
        rb->disabled = False;


	return;
}
